from kivy.core.text import DEFAULT_FONT
from kivy.uix.button import Button
from kivy.uix.behaviors import ButtonBehavior
from kivy.uix.label import Label
from kivy.uix.gridlayout import GridLayout
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.checkbox import CheckBox
from kivy.uix.togglebutton import ToggleButton
from kivy.graphics import Color, Line, RoundedRectangle, \
    GraphicException, Rectangle
from kivy.properties import NumericProperty, ListProperty, \
    StringProperty, VariableListProperty, DictProperty,\
    BooleanProperty, OptionProperty, ObjectProperty
from kivy.uix.widget import Widget
from kivy.uix.image import Image
from kivy.uix.listview import ListView, CompositeListItem, \
    ListItemButton, ListItemLabel
from kivy.uix.tabbedpanel import TabbedPanelHeader
from kivy.uix.bubble import Bubble
from kivy.animation import Animation
from kivy.uix.listview import SelectableView
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.modalview import ModalView
from kivy.uix.tabbedpanel import TabbedPanel
from .TinyLib import *
from . import images

def kStrRGB(Str, alpha=1):
    return StrRGB(Str) + (alpha,)
kXM_GREEN = kStrRGB("#CCFF26")
kXM_GREEN1 = kStrRGB("#CFF93E")
kXM_RED = kStrRGB("#E74C3C")
kRED = kStrRGB("F00")
k1_8_RED = kStrRGB("811")
k2_A_RED = kStrRGB("A22")
k36_C9_RED = kStrRGB("C93636")
k3F_F_RED = kStrRGB("FF3F3F")
kGREEN = kStrRGB("0F0")
k3F_F_GREEN = kStrRGB("3FFF3F")
k7_F_GREEN = kStrRGB("7F7")
kBLACK = kStrRGB("000")
k4_GREY = kStrRGB("444")
k4F_GREY = kStrRGB("4F4F4F")
k5_GREY = kStrRGB("555")
k7F_GREY = kStrRGB("7F7F7F")
kA_GREY = kStrRGB("AAA")
kC_GREY = kStrRGB("CCC")
kD0_GREY = kStrRGB("D0D0D0")
kDA_GREY = kStrRGB("#DADADA")
kD_GREY = kStrRGB("DDD")
kWHITE = kStrRGB("FFF")
kXM_LIST_BG = [kC_GREY, kDA_GREY]
stColors = ["FFF", "#CCFF26"]
cmpColors = ["FFF", "F00"]
cmpRGB = map(kStrRGB, cmpColors)
fromState = {"normal": 0, "down": 1}
toState = ["normal", "down"]
stRGB = map(kStrRGB, stColors)
defTxtStyle = dict(color=(1, 1, 1, 1), background_color=kStrRGB("444", 3))

# Widget mixin class to snap widgets to exact pixel boundaries
class SnappedWidget(object):
    def roundToInt(self, x, y):
        assert isinstance(self, Widget)
        x, y = self.to_window(*self.to_local(x, y))
        x, y = map(lambda z: int(z), [x, y])
        return self.to_parent(*self.to_widget(x, y))

    def __setattr__(self, attr, val):
        assert isinstance(self, Widget)
        if attr in {"pos", "size", "center", "text_size"}:
            newVal = self.roundToInt(*tuple(val))
            if isinstance(val, list): val = list(newVal)
        elif attr in {"x", "right", "width", "center_x"}:
            val = self.roundToInt(val, 0)[0]
        elif attr in {"y", "top", "height", "center_y"}:
            val = self.roundToInt(0, val)[1]
        super(SnappedWidget, self).__setattr__(attr, val)

# Widget mixin class to add borders into widget classes
#  bgc: background color
#  bdc: border color
#  bdw: border width
#  bdr: border radius
#  bdo: border offset
# unlike color properties for widgets from kivy these properties accept colors
#  of the form of (r, g, b) or (r, g, b, a) however like kivy these components
#  are still floating point numbers that range from 0.0 to 1.0
class BorderBehavior(SnappedWidget):
    bdw = NumericProperty(1)
    bdr = NumericProperty(8)
    # noinspection PyArgumentList
    bgc = ListProperty(StrRGB("404040"))
    # noinspection PyArgumentList
    bdc = ListProperty(StrRGB("d0d0d0"))
    bdo = NumericProperty(2)
    initDone = False

    def __init__(self, *args, **kwargs):
        assert isinstance(self, Widget)
        super(BorderBehavior, self).__init__(*args, **kwargs)
        draw = BorderBehavior.drawBkgr
        self.lstDrawObj = list()
        off = self.bdo
        pos = map(Stack(adder(-off), adder(.5), int), self.pos)
        size = map(Stack(adder(2 * off), int), self.size)
        with self.canvas.before:
            cur = ValuePack()
            cur.obj = Color(*self.bgc)
            cur.bgc = ["rgb", "rgba"]
            self.lstDrawObj.append(cur)
            cur = ValuePack()
            cur.obj = RoundedRectangle(
                pos=pos, size=size, radius=[(self.bdr,) * 2] * 4)
            cur[" func"] = self.doFillRect
            # cur.pos = ["pos"]
            # cur.size = ["size"]
            # cur.bdr = ["radius"]
            self.lstDrawObj.append(cur)
            cur = ValuePack()
            cur.obj = Color(*self.bdc)
            cur.bdc = ["rgb", "rgba"]
            self.lstDrawObj.append(cur)
            cur = ValuePack()
            # noinspection PyArgumentList
            cur.obj = Line(
                rounded_rectangle=tuple(pos + size + [self.bdr, self.bdr * 4]),
                width=self.bdw)
            cur[" func"] = self.doBorder
            cur.bdw = ["width"]
            self.lstDrawObj.append(cur)
        self.bind(
            pos=draw, size=draw, bgc=draw, bdc=draw, bdw=draw, bdr=draw, bdo=draw)
        self.initDone = True

    def doFillRect(self, obj):
        assert isinstance(self, Widget)
        off = self.bdo
        obj.pos = map(Stack(adder(-off), adder(.5), int), self.pos)
        obj.size = map(Stack(adder(2 * off), int), self.size)
        obj.radius = [(self.bdr,) * 2] * 4

    def doBorder(self, obj):
        assert isinstance(self, Widget)
        off = self.bdo
        pos = map(Stack(adder(-off), adder(.5), int), self.pos)
        size = map(Stack(adder(2 * off), int), self.size)
        if self.bdr == 0:
            obj.rectangle = tuple(pos + size)
        else:
            obj.rounded_rectangle = tuple(pos + size + [self.bdr, self.bdr * 4])

    def drawBkgr(self, val=None):
        del val
        if not self.initDone: return
        for cmd in self.lstDrawObj:
            obj = cmd.obj
            for k in cmd:
                if k == "obj": continue
                if k[0] == ' ':
                    if k[1:] == "func":
                        cmd[k](obj)
                    continue
                for attr in cmd[k]:
                    try:
                        setattr(obj, attr, getattr(self, k))
                        break
                    except (AttributeError, GraphicException):
                        pass
        doLayout = getattr(super(BorderBehavior, self), "do_layout", None)
        if doLayout is not None: doLayout()

    def add_widget(self, *args, **kwargs):
        assert isinstance(self, Widget)
        super(BorderBehavior, self).add_widget(*args, **kwargs)
        self.drawBkgr()


class SpacingHintBehavior(object):
    spacing_hint = VariableListProperty([0], length=2)

    def __init__(self, *args, **kwargs):
        assert isinstance(self, Widget)
        super(SpacingHintBehavior, self).__init__(*args, **kwargs)
        self.bind(size=SpacingHintBehavior.updateSize)

    def on_spacing_hint(self, obj, val):
        del obj, val
        assert isinstance(self, Widget)
        self.updateSize(self.size)

    def updateSize(self, val):
        del val
        assert isinstance(self, Widget)
        for c in xrange(2):
            self.spacing[c] = self.spacing_hint[c] * self.size[c]


class SpcHintGridLayout(SpacingHintBehavior, GridLayout): pass


class Bordered(BorderBehavior, FloatLayout): pass


class BorderedGridLayout(BorderBehavior, GridLayout): pass


class BorderedList(BorderBehavior, ListView): pass


class BorderedLabel(BorderBehavior, Label): pass


class LstItemBtnLabel(SnappedWidget, ButtonBehavior, ListItemLabel): pass


class ClkLabel(SnappedWidget, ButtonBehavior, Label): pass


class ClkCompLstItem(CompositeListItem, Button): pass


class SnappedCompositeListItem(SnappedWidget, CompositeListItem): pass


class SnappedListItemLabel(SnappedWidget, ListItemLabel): pass


class SnappedListItemButton(SnappedWidget, ListItemButton): pass


class SnappedFloatLayout(SnappedWidget, FloatLayout): pass


class ListItemCheckBox(SelectableView, CheckBox): pass


class ListItemClkLabel(ListItemLabel, ButtonBehavior): pass


class SnappedTabbedPanelHeader(SnappedWidget, TabbedPanelHeader): pass


class ClkImage(ButtonBehavior, Image):
    # noinspection PyArgumentList
    disabled_color = ListProperty([.2, .2, .2, 1.])
    # noinspection PyArgumentList
    enabled_color = ListProperty([1., 1., 1., 1.])
    # noinspection PyArgumentList
    down_color = ListProperty([.6, .6, .6, 1])

    def __init__(self, **kwargs):
        if "disabled_color" in kwargs:
            self.disabled_color = kwargs["disabled_color"]
        if "enabled_color" in kwargs:
            self.enabled_color = kwargs["enabled_color"]
        elif "color" in kwargs:
            self.enabled_color = kwargs["color"]
        super(ClkImage, self).__init__(**kwargs)
        ClkImage.updColors(self)
        self.bind(
            disabled=ClkImage.updColors,
            enabled_color=ClkImage.updColors,
            disabled_color=ClkImage.updColors,
            down_color=ClkImage.updColors,
            state=ClkImage.updColors)

    @staticmethod
    def updColors(self, val=None):
        del val
        if self.disabled:
            self.color = self.disabled_color
        else:
            self.color = [self.enabled_color, self.down_color][int(self.state == "down")]


# mixes color tuples c0 and c1 with each other by a ratio based on argument a
def mixColors(c0, c1, a):
    a = float(a)
    b = 1 - a
    return tuple([p0 * a + p1 * b for p0, p1 in zip(c0, c1)])


# makes the BrdLabel flash green then back to original color
#  when the text changes
def txtChgAnim(obj, val, c1_b, c1_t):
    del val
    # c1_b = obj.bdc
    # c1_t = obj.color
    obj.bdc = mixColors(kGREEN, c1_b, 1.)
    obj.color = mixColors(kGREEN, c1_t, .75)
    if obj.curAnim is not None: obj.curAnim.stop(obj)
    obj.curAnim = Animation(duration=.5, color=c1_t, bdc=c1_b)
    obj.curAnim.start(obj)


class BrdLabel(BorderBehavior, Label):
    def __init__(self, **kwargs):
        super(BrdLabel, self).__init__(**kwargs)
        self.curAnim = None
        if kwargs.get("doAnim", True):
            self.bind(text=Bind(txtChgAnim, c1_b=self.bdc, c1_t=self.color))


class ClkTxtIn(ButtonBehavior, BrdLabel):
    disabledMask = (.60, .60, .60)
    # noinspection PyArgumentList
    bgc_down = ListProperty([7., 7., 7., 1])
    # noinspection PyArgumentList
    bgc_normal = ListProperty([1, 1, 1, 1])

    def __init__(self, **kwargs):
        overrideDisabledColor = "disabled_color" in kwargs
        overrideColor = "color" in kwargs
        defs = {
            "color": (.0, .0, .0, 1.), "bgc": (1., 1., 1., 1.), "bdc": (.6, .6, .6, 1.),
            "disabled_color": (.4, .4, .4, 1.), "doAnim": False,
            "bgc_normal": kwargs.get("bgc", (1, 1, 1, 1)),
            "bgc_down": (.7, .7, .7, 1)}
        kwargs = dict(defs, **kwargs)
        super(ClkTxtIn, self).__init__(**kwargs)
        if self.disabled:
            self.disabledBgc = self.bgc
            self.enabledBgc = map(
                lambda z, y: 2 * z - y, self.bgc[0:3], self.disabledMask)
            self.enabledBgc = tuple(self.enabledBgc) + tuple(self.bgc[3:])
        else:
            self.enabledBgc = self.bgc
            self.disabledBgc = map(
                lambda x, y: .5 * (x + y), self.bgc[0:3], self.disabledMask)
            self.disabledBgc = tuple(self.disabledBgc) + tuple(self.bgc[3:])
        if not (overrideDisabledColor and overrideColor):
            if overrideDisabledColor:
                self.color = tuple(map(
                    lambda z, y: 2 * z - y, self.disabled_color[0:3],
                    self.disabledMask)) + tuple(self.disabled_color[3:])
            else:
                self.disabled_color = tuple(map(
                    lambda x, y: .5 * (x + y), self.color[0:3],
                    self.disabledMask)) + tuple(self.color[3:])
        self.bind(disabled=ClkTxtIn.onDisabled)
        self.bind(state=ClkTxtIn.updStateBgc)
        self.bind(bgc_down=ClkTxtIn.updStateBgc)
        self.bind(bgc_normal=ClkTxtIn.updStateBgc)

    def onDisabled(self, val):
        self.bgc = self.disabledBgc if val else self.enabledBgc

    @staticmethod
    def updStateBgc(self, val=None):
        del val
        if self.state == "down":
            self.bgc = self.bgc_down
        else:
            self.bgc = self.bgc_normal


class XMHomeItem(BoxLayout, ListItemButton):
    def __init__(self, **kwargs):
        kwargs["orientation"] = "horizontal"
        cls_dicts = kwargs["cls_dicts"]
        on_state = kwargs.get("on_state", None)
        if on_state is not None: del kwargs["on_state"]
        if "background_down" not in kwargs: kwargs["background_down"] = ""
        if "background_normal" not in kwargs: kwargs["background_normal"] = ""
        # if "background_disabled_down" not in kwargs: kwargs["background_disabled_down"] = ""
        # if "background_disabled_normal" not in kwargs: kwargs["background_disabled_normal"] = ""
        super(XMHomeItem, self).__init__(**kwargs)
        self.named_widgets = {}
        for cls_dict in cls_dicts:
            cls = cls_dict["cls"]
            kwargs1 = cls_dict["kwargs"]
            wid = cls(**kwargs1)
            if "id" in cls_dict:
                self.named_widgets[cls_dict["id"]] = wid
            self.add_widget(wid)
        if on_state is not None:
            self.bind(state=on_state)

    def get_widget(self, name):
        return self.named_widgets[name]


class XMLoneButton(Button):
    # noinspection PyArgumentList
    background_down_color = ListProperty([1., 1., 1., 1.])
    # noinspection PyArgumentList
    background_normal_color = ListProperty([1., 1., 1., 1.])
    # noinspection PyArgumentList
    down_color = ListProperty([1., 1., 1., 1.])
    # noinspection PyArgumentList
    normal_color = ListProperty([1., 1., 1., 1.])
    # noinspection PyArgumentList
    disabled_background_color = ListProperty([1., 1., 1., 1.])

    def __init__(self, **kwargs):
        assert isinstance(self, Button)
        fillDictDefs(kwargs, {
            "background": "",
            "background_normal": "",
            "background_down": "",
            "background_disabled": "",
            "background_disabled_down": "",
            "background_disabled_normal": ""})
        kwColor = None
        if "color" in kwargs:
            if "down_color" in kwargs or "normal_color" in kwargs:
                raise ValueError("Cannot set both normal_color/down_color and set the color itself")
            kwColor = kwargs["color"]
        super(XMLoneButton, self).__init__(**kwargs)
        self.background_color = self.background_normal_color
        if kwColor is not None:
            self.normal_color = kwColor
            self.down_color = kwColor
        self.color = self.normal_color

    def on_state(self, obj, val):
        del obj
        val = {"normal": 0, "down": 1}[val]
        bgc = self.background_down_color if val else self.background_normal_color
        if self.disabled:
            return
        self.background_color = bgc
        self.color = self.down_color if val else self.normal_color

    def on_disabled(self, obj, val):
        super(XMLoneButton, self).on_disabled(obj, val)
        if self.disabled:
            self.color = self.disabled_color
            self.background_color = self.disabled_background_color
        else:
            self.on_state(self, self.state)

    def on_disabled_background_color(self, obj, val):
        del obj, val
        if self.disabled:
            self.background_color = self.disabled_background_color

    def on_down_color(self, obj, val):
        del obj, val
        self.on_state(self, self.state)

    def on_normal_color(self, obj, val):
        del obj, val
        self.on_state(self, self.state)

    def on_background_down_color(self, obj, val):
        del obj, val
        self.on_state(self, self.state)

    def on_background_normal_color(self, obj, val):
        del obj, val
        self.on_state(self, self.state)


class XMButton(ListItemButton):
    # noinspection PyArgumentList
    background_down_color = ListProperty([1., 1., 1., 1.])
    # noinspection PyArgumentList
    background_normal_color = ListProperty([1., 1., 1., 1.])
    # noinspection PyArgumentList
    down_color = ListProperty([1., 1., 1., 1.])
    # noinspection PyArgumentList
    normal_color = ListProperty([1., 1., 1., 1.])

    def __init__(self, **kwargs):
        assert isinstance(self, Button)
        kwColor = None
        if "color" in kwargs:
            if "down_color" in kwargs or "normal_color" in kwargs:
                raise ValueError("Cannot set both normal_color/down_color and set the color itself")
            kwColor = kwargs["color"]
        on_is_selected = None
        if "on_is_selected" in kwargs:
            on_is_selected = kwargs["on_is_selected"]
            del kwargs["on_is_selected"]
        super(XMButton, self).__init__(**kwargs)
        if on_is_selected is not None:
            self.bind(is_selected=on_is_selected)
        self.background_color = self.background_normal_color
        if kwColor is not None:
            self.normal_color = kwColor
            self.down_color = kwColor
        self.color = self.normal_color

    def on_state(self, obj, val):
        val = {"normal": 0, "down": 1}[val]
        bgc = self.background_down_color if val else self.background_normal_color
        self.background_color = bgc
        self.selected_color = bgc
        self.deselected_color = bgc
        self.color = self.down_color if val else self.normal_color

    def select(self, *args):
        # bypass ListItemButton.select
        super(ListItemButton, self).select(*args)

    def deselect(self, *args):
        # bypass ListItemButton.select
        super(ListItemButton, self).deselect(*args)

    def on_down_color(self, obj, val):
        del obj, val
        self.on_state(self, self.state)

    def on_normal_color(self, obj, val):
        del obj, val
        self.on_state(self, self.state)

    def on_background_down_color(self, obj, val):
        del obj, val
        self.on_state(self, self.state)

    def on_background_normal_color(self, obj, val):
        del obj, val
        self.on_state(self, self.state)


class XMButton1(XMButton):
    selected_attrs = DictProperty(defaultvalue={})
    deselected_attrs = DictProperty(defaultvalue={})

    def __init__(self, **kwargs):
        super(XMButton1, self).__init__(**kwargs)
        self.bind(is_selected=self.updateAttrs)
        self.updateAttrs(self, self.is_selected)

    def updateAttrs(self, inst, val):
        del inst, val
        attrs = self.selected_attrs if self.is_selected else self.deselected_attrs
        for k in attrs:
            setattr(self, k, attrs[k])

    on_selected_attrs = updateAttrs
    on_deselected_attrs = updateAttrs
class XMStyleButton(BoxLayout, ListItemButton):
    widget_style = DictProperty(defaultvalue={})

    def __init__(self, **kwargs):
        fillDictDefs(kwargs, dict(
            orientation="horizontal",
            background_down="",
            background_normal=""))
        cls_dicts = kwargs.pop("cls_dicts", None)
        on_state = kwargs.pop("on_state", None)
        # if "background_disabled_down" not in kwargs: kwargs["background_disabled_down"] = ""
        # if "background_disabled_normal" not in kwargs: kwargs["background_disabled_normal"] = ""
        self.named_widgets = {}
        super(XMStyleButton, self).__init__(**kwargs)
        for cls_dict in cls_dicts:
            cls = cls_dict["cls"]
            kwargs1 = cls_dict["kwargs"]
            wid = cls(**kwargs1)
            if "id" in cls_dict:
                self.named_widgets[cls_dict["id"]] = wid
            self.add_widget(wid)
        if on_state is not None:
            self.bind(state=on_state)
        self.update_style(self, self.widget_style)
        self.bind(widget_style=self.update_style)
    @staticmethod
    def update_style(self, val):
        lstK = list(val)
        if "#" in lstK:
            lstK.remove("#")
            lstK.insert(0, "#")
        for k in lstK:
            cur = val[k]
            if k == "#":
                for wid in self.children:
                    for k0 in cur:
                        setattr(wid, k0, cur[k0])
            elif k.startswith("#"):
                wid = self.get_widget(k[1:])
                for k0 in cur:
                    setattr(wid, k0, cur[k0])
            else:
                setattr(self, k, cur)

    def get_widget(self, name):
        return self.named_widgets[name]


class XMFlagStyleButton(XMStyleButton):
    selected_down_style = DictProperty(defaultvalue={})
    selected_normal_style = DictProperty(defaultvalue={})
    deselected_down_style = DictProperty(defaultvalue={})
    deselected_normal_style = DictProperty(defaultvalue={})
    def __init__(self, **kwargs):
        super(XMFlagStyleButton, self).__init__(**kwargs)
        self.updateStyles(self)
        self.bind(is_selected=self.updateStyles)
        self.bind(state=self.updateStyles)
        self.bind(selected_down_style=self.updateStyles)
        self.bind(selected_normal_style=self.updateStyles)
        self.bind(deselected_down_style=self.updateStyles)
        self.bind(deselected_normal_style=self.updateStyles)
    @staticmethod
    def updateStyles(self, val=None):
        del val
        if self.is_selected:
            if self.state == "down":
                self.widget_style = self.selected_down_style
            else:
                self.widget_style = self.selected_normal_style
        else:
            if self.state == "down":
                self.widget_style = self.deselected_down_style
            else:
                self.widget_style = self.deselected_normal_style

class XMOptListItem1(BoxLayout, XMButton1):
    synch_child_colors = BooleanProperty(True)

    def __init__(self, **kwargs):
        kwargs["orientation"] = "horizontal"
        cls_dicts = kwargs["cls_dicts"]
        on_state = kwargs.get("on_state", None)
        if on_state is not None: del kwargs["on_state"]
        if "background_down" not in kwargs: kwargs["background_down"] = ""
        if "background_normal" not in kwargs: kwargs["background_normal"] = ""
        # if "background_disabled_down" not in kwargs: kwargs["background_disabled_down"] = ""
        # if "background_disabled_normal" not in kwargs: kwargs["background_disabled_normal"] = ""
        super(XMOptListItem1, self).__init__(**kwargs)
        self.colors_by_widname = {}
        self.named_widgets = {}
        for cls_dict in cls_dicts:
            cls = cls_dict["cls"]
            kwargs1 = cls_dict["kwargs"]
            wid = cls(**kwargs1)
            if "id" in cls_dict:
                self.named_widgets[cls_dict["id"]] = wid
                if "colors" in cls_dict:
                    self.colors_by_widname[cls_dict["id"]] = cls_dict["colors"]
            elif "colors" in cls_dict:
                raise ValueError("id must be included if colors is specified")
            self.add_widget(wid)
        if on_state is not None:
            self.bind(state=on_state)

    def on_synch_child_colors(self, obj, val):
        assert obj == self
        if not hasattr(self, "colors_by_widname"): return  # initialization check
        if val:
            for wid in self.children:
                if hasattr(wid, "color"):
                    wid.color = obj.color
                if hasattr(wid, "background_color"):
                    wid.background_color = obj.background_color
            index = int(self.is_selected)
            for k in self.colors_by_widname:
                wid = self.named_widgets[k]
                color, background_color = self.colors_by_widname.get(k, ((None, None), (None, None)))[index]
                if color is not None: wid.color = color
                if background_color is not None: wid.background_color = background_color

    def on_state(self, obj, val):
        super(XMOptListItem1, self).on_state(obj, val)
        self.on_synch_child_colors(self, self.synch_child_colors)

    def get_widget(self, name):
        return self.named_widgets[name]


def sw_on_rect_color(self, rect_color):
    """
    :param SeparatorWidget self:
    :param (int|long|float, int|long|float, int|long|float, int|long|float) rect_color:
    """
    self.canvas_color.rgba = rect_color


def sw_on_rect_width(self, rect_width):
    """
    :param SeparatorWidget self:
    :param int|long|float rect_width:
    """
    cvr = self.canvas_rect
    rsize = (rect_width, cvr.size[1])
    cvr.size = rsize
    sw_on_size_pos_rsize(self, self.size, self.pos, rsize)


def sw_on_rect_height(self, rect_height):
    """
    :param SeparatorWidget self:
    :param int|long|float rect_height:
    """
    cvr = self.canvas_rect
    rsize = (cvr.size[0], rect_height)
    cvr.size = rsize
    sw_on_size_pos_rsize(self, self.size, self.pos, rsize)


def sw_on_rect_x(self, rect_x):
    """
    :param SeparatorWidget self:
    :param int|long|float rect_x:
    """
    cvr = self.canvas_rect
    cvr.pos = rect_x, cvr.pos[1]


def sw_on_rect_y(self, rect_y):
    """
    :param SeparatorWidget self:
    :param int|long|float rect_y:
    """
    cvr = self.canvas_rect
    cvr.pos = cvr.pos[0], rect_y


def sw_on_size_pos_rsize(self, size, pos, rsize):
    """
    :param SeparatorWidget self:
    :param (int|long|float, int|long|float) size:
    :param (int|long|float, int|long|float) pos:
    :param (int|long|float, int|long|float) rsize:
    """
    off_type = self.off_type
    y_is_cy = bool(off_type & SeparatorWidget.OFF_Y_IS_CY)
    x_is_cx = bool(off_type & SeparatorWidget.OFF_X_IS_CX)
    if x_is_cx:
        self.rect_x = rsize[0] / 2.0 + pos[0]
    else:
        self.rect_x = (size[0] - rsize[0]) / 2.0 + pos[0]
    if y_is_cy:
        self.rect_y = rsize[1] / 2.0 + pos[1]
    else:
        self.rect_y = (size[1] - rsize[1]) / 2.0 + pos[1]


def sw_on_size(self, size):
    """
    :param SeparatorWidget self:
    :param (int|long|float, int|long|float) size:
    """
    sw_on_size_pos_rsize(self, size, self.pos, (self.rect_width, self.rect_height))


def sw_on_pos(self, pos):
    """
    :param SeparatorWidget self:
    :param (int|long|float, int|long|float) pos:
    """
    sw_on_size_pos_rsize(self, self.size, pos, (self.rect_width, self.rect_height))


class SeparatorWidget(Widget):
    OFF_Y_IS_CY = 1
    OFF_X_IS_CX = 2
    # noinspection PyArgumentList
    rect_color = ListProperty([1] * 4)
    rect_width = NumericProperty(0)
    rect_height = NumericProperty(0)
    rect_x = NumericProperty(0)
    rect_y = NumericProperty(0)
    canvas_color = None
    canvas_rect = None
    def __init__(self, **kwargs):
        self.off_type = SeparatorWidget.OFF_Y_IS_CY
        super(SeparatorWidget, self).__init__(**kwargs)
        with self.canvas:
            self.canvas_color = Color(rgba=self.rect_color)
            self.canvas_rect = Rectangle(
                pos=(self.rect_x, self.rect_y),
                size=(self.rect_width, self.rect_height))
        self.bind(
            rect_color=sw_on_rect_color,
            rect_width=sw_on_rect_width,
            rect_height=sw_on_rect_height,
            rect_x=sw_on_rect_x,
            rect_y=sw_on_rect_y,
            size=sw_on_size,
            pos=sw_on_pos
        )


# TODO: finish Popup with close to look like the following
"""
+------------------------------------+
|[---------title---------][close_btn]|
+=========widget "separator"=========+
|                                    |
|            content area            |
|        usually float layout        |
|                                    |
+------------------------------------+
"""
import kivy.metrics

class DoTheBind(object):
    def __init__(self, lstAttrObj, fn):
        """
        :param list[(Widget, list[str|unicode])] lstAttrObj:
        :param fn:
        """
        self.lstAttrObj = lstAttrObj
        self.fn = fn
        nAttr = 0
        for obj, lstAttr in lstAttrObj:
            obj.bind(**dict.fromkeys(lstAttr, self))
            nAttr += len(lstAttr)
        self.nAttr = nAttr
    def __call__(self, obj, v):
        args = [None] * self.nAttr
        c = 0
        for obj, lstAttr in self.lstAttrObj:
            for attr in lstAttr:
                args[c] = getattr(obj, attr)
                c += 1
        return self.fn(self, *args)


class PopupWithClose(ModalView):
    title = StringProperty('No title')
    title_size = NumericProperty('14sp')
    title_align = OptionProperty('left', options=['left', 'center', 'right', 'justify'])
    title_font = StringProperty(DEFAULT_FONT)
    content = ObjectProperty(None)
    # noinspection PyArgumentList
    title_color = ListProperty([1, 1, 1, 1])
    # noinspection PyArgumentList
    separator_color = ListProperty(list(kStrRGB("2FA7D4")))
    separator_height = NumericProperty('2dp')
    _container = ObjectProperty(None)

    def on_title(self, instance, title):
        del instance
        self.top_label.text = title

    def on_title_color(self, instance, title_color):
        del instance
        self.top_label.color = title_color

    def on_title_size(self, instance, title_size):
        del instance
        self.top_label.font_size = title_size

    def on_title_font(self, instance, title_font):
        del instance
        self.top_label.font_name = title_font

    def on_title_align(self, instance, title_align):
        del instance
        self.top_label.halign = title_align

    def on_separator_color(self, instance, separator_color):
        del instance
        self.separator_widget.rect_color = separator_color

    def on_separator_height(self, instance, separator_height):
        del instance
        self.separator_widget.rect_height = separator_height

    def __init__(self, **kwargs):
        self.main_layout = GridLayout(
            cols=1,
            size_hint=(None, None),
            padding=kivy.metrics.dp(12))
        self.top_layout = BoxLayout(
            orientation="horizontal",
            size_hint_x=1.0,
            size_hint_y=None
        )
        self.top_label = Label(
            size_hint=(0.8,1.0),
        )
        self.close_btn = XMLoneButton(
            on_release=self.dismiss, text="Cancel",
            size_hint=(0.2, 1.0), color=kWHITE,
            background_normal_color=k2_A_RED,
            background_down_color=k1_8_RED)
        self.top_layout.add_widget(self.top_label)
        self.top_layout.add_widget(self.close_btn)
        def fn1(obj, texture_size):
            obj.parent.height = texture_size[1] + kivy.metrics.dp(16)
        def fn2(obj, width):
            obj.text_size = width - kivy.metrics.dp(16), None
        self.top_label.bind(
            texture_size=fn1,
            width=fn2)
        self.separator_widget = SeparatorWidget(
            height=kivy.metrics.dp(4),
            size_hint_y=None,
            size_hint_x=1.0
        )
        def fn3(obj, width):
            obj.rect_width = width
        self.separator_widget.bind(width=fn3)
        self.main_layout.add_widget(self.top_layout)
        self.main_layout.add_widget(self.separator_widget)
        container = BoxLayout()
        self.main_layout.add_widget(container)
        set_rect_height = "separator_height" not in kwargs
        super(PopupWithClose, self).__init__(**kwargs)
        self.add_widget(self.main_layout)
        self._container = container
        if set_rect_height:
            self.on_separator_height(self, self.separator_height)

    def on_pos(self, instance, pos):
        del instance
        self.main_layout.pos = pos

    def on_size(self, instance, size):
        self.main_layout.size = size

    def add_widget(self, widget, index=0, canvas=None):
        if self._container:
            if self.content:
                raise ValueError(
                    'Popup can have only one widget as content')
            self.content = widget
        else:
            super(PopupWithClose, self).add_widget(widget, index)

    def on_content(self, instance, value):
        del instance
        if self._container:
            self._container.clear_widgets()
            self._container.add_widget(value)

    def on__container(self, instance, value):
        del instance
        if value is None or self.content is None:
            return
        self._container.clear_widgets()
        self._container.add_widget(self.content)

    def on_touch_down(self, touch):
        if self.disabled and self.collide_point(*touch.pos):
            return True
        return super(PopupWithClose, self).on_touch_down(touch)



# Widget that contains checkboxes that are checked to add up to a total that is
#  displayed at the bottom of the widget
class CheckBits(Bordered):
    """
    :type lstChkBox: list[CheckBox]
    :type lstChkLbl: list[Label]
    :type lstVals: list[int|long|float]
    :type prepend: str|unicode
    :type append: str|unicode
    :type theGrid: GridLayout
    :type onChg: (CheckBits, int|long) -> any
    :type topLbl: Label
    :type totalLbl: Label
    :type lstBits: list[int]
    :type myWidth: int|long
    :type lstBlank: list[(Label, Label)]
    """
    # onChg callback is called as onChg(self, index)
    def __init__(self, app, lbl, loadDat, lstVals, onChg, prepend, append="", **kwargs):
        """
        :param app:
        :param str|unicode lbl:
        :param list[int|long] loadDat:
        :param list[int|long|float] lstVals:
        :param (CheckBits, int|long) -> any onChg:
        :param str|unicode prepend:
        :param str|unicode append:
        :param kwargs:
        """
        super(CheckBits, self).__init__(**kwargs)
        self.app = app
        self.theGrid = GridLayout(
            cols=2, rows=5, pos_hint={'x': 0.0, "top": .95}, size_hint=(1.0, .89))
        theGrid = self.theGrid
        self.onChg = onChg
        self.lstVals = lstVals
        lstChkBox = [None] * len(lstVals)
        lstChkLbl = [None] * len(lstVals)
        self.lstBits = list(loadDat)
        self.append = append
        self.prepend = prepend
        self.topLbl = Label(text=lbl, pos_hint={"top": .99, 'x': 0.0}, size_hint_y=.05)
        self.add_widget(self.topLbl)
        for c in xrange(len(lstVals)):
            onChgChk = self.mkChgFunc(c)
            lstChkBox[c] = CheckBox(state=toState[loadDat[c]])
            lstChkBox[c].bind(state=onChgChk)
            theGrid.add_widget(lstChkBox[c])
            lstChkLbl[c] = Label(
                text="+" + str(self.lstVals[c]), color=stRGB[self.lstBits[c]])
            theGrid.add_widget(lstChkLbl[c])
        self.lstChkBox = lstChkBox
        self.lstChkLbl = lstChkLbl
        self.lstBlank = list()
        lstBlank = self.lstBlank
        for c in xrange(len(lstVals), 5):
            lstBlank.append((Label(text=""), Label(text="")))
            theGrid.add_widget(lstBlank[-1][0])
            theGrid.add_widget(lstBlank[-1][1])
        maxNum = sum(lstVals)
        self.myWidth = len(str(maxNum)) + len(prepend) + len(append)
        w = self.myWidth * 8
        self.totalLbl = Label(
            text="", size_hint_y=.05, pos_hint={'y': 0.01, 'x': 0.01},
            size_hint_x=None, width=w, halign="left",
            text_size=(w, None))
        total = sum([self.lstVals[c] for c in xrange(len(self.lstVals)) if self.lstBits[c]])
        self.totalLbl.text = self.prepend + str(total) + self.append
        self.add_widget(self.totalLbl)
        self.add_widget(theGrid)

    def mkChgFunc(self, num):
        def onChgChk(inst, val):
            del inst
            self.lstBits[num] = fromState[val]
            total = sum([self.lstVals[c] for c in xrange(len(self.lstVals)) if self.lstBits[c]])
            self.totalLbl.text = self.prepend + str(total) + self.append
            self.lstChkLbl[num].color = stRGB[self.lstBits[num]]
            if self.onChg is not None:
                try:
                    self.onChg(self, num)
                except StandardError:
                    sys.stderr.write(traceback.format_exc())
                    sys.stderr.flush()

        return onChgChk


# calls function with name funcName on the super class only if the super class
#  has that function otherwise returns without error
def callSuper(cls, self, funcName, *args, **kwargs):
    Super = super(cls, self)
    if hasattr(Super, funcName):
        return True, getattr(Super, funcName)(*args, **kwargs)
    else:
        return False, None


# widget that is a error message bubble that shows up when its designated parent
#  is touched while it is disabled
class TouchDownMsg(Bubble):
    defBubble = dict(
        background_image=images.bubbleBG, arrow_image=images.bubbleArrow,
        background_color=(1.75, 1.75, 1.75, 1.25), size_hint=(None, None))
    text = StringProperty("")
    allShown = set()

    def __init__(self, **kwargs):
        fillDictDefs(kwargs, self.defBubble)
        self.lbl = Label(text="", color=(0., 0., 0., 1.))
        self.lbl.disabled_color = self.lbl.color
        super(TouchDownMsg, self).__init__(**kwargs)
        self.text = kwargs["text"]
        self.add_widget(self.lbl)
        self.exclusive = True
        self.disabled_background_color = self.background_color
        self.prevParent = None

    def parentOnDisable(self, obj, disabled):
        del obj
        if not disabled:
            self.parent.remove_widget(self)

    def on_text(self, obj, text):
        del obj
        self.lbl.text = text
        lines = text.split("\n")
        self.width = max(map(len, lines)) * 7.68
        # self.height = len(lines) * 12

    def on_size(self, obj, size):
        if self.parent is not None:
            inst = self.parent
            self.pos = (
                .5 * (inst.size[0] - size[0]) + inst.pos[0],
                inst.size[1] + inst.pos[1])
        callSuper(TouchDownMsg, self, "on_size", obj, size)

    def on_parent(self, obj, parent):
        callSuper(TouchDownMsg, self, "on_parent", obj, parent)
        if parent is None:
            self.allShown.discard(self)
            self.prevParent.unbind(disabled=self.parentOnDisable)
            self.prevParent = None
            return
        else:
            if self.exclusive or any(map(lambda x: x.exclusive, self.allShown)):
                while len(self.allShown) > 0:
                    inst = self.allShown.pop()
                    if inst.parent is not None: inst.parent.remove_widget(inst)
            self.allShown.add(self)
        inst = self.parent
        self.pos = (
            .5 * (inst.size[0] - self.size[0]) + inst.pos[0], inst.size[1] + inst.pos[1])
        self.prevParent = self.parent
        self.parent.bind(disabled=self.parentOnDisable)

    def on_touch_down(self, touch):
        if not self.collide_point(*touch.pos):
            if self.parent is not None: self.parent.remove_widget(self)
            # log.Log("YES: " + repr(touch.pos), 3)
        #else: log.Log("NO: " + repr(self.text), 3)
        callSuper(TouchDownMsg, self, "on_touch_down", touch)
        return False


class ErrBubble(Bubble):
    defBubble = dict(
        background_image=images.bubbleBG, arrow_image=images.bubbleArrow,
        background_color=(1.75, 1.75, 1.75, 1.25), size_hint=(None, None))
    text = StringProperty("")
    side_pos = OptionProperty("middle", options=["left", "middle", "right"])
    side_anchor = OptionProperty("bottom", options=["bottom", "top", "left", "right"])

    def __init__(self, **kwargs):
        fillDictDefs(kwargs, self.defBubble)
        self.lbl = Label(text="", color=(0., 0., 0., 1.))
        self.lbl.disabled_color = self.lbl.color
        self.num_ignore_touch = kwargs.pop("num_ignore_touch", 0)
        super(ErrBubble, self).__init__(**kwargs)
        self.text = kwargs["text"]
        self.add_widget(self.lbl)
        self.disabled_background_color = self.background_color
        self.prevParent = None
        self.unbindLst = None

    def addUnbindAtRemove(self, obj, k, fn):
        if self.unbindLst is None:
            self.unbindLst = [(obj, k, fn)]
        else:
            self.unbindLst.append((obj, k, fn))

    def do_positioning(self):
        inst = self.parent
        if inst is None:
            return
        x0, y0 = self.pos
        w0, h0 = self.size
        x1, y1 = inst.pos
        w1, h1 = inst.size
        side_pos = (x0, x1, w0, w1)
        vert_pos = (y0, y1, h0, h1)
        if self.side_anchor in ["left", "right"]:
            side_pos = (y0, y1, h0, h1)
            vert_pos = (x0, x1, w0, w1)
        x0_o, x1_o, w0_o, w1_o = side_pos
        y0_o, y1_o, h0_o, h1_o = vert_pos
        z = 1
        #-----self-----
        #   --inst--
        # goal place inst.center_x at self.x + self.w * z
        # self.x = inst.x + inst.w / 2 - self.w * z
        if self.side_pos == "left":
            z = 1./6
        elif self.side_pos == "middle":
            z = .5
        elif self.side_pos == "right":
            z = 5./6
        if self.side_anchor in ["left", "right"]:
            z = 1 - z
        x0_o = x1_o + .5 * w1_o - w0_o * z
        y0_o = y1_o + (h1_o if self.side_anchor in ["bottom", "left"] else -h0_o)
        if self.side_anchor in ["left", "right"]:
            x0, y0 = y0_o, x0_o
        else:
            x0, y0 = x0_o, y0_o
        self.pos = x0, y0
        side_pos = self.side_pos
        side_anchor = self.side_anchor
        if side_anchor in ["left", "right"]:
            if side_pos == "left":
                side_pos = "top"
            elif side_pos == "right":
                side_pos = "bottom"
        if side_pos == "middle":
            side_pos = "mid"
        self.arrow_pos = side_anchor + "_" + side_pos

    def on_side_anchor(self, obj, anchor):
        del obj, anchor
        self.do_positioning()

    def on_side_pos(self, obj, pos):
        del obj, pos
        self.do_positioning()

    def on_text(self, obj, text):
        del obj
        self.lbl.text = text
        lines = text.split("\n")
        self.width = max(map(len, lines)) * 7.68

    def on_parent(self, obj, parent):
        callSuper(ErrBubble, self, "on_parent", obj, parent)
        if parent is None:
            self.prevParent = None
            if self.unbindLst is not None:
                for obj, k, fn in self.unbindLst:
                    obj.funbind(k, fn)
            return
        #inst = self.parent
        #self.pos = (.5 * (inst.size[0] - self.size[0]) + inst.pos[0], inst.size[1] + inst.pos[1])
        self.do_positioning()
        self.prevParent = self.parent

    def on_size(self, obj, size):
        if self.parent is not None:
            # inst = self.parent
            self.do_positioning()
        callSuper(ErrBubble, self, "on_size", obj, size)

    def on_touch_down(self, touch):
        if not (self.collide_point(*touch.pos) or self.parent is None):
            if self.num_ignore_touch > 0:
                self.num_ignore_touch -= 1
            else:
                self.parent.remove_widget(self)
        callSuper(ErrBubble, self, "on_touch_down", touch)
        return False


# used for the widget.disabled==True touch info
def touchDownBubble(parent, obj, touch):
    if touch is None or not obj.collide_point(touch.x, touch.y):
        parent.remove_widget(obj)


def touchDownMsg(disableMsg, inst, touch):
    if inst.disabled and inst.collide_point(touch.x, touch.y):
        lines = disableMsg.split("\n")
        maxW = int(7.68 * max(map(len, lines)))
        # maxH = int(12 * len(lines))  # ,
        #    pos_hint={'center_x':0.5, 'y':1.0}
        bubble = Bubble(
            background_image=images.bubbleBG, arrow_image=images.bubbleArrow,
            width=maxW, size_hint=(None, .01),
            background_color=(1.5, 1.5, 1.5, 1.25),
            pos=(.5 * (inst.size[0] - maxW) + inst.pos[0],
                 inst.size[1] + inst.pos[1]))
        bubble.bind(on_touch_down=Bind(touchDownBubble, inst))
        bubble.add_widget(Label(text=disableMsg, disabled_color=(0., 0., 0., 1.)))
        inst.add_widget(bubble)


def tchDownBubble(bubble, inst, touch, after=False):
    del after
    oldParent = bubble.parent
    if oldParent is not None: oldParent.remove_widget(bubble)
    if inst.disabled and inst.collide_point(*touch.pos):
        if oldParent != inst: inst.add_widget(bubble)
    return False


def radioBtnChg(theObj, index, inst, val):
    del inst
    val = fromState[val]
    theObj.opts[index][1].color = stRGB[val]
    if not val: return
    theObj.val = index
    if hasattr(theObj, "chgVal"): theObj.chgVal(index)


def bitsToTotal(theObj):
    curVal = theObj.lsbVal
    rtn = 0
    for bit in theObj.data:
        rtn += bit * curVal
        curVal *= 2
    theStr = theObj.prepend.replace("%", "%%") + "%.4g"
    theStr += theObj.append.replace("%", "%%")
    return theStr % rtn


def chkBoxChg(theObj, index, inst, val):
    del inst
    val = fromState[val]
    theObj.data[index] = int(val)
    theObj.widgets[index][1].color = stRGB[val]
    theObj.widgets[-1].text = bitsToTotal(theObj)
    theObj.onChgAny(theObj)


def togChkBox(chkBox, inst):
    del inst
    if chkBox.group is None:
        chkBox.state = toState[not fromState[chkBox.state]]
        return
    if not chkBox.allow_no_selection and fromState[chkBox.state]: return
    lstWid = ToggleButton.get_widgets(chkBox.group)
    if len(lstWid) == 0:
        chkBox.state = toState[not fromState[chkBox.state]]
        del lstWid
        return
    elif len(lstWid) == 1:
        val = not fromState[chkBox.state]
        if not chkBox.allow_no_selection: val = True
        chkBox.state = toState[val]
        del lstWid
        return
    val = fromState[chkBox.state]
    val0 = toState[False]
    for wid in lstWid: wid.state = val0
    chkBox.state = toState[not val]


class ChkBoxListStorage(object):
    """
    :type layout: Layout
    :type widgets: list[None|(CheckBox, Label)|Label]
    :type lsbVal: float|int|long
    :type data: list[int]
    :type prepend: str|unicode
    :type append: str|unicode
    :type gLayout: GridLayout
    :type onChgAny: (ChkBoxListStorage) -> any
    """
    __slots__ = ["layout", "widgets", "lsbVal", "data", "prepend", "append", "gLayout", "onChgAny"]
    def __init__(self, dctAttr):
        """
        :param dict dctAttr:
        """
        self.layout = dctAttr["layout"]
        self.widgets = dctAttr["widgets"]
        self.lsbVal = dctAttr["lsbVal"]
        self.data = dctAttr["data"]
        self.prepend = dctAttr["prepend"]
        self.append = dctAttr["append"]
        self.gLayout = dctAttr["gLayout"]
        self.onChgAny = dctAttr["onChgAny"]



def mkChkBoxList(kwargsLayout, numBox, lsbVal, data, prepend="", append="", title=""):
    widgets = [None] * (numBox + 1)
    layout = Bordered(**kwargsLayout)
    topSz = (len(title) * .03, .05)
    layout.add_widget(Label(
        size_hint=topSz, text=title, pos_hint={'x': .5 - topSz[0] / 2, "top": .98}))
    gLayout = GridLayout(cols=4, rows=(numBox + 1) / 2 + 1, size_hint=(1.0, .83),
                         pos_hint={'x': 0.0, 'top': .93})
    theObj = ChkBoxListStorage({
        "layout": layout,
        "widgets": widgets,
        "lsbVal": lsbVal,
        "data": data,
        "prepend": prepend,
        "append": append,
        "gLayout": gLayout,
        "onChgAny": lambda theObj1: None})
    curVal = lsbVal
    for c in xrange(numBox):
        curCol = stRGB[data[c]]
        box, lbl = (
            CheckBox(state=toState[data[c]], size_hint_x=None, width=40),
            sizedCls(
                ClkLabel, text="+%.4g" % curVal, size_hint=(1.0, 1.), color=curCol))
        lbl.bind(on_press=Bind(togChkBox, box))
        gLayout.add_widget(box)
        gLayout.add_widget(lbl)
        widgets[c] = box, lbl
        box.bind(state=Bind(chkBoxChg, theObj, c))
        curVal *= 2
    widgets[-1] = Label(text=bitsToTotal(theObj), **defTxtStyle)
    layout.add_widget(gLayout)
    layout.add_widget(widgets[-1])
    lbl = widgets[-1]
    szX = .01 * len(lbl.text)
    szY = .05
    lbl.pos_hint = {'x': .5 - szX / 2, 'y': 0.025}
    lbl.size_hint = (szX, szY)
    return theObj


def setTextSz(lbl, val):
    lbl.text_size = val


defSized = {"halign": "left", "valign": "middle", "size_hint": (None, None)}


def sizedCls(cls, **kwargs):
    obj = cls(**dict(defSized, **kwargs))
    obj.texture_update()
    obj.text_size = obj.texture_size
    obj.size = obj.texture_size
    obj.bind(size=setTextSz)
    return obj


def sizedCls1(cls, **kwargs):
    obj = cls(**dict(defSized, **kwargs))
    obj.bind(size=setTextSz)
    return obj


def setLblSz(lbl, val):
    lbl.size = val


def sizedLabel(**kwargs): return sizedCls(Label, **kwargs)


def sizedListItemLabel(**kwargs): return sizedCls(ListItemLabel, **kwargs)


def sizedButton(**kwargs): return sizedCls(Button, **kwargs)


# makes a bordered layout with a label at the top of it
def mkLblBorder(clsLabel, clsInner, kwargsOuter, kwargsLabel, kwargsInner):
    layout = Bordered(**kwargsOuter)
    layout.add_widget(clsLabel(**kwargsLabel))
    inner = clsInner(**kwargsInner)
    layout.add_widget(inner)
    return layout, inner


def mkErrorBubble(popup, errMsg):
    bubble = Bubble(
        background_image=images.bubbleBG,
        arrow_image=images.bubbleArrow,
        background_color=(1., .75, .75, 1.25),
        pos_hint={'center_x': 0.5, 'y': 1.2},
        size_hint=(None, .2), width=int(7.68 * len(errMsg)))
    bubble.add_widget(Label(text=errMsg))
    popup.content.add_widget(bubble)
    fnTouch = Bind(touchDownBubble, popup.content)
    bubble.bind(on_touch_down=fnTouch)

    def fnOnDismiss(obj):
        obj.unbind(on_dismiss=fnOnDismiss)
        fnTouch(bubble, None)

    popup.bind(on_dismiss=fnOnDismiss)

class SnappedTabbedPanel(SnappedWidget, TabbedPanel): pass