from .args_converters import *

from kivy.adapters.listadapter import ListAdapter
from kivy.uix.popup import Popup
from kivy.uix.textinput import TextInput
from kivy.uix.scrollview import ScrollView
from . import globalVars
import sys


minLenResize = 22


def updTxtPopup(useLst, txtIn, adapter, val=None):
    del val
    if len(adapter.selection) == 0:
        txtIn.text = "NONE"
    try:
        txtIn.text = useLst[adapter.selection[0].index][0]
    except StandardError:
        txtIn.text = "Select"


# command oriented text manipulation function
# commands are specified by characters in the text with a value less than 32
#  and indicate what to do with the existing text
# other characters are just appended to txtIn.text
# commands:
#  '\0': clear
#  '\b': backspace
#  '\1': at beginning of strAdd,
#    toggles whether strAdd is present at the beginning
#      example: minus sign button on integer numpad
def addTxt(txtIn, strAdd, obj):
    del obj
    newTxt = txtIn.text
    if strAdd[0] == '\1':
        if newTxt[0:len(strAdd) - 1] == strAdd[1:]:
            newTxt = newTxt[len(strAdd) - 1:]
        else:
            newTxt = strAdd[1:] + newTxt
        txtIn.text = newTxt
        return
    for ch in strAdd:
        if ch == '\b':
            newTxt = newTxt[:-1]
        elif ch == '\0':
            newTxt = ""
        else:
            newTxt += ch
    txtIn.text = newTxt


defTxtIn = dict(
    pos_hint={"top": 1.0, 'x': 0.0}, size_hint_y=.2, readonly=True, font_size=36,
    use_handles=False, use_bubble=False)
defDone = dict(text="Done", pos_hint={'y': 0.0, 'x': 0.0}, size_hint_y=.15)
defPopup = dict(title="", title_size="18sp", title_align="center",
                separator_color=kD0_GREY, size_hint=(.4, .7))


# makes a popup box usually used for numpads or list selection
def mkPopup(getContent, kwargsPopup=None, kwargsTxtIn=None, addContent=(), kwargsDone=None):
    if kwargsDone is None:
        kwargsDone = defDone
    else:
        kwargsDone = dict(defDone, **kwargsDone)
    if kwargsTxtIn is None:
        kwargsTxtIn = defTxtIn
    else:
        kwargsTxtIn = dict(defTxtIn, **kwargsTxtIn)
    if kwargsPopup is None:
        kwargsPopup = defPopup
    else:
        kwargsPopup = dict(defPopup, **kwargsPopup)
    done = Button(**kwargsDone)
    txtIn = TextInput(**kwargsTxtIn)
    popup = Popup(**kwargsPopup)
    content = FloatLayout(size_hint=(1.0, 1.0), pos_hint={'x': 0.0, "top": 1.0})
    content.add_widget(txtIn)
    content.add_widget(done)
    popup.content = content
    calcTop = txtIn.pos_hint.get('y', None)
    if calcTop is None:
        calcTop = txtIn.pos_hint.get("top", None)
        if calcTop is None:
            calcTop = .8
        else:
            calcTop -= txtIn.size_hint_y
    doneTop = done.pos_hint.get("top", None)
    if doneTop is None:
        doneTop = done.pos_hint.get('y', None)
        if doneTop is None:
            doneTop = .1
        else:
            doneTop += done.size_hint_y
    calcSz = (1.0, calcTop - doneTop)
    kwargsInner = {"size_hint": calcSz, "pos_hint": {'x': 0.0, "top": calcTop}}
    for part in addContent:
        content.add_widget(part)

    def on_dismiss(inst):
        for part1 in addContent:
            inst.content.remove_widget(part1)

    content.add_widget(getContent(done, txtIn, popup, kwargsInner))
    popup.bind(on_dismiss=on_dismiss)
    return popup


class PopupMaker(object):
    def __init__(self, kwargsPopup=None, kwargsTxtIn=None, addContent=(), kwargsDone=None):
        if kwargsDone is None:
            kwargsDone = defDone
        else:
            kwargsDone = dict(defDone, **kwargsDone)
        if kwargsTxtIn is None:
            kwargsTxtIn = defTxtIn
        else:
            kwargsTxtIn = dict(defTxtIn, **kwargsTxtIn)
        if kwargsPopup is None:
            kwargsPopup = defPopup
        else:
            kwargsPopup = dict(defPopup, **kwargsPopup)
        popup = Popup(**kwargsPopup)
        done = Button(**kwargsDone)
        txtIn = TextInput(**kwargsTxtIn)
        self.txtIn = txtIn
        content = FloatLayout(size_hint=(1.0, 1.0), pos_hint={'x': 0.0, "top": 1.0})
        content.add_widget(txtIn)
        content.add_widget(done)
        popup.content = content
        calcTop = txtIn.pos_hint.get('y', None)
        if calcTop is None:
            calcTop = txtIn.pos_hint.get("top", None)
            if calcTop is None:
                calcTop = .8
            else:
                calcTop -= txtIn.size_hint_y
        doneTop = done.pos_hint.get("top", None)
        if doneTop is None:
            doneTop = done.pos_hint.get('y', None)
            if doneTop is None:
                doneTop = .1
            else:
                doneTop += done.size_hint_y
        calcSz = (1.0, calcTop - doneTop)
        kwargsInner = {"size_hint": calcSz, "pos_hint": {'x': 0.0, "top": calcTop}}
        for part in addContent: content.add_widget(part)
        content.add_widget(self.getContent(done, txtIn, popup, kwargsInner))
        self.popup = popup

    def getContent(self, done, txtIn, popup, kwargsInner):
        return Label(text="UNNAMED POPUP", **kwargsInner)

    def showDialog(self, func, val=None, title=""):
        self.popup.title = title
        self.popup.open()


# new class for SelectPopupMaker and eventually get_enterSetLst (but not yet)
# support for this is possibly limited. If you want to use this check and add support for it in the code
# confirmed support in the following files
#   part_HMC_fDiv.py
class OrderedOptMap(object):
    """
    :type obj: list[(str|unicode,int|long)]|dict[int|long,str|unicode]
    :type ordering: list[int|long]|None
    :type useLst: list[(str|unicode,int|long)]
    """
    def __init__(self, obj, ordering=None):
        """
        :param list[(str|unicode,int|long)]|dict[int|long,str|unicode] obj:
        :param list[int|long]|None ordering:
        """
        self.obj = obj
        self.ordering = ordering
        if isinstance(obj, dict):
            assert ordering is not None, "'ordering' must be specified since it is not with dict 'obj' alone"
            for k in obj:
                assert isinstance(k, (int, long)),\
                    "each option key in 'obj' must be an integer, got %r" % k
                assert isinstance(obj[k], (str, unicode)),\
                    "each option value in 'obj' must be str or unicode, got %r" % obj[k]
            self.useLst = list(map(lambda k1: (obj[k1], k1), ordering))
        elif isinstance(obj, list):
            assert ordering is None, "ordering cannot be specified if implicitly done so with a useLst in 'obj'"
            for elem in obj:
                assert isinstance(elem, (tuple, list)),\
                    "each element of useLst 'obj' must be a 2-tuple (txt, optNum), got %r" % elem
            self.useLst = obj
        # prev = self.useLst[0][1]
        if self.ordering is None:
            ordering = [self.useLst[0][1]] * len(self.useLst)
            calcOrdering = False
            for c in xrange(1, len(self.useLst)):
                ordering[c] = self.useLst[c][1]
                if ordering[c] < ordering[c - 1]:
                    calcOrdering = True
            if calcOrdering:
                self.ordering = ordering

class SelectPopupMaker(PopupMaker):
    def __init__(self, lstOpts, defVal, kwargsPopup=None, kwargsTxtIn=None, addContent=(), kwargsDone=None):
        useLst = None
        self.orderMap = None
        if isinstance(lstOpts, (tuple, list)):
            useLst = [None] * reduce(lambda total, x: total + int(x[0]), lstOpts, 0)
            cur = 0
            for c in xrange(len(useLst)):
                while not lstOpts[cur][0]: cur += 1
                useLst[c] = (lstOpts[cur][1], cur)
                cur += 1
        elif isinstance(lstOpts, dict):
            iterLst = sorted(lstOpts)
            useLst = [(lstOpts[k], k) for k in iterLst]
        elif isinstance(lstOpts, OrderedOptMap):
            useLst = lstOpts.useLst
            self.orderMap = {x: c for c, x in enumerate(lstOpts.ordering)}
        if useLst is None: raise TypeError("lstOpts must be instance of dict, tuple or list")
        self.useLst = useLst
        self.defVal = defVal
        self.func = None
        self.allowEmptySel = False
        self.adapter = None
        super(SelectPopupMaker, self).__init__(kwargsPopup, kwargsTxtIn, addContent, kwargsDone)

    def getIndexFromVal(self, val):
        curSel = -1 if self.allowEmptySel else 0
        if val >= 0:
            Tmp = len(self.useLst)
            if self.orderMap is None:
                Tmp = BinaryApprox1(DynamicMap(lambda x: x[1], self.useLst), val)
            else:
                Tmp = self.orderMap.get(val, Tmp)
            if Tmp < len(self.useLst):
                curSel = Tmp
        return curSel

    def showDialog(self, func, val=None, title=""):
        if val is None: val = self.defVal
        curSel = self.getIndexFromVal(val)
        adapter = self.adapter
        self.func = func
        map(adapter.deselect_item_view, adapter.selection)
        if curSel != -1:
            adapter.select_item_view(adapter.get_view(curSel))
        adapter.allow_empty_selection = self.allowEmptySel
        super(SelectPopupMaker, self).showDialog(func, val, title)
        updTxtPopup(self.useLst, self.txtIn, adapter, val)

    def onDone(self, obj):
        if callable(self.func):
            self.func(self.useLst, self.adapter, self.popup, obj)

    def getContent(self, done, txtIn, popup, kwargsContent):
        maxLen = max(map(lambda x: len(x[0]), self.useLst))
        adapter = ListAdapter(
            cls=XMButton1, allow_empty_selection=self.allowEmptySel, selection_mode="single",
            data=self.useLst, args_converter=convLstOpts)
        txtIn.font_size = 24
        lst = ListView(adapter=adapter, **kwargsContent)
        map(adapter.deselect_item_view, adapter.selection)
        scroll = filter(lambda x: isinstance(x, ScrollView), lst.children)[0]
        assert isinstance(scroll, ScrollView)
        scroll.bind(on_scroll_start=ScrollAllower)
        # scroll.scroll_timeout = int(scroll.scroll_timeout * 30 / 48.)
        updTxtPopup(self.useLst, txtIn, adapter)
        # noinspection PyArgumentList
        adapter.bind(on_selection_change=Bind(updTxtPopup, self.useLst, txtIn))
        self.adapter = adapter
        maxLen -= minLenResize
        done.bind(on_release=self.onDone)
        if maxLen > 0:
            szX = popup.size_hint_x
            szX += .015 * maxLen
            popup.size_hint_x = szX
        return lst


# widget mixin class that will make a popup when pressed
class ClickPopup(object):
    kwargsPopup = dict(
        title="", title_size="18sp", title_align="center",
        separator_color=kD0_GREY, size_hint=(.4, .7))

    def __init__(self, **kwargs):
        assert isinstance(self, Widget)
        if "layout" in kwargs:
            self.layout = kwargs["layout"]
            del kwargs["layout"]
        super(ClickPopup, self).__init__(**kwargs)

    def on_release(self):
        assert isinstance(self, Widget)
        super(ClickPopup, self).on_release()
        popup = Popup(**self.kwargsPopup)
        curLayout = self.layout({"clkSelf": self, "popup": popup})
        popup.content = curLayout.layoutWidget


numpadDone = dict(pos_hint={'y': 0.0, 'x': .5}, size_hint_x=.5)


def getNumpad(done, txtIn, popup, kwargsContent):
    del done
    grid = GridLayout(cols=3, **kwargsContent)
    for txt in map(str, xrange(1, 10)) + [".", "0"]:
        grid.add_widget(Button(text=txt, on_release=Bind(addTxt, txtIn, txt)))
    grid.add_widget(Button(text="Backspace", on_release=Bind(addTxt, txtIn, "\b")))
    popup.size_hint_x = .8
    # grid.add_widget(Button(text="Clear",on_release=Bind(addTxt,txtIn,"\0")))
    return grid


def getNumpad1(done, txtIn, popup, kwargsContent):
    del done, popup
    grid = GridLayout(cols=3, **kwargsContent)
    for txt in map(str, xrange(1, 10)) + ["-", "0"]:
        txtCmd = txt if txt != "-" else "\1" + txt
        grid.add_widget(Button(text=txt, on_release=Bind(addTxt, txtIn, txtCmd)))
    grid.add_widget(Button(text="Backspace", on_release=Bind(addTxt, txtIn, "\b")))
    # grid.add_widget(Button(text="Clear",on_release=Bind(addTxt,txtIn,"\0")))
    return grid


def mkNumpad(getContent, kwargsPopup=None, kwargsTxtIn=None, addContent=(), kwargsDone=None):
    clearBtn = Button(text="Clear", pos_hint={'x': .0, 'y': .0}, size_hint_x=.5)

    def myGetContent(done, txtIn, popup, kwargsContent):
        clearBtn.size_hint_y = done.size_hint_y
        clearBtn.bind(on_release=Bind(addTxt, txtIn, "\0"))
        done.pos_hint = {'x': .5, 'y': .0}
        done.size_hint_x = .5
        return getContent(done, txtIn, popup, kwargsContent)

    return mkPopup(
        myGetContent, kwargsPopup, kwargsTxtIn, (clearBtn,) + addContent, kwargsDone)


class NumpadPopupMaker(PopupMaker):  # TODO: finish showDialog (I think it is finished?)
    def __init__(self, defVal, kwargsPopup=None, kwargsTxtIn=None, addContent=(), kwargsDone=None):
        self.clearBtn = Button(text="Clear", pos_hint={'x': .0, 'y': .0}, size_hint_x=.5)
        self.defVal = defVal
        self.func = None
        super(NumpadPopupMaker, self).__init__(
            kwargsPopup, kwargsTxtIn, (self.clearBtn,) + addContent, kwargsDone)

    # call this through super before the rest of getContent
    def getContent(self, done, txtIn, popup, kwargsInner):
        self.clearBtn.size_hint_y = done.size_hint_y
        if done.size_hint_y is None: self.clearBtn.height = done.height
        self.clearBtn.bind(on_release=Bind(addTxt, txtIn, "\0"))
        done.pos_hint = {'x': .5, 'y': .0}
        done.size_hint_x = .5

    def showDialog(self, func, val=None, title=""):
        if val is None: val = self.defVal
        self.func = func
        super(NumpadPopupMaker, self).showDialog(func, val, title)
        self.txtIn.text = str(val)


class NumpadPopupMaker1(PopupMaker):  # TODO: finish showDialog (I think it is finished?)
    def __init__(self, defVal, kwargsPopup=None, kwargsTxtIn=None, addContent=(), kwargsDone=None):
        self.bkspcBtn = Button(text="Backspace", pos_hint={'x': .0, 'y': .0}, size_hint_x=1. / 3)
        self.clearBtn = Button(text="Clear", pos_hint={'x': 1. / 3, 'y': .0}, size_hint_x=1. / 3)
        self.defVal = defVal
        self.func = None
        super(NumpadPopupMaker1, self).__init__(
            kwargsPopup, kwargsTxtIn, (self.clearBtn, self.bkspcBtn) + addContent, kwargsDone)

    # call this through super before the rest of getContent
    def getContent(self, done, txtIn, popup, kwargsInner):
        self.clearBtn.size_hint_y = done.size_hint_y
        self.bkspcBtn.size_hint_y = done.size_hint_y
        if done.size_hint_y is None: self.clearBtn.height = done.height
        self.clearBtn.bind(on_release=Bind(addTxt, txtIn, "\0"))
        self.bkspcBtn.bind(on_release=Bind(addTxt, txtIn, "\b"))
        done.pos_hint = {'x': 2. / 3, 'y': .0}
        done.size_hint_x = 1. / 3

    def showDialog(self, func, val=None, title=""):
        if val is None: val = self.defVal
        self.func = func
        super(NumpadPopupMaker1, self).showDialog(func, val, title)
        self.txtIn.text = str(val)


class IntNumpadPopup(PopupMaker):
    def __init__(self, defVal, kwargsPopup=None, kwargsTxtIn=None, addContent=(), kwargsDone=None):
        self.defVal = defVal
        self.func = None
        super(IntNumpadPopup, self).__init__(kwargsPopup, kwargsTxtIn, addContent, kwargsDone)

    def getContent(self, done, txtIn, popup, kwargsContent):
        grid = GridLayout(cols=3, **kwargsContent)
        for txt in map(str, xrange(1, 10)):
            grid.add_widget(Button(text=txt, on_release=Bind(addTxt, txtIn, txt)))
        grid.add_widget(Button(text="Clear", on_release=Bind(addTxt, txtIn, "\0")))
        grid.add_widget(Button(text="0", on_release=Bind(addTxt, txtIn, "0")))
        grid.add_widget(Button(text="Backspace", on_release=Bind(addTxt, txtIn, "\b")))
        done.bind(on_release=self.onDone)
        return grid

    def showDialog(self, func, val=None, title=""):
        if val is None: val = self.defVal
        self.func = func
        super(IntNumpadPopup, self).showDialog(func, val, title)
        self.txtIn.text = str(val)

    def onDone(self, obj):
        del obj
        rtn = self.func(self.txtIn.text)
        if isinstance(rtn, (str, unicode)):
            if len(rtn):
                mkErrorBubble(self.popup, rtn)
                rtn = False
            else:
                rtn = True
        if rtn:
            self.popup.dismiss()


class FloatNumpadPopup(NumpadPopupMaker):
    def __init__(self, defVal, kwargsPopup=None, kwargsTxtIn=None, addContent=(), kwargsDone=None):
        super(FloatNumpadPopup, self).__init__(defVal, kwargsPopup, kwargsTxtIn, addContent, kwargsDone)

    def getContent(self, done, txtIn, popup, kwargsContent):
        super(FloatNumpadPopup, self).getContent(done, txtIn, popup, kwargsContent)
        grid = GridLayout(cols=3, **kwargsContent)
        for txt in map(str, xrange(1, 10)):
            grid.add_widget(Button(text=txt, on_release=Bind(addTxt, txtIn, txt)))
        grid.add_widget(Button(text=".", on_release=Bind(addTxt, txtIn, ".")))
        grid.add_widget(Button(text="0", on_release=Bind(addTxt, txtIn, "0")))
        grid.add_widget(Button(text="Backspace", on_release=Bind(addTxt, txtIn, "\b")))
        popup.size_hint_x = .8
        done.bind(on_release=self.onDone)
        return grid

    def showDialog(self, func, val=None, title=""):
        if val is None: val = self.defVal
        self.func = func
        super(FloatNumpadPopup, self).showDialog(func, val, title)
        self.txtIn.text = "%.15g" % val

    def onDone(self, obj):
        del obj
        rtn = self.func(self.txtIn.text)
        if isinstance(rtn, (str, unicode)):
            if len(rtn):
                mkErrorBubble(self.popup, rtn)
                rtn = False
            else:
                rtn = True
        if rtn:
            self.popup.dismiss()


class FloatNumpadPopup1(NumpadPopupMaker1):
    def __init__(self, defVal, kwargsPopup=None, kwargsTxtIn=None, addContent=(), kwargsDone=None):
        super(FloatNumpadPopup1, self).__init__(defVal, kwargsPopup, kwargsTxtIn, addContent, kwargsDone)

    def getContent(self, done, txtIn, popup, kwargsContent):
        super(FloatNumpadPopup1, self).getContent(done, txtIn, popup, kwargsContent)
        grid = GridLayout(cols=3, **kwargsContent)
        for txt in map(str, xrange(1, 10)):
            grid.add_widget(Button(text=txt, on_release=Bind(addTxt, txtIn, txt)))
        grid.add_widget(Button(text=".", on_release=Bind(addTxt, txtIn, ".")))
        grid.add_widget(Button(text="0", on_release=Bind(addTxt, txtIn, "0")))
        grid.add_widget(Button(text="-", on_release=Bind(addTxt, txtIn, "\1-")))
        popup.size_hint_x = .8
        done.bind(on_release=self.onDone)
        return grid

    def showDialog(self, func, val=None, title=""):
        if val is None: val = self.defVal
        self.func = func
        super(FloatNumpadPopup1, self).showDialog(func, val, title)
        self.txtIn.text = "%.15g" % val

    def onDone(self, obj):
        del obj
        rtn = self.func(self.txtIn.text)
        if isinstance(rtn, (str, unicode)):
            if len(rtn):
                mkErrorBubble(self.popup, rtn)
                rtn = False
            else:
                rtn = True
        if rtn:
            self.popup.dismiss()


def getIntNumpad(done, txtIn, popup, kwargsContent):
    del done, popup
    grid = GridLayout(cols=3, **kwargsContent)
    for txt in map(str, xrange(1, 10)):
        grid.add_widget(Button(text=txt, on_release=Bind(addTxt, txtIn, txt)))
    grid.add_widget(Button(text="Clear", on_release=Bind(addTxt, txtIn, "\0")))
    grid.add_widget(Button(text="0", on_release=Bind(addTxt, txtIn, "0")))
    grid.add_widget(Button(text="Backspace", on_release=Bind(addTxt, txtIn, "\b")))
    return grid


def getHexNumpad(done, txtIn, popup, kwargsContent):
    del done, popup
    grid = GridLayout(cols=4, **kwargsContent)
    for txt in DIGITS:
        grid.add_widget(Button(text=txt, on_release=Bind(addTxt, txtIn, txt)))
    grid.add_widget(Button(text="Clear", on_release=Bind(addTxt, txtIn, "\0")))
    grid.add_widget(Button(text="Backspace", on_release=Bind(addTxt, txtIn, "\b")))
    return grid


def getTxtSizing(txt, chH=16, chW=8):
    maxChX = max(map(len, txt.split("\n")))
    numLn = txt.count("\n") + 1
    return map(int.__mul__, [maxChX, numLn], [chW, chH])


def getLstContent(lstOpts, curOpt, setFunc, done, txtIn, popup, kwargsContent):
    useLst = None
    curSel = -1
    if isinstance(lstOpts, (tuple, list)):
        useLst = [None] * reduce(lambda total, x: total + int(x[0]), lstOpts, 0)
        cur = 0
        for c in xrange(len(useLst)):
            while not lstOpts[cur][0]: cur += 1
            if curOpt <= cur and curSel == -1: curSel = c
            useLst[c] = (lstOpts[cur][1], cur)
            cur += 1
    elif isinstance(lstOpts, dict):
        iterLst = sorted(lstOpts)
        useLst = [(lstOpts[k], k) for k in iterLst]
        try:
            curSel = iterLst.index(curOpt)
        except ValueError:
            globalVars.log.Log(
                "ERRORS: " + ", ".join(map(str, (useLst, iterLst, curOpt))), 2)
    if useLst is None:
        raise TypeError("lstOpts must be instance of dict, tuple or list")
    maxLen = max(map(lambda x: len(x[0]), useLst))
    adapter = ListAdapter(
        cls=XMButton1, allow_empty_selection=False, selection_mode="single",
        data=useLst, args_converter=convLstOpts)
    map(adapter.deselect_item_view, adapter.selection)
    adapter.select_item_view(adapter.get_view(curSel))
    txtIn.font_size = 24
    lst = ListView(adapter=adapter, **kwargsContent)
    scroll = filter(lambda x: isinstance(x, ScrollView), lst.children)[0]
    assert isinstance(scroll, ScrollView)
    # scroll.scroll_timeout = int(scroll.scroll_timeout * 30 / 48.)
    # scroll.bind(size=lambda obj, *args: setattr(obj, "scroll_distance", 7200./obj.size[1]))
    scroll.bind(on_scroll_start=ScrollAllower)
    updTxtPopup(useLst, txtIn, adapter)
    # noinspection PyArgumentList
    adapter.bind(on_selection_change=Bind(updTxtPopup, useLst, txtIn))
    done.bind(on_release=Bind(setFunc, useLst, adapter, popup))
    maxLen -= minLenResize
    if maxLen > 0:
        szX = popup.size_hint_x
        szX += .015 * maxLen
        popup.size_hint_x = szX
    return lst


def getLstInfContent(lstOpts, lstInfs, curOpt, setFunc, done, txtIn, popup, kwargsContent):
    useLst = None
    curSel = -1
    if isinstance(lstOpts, (tuple, list)):
        useLst = [None] * reduce(lambda total, x: total + int(x[0]), lstOpts, 0)
        cur = 0
        for c in xrange(len(useLst)):
            while not lstOpts[cur][0]: cur += 1
            if curOpt <= cur and curSel == -1: curSel = c
            useLst[c] = (lstOpts[cur][1], cur)
            cur += 1
    elif isinstance(lstOpts, dict):
        iterLst = sorted(lstOpts)
        useLst = [(lstOpts[k], k) for k in iterLst]
        try:
            curSel = iterLst.index(curOpt)
        except ValueError:
            globalVars.log.Log(
                "ERRORS: " + ", ".join(map(str, (useLst, iterLst, curOpt))), 2)
    else:
        raise TypeError("lstOpts must be instance of dict, tuple or list")
    infLst = lstInfs
    if isinstance(infLst, dict):
        infLst = DctListRepr(infLst, max(map(lambda x: x[1], useLst)) + 1, "NONE")
    maxLen = max(map(lambda x: len(x[0]), useLst))
    """def onEach(cur):
        curLst = cur.children
        curLst[0].bind(on_press=lambda *args: curLst[1].on_press(curLst[1]))
        curLst[1].bind(
            on_state=lambda obj, val: (
                cur.select() if fromState[val] else cur.deselect()))"""
    grid1 = GridLayout(cols=2, size_hint=(1., 1.), spacing=[0, 0])
    lstChk = [None] * len(useLst)
    data = ValuePack({"curSel": curSel})
    lblInf = sizedLabel(text=infLst[curSel], size_hint=(1., 1.,))

    def onState1(index, obj, val):
        del obj
        if fromState[val]:
            data.curSel = index
            txtIn.text = useLst[index][0]
            lblInf.text = infLst[data.curSel]

    lblHeight = 30
    for c in xrange(len(useLst)):
        lstChk[c] = CheckBox(
            group=popup, state=toState[False],
            size_hint_x=None, width=30, allow_no_selection=False)
        lbl = sizedCls(ClkLabel, text=useLst[c][0])
        lbl.height += 20
        lstChk[c].bind(state=Bind(onState1, c))
        lbl.bind(on_press=Bind(togChkBox, lstChk[c]))
        grid1.add_widget(lstChk[c])
        grid1.add_widget(lbl)
        lblHeight = max(lblHeight, lbl.height)
    txtIn.font_size = 24
    grid = GridLayout(cols=2, spacing=[10, 0], **kwargsContent)
    scrl = ScrollView(size_hint=(1.0, 1.0))
    # scrl.scroll_timeout = int(scrl.scroll_timeout * lblHeight / 48.)
    scrl.add_widget(grid1)
    scrl.bind(on_scroll_start=ScrollAllower)
    grid.add_widget(scrl)
    # grid.add_widget(grid1)
    grid.add_widget(lblInf)
    # noinspection PyUnresolvedReferences
    lstChk[data.curSel].state = toState[True]
    done.bind(on_release=Bind(setFunc, useLst, data, popup))
    popup.size_hint_x += .2
    maxLen -= minLenResize
    if maxLen > 0:
        szX = popup.size_hint_x
        szX += .015 * maxLen
        popup.size_hint_x = szX
    return grid


def addFunc(obj, func):
    """
    :param obj:
    :param any func:
    """
    assert callable(func)
    assert hasattr(func, "__name__")
    setattr(obj, func.__name__, func)


# example       (ProgADF41XX,Muxout)
def get_enterSetLst(cls, funcName, updName=None):
    datName = "sel" + funcName
    lstOptName = "lst" + funcName + "Opts"
    attrName = "in" + funcName
    SetName = "set" + funcName

    def setName(self, useLst, adapter, popup, obj=None):
        del obj
        setattr(self.data, datName, useLst[adapter.selection[0].index][1])
        obj1 = getattr(self, attrName)
        try:
            obj1.text = useLst[adapter.selection[0].index][0]
        except StandardError:
            obj1.text = "ERROR"
        popup.dismiss()
        if updName is not None: getattr(self, updName)()

    def enterName(self, obj=None):
        del obj
        c = getattr(self.data, datName)
        mkPopup(Bind(
            getLstContent, getattr(self, lstOptName), c,
            getattr(self, SetName))).open()

    enterName.__name__ = "enter" + funcName
    setName.__name__ = SetName
    addFunc(cls, enterName)
    addFunc(cls, setName)


def get_enterSetLstInf(cls, funcName, updName=None):
    datName = "sel" + funcName
    lstOptName = "lst" + funcName + "Opts"
    lstInfName = "lst" + funcName + "Infs"
    attrName = "in" + funcName
    SetName = "set" + funcName

    def setName(self, useLst, data, popup, obj=None):
        del obj
        setattr(self.data, datName, useLst[data.curSel][1])
        obj1 = getattr(self, attrName)
        try:
            obj1.text = useLst[data.curSel][0]
        except StandardError:
            obj1.text = "ERROR"
        popup.dismiss()
        if updName is not None: getattr(self, updName)()

    def enterName(self, obj=None):
        del obj
        c = getattr(self.data, datName)
        mkPopup(Bind(
            getLstInfContent, getattr(self, lstOptName),
            getattr(self, lstInfName), c, getattr(self, SetName))).open()

    enterName.__name__ = "enter" + funcName
    setName.__name__ = SetName
    addFunc(cls, enterName)
    addFunc(cls, setName)


def txtChanger(varRange, conv, inst, val, extData=None):
    if len(val) == 0: return
    maxLen = varRange.maxLen(conv, extData)
    lenDVal, lenVal, lastDig = reduce(
        lambda tot, x: (tot[0] + int(x.isalnum()), tot[1] + 1,
                        tot[1] if x.isalnum() else tot[2]), val, (0, 0, 0))
    if lenDVal > maxLen: inst.text = val[:lastDig]


# creates a set of functions and puts them in cls
# this set of functions is used to get numeric data from the user using a popup
#  varRange: an instance of BaseRange or a 2-tuple of (min, max) is used to
#    validate the user data before accepting it
#  title: title that is displayed at the top of the popup
#  conv: the 2-tuple of converter format strings or functions to be used by
#    varRange to get maxLen and to assign the user data to the
#    self.data["num" + funcName]
#  updName: attribute name of the function in self that is called after the data
#    is successfully validated
#  unit: the unit string that is displayed to the right of the text box
#  getContent: function that will fill in the body of the popup
#    function prototype: func(done, txtIn, popup, kwargsContent) -> grid
#  popupType: the popup maker function to use (usually mkPopup or mkNumpad)
#  cls: class to place the functions in
def get_enterSetNum(cls, popupType, funcName, getContent, conv, varRange, title, updName=None, unit=None):
    assert isinstance(funcName, (str, unicode))
    assert isinstance(title, (str, unicode))
    assert updName is None or isinstance(updName, (str, unicode))
    assert unit is None or isinstance(unit, (str, unicode))
    assert callable(getContent)
    assert isinstance(popupType, int)
    if not isinstance(varRange, BaseRange): varRange = NumRange(*varRange)
    datName = "num" + funcName
    attrName = "txt" + funcName
    SetName = "set" + funcName
    GetName = "get" + funcName
    popupMaker = mkPopup
    if popupType == 1: popupMaker = mkNumpad
    lstOther = []
    kwargsTxt = {}
    if unit is not None:
        lstOther.append(Label(
            text=unit, pos_hint={'x': .8, "top": 1.0}, size_hint=(.2, .2)))
        kwargsTxt["size_hint_x"] = .8
    curTxtChanger = Bind(txtChanger, varRange, conv)

    def setName(self, txtIn, popup, obj):
        num = None
        txt = txtIn.text
        if not (varRange.noneable and len(txt) == 0):
            try:
                num = conv[0](txt)
                if not varRange.satisfies(num, self):
                    raise ValueError(varRange.userInfo(title, conv, self))
            except StandardError as Exc:
                sys.stderr.write("txt=%r\n"%txt)
                sys.stderr.write(traceback.format_exc())
                txtMsg = str(Exc)
                bubble = Bubble(
                    background_image=images.bubbleBG,
                    arrow_image=images.bubbleArrow,
                    background_color=(1., .75, .75, 1.25),
                    pos_hint={'center_x': 0.5, 'y': 1.2},
                    size_hint=(None, .2), width=int(7.68 * len(txtMsg)))
                bubble.add_widget(Label(text=txtMsg))
                popup.content.add_widget(bubble)
                bubble.bind(on_touch_down=Bind(touchDownBubble, popup.content))
                return
        self.data[datName] = num
        getattr(self, attrName).text = conv[1](num) if num is not None else ""
        popup.dismiss(obj)
        if updName is not None: getattr(self, updName)()

    def getName(self, done, txtIn, popup, kwargsContent):
        grid = getContent(done, txtIn, popup, kwargsContent)
        done.bind(on_release=Bind(getattr(self, SetName), txtIn, popup))
        if varRange.verifyOnChg:
            if varRange.wantsExtData:
                txtIn.bind(
                    text=lambda *args: curTxtChanger(*args, extData=self))
            else:
                txtIn.bind(text=curTxtChanger)
        return grid

    def enterName(self, obj):
        del obj
        popupMaker(
            getattr(self, GetName), {"title": title},
            dict(kwargsTxt, text=getattr(self, attrName).text),
            tuple(lstOther)).open()

    enterName.__name__ = "enter" + funcName
    getName.__name__ = GetName
    setName.__name__ = SetName
    addFunc(cls, enterName)
    addFunc(cls, setName)
    addFunc(cls, getName)


# floating point number specific version of get_enterSetNum
def get_enterSetFloat(cls, funcName, varRange, title, updName=None, unit=None):
    func, popupType = getNumpad, 1
    get_enterSetNum(
        cls, popupType, funcName, func, (float, fromFloat),
        varRange, title, updName, unit)


# integer number specific version of get_enterSetNum
def get_enterSetInt(cls, funcName, varRange, title, updName=None):
    if not isinstance(varRange, BaseRange): varRange = NumRange(*varRange)
    func, popupType = getIntNumpad, 0
    if varRange.isNeg: func, popupType = getNumpad1, 1
    get_enterSetNum(
        cls, popupType, funcName, func, (int, str), varRange, title, updName)


def get_enterSetDecimal(cls, funcName, varRange, title, updName=None, unit=None):
    func, popupType = getNumpad, 1
    get_enterSetNum(
        cls, popupType, funcName, func, (Decimal, fromDecimal),
        varRange, title, updName, unit)


# hexadecimal number specific version of get_enterSetNum
def get_enterSetHex(cls, funcName, varRange, title, updName=None):
    if not isinstance(varRange, BaseRange): varRange = NumRange(*varRange)
    conv = (fromHexStr, toHexBasic)
    padZero = varRange.maxLen(conv)
    if padZero != -1:
        fmtStr = "%%0%uX" % padZero
        toHexStr = lambda num: fmtStr % num
        conv = (fromHexStr, toHexStr)
    func, popupType = getHexNumpad, 0
    get_enterSetNum(
        cls, popupType, funcName, func, conv, varRange, title, updName)


# a more contained version of the class attribute fillers of get_enterSet***,
class GetEnterSet_NumFunc(object):
    def __init__(self, widget, kwargs, varRange, name, getContent, conv, title, unit=None):
        self.updCallable = None
        self.widget = widget
        self.title = title
        self.titleLbl = sizedLabel(text=title, size_hint=(None,None))
        if unit is not None:
            self.grid = GridLayout(cols=3, spacing=[10,5], **kwargs)
            self.grid.add_widget(self.titleLbl)
            self.grid.add_widget(self.widget)
            self.grid.add_widget(sizedLabel(text=unit, size_hint=(None,None)))
        else:
            self.grid = GridLayout(cols=2, spacing=[10,5], **kwargs)
            self.grid.add_widget(self.titleLbl)
            self.grid.add_widget(self.widget)
        self.varRange = varRange
        self.name = name
        self.data = 0
        self.getContent = getContent
        self.curTxtChanger = None
        self.conv = conv
        self.lstOther = []
        if varRange.verifyOnChg:
            self.curTxtChanger = Bind(txtChanger, varRange, conv)
        widget.bind(on_release=self.enterName)
        self.popupMaker = mkPopup
        self.kwargsTxt = {}
    def setData(self, data):
        self.data = data
        self.widget.text = "" if data is None else self.conv[1](data)
    def setName(self, txtIn, popup, obj):
        name, varRange, conv = self.name, self.varRange, self.conv
        title = self.title
        num = None
        txt = txtIn.text
        if not(varRange.noneable and len(txt) == 0):
            try:
                num = conv[0](txt)
                if not varRange.satisfies(num):
                    raise ValueError(self.name + " must satisfy " + str(varRange))
            except Exception as Msg:
                sys.stderr.write(txt + "\n")
                sys.stderr.write(traceback.format_exc())
                txtMsg = str(Msg)
                if txtMsg.startswith(name):
                    txtMsg = varRange.userInfo(title)
                bubble = Bubble(
                    background_image=images.bubbleBG,
                    arrow_image=images.bubbleArrow,
                    background_color=(1.,.75,.75,1.25),
                    pos_hint={'center_x':0.5, 'y':1.2},
                    size_hint=(None,.2),width=int(7.68*len(txtMsg)))
                bubble.add_widget(Label(text=txtMsg))
                popup.content.add_widget(bubble)
                bubble.bind(on_touch_down=Bind(touchDownBubble,popup.content))
                return
        self.data = num
        self.widget.text = "" if num is None else conv[1](num)
        self.updFunc()
        popup.dismiss(obj)
    def getName(self, done, txtIn, popup, kwargsContent):
        grid = self.getContent(done, txtIn, popup, kwargsContent)
        done.bind(on_release=Bind(self.setName, txtIn, popup))
        if self.curTxtChanger is not None: txtIn.bind(text=self.curTxtChanger)
        return grid
    def enterName(self, obj):
        del obj
        self.popupMaker(
            self.getName,{"title":self.title},
            dict(self.kwargsTxt, text=self.widget.text),
            tuple(self.lstOther)).open()
    def updFunc(self):
        if self.updCallable is not None: self.updCallable()


class MultipleDivConv(object):
    def __init__(self, mul, baseConv):
        self.mul = mul
        self.baseFromStr, self.baseToStr = baseConv
    def toStr(self, num):
        return self.baseToStr(num * self.mul)
    def fromStr(self, Str):
        return self.baseFromStr(Str) / self.mul
    def getConvTuple(self):
        return self.fromStr, self.toStr


class CheckBitsPanel(object):
    def __init__(self, title, fmtVal, chkFmt, lsbVal, nBits, defVal, onChgCB, **kwargs):
        assert callable(fmtVal) or isinstance(fmtVal, (str, unicode))
        self.fmt = fmtVal
        self.lsbVal = lsbVal
        self.word = bitMapInt(defVal, nBits)
        self.topLbl = Label(text=title, size_hint=(1.0, .1), pos_hint={"center_x": .5, "top": 1.0})
        self.botLbl = Label(text=self.GetBottomText(), size_hint=(1.0, .1), pos_hint={"center_x": .5, "top": .1})
        self.layout = Bordered(**kwargs)
        self.grid = GridLayout(cols=4, size_hint=(1.0, .8), pos_hint={'x': 0.0, "top": .9})
        self.layout.add_widget(self.topLbl)
        self.layout.add_widget(self.grid)
        self.layout.add_widget(self.botLbl)
        self.lstChk = [None] * nBits
        self.lstLbl = [None] * nBits
        self.onChgCB = onChgCB
        val = lsbVal
        for c in xrange(nBits):
            chk = CheckBox(active=self.word[c], size_hint=(None, None), size=(32, 32))
            chk.bind(active=self.onChkBoxChg)
            lbl = ClkLabel(
                text=chkFmt % val if isinstance(chkFmt, (str, unicode)) else chkFmt(val),
                halign="left", valign="middle", on_release=self.onLblClk)
            lbl.bind(size=setTextSz)
            self.lstChk[c] = chk
            self.lstLbl[c] = lbl
            self.grid.add_widget(chk)
            self.grid.add_widget(lbl)
            val *= 2
    def onLblClk(self, lbl):
        try:
            pos = self.lstLbl.index(lbl)
            chk = self.lstChk[pos]
            chk.active = not chk.active
        except ValueError:
            return
    def onChkBoxChg(self, chk, val):
        try:
            pos = self.lstChk.index(chk)
            self.lstLbl[pos].color = stRGB[val]
            self.word[pos] = val
            self.botLbl.text = self.GetBottomText()
            if self.onChgCB is not None:
                self.onChgCB(self)
        except ValueError:
            return

    def GetBottomText(self):
        val = self.word.num * self.lsbVal
        if isinstance(self.fmt, (str, unicode)):
            return self.fmt % val
        elif callable(self.fmt):
            return self.fmt(val)


def get_enterSetGeneric(cls, funcName, varRange, title, conv, updName=None, popupDesc=None, unit=None):
    if not isinstance(varRange, BaseRange): varRange = NumRange(*varRange)
    assert popupDesc is None or (isinstance(popupDesc, (tuple, list)) and len(popupDesc) == 2)
    func, popupType = (
        (
            (getNumpad1, 1)
            if varRange.isNeg else
            (getIntNumpad, 0)
        )
        if popupDesc is None else
        popupDesc
    )
    get_enterSetNum(cls, popupType, funcName, func, conv, varRange, title, updName, unit)