def GetStrBytesLong(Long):
    """
    :type Long: int|long
    :rtype: str
    """
    if Long == 0:
        return "\0"
    Rtn = bytearray("\0"*((Long.bit_length() + 7)/8))
    for c in xrange(len(Rtn)):
        Rtn[c] = Long & 0xFF
        Long >>= 8
    return str(Rtn)
def GetLongStrBytes(StrBytes):
    """
    :type StrBytes: str
    :rtype: int|long
    """
    Power = 1
    Rtn = long()
    for Ch in StrBytes:
        Rtn += Power * ord(Ch)
        Power <<= 8
    return Rtn
def AlignStrBytesLong(Long, Align):
    """
    :type Long: int|long
    :type Align: int
    :rtype: str
    """
    Rtn = bytearray("\0"*Align)
    Long = long(Long)
    for c in xrange(len(Rtn)):
        Rtn[c] = Long & 0xFF
        Long >>= 8
    return str(Rtn)
class AnyFile(object):
    def __init__(self, Obj):
        self.IsIter = True
        try:
            Obj[0:len(Obj)]
        except StandardError:
            self.IsIter = False
        self.Obj = Obj
        self.Pos = 0
    def read(self, Num=None):
        if Num is None:
            if self.IsIter:
                Rtn = self.Obj[self.Pos:]
                self.Pos += len(Rtn)
                return Rtn
            else: return self.Obj.read()
        if self.IsIter:
            if self.Pos >= len(self.Obj):
                return ""
            Orig = self.Pos
            self.Pos += Num
            if self.Pos > len(self.Obj):
                self.Pos = len(self.Obj)
            return self.Obj[Orig:self.Pos]
        return self.Obj.read(Num)
    def write(self, Data, KeepLast = False):
        if self.IsIter:
            if KeepLast:
                self.Obj = self.Obj[0:self.Pos] + Data + self.Obj[self.Pos:]
            else:
                self.Obj = self.Obj[0:self.Pos] + Data
            self.Pos += len(Data)
            return len(Data)
        return self.Obj.write(Data)
    def tell(self): return self.Pos
    def seek(self, offset, whence):
        if self.IsIter:
            if whence == 0: pass# SEEK_SET
            elif whence == 1: offset += self.Pos # SEEK_CUR
            elif whence == 2: offset += len(self.Obj) # SEEK_END
            else: raise IOError("Invalid whence argument seek")
            self.Pos = max(0, min(len(self.Obj), offset))
        else: self.Obj.seek(offset, whence)
    def GetLen(self):
        if self.IsIter:
            return len(self.Obj)
        self.Pos = self.Obj.tell()
        self.Obj.seek(0, 2)
        Rtn = self.Obj.tell()
        self.Obj.seek(self.Pos)
        return Rtn
DATA_OBJ = 0
DATA_LONG = 1
DATA_INT = 2
DATA_BYTES = 3
DATA_VAR_BYTES = 4
DATA_ARRAY = 5
DATA_BYTES_LEFT = 6
DATA_COND_FMT = 7
DATA_VAR_FMT = 8
DATA_DIRECT = 9
DATA_NONE = 10

def PackNone(Fl, Data):
    assert Data is None


def UnpackNone(Fl):
    return None


class DataPacker(object):
    Types = {
        DATA_OBJ, DATA_LONG, DATA_INT, DATA_BYTES, DATA_VAR_BYTES,
        DATA_ARRAY, DATA_BYTES_LEFT, DATA_COND_FMT, DATA_VAR_FMT, DATA_DIRECT}
    MetaTypes = {DATA_COND_FMT, DATA_VAR_FMT}
    ExtTypes = {
        DATA_OBJ: lambda Arg:(Arg[0], DataPacker(*Arg[1:])),
        # Type, header length = 4, header number offset = 0, number offset = 0
        DATA_LONG: lambda Arg:(
            Arg[0], 4 if len(Arg) < 2 else Arg[1], 0 if len(Arg) < 3 else Arg[2], 0 if len(Arg) < 4 else Arg[3]),
        # Type, header length = 4, number offset = 0
        DATA_INT:lambda Arg:(Arg[0], 4 if len(Arg) < 2 else Arg[1], 0 if len(Arg) < 3 else Arg[2]),
        # Type, number of bytes
        DATA_BYTES:lambda Arg:(Arg[0], Arg[1]),
        # Type, header length = 4, header num offset = 0
        DATA_VAR_BYTES:lambda Arg:(Arg[0], 4 if len(Arg) < 2 else Arg[1], 0 if len(Arg) < 3 else Arg[2]),
        # Type, array type, header length = 4, header num offset = 0
        DATA_ARRAY:lambda Arg:(
            Arg[0], DataPacker.ExtTypes[Arg[1][0]](Arg[1]),
            4 if len(Arg) < 3 else Arg[2], 0 if len(Arg) < 4 else Arg[3]),
        # Type,
        DATA_BYTES_LEFT:lambda Arg: (Arg[0],),
        # Type, condition function, ...list of type tuples to select from
        DATA_COND_FMT:lambda Arg: (
            Arg[0], Arg[1], map(lambda x: DataPacker.ExtTypes[x[0]](x), Arg[2:])),
        # Type, FmtMakerFunc function
        DATA_VAR_FMT: lambda Arg: (
            Arg[0], Arg[1]),
        # Type, Obj To Fl Func, Fl to Obj Func, args, kwargs
        DATA_DIRECT: lambda Arg: (Arg[0], Arg[1], Arg[2], Arg[3], Arg[4]),
        DATA_NONE: lambda Arg: (Arg[0],)
    }
    TypePacker = {
        DATA_OBJ: lambda Fl, Data, Packer: Packer.Pack(Data, Fl),
        # Type, header length, length number offset, number offset
        DATA_LONG: lambda Fl, Data, HeadLen, HeadNumOff, NumOff:(
            DataPacker.TypePacker[DATA_VAR_BYTES](Fl, GetStrBytesLong(Data-NumOff),HeadLen,HeadNumOff)),
        DATA_INT: lambda Fl, Data, IntSize, NumOff: Fl.write(AlignStrBytesLong(Data-NumOff, IntSize)),
        DATA_BYTES: lambda Fl, Data, ByteSize: Fl.write((Data+"\0"*max(0,ByteSize-len(Data)))[:ByteSize]),
        DATA_VAR_BYTES: lambda Fl, Data, HeadLen, HeadNumOff: (
            Fl.write(AlignStrBytesLong(len(Data)-HeadNumOff,HeadLen)+Data)),
        DATA_ARRAY: lambda Fl, Data, Type, HeadLen, HeadNumOff: (
            DataPacker.TypePacker[DATA_INT](Fl,len(Data),HeadLen, HeadNumOff),
            map(lambda Data1: DataPacker.TypePacker[Type[0]](Fl, Data1, *Type[1:]), Data)),
        DATA_BYTES_LEFT: lambda Fl, Data: Fl.write(Data),
        # meta-type
        DATA_COND_FMT: lambda Fl, PrevData, Data, CondFunc, LstTypes: (
            DataPacker.PackType(Fl, LstTypes[CondFunc(PrevData)], Data, PrevData)),
        DATA_VAR_FMT: lambda Fl, PrevData, Data, FmtMakerFunc: (
            DataPacker.PackType(Fl, FmtMakerFunc(PrevData), Data, PrevData)),
        DATA_DIRECT: lambda Fl, Data, ObjToFl, FlToObj, args, kwargs: (
            ObjToFl(Fl, Data, *args, **kwargs)),
        DATA_NONE: PackNone
    }
    TypeUnpacker = {
        DATA_OBJ: lambda Fl, Packer: Packer.Unpack(Fl),
        # Type, header length, length number offset, number offset
        DATA_LONG: lambda Fl, HeadLen, HeadNumOff, NumOff: (
            GetLongStrBytes(DataPacker.TypeUnpacker[DATA_VAR_BYTES](Fl, HeadLen, HeadNumOff))+NumOff),
        DATA_INT: lambda Fl, IntSize, NumOff: GetLongStrBytes(Fl.read(IntSize)) + NumOff,
        DATA_BYTES: lambda Fl, ByteSize: Fl.read(ByteSize),
        DATA_VAR_BYTES: lambda Fl, HeadLen, HeadNumOff: Fl.read(GetLongStrBytes(Fl.read(HeadLen))+HeadNumOff),
        DATA_ARRAY: lambda Fl, Type, HeadLen, HeadNumOff: (
            [DataPacker.TypeUnpacker[Type[0]](Fl, *Type[1:])
             for _ in xrange(DataPacker.TypeUnpacker[DATA_INT](Fl, HeadLen, HeadNumOff))]),
        DATA_BYTES_LEFT: lambda Fl: Fl.read(),
        # meta-type
        DATA_COND_FMT: lambda Fl, PrevData, CondFunc, LstTypes: (
            DataPacker.UnpackType(Fl, LstTypes[CondFunc(PrevData)], PrevData)),
        # meta-type
        DATA_VAR_FMT: lambda Fl, PrevData, FmtMakerFunc: (
            DataPacker.UnpackType(Fl, FmtMakerFunc(PrevData), PrevData)),
        DATA_DIRECT: lambda Fl, ObjToFl, FlToObj, args, kwargs: (
            FlToObj(Fl, *args, **kwargs)),
        DATA_NONE: UnpackNone
    }
    @classmethod
    def ProcExtType(cls, Arg):
        return cls.ExtTypes[Arg[0]](Arg)
    # watch out for default arguments on meta-types
    @classmethod
    def PackType(cls, Fl, TypeArgs, Data, PrevData=None):
        if TypeArgs[0] in cls.MetaTypes:
            cls.TypePacker[TypeArgs[0]](Fl, PrevData, Data, *TypeArgs[1:])
        else:
            cls.TypePacker[TypeArgs[0]](Fl, Data, *TypeArgs[1:])
    # watch out for default arguments on meta-types
    @classmethod
    def UnpackType(cls, Fl, TypeArgs, PrevData=None):
        if TypeArgs[0] in cls.MetaTypes:
            return cls.TypeUnpacker[TypeArgs[0]](Fl, PrevData, *TypeArgs[1:])
        else:
            return cls.TypeUnpacker[TypeArgs[0]](Fl, *TypeArgs[1:])
    def __init__(self, *args):
        self.TypeLst = [None] * len(args)
        for c in xrange(len(args)):
            Arg = args[c]
            if isinstance(Arg, (tuple, list)):
                BaseType = Arg[0]
                if BaseType not in DataPacker.ExtTypes:
                    raise ValueError("Invalid Type argument: %s" % repr(Arg))
                self.TypeLst[c] = DataPacker.ExtTypes[BaseType](Arg)
            else:
                Arg = (Arg,)
                self.TypeLst[c] = DataPacker.ExtTypes[Arg[0]](Arg)
            if Arg[0] in self.MetaTypes:
                assert c > 0, "Meta types must have types preceding them"
    def Pack(self, LstData, Fl = None):
        assert len(LstData) == len(self.TypeLst)
        RtnStr = Fl is None
        if RtnStr: Fl = AnyFile("")
        for Dat, Typ, c in zip(LstData, self.TypeLst, xrange(len(LstData))):
            if Typ[0] not in self.MetaTypes: DataPacker.TypePacker[Typ[0]](Fl, Dat, *Typ[1:])
            else: DataPacker.TypePacker[Typ[0]](Fl, LstData[c-1], Dat, *Typ[1:])
        if hasattr(Fl, "flush"): Fl.flush()
        if RtnStr: return Fl.Obj
    def Unpack(self, Data):
        if isinstance(Data, basestring): Data = AnyFile(Data)
        else: assert hasattr(Data, "read") and callable(Data.read)
        Rtn = [None] * len(self.TypeLst)
        for c in xrange(len(Rtn)):
            Type = self.TypeLst[c]
            Unpacker = DataPacker.TypeUnpacker[Type[0]]
            if Type[0] not in self.MetaTypes:
                Rtn[c] = Unpacker(Data, *Type[1:])
            else:
                Rtn[c] = Unpacker(Data, Rtn[c-1], *Type[1:])
        return Rtn
def DataObj(*FmtTuples):
    return (DATA_OBJ,) + tuple(FmtTuples)
def DataLong(HeadLen = 4, HeadNumOff = 0, NumOff = 0):
    return (DATA_LONG, HeadLen, HeadNumOff, NumOff)
def DataInt(HeadLen = 4, NumOff = 0):
    return (DATA_INT, HeadLen, NumOff)
def DataBytes(NumBytes):
    return (DATA_BYTES, NumBytes)
def DataVarBytes(HeadLen = 4, HeadNumOff = 0):
    return (DATA_VAR_BYTES, HeadLen, HeadNumOff)
def DataArray(ArrayType, HeadLen = 4, HeadNumOff = 0):
    return (DATA_ARRAY, ArrayType, HeadLen, HeadNumOff)
def DataBytesLeft():
    return (DATA_BYTES_LEFT,)
def DataCondFmt(CondFunc, *FmtTuples):
    return (DATA_COND_FMT, CondFunc) + tuple(FmtTuples)
def DataVarFmt(FmtMakerFunc):
    return (DATA_VAR_FMT, FmtMakerFunc)
def DataDirect(ObjToFl, FlToObj, *args, **kwargs):
    return (DATA_DIRECT, ObjToFl, FlToObj, args, kwargs)
def DataNone():
    return (DATA_NONE,)
