from .LibBaseScreen import BaseScreenUSB, BaseScreen, ScreenStorage
from .DataEntryHelpers import *
from . import globalVars
from watchdog import events
import subprocess
from sys import executable as exec_path
from .LibTask import ImmediateAsynch, Task
from .LibBaseFileLoading import LoadError
from .LibGetTabPanel import getTabPanel
from .LibPopup import SimpleTextPopup
from .LibMessageBox import IDYES
import shutil
import tarfile
import json
import threading


def getIPv4And6Addr():
    ipv4 = "None"
    ipv6 = "None"
    if os.name == "nt":
        for line in subprocess.check_output("ipconfig").split("\r\n"):
            if line.find("IPv4 Address") != -1:
                ipv4 = line[line.find(":") + 1:].strip()
            elif line.find("IPv6 Address") != -1:
                ipv6 = line[line.find(":") + 1:].strip()
    elif os.name == "posix":
        entries = [{}]
        orderMap = {
            "Ethernet": 0,
            "Local Loopback": 2
        }
        for line in subprocess.check_output("ifconfig").split("\n"):
            pos = line.find("encap:")
            if pos != -1:
                entries[-1]["encap"] = line[pos + 6:line.find("  ", pos + 6)]
                continue
            if line.find("UP") != -1:
                entries[-1]["RUNNING"] = line.find("RUNNING") != -1
                continue
            pos = line.find("inet addr:")
            if pos != -1:
                entries[-1]["inet"] = line[pos + 10:line.find(" ", pos + 10)]
                continue
            pos = line.find("inet6 addr: ")
            if pos != -1:
                entries[-1]["inet6"] = line[pos + 12:line.find(" ", pos + 12)]
                continue
            if line.isspace() or len(line) == 0:
                entries.append({})
        entries = filter(lambda x: x.get("RUNNING", False), entries)
        entries = sorted(entries, key=lambda x: orderMap.get(x["encap"], 1))
        if len(entries):
            ipv4 = entries[0].get("inet", ipv4)
            ipv6 = entries[0].get("inet6", ipv6)
    return ipv4, ipv6


class FileListDir(object):
    def __init__(self, baseDir1, lstFiles):
        self.lstFiles = lstFiles
        self.baseDir = baseDir1
    def Filter(self, fileFilter):
        """
        :param BaseFileFilter fileFilter:
        :rtype: list[any]
        """
        match = [fileFilter.Match(self.baseDir, self.lstFiles[c]) for c in xrange(len(self.lstFiles))]
        """:type: list[any|None]"""
        if fileFilter.exclusive:
            lstPosMatch = filter(lambda x: x[1] is not None, list(enumerate(match)))
            """:type: list[(int,any)]"""
            rtn = [None] * len(lstPosMatch)
            for c in xrange(len(rtn)):
                c0 = len(rtn) - 1 - c
                cDel, rtn[c0] = lstPosMatch[c0]
                self.lstFiles.pop(cDel)
            return rtn
        else:
            return filter(None, match)
    def Narrow(self, fileFilter):
        lstPosMatch = [
            c
            for c in xrange(len(self.lstFiles))
            if fileFilter.Match(self.baseDir, self.lstFiles[c]) is None][::-1]
        return [self.lstFiles.pop(x) for x in lstPosMatch][::-1]


class BaseFileFilter(object):
    def __init__(self):
        self.exclusive = True
    def Match(self, Dir, fName):
        return None


class DefFileFilter(BaseFileFilter):
    def __init__(self, ext):
        super(DefFileFilter, self).__init__()
        self.exclusive = False
        self.cmpExt = "." + ext
        self.ext = ext
    def Match(self, Dir, fName):
        if fName.startswith("."):
            return None
        return fName if fName.endswith(self.cmpExt) else None


log = globalVars.log


class UsbHandler(events.FileSystemEventHandler):
    def __init__(self, app):
        """
        :param AppData app:
        """
        self.app = app
        self.watchDir = "/dev" if os.name == "posix" else None
        self.firstPart = "/sd"
        self.mounted = []
        self.startsWith = ""
        if self.watchDir is not None:
            self.startsWith = self.watchDir + self.firstPart
        self.Lk = threading.RLock()
    def refresh_synch_main(self):
        if self.watchDir is None: return
        with self.Lk:
            curDevs = set(map(lambda x: x[1], self.mounted))
            tmp = self.firstPart[1:]
            allDevs = set(map(
                lambda x: self.watchDir + "/" + x,
                filter(
                    lambda x: x.startswith(tmp),
                    os.listdir(self.watchDir))))
            addDevs = filter(lambda x: x not in curDevs, allDevs)
            remDevs = filter(lambda x: x not in allDevs, curDevs)
            for dev_name in remDevs:
                self.process_rem_dev(None, dev_name, ImmediateAsynch)
            for dev_name in addDevs:
                self.process_add_dev(dev_name, ImmediateAsynch)

    def process_add_dev(self, dev_name, async):
        assert isinstance(dev_name, (unicode, str))
        assert callable(async.schedule_once)
        with self.Lk:
            if len(self.mounted): return
            if not dev_name.startswith(self.startsWith): return
            if os.name != "posix": return
            txt = None
            try:
                task = Task(subprocess.check_output,
                            ["sudo", "lsblk", "-o", "kname,label,type,rm,uuid", dev_name])
                async.schedule_once(lambda dt=None: task.run())
                txt = task.getReturn()
            except subprocess.CalledProcessError as Exc:
                log.Log("bad lsblk %s" % Exc, 2)
            if txt is None: return
            text_grid = filter(len, txt.decode("utf-8").split("\n"))
            """:type: list[str|unicode]"""
            if any(map(lambda x: x.startswith("lsblk:"), text_grid)):
                log.Log("lsblk Error: %s" % txt, 2)
                return  # error
            lbls_offs = [["", 0]]
            for c, ch in enumerate(text_grid[0]):
                if ch.isspace():
                    if len(lbls_offs[-1][0]):
                        lbls_offs.append(["", c])
                else:
                    if len(lbls_offs[-1][0]) == 0:
                        lbls_offs[-1][1] = c
                    lbls_offs[-1][0] += ch
            for c in xrange(len(lbls_offs) - 1):
                lbls_offs[c].append(lbls_offs[c + 1][1])
            lbls_offs[-1].append(max(map(len, text_grid)))
            dev_infs = []
            lStripLbls = {"RM"}
            for line in text_grid[1:]:
                dev_inf = {}
                for lbl, off, off1 in lbls_offs:
                    cur = line[off:off1]
                    if lbl in lStripLbls:
                        cur = cur.lstrip(" ")
                    else:
                        cur = cur.rstrip(" ")
                    dev_inf[lbl] = cur
                dev_infs.append(dev_inf)
            if len(dev_infs) > 1:
                log.Log("not a single partition %r" % dev_infs, 3)
                return  # not a single partition
            elif len(dev_infs) == 0:
                log.Log("no results for lsblk <dev_name=%r>" % dev_name, 3)
                return
            dev_inf = dev_infs[0]
            if dev_inf["UUID"] == "6027-4CD3":
                log.Log("UUID blacklist %r" % dev_infs[0], 3)
                return
            if "USB30FD" in dev_inf["LABEL"]:
                log.Log("USB30FD still got through", 2)
                return
            usr_name = os.environ["USER"] if "USER" in os.environ else os.environ["USERNAME"]
            label = dev_inf["LABEL"]
            dev_name = "/dev/" + dev_inf["KNAME"]
            mnt_dir = "/media/%s/%s" % (usr_name, label)
            if not os.path.isdir(mnt_dir):
                os.mkdir(mnt_dir)
            args = ["sudo", "mount", dev_name, mnt_dir, "-o", "uid=%u,gid=%u,flush" % (os.getuid(), os.getgid())]
            log.Log("mount args: %r" % args, 3)
            task = Task(self.app.tryMountUsbDrive, dev_name, mnt_dir, dev_inf, args)
            async.schedule_once(lambda dt=None: task.run())
            code, txt = task.getReturn()
            if code == 0:
                self.mounted.append((dev_inf, dev_name, mnt_dir))
                log.Log("Auto mounted \"%s\" into \"%s\"; output:\n%s" % (dev_name, mnt_dir, txt), 3)
            else:
                log.Log("Could not mount \"%s\" into \"%s\"; output:\n%s" % (dev_name, mnt_dir, txt), 2)

    def on_created(self, event):
        self.process_add_dev(event.src_path, Clock)

    def process_rem_dev(self, mnt_dir, dev_name, asynch):
        if os.name == "nt":
            asynch.schedule_once(lambda dt=None: self.app.remWinUsbDrive(mnt_dir), 0)
            return True
        assert isinstance(dev_name, (unicode, str, None.__class__))
        assert isinstance(mnt_dir, (unicode, str, None.__class__))
        assert callable(asynch.schedule_once)
        with self.Lk:
            if len(self.mounted) == 0:
                log.Log("len(self.mounted) == 0; cannot eject", 2)
                return False
            if os.name != "posix": return False
            Pos = len(self.mounted)
            for c, (dev_inf, devName, mntDir) in enumerate(self.mounted):
                if ((mnt_dir is not None and is_path_equal(mntDir, mnt_dir)) or
                        (devName is not None and is_path_equal(devName, dev_name))):
                    Pos = c
                    break
            if Pos < len(self.mounted):
                dev_inf, dev_name, mnt_dir = self.mounted[Pos]
                args = ["sudo", "umount", mnt_dir]
                task = Task(self.app.tryDismountDrive, dev_name, mnt_dir, dev_inf, args, lambda x: x[0] == 0)
                asynch.schedule_once(lambda dt=None: task.run())
                code, txt = task.getReturn()
                if code == 0:
                    log.Log("successfully dismounted \"%s\" from \"%s\"; output: %s" % (dev_name, mnt_dir, txt), 3)
                    self.mounted.pop(Pos)
                    return True
                else:
                    log.Log("error, could not dismount \"%s\" from \"%s\"; output: %s" % (dev_name, mnt_dir, txt), 2)
        return False
    def eject_device(self, mnt_dir, dev_name=None):
        """
        :param NoneType|str|unicode mnt_dir:
        :param NoneType|str|unicode dev_name:
        :return: bool
        """
        return self.process_rem_dev(mnt_dir, dev_name, Clock)
    def process_rem_no_umount(self, mnt_dir, dev_name, asynch):
        assert isinstance(dev_name, (unicode, str, None.__class__))
        assert isinstance(mnt_dir, (unicode, str, None.__class__))
        assert callable(asynch.schedule_once)
        with self.Lk:
            if len(self.mounted) == 0:
                log.Log("len(self.mounted) == 0; cannot remove", 2)
                return False
            if os.name != "posix": return False
            Pos = len(self.mounted)
            for c, (dev_inf, devName, mntDir) in enumerate(self.mounted):
                if mntDir == mnt_dir or devName == dev_name:
                    Pos = c
                    break
            if Pos < len(self.mounted):
                dev_inf, dev_name, mnt_dir = self.mounted.pop(Pos)
                task = Task(self.app.remUsbDrive, dev_name, mnt_dir, dev_inf)
                asynch.schedule_once(lambda dt=None: task.run())
                task.getReturn()
                return True
        return False
    def rem_NoUMount(self, mnt_dir, dev_name=None):
        self.process_rem_no_umount(mnt_dir, dev_name, Clock)

    def on_deleted(self, event):
        self.eject_device(None, event.src_path)


class FileSelector(object):
    defDataMapper = lambda x: (x, True)
    def __init__(self, baseDir1, topLabel, ext, dataMapper=defDataMapper, flLstDir=None):
        self.ext = ext
        self.flLstDir = flLstDir
        self.baseDir = baseDir1
        self.layout = self.layout = Bordered()
        self.dataMapper = dataMapper
        self.lstOpts = lstOpts = self.mkLstOpts()
        data = map(self.dataMapper, lstOpts)
        self.adapter = adapter = ListAdapter(
            cls=XMButton1, allow_empty_selection=True, selection_mode='single',
            data=data, args_converter=convLstOpts)
        self.theList = BorderedList(
            adapter=adapter, size_hint=(.8, .5), pos_hint={"center_x": .5, "center_y": .5})
        scroll = filter(lambda x: isinstance(x, ScrollView), self.theList.children)[0]
        scroll.bind(on_scroll_start=ScrollAllower)
        self.topLbl = Label(text=topLabel, size_hint=(1., .1), pos_hint={"center_x": .5, "top": 1.0})
        self.layout.add_widget(self.topLbl)
        self.layout.add_widget(self.theList)
    def mkLstOpts(self):
        lstFiles = filter(lambda x1: not x1.startswith("."), os.listdir(self.baseDir))
        if self.ext == "": return lstFiles
        ext = "." + self.ext
        return sorted([x[:-len(ext)] for x in lstFiles if x.endswith(ext)])
    def reload(self):
        self.lstOpts = lstOpts = self.mkLstOpts()
        data = map(self.dataMapper, lstOpts)
        self.adapter.data = data
        self.theList.populate()

    def add_eject_btn(self, app, **kwargs):
        """
        :param AppData app:
        """
        fillDictDefs(kwargs, dict(
            size_hint_x=.2, size_hint_y=.18, pos_hint={"right":1.0, "top": .94},
            color=k7_F_GREEN))
        btn = Button(
            on_release=lambda obj: (
                app.usbEjectPopup.open(obj)
                if
                app.usbScreenStorage.usbHandler.process_rem_dev(self.baseDir, None, ImmediateAsynch)
                else
                app.usbEjectFailPopup.open(obj)),
            text="Eject\nUSB", **kwargs)
        self.layout.add_widget(btn)
        return self

    banCh = set("<>:\"/\\|?*") | set(map(chr, xrange(32)))
    banNames = (
        {"CON", "PRN", "AUX", "NUL", "..", "."} |
        set(map(getFormatFn("LPT%u"), xrange(1, 10))) |
        set(map(getFormatFn("COM%u"), xrange(1, 10))))

    # returns bool, reason
    @classmethod
    def isBannedFile(cls, fName):
        if len(fName) == 0:
            return False, "cannot use blank file name"
        for ch in fName:
            if ch in cls.banCh:
                val = ord(ch)
                return False, "Unsupported character '%s'" % ("\\x%02X" % val if val < 32 else ch)
        if fName.upper() in cls.banNames:
            return False, "file name \"%s\" is not allowed" % fName
        elif fName.count('.') > 0:
            if fName.rsplit('.', 1)[0].upper() in cls.banNames:
                return False, "file name \"%s\" is not allowed" % fName
        return True, ""


def FilterFilesAndOld(lstFiles, ext):
    lstFiles = filter(lambda x: not x.startswith("."), lstFiles)
    if ext == "":
        return lstFiles
    lstFiles1 = map(
        lambda (c, x): (
            c, x, len(x) == 2,
            len(x) == 2 and x[1][0:3] == "old" and (len(x[1]) == 3 or x[1][3:].isdigit())),
        map(lambda (c, x): (c, x.rsplit('.', 1)), enumerate(lstFiles)))
    lstFiles1 = filter(lambda (c, x, b0, b1): b0, lstFiles1)
    lstFiles1 = map(lambda (c, x, b0, b1): (c, x[0] if b1 else lstFiles[c]), lstFiles1)
    ext = "." + ext
    lstFiles1 = filter(lambda (c, x): x.endswith(ext), lstFiles1)
    return map(lambda (c, x): lstFiles[c], lstFiles1)


class LogFileSelector(FileSelector):
    def mkLstOpts(self):
        return FilterFilesAndOld(os.listdir(self.baseDir), self.ext)


class BaseFileManager(FileSelector):
    def __init__(self, baseDir1, topLabel, ext, other=None):
        super(BaseFileManager, self).__init__(baseDir1, topLabel, ext)
        self.togSelBtn = CheckBox(size_hint=(.05, .1), pos_hint={'x': .075, "top": .875})
        self.togSelLbl = Label(
            size_hint=(.575, .1), text="Check to select more than one file",
            pos_hint={'x':.15, "top": .875})

        def onState(obj, val):
            del obj
            val = fromState[val]
            for view in self.adapter.selection:
                if view.is_selected:
                    view.is_selected = False
                    view.deselect()
            self.adapter.selection = []
            self.adapter.selection_mode = 'multiple' if val else 'single'

        self.togSelBtn.bind(state=onState)
        self.layout.add_widget(self.togSelBtn)
        self.layout.add_widget(self.togSelLbl)
        self.other = other
        self.btnPanel = BorderedGridLayout(rows=1, size_hint=(.95, .2), pos_hint={"center_x": .5, "top": .22})
        self.layout.add_widget(self.btnPanel)

        self.renamePopup = SimpleTextPopup(
            "Rename", self.isBannedFile, self.renameAction,
            size_hint=(.5, .5), pos_hint={"center_x": .5, "top": 1.0})

        self.lstBtns = list(map(lambda x: Button(text=x[0], on_release=x[1]), [
            ("Rename", self.renameFile),
            ("Move", self.moveFiles),
            ("Copy", self.copyFiles),
            ("Delete", self.deleteFiles)
        ]))
        # noinspection PyArgumentList
        self.adapter.bind(selection=self.onSelChg)

    def reload(self):
        super(BaseFileManager, self).reload()
        self.togSelBtn.active = False

    def renameAction(self, newName):
        oldName = self.baseDir + self.adapter.selection[0].text + "." + self.ext
        newName = self.baseDir + newName + "." + self.ext
        os.rename(oldName, newName)
        self.reload()

    def renameFile(self, obj=None):
        self.renamePopup.show(self.adapter.selection[0].text)
        obj.add_widget(ErrBubble(
            text="Successfully renamed file", background_color=(1, 1.75, 1, 1.25)))

    def moveFiles(self, obj=None):
        if self.other is None:
            if obj is not None:
                obj.add_widget(ErrBubble(
                    text="USB Drive must be plugged in to Move to", background_color=(1.75, 1, 1, 1.25)))
            return
        dest = self.other.baseDir
        for view in self.adapter.selection:
            oldName = self.baseDir + view.text + "." + self.ext
            shutil.move(oldName, dest)
        self.reload()
        self.other.reload()

    def copyFiles(self, obj=None):
        if self.other is None:
            if obj is not None:
                obj.add_widget(ErrBubble(
                    text="USB Drive must be plugged in to Copy to", background_color=(1.75, 1, 1, 1.25)))
            return
        dest = self.other.baseDir
        for view in self.adapter.selection:
            oldName = self.baseDir + view.text + "." + self.ext
            shutil.copy2(oldName, dest)
        self.other.reload()
        obj.add_widget(ErrBubble(
            text="Successfully copied file(s)", background_color=(1, 1.75, 1, 1.25)))

    def deleteFiles(self, obj=None):
        del obj
        for view in self.adapter.selection:
            fName = self.baseDir + view.text + "." + self.ext
            os.remove(fName)
        self.reload()

    def onSelChg(self, obj, val):
        del obj
        btnPanel, lstBtns = self.btnPanel, self.lstBtns
        btnPanel.clear_widgets()
        if len(val) == 1:
            btnPanel.add_widget(lstBtns[0])
        if len(val):
            btnPanel.add_widget(lstBtns[1])
            btnPanel.add_widget(lstBtns[2])
            btnPanel.add_widget(lstBtns[3])

    def updateOther(self, other):
        self.other = other


class SaveFileManager(BaseFileManager):
    def __init__(self, baseDir1, topLabel, other=None):
        super(SaveFileManager, self).__init__(baseDir1, topLabel, "xmdat", other)


class SaveLogsManager(BaseFileManager):
    def __init__(self, baseDir1, topLabel, other=None):
        super(SaveLogsManager, self).__init__(baseDir1, topLabel, "xmlog", other)


class FileChooser(FileSelector):
    def __init__(self, topText, btnText, actionIt, curBaseDir, curProjName, setCurProjName, preActionCheck, ext,
                 right=False):
        """
        :param str|unicode topText:
        :param str|unicode btnText:
        :param (str|unicode) -> str|unicode|None actionIt: callback action
        :param str|unicode curBaseDir:
        :param str|unicode curProjName:
        :param (str|unicode) -> any setCurProjName:
        :param (str|unicode, str|unicode) -> (bool, str|unicode) preActionCheck: check executed before taking action
        :param str|unicode ext: file extension
        :param bool right: layout to the right
        """
        self.curProjName = curProjName
        self.setCurProjName = setCurProjName
        self.actionIt = actionIt
        if not curBaseDir.endswith("/"): curBaseDir += "/"
        super(FileChooser, self).__init__(curBaseDir, topText, ext)
        self.curBaseDir = self.baseDir
        self.preActionCheck = preActionCheck
        size = (.48, .89)
        if right is not None:
            self.layout.size_hint = size
            self.layout.pos_hint = {"x": 0.505 if right else 0.005, "y": 0.01}
        self.topLbl.size_hint = (.2, .08)
        self.topLbl.pos_hint = {"top": 1.0, "center_x": .5}
        self.txtIn = TextInput(
            size_hint=(.2 / size[0], .08 / size[1]),
            pos_hint={"top": .9, "center_x": .5}, multiline=False)
        self.actBtn = Button(
            text=btnText, pos_hint={"center_x": .5, 'y': .1},
            size_hint=(.2 / size[0], .1 / size[1]))
        lstOpts = self.lstOpts
        adapter = self.adapter
        adapter.selection_mode = 'multiple'
        self.txtIn.bind(text=self.onText)
        self.txtIn.bind(on_text_validate=lambda obj, *args: self.tryActionIt(obj, *args))
        self.actBtn.bind(on_release=lambda obj, *args: self.tryActionIt(obj, *args))
        for view in map(adapter.get_view, xrange(len(lstOpts))):
            view.bind(on_release=lambda x: self.selectSpecific(x))
        self.theList.pos_hint = {'center_x': .5, 'top': .8}
        self.theList.size_hint = (.4 / size[0], .5 / size[1])
        if self.curProjName in lstOpts:
            self.selectSpecific(adapter.get_view(lstOpts.index(self.curProjName)))
        self.layout.add_widget(self.txtIn)
        self.layout.add_widget(self.actBtn)

    def reload(self):
        super(FileChooser, self).reload()
        lstOpts = self.lstOpts
        for view in map(self.adapter.get_view, xrange(len(lstOpts))):
            view.bind(on_release=lambda x: self.selectSpecific(x))
        if self.curProjName in lstOpts:
            self.selectSpecific(self.adapter.get_view(lstOpts.index(self.curProjName)))

    def selectSpecific(self, selView):
        self.txtIn.text = selView.text
        adapter = self.adapter
        for view in list(adapter.selection):
            adapter.deselect_item_view(view)

    def onText(self, inst, txt):
        del inst
        adapter = self.adapter
        lstOpts = self.lstOpts
        theRange = xrange(len(lstOpts))
        views = map(adapter.get_view, theRange)
        selViews = map(views.__getitem__, filter(lambda x: len(txt) and lstOpts[x].startswith(txt), theRange))
        adapter.select_list(selViews, False)
        for view in views:
            if view not in adapter.selection:
                view.deselect()
                view.is_selected = False
        """for view in views:
            view.deselect()
        for view in selViews:
            view.select()"""

    @staticmethod
    def actionIt(fName):
        raise IOError(22, "I don't like that filename: \"%s\"" % fName)

    def setCurProjectName(self, name):
        self.curProjName = name
        self.setCurProjName(name)

    def tryActionIt(self, obj, *args):
        del obj, args
        text = self.txtIn.text
        check, reason = self.isBannedFile(text)
        if not check:
            self.actBtn.add_widget(ErrBubble(
                text=reason, background_color=(1.75, 1, 1, 1.25)))
            return
        fName = self.curBaseDir + "%s.%s" % (text, self.ext)
        ok, reason = self.preActionCheck(fName, text)
        if not ok:
            self.actBtn.add_widget(ErrBubble(
                text=reason, background_color=(1.75, 1, 1, 1.25)))
            return
        try:
            msg = self.actionIt(fName)
        except EnvironmentError as Exc:
            log.Log(traceback.format_exc() + "File action failed", 2)
            self.actBtn.add_widget(ErrBubble(
                text=Exc.message, background_color=(1.75, 1, 1, 1.25)))
        except LoadError as Exc:
            log.Log(traceback.format_exc() + "File action failed", 2)
            self.actBtn.add_widget(ErrBubble(
                side_pos="left" if self.layout.pos_hint['x'] < .5 else "right", side_anchor="bottom",
                text=Exc.StrNoFName(), background_color=(1.75, 1, 1, 1.25)))
        else:
            self.setCurProjectName(text)
            if msg is not None:
                self.actBtn.add_widget(ErrBubble(
                    text=msg % text, background_color=(1, 1.75, 1, 1.25)))
        self.reload()


class SaveLoadScreen(BaseScreenUSB):
    helpTxt = u"\n".join([
        u"{Type} Configuration Help:",
        u"  each panel consists of the following",
        u"    text box at the top",
        u"    selectable list in the middle",
        u"    button at the bottom",
        u"the following are the 2 possible panels on the screen",
        u"\u2022 left: {title} internal SD Card (always visible)",
        u"\u2022 right: {title} USB Drive (name) (only visible if a USB drive is recognized)",
        u"how to {type} a file:",
        u"\u2022 Option 1:",
        u"    - Click one of the items in the list and that item will become",
        u"      highlighted and the name will be filled in the text box.",
        u"\u2022 Option 2:",
        u"    - Click on the text entry box, start typing and items that start ",
        u"      with the text in the box will be highlighted.",
        u"Click {cmd} to {type} the file, the file will be named with the text in the text box",

    ])
    def __init__(self, isSave):
        self.usbFileChoosers = []
        self.isSave = isSave
        self.name = "Save" if isSave else "Load"
        title = "Save File To" if isSave else "Load File From"
        cmd = "Save To File" if isSave else "Load From File"
        self.helpTxt = self.helpTxt.format(title=title, Type=self.name, type=self.name.lower(), cmd=cmd)

    def getName(self):
        return self.name

    def build(self, app, ext_data=None):
        """
        :param AppData app:
        :param None ext_data:
        """
        a = app.BaseDirFileChooser(self.isSave)
        app.wnd.add_widget(a.layout)
        if len(app.usbScreenStorage.usbDirs):
            mnt_dir, dev_name, dev_inf = app.usbScreenStorage.usbDirs[0]
            label = dev_inf["LABEL"]
            b = app.getUsbFileChooser(mnt_dir, label, self.isSave)
            self.usbFileChoosers.append(b)
            app.wnd.add_widget(b.layout)
        super(SaveLoadScreen, self).build(app)

    def hide(self, app):
        """

        :param AppData app:
        :return:
        """
        super(SaveLoadScreen, self).hide(app)
        while len(self.usbFileChoosers):
            a = self.usbFileChoosers.pop()
            app.wnd.remove_widget(a.layout)

    def getHelp(self, app):
        """

        :param AppData app:
        :rtype: (str|unicode, bool)
        """
        return self.helpTxt, False

    def addUsbDrive(self, app):
        """

        :param AppData app:
        :return:
        """
        if len(app.usbScreenStorage.usbDirs) > 1: return
        if not self.showing: return
        mnt_dir, dev_name, dev_inf = app.usbScreenStorage.usbDirs[-1]
        label = dev_inf["LABEL"]
        a = app.getUsbFileChooser(mnt_dir, label, self.isSave)
        self.usbFileChoosers.append(a)
        app.wnd.add_widget(a.layout)

    def remUsbDrive(self, app, drvInfo):
        """

        :param AppData app:
        :param (str|unicode, str|unicode, any) drvInfo:
        :return:
        """
        if not self.showing: return
        mnt_dir, dev_name, dev_inf = drvInfo
        while mnt_dir.endswith("/"): mnt_dir = mnt_dir[:-1]
        for a in self.usbFileChoosers:
            assert isinstance(a, FileChooser)
            cmpDir = a.curBaseDir
            while cmpDir.endswith("/"): cmpDir = cmpDir[:-1]
            if cmpDir == mnt_dir:
                app.wnd.remove_widget(a.layout)
                self.usbFileChoosers.remove(a)
                return


class FileManagementScreen(BaseScreenUSB):
    helpTxt = u"\n".join([
        u"File Management Help:",
        u"\u2022 Save File Management",
        u"    - left: internal SD Card",
        u"    - right: USB drive",
        u"    - both sides: the other side is the opposite side (left->right or right->left)",
        u"        press select to change selection behavior:",
        u"          select one at a time with select: off",
        u"          select multiple with select: on",
        u"        you can move and copy multiple files to the other side at a time",
        u"        you can delete multiple files at a time",
        u"        you can only rename one file at a time",
        u"\u2022 Write Log Management",
        u"    same as File Management except it shows the write logs that were saved instead of configurations",
        u"\u2022 Update",
        u"    plug in USB drive, select desired version, and click update",
        u"\u2022 DevLog Management"
    ])
    filesTabLayout = [
        ("Save File Management", None, (
            BorderedGridLayout, "gridSaveFiles", dict(cols=2, spacing=[5, 5], padding=[5]*4), [
                (Label, "dummySave", {"text":"No USB drive detected."})
            ])),
        ("Write Log Management", None, (
            BorderedGridLayout, "gridWLogFiles", dict(cols=2, spacing=[5, 5], padding=[5]*4), [
                (Label, "dummySave1", {"text": "No USB drive detected."})
            ])),
        ("Update", None, (
            BorderedGridLayout, "gridUpdate", dict(cols=2, spacing=[5, 5], padding=[5]*4), [
                (Label, "verLbl", {"text": "current version: %s" % globalVars.version}),
                (Label, "dummyUpd", {"text": "\n".join([
                    "(step 1) Download the latest firmware update from",
                    "  www.xmicrowave.com",
                    "(step 2) Copy the update file to the root folder",
                    "  of a USB drive.",
                    "(step 3) Insert the flash drive into the controller.",
                    "  NOTE: flash drive could take 5 or more seconds to",
                    "  be recognized.",
                    "(step 4) Select the desired firmware update version."])})
            ])),
        ("OS Update", None, (
            BorderedGridLayout, "gridOS_Update", dict(cols=2, spacing=[5, 5], padding=[5]*4), [
                (Label, "verOS_Lbl", {"text": "current os version: %s" % globalVars.metainf["os_pkg_ver"]}),
                (Label, "dummyOS_Upd", {"text": "\n".join([
                    "(step 1) Download the latest OS update from",
                    "  www.xmicrowave.com",
                    "(step 2) Copy the OS update file to the root folder",
                    "  of a USB drive.",
                    "(step 3) Insert the flash drive into the controller.",
                    "  NOTE: flash drive could take 5 or more seconds to",
                    "  be recognized.",
                    "(step 4) Select the desired OS update version."])})
            ])),
        ("Dev Log Management", None, (
            BorderedGridLayout, "gridLogs", dict(cols=2, spacing=[5, 5], padding=[5, 5, 5, 5]), [
                (Bordered, "exportPanel", {}, [
                    (Button, "exportAllBtn", {
                        "text": "Export All Logs", "size_hint": (.4,.1),
                        "pos_hint": {"center_x":.5, "top": .9}}),
                    (Button, "exportCrashBtn", {
                        "text": "Export Crash logs", "size_hint": (.4,.1),
                        "pos_hint": {"center_x":.5, "top": .75}}),
                    (Button, "exportNonCrashBtn", {
                        "text": "Export Non-Crash logs", "size_hint": (.4,.1),
                        "pos_hint": {"center_x":.5, "top": .6}})]),
                (Label, "devLogsUSBLabel", {"text": "Plug in Flash Drive"})]))]

    def __init__(self):
        self.tabPanel, dctNamed = getTabPanel(
            self.filesTabLayout, size_hint=(.99, .89), pos_hint={"center_x": .5, "top": .9}, tab_width=175)
        gridSaveFiles = dctNamed["gridSaveFiles"]
        gridWLogFiles = dctNamed["gridWLogFiles"]
        gridUpdate = dctNamed["gridUpdate"]
        gridOS_Update = dctNamed["gridOS_Update"]
        gridLogs = dctNamed["gridLogs"]
        devLogsUSBLabel = dctNamed["devLogsUSBLabel"]
        exportAllBtn = dctNamed["exportAllBtn"]
        exportCrashBtn = dctNamed["exportCrashBtn"]
        exportNonCrashBtn = dctNamed["exportNonCrashBtn"]
        assert isinstance(gridSaveFiles, BorderedGridLayout)
        assert isinstance(gridWLogFiles, BorderedGridLayout)
        assert isinstance(gridUpdate, BorderedGridLayout)
        assert isinstance(gridOS_Update, BorderedGridLayout)
        assert isinstance(gridLogs, BorderedGridLayout)
        assert isinstance(devLogsUSBLabel, Label)
        assert isinstance(exportAllBtn, Button)
        assert isinstance(exportCrashBtn, Button)
        assert isinstance(exportNonCrashBtn, Button)
        self.exportAllBtn = exportAllBtn
        self.exportCrashBtn = exportCrashBtn
        self.exportNonCrashBtn = exportNonCrashBtn
        exportAllBtn.bind(on_release=self.exportAllLogs)
        exportCrashBtn.bind(on_release=self.exportCrashLogs)
        exportNonCrashBtn.bind(on_release=self.exportNonCrashLogs)
        self.gridSaveFiles = gridSaveFiles
        self.gridWLogFiles = gridWLogFiles
        self.gridUpdate = gridUpdate
        self.gridLogs = gridLogs
        self.gridOS_Update = gridOS_Update
        self.dummySave = dctNamed["dummySave"]
        self.dummySave1 = dctNamed["dummySave1"]
        self.dummyUpd = dctNamed["dummyUpd"]
        self.dummyOS_Upd = dctNamed["dummyOS_Upd"]
        self.verLbl = dctNamed["verLbl"]
        self.verOS_Lbl = dctNamed["verOS_Lbl"]
        self.devLogsUSBLabel = devLogsUSBLabel
        self.usbUpdateSelectors = []
        self.usbFileSelectors = []
        self.usbFileSelectors1 = []
        self.usbLogSelectors = []
        self.usbFileSelectors_1 = []
        curBaseDir = globalVars.logCtx.baseDir if len(globalVars.logCtx.baseDir) else "./"
        self.homeMngr = SaveFileManager(curBaseDir, "Internal SD Card")
        self.homeMngr1 = SaveLogsManager(curBaseDir, "Internal SD Card")
        gridSaveFiles.add_widget(self.homeMngr.layout, 1)
        gridWLogFiles.add_widget(self.homeMngr1.layout, 1)

    def exportAllLogs(self, obj=None):
        del obj
        if len(self.usbFileSelectors_1) == 0: return
        curBaseDir = globalVars.logCtx.baseDir if len(globalVars.logCtx.baseDir) else "./"
        a = self.usbFileSelectors_1[0][2]
        assert isinstance(a, LogFileSelector)
        if not curBaseDir.endswith("/"): curBaseDir += "/"
        lstFiles = FilterFilesAndOld(os.listdir(curBaseDir), "log")
        for fName in lstFiles:
            shutil.copy2(curBaseDir + fName, a.baseDir)
        a.reload()

    def exportCrashLogs(self, obj=None):
        del obj
        if len(self.usbFileSelectors_1) == 0: return
        curBaseDir = globalVars.logCtx.baseDir if len(globalVars.logCtx.baseDir) else "./"
        a = self.usbFileSelectors_1[0][2]
        assert isinstance(a, LogFileSelector)
        if not curBaseDir.endswith("/"): curBaseDir += "/"
        lstFiles = filter(
            lambda x: x.startswith("crash-"),
            FilterFilesAndOld(os.listdir(curBaseDir), "log"))
        for fName in lstFiles:
            shutil.copy2(curBaseDir + fName, a.baseDir)
        a.reload()

    def exportNonCrashLogs(self, obj=None):
        del obj
        if len(self.usbFileSelectors_1) == 0: return
        curBaseDir = globalVars.logCtx.baseDir if len(globalVars.logCtx.baseDir) else "./"
        a = self.usbFileSelectors_1[0][2]
        assert isinstance(a, LogFileSelector)
        if not curBaseDir.endswith("/"): curBaseDir += "/"
        lstFiles = filter(
            lambda x: not x.startswith("crash-"),
            FilterFilesAndOld(os.listdir(curBaseDir), "log"))
        for fName in lstFiles:
            shutil.copy2(curBaseDir + fName, a.baseDir)
        a.reload()

    def getName(self):
        return "File Management"

    def addFileManagers(self, app):
        """
        :param AppData app:
        """
        layout = self.gridSaveFiles
        layout1 = self.gridWLogFiles
        layout2 = self.gridLogs
        layout3 = self.gridUpdate
        layout4 = self.gridOS_Update
        h = self.homeMngr
        h1 = self.homeMngr1
        mnt_dir, dev_name, dev_inf = app.usbScreenStorage.usbDirs[0]
        usbTopLabel = "USB(%s)" % dev_inf["LABEL"]
        usbBaseDir = mnt_dir
        if not usbBaseDir.endswith("/"): usbBaseDir += "/"
        # SaveFileManager
        a = SaveFileManager(usbBaseDir, usbTopLabel).add_eject_btn(app)
        a.updateOther(h)
        h.updateOther(a)
        layout.add_widget(a.layout)
        # SaveLogsManager
        b = SaveLogsManager(usbBaseDir, usbTopLabel).add_eject_btn(app)
        b.updateOther(h1)
        h1.updateOther(b)
        layout1.add_widget(b.layout)
        # LogFileSelector
        c = LogFileSelector(usbBaseDir, usbTopLabel, "log").add_eject_btn(app)
        layout2.add_widget(c.layout)
        flLstDir = FileListDir(usbBaseDir, os.listdir(usbBaseDir))
        flLstDir.Narrow(DefFileFilter("tar.gz"))
        # UpdateFileManager
        d = UpdateFileManager(usbBaseDir, usbTopLabel, app, flLstDir).add_eject_btn(app)
        layout3.add_widget(d.layout)
        # OS_UpdateFileManager
        e = OS_UpdateFileManager(usbBaseDir, usbTopLabel, app, flLstDir).add_eject_btn(app)
        layout4.add_widget(e.layout)
        self.usbFileSelectors_1.append((a,b,c,d,e))
    def addDummyWidgets(self, app):
        """
        :param AppData app:
        """
        del app
        layout = self.gridSaveFiles
        layout1 = self.gridWLogFiles
        layout2 = self.gridLogs
        layout3 = self.gridUpdate
        layout4 = self.gridOS_Update
        layout.add_widget(self.dummySave)
        layout1.add_widget(self.dummySave1)
        layout2.add_widget(self.devLogsUSBLabel)
        layout3.add_widget(self.dummyUpd)
        layout4.add_widget(self.dummyOS_Upd)
    def remDummyWidgets(self, app):
        """
        :param AppData app:
        """
        del app
        layout = self.gridSaveFiles
        layout1 = self.gridWLogFiles
        layout2 = self.gridLogs
        layout3 = self.gridUpdate
        layout4 = self.gridOS_Update
        layout.remove_widget(self.dummySave)
        layout1.remove_widget(self.dummySave1)
        layout2.remove_widget(self.devLogsUSBLabel)
        layout3.remove_widget(self.dummyUpd)
        layout4.remove_widget(self.dummyOS_Upd)
    def remFileManagers(self, app, specific=None):
        """
        :param AppData app:
        :param None|str|unicode specific:
        """
        del app
        layout = self.gridSaveFiles
        layout1 = self.gridWLogFiles
        layout2 = self.gridLogs
        layout3 = self.gridUpdate
        layout4 = self.gridOS_Update
        for tup in self.usbFileSelectors_1:
            a, b, c, d, e = tup
            assert isinstance(a, SaveFileManager)
            assert isinstance(b, SaveLogsManager)
            assert isinstance(c, LogFileSelector)
            assert isinstance(d, UpdateFileManager)
            assert isinstance(e, OS_UpdateFileManager)
            if specific is not None:
                lst = map(lambda x: is_path_equal(x.baseDir, specific), [a,b,c,d])
                if not all(lst):
                    assert not any(lst), "all must have same directory"
                    continue
            layout.remove_widget(a.layout)
            h = a.other
            if h is not None:
                h.updateOther(None)
                a.updateOther(None)
            layout1.remove_widget(b.layout)
            h1 = b.other
            if h1 is not None:
                h1.updateOther(None)
                b.updateOther(None)
            layout2.remove_widget(c.layout)
            layout3.remove_widget(d.layout)
            layout4.remove_widget(e.layout)
            if specific is not None:
                self.usbFileSelectors_1.remove(tup)
                break
        if specific is None:
            self.usbFileSelectors_1 = []

    def build(self, app, ext_data=None):
        """
        :param AppData app:
        :param None ext_data:
        """
        app.wnd.add_widget(self.tabPanel)
        self.homeMngr.reload()
        self.homeMngr1.reload()
        if len(app.usbScreenStorage.usbDirs):
            self.remDummyWidgets(app)
            self.addFileManagers(app)
        super(FileManagementScreen, self).build(app)

    def hide(self, app):
        """
        :param AppData app:
        """
        super(FileManagementScreen, self).hide(app)
        app.wnd.remove_widget(self.tabPanel)
        a = len(self.usbFileSelectors_1)
        if a:
            self.remFileManagers(app)
            self.addDummyWidgets(app)

    def getHelp(self, app):
        """

        :param AppData app:
        :rtype: (str|unicode,bool)|None
        """
        return self.helpTxt, False

    def addUsbDrive(self, app):
        """

        :param AppData app:
        :return:
        """
        if len(app.usbScreenStorage.usbDirs) > 1: return
        if not self.showing: return
        if len(self.usbFileSelectors_1) == 0:
            self.remDummyWidgets(app)
        self.addFileManagers(app)

    def remUsbDrive(self, app, drvInfo):
        """

        :param AppData app:
        :param (str|unicode, str|unicode, any) drvInfo:
        :return:
        """
        if not self.showing: return
        mnt_dir, dev_name, dev_inf = drvInfo
        self.remFileManagers(app, mnt_dir)
        if len(self.usbFileSelectors_1) == 0:
            self.addDummyWidgets(app)


class UpdateInfo(object):
    def __init__(self, fName):
        self.fName = fName
        self.name = fName[fName.rfind("/")+1:]
        with tarfile.open(fName, "r:gz") as fl:
            base = None
            for name in ["src", "bin"]:
                try:
                    fl.getmember(name)
                except KeyError:
                    pass
                else:
                    base = name
                    break
            if base is None:
                raise ValueError("invalid update file")
            print "base is", base
            try:
                flObj = fl.extractfile(fl.getmember(base + "/metainf.json"))
                try:
                    self.metainf = json.load(flObj)
                except ValueError:
                    pass
                flObj.close()
            except KeyError:
                self.metainf = None
        end = "update.tar.gz"
        if self.metainf is None and self.name.endswith(end):
            version = self.name[:len(self.name)-len(end)]
            self.metainf = {
                "version": version[version.rfind("v") + 1:],
                "os_pkg_ver": 0}
        elif self.metainf is not None and "version" not in self.metainf:
            raise ValueError("Not a Simple Update File")
        if self.metainf is None:
            self.dispTxt = "<%s>" % self.name
        else:
            self.dispTxt = self.metainf["version"]


class UpdateFileManager(FileSelector):
    mbCheck = None
    mbOk = None
    @classmethod
    def initCls(cls, app):
        """
        :param AppData app:
        """
        cls.mbCheck = app.mbCheck
        cls.mbOk = app.mbOk
    def __init__(self, baseDir1, topLabel, app, flLstDir=None):
        """
        :param str|unicode baseDir1:
        :param str|unicode topLabel:
        :param FileListDir|None flLstDir:
        :param AppData app:
        """
        super(UpdateFileManager, self).__init__(
            baseDir1, topLabel, "tar.gz",
            lambda x: (x.dispTxt, True), flLstDir)
        self.theList.allow_empty_selection = False
        self.actBtn = Button(
            text="Update", pos_hint={"center_x": .5, 'y': .1},
            size_hint=(.25, .1))
        self.actBtn.bind(on_release=lambda obj, *args: self.tryActionIt(obj, *args))
        self.layout.add_widget(self.actBtn)
        self.infLbl = sizedLabel(
            halign='left', valign='top',
            text="Info:", pos_hint={"center_x":.5, "top": .93}, size_hint=(.8,.18))
        self.layout.add_widget(self.infLbl)
        # noinspection PyArgumentList
        self.adapter.bind(selection=self.onChgSel)
        self.app = app

    def onChgSel(self, obj, selection):
        del obj
        if len(selection) > 0:
            updInf = self.lstOpts[selection[0].index]
            assert isinstance(updInf, UpdateInfo)
            self.infLbl.text = "Info:\n  Required OS version: %u\n  Filename: %s" % (updInf.metainf["os_pkg_ver"], updInf.name)

    def tryActionIt(self, obj, *args):
        del obj, args
        if len(self.adapter.selection) == 0:
            self.mbOk.show(
                "Select Update",
                "Select which update you want to update to.",
                lambda x: None
            )
            return
        selUpdInfo = self.lstOpts[self.adapter.selection[0].index]
        assert isinstance(selUpdInfo, UpdateInfo)
        if selUpdInfo.metainf is not None:
            if selUpdInfo.metainf["os_pkg_ver"] > globalVars.metainf["os_pkg_ver"]:
                self.mbOk.show(
                    "OS Package Outdated",
                    "In order to update to this version, you need to have OS package version %u or later" % (
                        selUpdInfo.metainf["os_pkg_ver"]), lambda x: None)
                return
        text, updPath = selUpdInfo.dispTxt, selUpdInfo.fName
        homePath = globalVars.logCtx.baseDir
        if not homePath.endswith("/"): homePath += "/"
        homePath += ".."
        homePath = os.path.abspath(homePath).replace("\\", "/")
        updaterPath = homePath + "/Updater.py"
        startPath = homePath + "/Start.py"
        if not os.path.isfile(updaterPath):
            updaterPath += "c"
        if not os.path.isfile(startPath):
            startPath += "c"
        assert os.path.isfile(updaterPath)
        assert os.path.isfile(startPath)
        exeFile = exec_path.replace("\\", "/")
        if os.name == "nt":
            pos = exeFile.rfind("/")
            if pos != -1:
                exeFile = exeFile[:pos+1]
                exeFile += "python.exe"
        def doIt(mbRtn):
            if mbRtn == IDYES:
                self.actionIt([exeFile, updaterPath, updPath, homePath, exec_path, startPath])
                #self.actionIt([exeFile, updaterPath, updPath, homePath, "sudo", "reboot"])
        self.mbCheck.show(
            "Confirm Update",
            "Are you sure you want to update from %s to %s?" % (globalVars.version, text), doIt)
    def actionIt(self, args):
        selfDir = self.baseDir
        if selfDir.endswith("/"):
            selfDir = selfDir[:-1]
        mnt_dir, dev_name, dev_inf = None, None, None
        for mntDir, devName, devInf in self.app.usbScreenStorage.usbDirs:
            cmpDir = mntDir[:-1] if mntDir.endswith("/") else mntDir
            if cmpDir == selfDir:
                mnt_dir, dev_name, dev_inf = mntDir, devName, devInf
                break
        globalVars.execExit = args
        globalVars.execExit[4:4] = ["--wait", str(os.getpid())]
        if mnt_dir is not None:
            globalVars.execExit[4:4] = ["--umount", mnt_dir]
        log.Log("Updating: args=%r"%globalVars.execExit, 3)
        # don't dismount here (updater does that)
        self.app.usbScreenStorage.usbHandler.process_rem_no_umount(None, dev_name, ImmediateAsynch)
        self.app.stopApp()
    def reload(self):
        raise NotImplementedError("Not Implemented")
    def mkLstOpts(self):
        # lst = super_call
        #rtn = [None] * len(lst)
        #fmt = self.baseDir
        #if not fmt.endswith("/"): fmt += "/"
        #fmt += "%%s.%s" % self.ext
        #for c in xrange(len(lst)):
        #    try:
        #        rtn[c] = UpdateInfo(fmt%lst[c])
        #    except ValueError:
        #        rtn[c] = None
        #return filter(None, rtn)
        return self.flLstDir.Filter(UpdateFileFilter(UpdateInfo))


class UpdateFileFilter(BaseFileFilter):
    def __init__(self, cls):
        """
        :param (str|unicode) -> T cls:
        """
        super(UpdateFileFilter, self).__init__()
        self.cls = cls
    def Match(self, Dir, fName):
        """
        :param str|unicode Dir:
        :param str|unicode fName:
        :rtype: T|None
        """
        if not Dir.endswith("/"): Dir += "/"
        Path = Dir + fName
        try:
            return self.cls(Path)
        except ValueError:
            return None
        except StandardError:
            log.Log("Unrecognized error: " + traceback.format_exc()[:-1], 2)
            return None


class OS_UpdateInfo(object):
    def __init__(self, fName):
        self.fName = fName
        self.name = fName[fName.rfind("/")+1:]
        with tarfile.open(fName, "r:gz") as fl:
            try:
                flObj = fl.extractfile(fl.getmember("metainf.json"))
                try:
                    self.metainf = json.load(flObj)
                except ValueError:
                    pass
                flObj.close()
            except KeyError:
                raise ValueError("Not a valid OS_update file (missing metainf.json)")
        self.dispTxt = "OS Version: %u" % self.metainf["os_pkg_ver"]


class OS_UpdateFileManager(FileSelector):
    mbCheck = None
    mbOk = None

    @classmethod
    def initCls(cls, app):
        """
        :param ProgramItApp app:
        """
        cls.mbCheck = app.mbCheck
        cls.mbOk = app.mbOk

    def __init__(self, baseDir1, topLabel, app, flLstDir=None):
        """
        :param str|unicode baseDir1:
        :param str|unicode topLabel:
        :param AppData app:
        """
        super(OS_UpdateFileManager, self).__init__(baseDir1, topLabel, "tar.gz", lambda x: (x.dispTxt, True), flLstDir)
        self.theList.allow_empty_selection = False
        self.actBtn = Button(
            text="Update", pos_hint={"center_x": .5, 'y': .1},
            size_hint=(.25, .1))
        self.actBtn.bind(on_release=lambda obj, *args: self.tryActionIt(obj, *args))
        self.layout.add_widget(self.actBtn)
        self.infLbl = sizedLabel(
            halign='left', valign='top',
            text="Info:", pos_hint={"center_x":.5, "top": .93}, size_hint=(.8,.18))
        self.layout.add_widget(self.infLbl)
        # noinspection PyArgumentList
        self.adapter.bind(selection=self.onChgSel)
        self.app = app

    def onChgSel(self, obj, selection):
        del obj
        if len(selection) > 0:
            updInf = self.lstOpts[selection[0].index]
            assert isinstance(updInf, OS_UpdateInfo)
            self.infLbl.text = "Info:\n  Filename: %s" % updInf.name

    def tryActionIt(self, obj, *args):
        del obj, args
        selUpdInfo = self.lstOpts[self.adapter.selection[0].index]
        assert isinstance(selUpdInfo, OS_UpdateInfo)
        if selUpdInfo.metainf is not None:
            if selUpdInfo.metainf["os_pkg_ver"] < globalVars.metainf["os_pkg_ver"]:
                self.mbOk.show(
                    "Cannot Downgrade OS",
                    "The current OS version is newer than %u" % (
                        selUpdInfo.metainf["os_pkg_ver"]), lambda x: None)
                return
        text, updPath = selUpdInfo.dispTxt, selUpdInfo.fName
        homePath = globalVars.logCtx.baseDir
        if not homePath.endswith("/"): homePath += "/"
        homePath += ".."
        homePath = os.path.abspath(homePath).replace("\\", "/")
        extraction_point = os.environ.get("HOME", os.environ.get("USERPROFILE", None)).replace("\\", "/")
        extraction_point = "/".join([extraction_point, "OS_Update_Extraction"])
        if not os.path.isdir(extraction_point):
            os.mkdir(extraction_point)
        updaterPath = homePath + "/OS_Updater.py"
        startPath = homePath + "/Start.py"
        if not os.path.isfile(updaterPath):
            updaterPath += "c"
        if not os.path.isfile(startPath):
            startPath += "c"
        assert os.path.isfile(updaterPath), "updaterPath must be file: %s" % updaterPath
        assert os.path.isfile(startPath), "startPath must be file: %s" % startPath
        assert os.path.isdir(extraction_point), "extraction_point must be a directory: %s" % extraction_point
        exeFile = exec_path.replace("\\", "/")
        if os.name == "nt":
            pos = exeFile.rfind("/")
            if pos != -1:
                exeFile = exeFile[:pos + 1]
                exeFile += "python.exe"

        def doIt(mbRtn):
            if mbRtn == IDYES:
                self.actionIt([
                    "sudo", exeFile, updaterPath, updPath, extraction_point,
                    "sudo", "-u", "#%u"%os.getuid(), exec_path, startPath])
                # self.actionIt([exeFile, updaterPath, updPath, homePath, "sudo", "reboot"])

        self.mbCheck.show(
            "Confirm Update",
            "Are you sure you want to update from %s to %s?" % (
                globalVars.metainf["os_pkg_ver"], selUpdInfo.metainf["os_pkg_ver"]), doIt)

    def actionIt(self, args):
        selfDir = self.baseDir
        if selfDir.endswith("/"):
            selfDir = selfDir[:-1]
        mnt_dir, dev_name, dev_inf = None, None, None
        for mntDir, devName, devInf in self.app.usbScreenStorage.usbDirs:
            cmpDir = mntDir[:-1] if mntDir.endswith("/") else mntDir
            if cmpDir == selfDir:
                mnt_dir, dev_name, dev_inf = mntDir, devName, devInf
                break
        globalVars.execExit = args
        globalVars.execExit[5:5] = ["--wait", str(os.getpid())]
        if mnt_dir is not None:
            globalVars.execExit[5:5] = ["--umount", mnt_dir]
        log.Log("Updating: args=%r" % globalVars.execExit, 3)
        # don't dismount here (OS_updater does that)
        self.app.usbScreenStorage.usbHandler.process_rem_no_umount(None, dev_name, ImmediateAsynch)
        self.app.stopApp()

    def reload(self):
        raise NotImplementedError("Not Implemented")

    def mkLstOpts(self):
        # blank comment
        #lst = super(OS_UpdateFileManager, self).mkLstOpts()
        #rtn = [None] * len(lst)
        #fmt = self.baseDir
        #if not fmt.endswith("/"): fmt += "/"
        #fmt += "%%s.%s" % self.ext
        #for c in xrange(len(lst)):
        #    try:
        #        rtn[c] = OS_UpdateInfo(fmt % lst[c])
        #    except ValueError:
        #        rtn[c] = None
        #return filter(None, rtn)
        return self.flLstDir.Filter(UpdateFileFilter(OS_UpdateInfo))


class UsbScreenStorage(ScreenStorage):
    __slots__ = ["usbDirs", "usbHandler",]
    """
    :type usbDirs: list[(str|unicode, str|unicode, dict)]
    :type usbHandler: UsbHandler
    """

    def __init__(self, screens, usbHandler):
        """
        :param list[BaseScreen] screens:
        :param UsbHandler usbHandler:
        """
        super(UsbScreenStorage, self).__init__(screens)
        self.usbDirs = []
        self.usbHandler = usbHandler

from .LibAppData import AppData