from .LibIndAppData import ProgStyleStorage
from .LibBaseScreen import BaseScreenUSB
from .LibMessageBox import MessageBox
from .LibPopup import SelectPopupMaker
from .Logging import Logger
from kivy.uix.floatlayout import FloatLayout
from .LibIndAppData import preLoadCheck, preSaveCheck
from .TinyLib import is_path_equal, mkNormPath
from os.path import exists, isfile, isdir
from os import mkdir
import subprocess


class AppData(object):
    __slots__ = [
        "defStyle", "SpiChains", "wnd", "mbCheck", "mbOk", "stopApp",
        "usbEjectPopup", "usbEjectFailPopup", "lstColW", "progStyle",
        "lstProgs", "logObj", "usbScreenStorage", "baseDir", "saveItCB",
        "loadItCB", "curProjName", "chainProgramIt"
    ]
    """
    :type curProjName: str|unicode
    """

    def __init__(self, defStyle, SpiChains, wnd, mbCheck, mbOk, stopApp, usbEjectPopup, usbEjectFailPopup, lstColW, progStyle, lstProgs, logObj, usbScreenStorage, baseDir, saveItCB, loadItCB, chainProgramIt):
        """
        :param defStyle: dict
        :param dict[(int, int), list] SpiChains:
        :param FloatLayout|None wnd:
        :param MessageBox|None mbCheck:
        :param MessageBox|None mbOk:
        :param stopApp:
        :param SelectPopupMaker|None usbEjectPopup:
        :param SelectPopupMaker|None usbEjectFailPopup:
        :param list[float|int|long|None]|None lstColW:
        :param ProgStyleStorage progStyle:
        :param list lstProgs:
        :param Logger logObj:
        :param UsbScreenStorage|None usbScreenStorage:
        :param str|unicode baseDir:
        :param saveItCB:
        :param loadItCB:
        :param chainProgramIt:
        """
        self.defStyle = defStyle
        self.SpiChains = SpiChains
        self.wnd = wnd
        self.mbCheck = mbCheck
        self.mbOk = mbOk
        self.stopApp = stopApp
        self.usbEjectPopup = usbEjectPopup
        self.usbEjectFailPopup = usbEjectFailPopup
        self.lstColW = lstColW
        self.progStyle = progStyle
        self.lstProgs = lstProgs
        self.logObj = logObj
        self.usbScreenStorage = usbScreenStorage
        self.baseDir = baseDir
        self.saveItCB = saveItCB
        self.loadItCB = loadItCB
        self.chainProgramIt = chainProgramIt
        self.curProjName = ""

    def setCurProjName(self, v):
        """
        :param str|unicode v:
        """
        self.curProjName = v

    def BaseDirFileChooser(self, isSave):
        args = (
            ["Save File To Internal SD card", "Save To File", self.saveItCB, preSaveCheck]
            if isSave else
            ["Load File From Internal SD card", "Load From File", self.loadItCB, preLoadCheck])
        return FileChooser(
            args[0], args[1], args[2],
            self.baseDir if len(self.baseDir) else "./", self.curProjName,
            self.setCurProjName, args[3], "xmdat")

    def getUsbFileChooser(self, baseDir1, label, isSave):
        chgArgs = (
            ("Save File To USB (%s)" % label, "Save To File", self.saveItCB, preSaveCheck)
            if isSave else
            ("Load File From USB (%s)" % label, "Load From File", self.loadItCB, preLoadCheck))
        return FileChooser(
            chgArgs[0], chgArgs[1], chgArgs[2], baseDir1, self.curProjName,
            self.setCurProjName, chgArgs[3], "xmdat", True).add_eject_btn(
            self, size_hint_y=.13)

    def tryMountUsbDrive(self, dev_name, mnt_dir, dev_inf, args):
        rtn = (0, "")
        if not exists(mnt_dir):
            mkdir(mnt_dir)
        elif not isdir(mnt_dir):
            rtn = (1, "'%s' exists but is not directory\n" % mnt_dir)
        if rtn[0] != 0:
            return rtn
        try:
            rtn = (0, subprocess.check_output(args))
        except subprocess.CalledProcessError as Exc:
            return Exc.returncode, Exc.output
        self.addUsbDrive(mnt_dir, dev_name, dev_inf)
        return rtn

    def tryDismountDrive(self, dev_name, mnt_dir, dev_inf, args, success_check=None):
        try:
            rtn = (0, subprocess.check_output(args))
        except subprocess.CalledProcessError as Exc:
            rtn = (Exc.returncode, Exc.output)
        if success_check is None or success_check(rtn):
            self.remUsbDrive(mnt_dir, dev_name, dev_inf)
        return rtn

    def addUsbDrive(self, mnt_dir, dev_name, dev_inf):
        scrnStor = self.usbScreenStorage
        scrnStor.usbDirs.append((mnt_dir, dev_name, dev_inf))
        if scrnStor.curScreenId:
            scrn = scrnStor.screens[scrnStor.curScreenId]
            assert isinstance(scrn, BaseScreenUSB)
            scrn.addUsbDrive(self)
            return
        # label = dev_inf["LABEL"]
        # TODO: finish?

    def remUsbDrive(self, mnt_dir, dev_name, dev_inf):
        scrnStor = self.usbScreenStorage
        c = 0
        while c < len(scrnStor.usbDirs):
            mntDir, devName, devInf = scrnStor.usbDirs[c]
            if is_path_equal(mntDir, mnt_dir) and devName == dev_name:
                scrnStor.usbDirs.pop(c)
            else:
                c += 1
        if scrnStor.curScreenId:
            scrn = scrnStor.screens[scrnStor.curScreenId]
            if isinstance(scrn, BaseScreenUSB):
                scrn.remUsbDrive(self, (mnt_dir, dev_name, dev_inf))
            return

    def addWinUsbDrive(self, mnt_dir):
        mnt_dir = mkNormPath(mnt_dir)
        self.addUsbDrive(mnt_dir, mnt_dir.split(':',1)[0], {"LABEL": mnt_dir})

    def remWinUsbDrive(self, mnt_dir):
        mnt_dir = mkNormPath(mnt_dir)
        self.remUsbDrive(mnt_dir, mnt_dir.split(':',1)[0], {"LABEL": mnt_dir})


from .FileManagement import FileChooser, UsbScreenStorage # import for type hint