from . import LibIO
from .LibIO import IO_SPI, IO_PIN, IO_NONE, spiIO_Object, PinIO_Object, enumeratePins, getPinNumFromBCM_ID, getSpiBusInfo
from . import globalVars
globalVars.log.Log("Importing PartCats", 3)
from .PartCats import *
globalVars.log.Log("Importing ScatterLayout", 3)
from kivy.uix.scatterlayout import ScatterLayout
globalVars.log.Log("Importing Matrix", 3)
from kivy.graphics.transformation import Matrix
globalVars.log.Log("Importing LibGetTabPanel", 3)
from .LibGetTabPanel import *
globalVars.log.Log("Importing LibAppData", 3)
from .LibAppData import AppData


Hittite = "Analog Devices"
defGreen = dict(defTxtStyle, color=kXM_GREEN)


# makes the text vertically fill the widget
def fillBtnTxt(inst, val):
    del val
    inst.font_size = 10 + .5 * inst.size[1]


def numShiftSlice(slc, num, sz):
    if not isinstance(slc, slice):
        return [(slc, num)]
    lstBits = xrange(*slc.indices(sz))
    return [(lstBits[x], bool(num & 2 ** x)) for x in xrange(len(lstBits))]


def spiPulseData(spiBus, lstVals):
    spiBus.cshigh = False
    spiBus.writebytes(lstVals)
    spiBus.cshigh = True

class ChainSpiBus(object):
    def __init__(self):
        self.bits = bitMapInt(0, 0)

    def addNum(self, num, nBits):
        self.bits += (num, nBits)


def spiWriteNum(num, nBits, spiBus):
    if isinstance(spiBus, ChainSpiBus):
        spiBus.addNum(num, nBits)
        return
    Len = (nBits + 7) / 8
    rtn = [0] * Len
    tmp = num
    lsbfirst = spiBus.lsbfirst
    # get a list of the bytes to send
    for c in xrange(Len):
        c = Len - c - 1
        rtn[c] = tmp & 0xFF
        # if lsbfirst: swap the order of the bits in each byte
        #   each byte still is kept in the same order however
        if lsbfirst: rtn[c] = bitMapInt(rtn[c], 8)[::-1]
        tmp >>= 8
    # make sure all the bytes are ints (spidev will fail with longs)
    rtn = map(int, rtn)
    spiPulseData(spiBus, rtn)


def SimpleBitComp(lstStartEnds, bitLen, *args):
    num = 0
    for c, startEnd in enumerate(lstStartEnds):
        start, end, off = None, bitLen, 0
        if isinstance(startEnd, (tuple, list)):
            start = startEnd[0]
            if len(startEnd) >= 2:
                end = startEnd[1]
            if len(startEnd) >= 3:
                off = startEnd[2]
        else:
            start = startEnd
            end = start + 1
        Mask = (1 << (end - start)) - 1
        Mask <<= off
        num |= ((args[c] & Mask) >> off) << start
    return num


def lstBoxChg(theObj, inst, val):
    del inst
    theObj.onPreChg(theObj, val)
    curVal = int(val / theObj.lsbVal)
    for c in xrange(len(theObj.data)):
        theObj.data[c] = int(curVal % 2)
        curVal /= 2
    theObj.mainBtn.text = bitsToTotal(theObj)
    theObj.onChgAny(theObj)

SpecialAllocDct = {}


custom_name_packer = DataPacker(
    DataVarBytes(1)
)


# Base Programmable Part
class BaseProgPart(object):
    XM_partName = "NONE"
    mfgPartName = "None"
    _custom_name = ""
    @property
    def custom_name(self):
        return self._custom_name
    @custom_name.setter
    def custom_name(self, v):
        if v == self._custom_name:
            return
        self.on_chg_custom_name(v)
        self._custom_name = v
    def set_custom_name(self, v):
        if v == self._custom_name:
            return
        self.on_chg_custom_name(v)
        self._custom_name = v

    def on_chg_custom_name(self, v):
        """
        called right before the change is made so that
        self.custom_name contains the old value and
        v contains the new one
        :param str|unicode v:
        """
        pass
    userHelp = None
    DevOnly = False
    catId = CAT_NONE
    ioType = IO_SPI
    mfgName = "None"
    interface = "None"
    TopInfLbls = ["X-MW PartNumber", "Manufacturer Part Number", "Manufacturer"]
    brdImgW = .08
    genInfW = .75
    genInfH = .08 * .8
    genInfX = .055
    boardImageName = None
    allowChain = False
    spiBus = None
    nPins = 0
    brdImgScatter = None
    _app = None
    @property
    def app(self):
        return self._app
    @app.setter
    def app(self, v):
        assert isinstance(v, AppData)
        self._app = v
    genInf = None

    @staticmethod
    def Allocate(inst):
        mainName = inst.XM_partName
        mainName = mainName[mainName.find("-") + 1:mainName.rfind("-")]
        if LibIO.isMultiIO(inst.ioType):
            multiArgs = SpecialAllocDct[(inst.ioType, mainName)]
            if inst.ioType == IO_PIN:
                assert inst.nPins == len(multiArgs)
            ioCls = LibIO.dctClsIO[inst.ioType]
            return [ioCls(*args) for args in multiArgs]
        else:
            args = SpecialAllocDct[(inst.ioType, mainName)]
            return LibIO.dctClsIO[inst.ioType](*args)

    @staticmethod
    def SpecialAlloc(inst):
        mainName = inst.XM_partName
        mainName = mainName[mainName.find("-") + 1:mainName.rfind("-")]
        return (inst.ioType, mainName) in SpecialAllocDct

    @classmethod
    def initCls(cls, numBuses):
        pass

    def getBoardImg(self, **kwargs):
        if self.boardImageName is None: return None
        content = FloatLayout(
            size_hint=(1., 1.), pos_hint={'x': 0.0, "top": 1.0})
        popup = Popup(
            title="Board Layout Image", content=content, size_hint=(.9, .9),
            color=(.5, .5, .5, 1.),
            **Filter(defPopup, InListRange(["title", "size_hint"], 1)))
        Img = Image(
            source=self.boardImageName, size_hint=(1., .9))
        scatter = ScatterLayout(
            do_rotation=True, do_translation=True, do_scale=True)
        scatter.add_widget(Img)
        content.add_widget(scatter)
        # content.add_widget(Img)
        content.add_widget(Button(
            on_release=popup.dismiss, text="Close", size_hint=(.2, .1),
            pos_hint={'x': 0.0, 'y': 0.0}))
        scatter.transform = Matrix().translate(52, 75, 0)
        self.brdImgScatter = scatter
        origTransform = scatter.transform
        content.add_widget(Button(
            on_release=lambda inst: (
                setattr(scatter, "transform", origTransform)),
            text="Reset Position", size_hint=(.4, .1),
            pos_hint={'x': .2, 'y': 0.0}))
        content.add_widget(Button(
            on_release=lambda inst: (
                setattr(scatter, "rotation", (scatter.rotation + 90) % 360)),
            text="Rotate 90\xc2\xb0", size_hint=(.4, .1),
            pos_hint={'x': .6, 'y': 0.0}))
        return ClkImage(
            on_release=lambda obj: popup.open(),
            source=self.boardImageName, **kwargs)

    def getDispCols(self):
        port, cs = "", ""
        if self.ioType == IO_SPI:
            port, cs = getSpiBusInfo(self.spiBus)
            multi = self.spiBus.multi
            spiInf = (port, cs) if multi is None else (port, cs, multi)
            port = "SPI " + str(port)
            cs = "LE " + str(cs)
            if hasattr(self, "app") and self.app is not None and self in self.app.SpiChains[spiInf]:
                port += "\nChained"
                cs += "\nIndex:%u" % self.app.SpiChains[spiInf].index(self)
            if multi is not None:
                port = "Muxed\n" + port
        elif hasattr(self, "pin"):
            port, cs = getSpiBusInfo(self.pin)
            port = "SPI " + str(port)
            cs = "LE " + str(cs)
        return [
            globalVars.partCats[self.catId].abrev,
            "%s\n(%s)" % (self.custom_name, self.XM_partName) if len(self.custom_name) else self.XM_partName,
            self.mfgName,
            self.mfgPartName, port, cs]

    def setSpiBus(self, spiBus):
        if self.spiBus is not None and self.spiBus is not spiBus and isinstance(self.spiBus, spiIO_Object):
            self.spiBus.close()
        self.spiBus = spiBus
    def setPins(self, pins):
        if self.pins is not None and len(self.pins) > 0:
            for pin in self.pins:
                if pin not in pins:
                    pin.close()
        self.pins = pins
    def setSinglePin(self, pinIndex, pinObj):
        oldPin = self.pins[pinIndex]
        self.pins[pinIndex] = pinObj
        if oldPin not in self.pins:
            oldPin.close()
    def getPinNames(self):
        return map(str, xrange(len(self.pins)))

    # noinspection PyUnusedLocal
    def __init__(self, fl=None, ver=0):
        self.beenProg = False
        self.progBtn = ValuePack()
        self.chainLbl = ValuePack()
        if fl is not None:
            print("loading version: %u" % ver)
            ioObj = LibIO.deserializeIO(fl)
            if isinstance(ioObj, LibIO.BaseIO_Object):
                self.ioType = ioObj.ioType
            elif len(ioObj):
                self.ioType = ioObj[0].ioType
            if self.ioType == IO_NONE:
                globalVars.log.Log("Usage of ioType=IO_NONE for class=%s" % self.__class__.__name__, 2)
            elif self.ioType == IO_SPI:
                self.spiBus = ioObj
            elif self.ioType == IO_PIN:
                self.pins = ioObj
            if ver >= 3:
                self.custom_name = custom_name_packer.Unpack(fl)[0]
        elif self.SpecialAlloc(self):
            self.setSpiBus(self.Allocate(self))
        elif self.ioType == IO_SPI:
            self.setSpiBus(spiIO_Object(*getSpiBusInfo(globalVars.spi_obj[1])))
        elif self.ioType == IO_PIN:
            # if self.nPins != 1: log.Log("Only 1 pin is allowed right now %s"%self.__class__.__name__, 2)
            self.pins = [PinIO_Object(c) for c in enumeratePins(getPinNumFromBCM_ID)[:max(1, self.nPins)]]
        tmp = self.genInfX + self.genInfW - self.brdImgW
        self.brdImg = self.getBoardImg(
            pos_hint={'x': tmp, "top": .99}, size_hint=(self.brdImgW, self.genInfH))

    def getIO_Obj(self):
        if self.ioType == IO_SPI:
            return self.spiBus
        elif self.ioType == IO_PIN:
            return self.pins
        else:
            return None

    def serialize(self, fl, ver):
        print("saving version: %u" % ver)
        LibIO.serializeIO(self.ioType, self.getIO_Obj(), fl)
        if ver >= 3:
            custom_name_packer.Pack([self.custom_name], fl)

    def destroy(self, app):
        assert isinstance(app, AppData)
        if self.ioType == IO_SPI and self.allowChain:
            spiChain = app.SpiChains[tuple(getSpiBusInfo(self.spiBus))]
            if self in spiChain: spiChain.remove(self)

    def onhide(self, app):
        assert isinstance(app, AppData)
        self.progBtn = ValuePack()
        if self.allowChain:
            self.chainLbl = ValuePack()

    def build(self, app):
        assert isinstance(app, AppData)
        # app.topMenu()
        # noinspection PyArgumentList
        layout = SnappedFloatLayout(**app.defStyle)
        self.app = app
        #szLbl = .1
        genInfStyle = dict(
            cols=3, rows=2, pos_hint={'x': self.genInfX, "top": .99}, bdr=8)
        if self.brdImg is None:
            self.genInf = BorderedGridLayout(
                size_hint=(self.genInfW, self.genInfH), **genInfStyle)
        else:
            self.genInf = BorderedGridLayout(
                size_hint=(self.genInfW - self.brdImgW, self.genInfH), **genInfStyle)
            app.wnd.add_widget(self.brdImg)
        app.wnd.add_widget(self.genInf)
        partInf = self.dispGeneral()
        green = kXM_GREEN
        self.genInf.add_widget(Label(text=partInf[0][0]))
        self.genInf.add_widget(Label(text=partInf[1][0]))
        self.genInf.add_widget(Label(text=partInf[2][0]))
        self.genInf.add_widget(Label(text=partInf[0][1], color=green))
        self.genInf.add_widget(Label(text=partInf[1][1], color=green))
        self.genInf.add_widget(Label(text=partInf[2][1], color=green))
        # btnStyle = {"color":k36_C9_RED}
        btnStyle = {"color": kRED}
        if self.beenProg: btnStyle = defGreen
        btnColor = btnStyle["color"]
        self.progBtn = Button(
            text="Write", pos_hint={'x': 0.1, 'y': 0.35},
            size_hint=(.175, .1), color=btnColor, bold=True)
        self.progBtn.on_release = self.programIt
        self.progBtn.bind(size=fillBtnTxt)
        if self.allowChain:
            self.chainLbl = Label(
                text="Chain Position: " + self.getDispChained(),
                size_hint=(.1, .05),
                pos_hint={"center_x": .5, "top": 1.0},
                font_size=24)
            layout.add_widget(self.chainLbl)
        layout.add_widget(self.progBtn)
        app.wnd.add_widget(layout)
        return layout

    def getDispChained(self):
        if not hasattr(self, "app"): return "UNKNOWN"
        spiInf = getSpiBusInfo(self.spiBus)
        try:
            return str(self.app.SpiChains[spiInf].index(self))
        except ValueError:
            return "not set"

    def programIt(self, obj=None):
        self.beenProg = True
        self.progBtn.color = defGreen["color"]
        # for k in defGreen: setattr(self.progBtn, k, defGreen[k])

    def onProgChg(self):
        self.beenProg = False
        self.progBtn.color = kRED
        # self.progBtn.background_color = kStrRGB("C93636", 8)
        # self.progBtn.background_color = kRED
        # self.progBtn.background_normal = "./btnImg.png"
        # for k in defTxtStyle: setattr(self.progBtn, k, defTxtStyle[k])

    def dispName(self):
        return self.XM_partName

    def dispGeneral(self):
        return zip(self.TopInfLbls,
                   (self.XM_partName, self.mfgPartName, self.mfgName), )

    def executeCMD(self, s):
        if s.lower() == "help":
            return "S" + "\n".join([
                "list of commands",
                "  write  -  writes the data to the device",
                "  help  -  shows this help"
            ])
        elif s.lower() == "write":
            self.programIt(None)
            return "S" + "wrote data"
        else:
            return "E" + "unrecognized command, type help to see list of commands"

class TopCustomNameStorage(object):
    __slots__ = ["lbl", "txt_in", "edit_btn"]
    def __init__(self, prog, txt_in=None, edit_btn=None, lbl_sz_hint=(0.13, 0.07), txt_sz_hint=(0.6, 0.07), btn_sz_hint=(0.07, 0.07), top_hint=1.08, pos_x_hint=0.1):
        """
        :param TextInput|None txt_in:
        :param Button|None edit_btn:
        :param BaseProgPart prog:
        :param (float, float) lbl_sz_hint:
        :param (float, float) txt_sz_hint:
        :param (float, float) btn_sz_hint:
        :param float top_hint:
        :param float pos_x_hint:
        """
        self.lbl = Label(text="Custom Name")
        if txt_in is None:
            txt_in = TextInput(
                disabled=True,
                text=prog.custom_name
            )
        if edit_btn is None:
            edit_btn = Button(text="edit")
        self.txt_in = txt_in
        self.edit_btn = edit_btn
        self.lbl.pos_hint = {
            "top": top_hint,
            "x": pos_x_hint
        }
        self.lbl.size_hint = lbl_sz_hint
        pos_x_hint += lbl_sz_hint[0]
        txt_in.multiline = False
        txt_in.pos_hint = {
            "top": top_hint,
            "x": pos_x_hint
        }
        pos_x_hint += txt_sz_hint[0]
        txt_in.size_hint = txt_sz_hint
        edit_btn.pos_hint = {
            "top": top_hint,
            "x": pos_x_hint
        }
        edit_btn.size_hint = btn_sz_hint
        txt_in.bind(text=lambda obj, v: prog.set_custom_name(v))
        edit_btn.bind(on_release=Bind(TopCustomNameStorage.enable_editing, self))
        txt_in.bind(on_enter=Bind(TopCustomNameStorage.disable_editing, self))

    def enable_editing(self, obj):
        del obj
        self.txt_in.disabled = False

    def disable_editing(self, obj):
        del obj
        self.txt_in.disabled = True

    def remove(self, layout):
        """
        :param Layout layout:
        """
        layout.remove_widget(self.lbl)
        layout.remove_widget(self.txt_in)
        layout.remove_widget(self.edit_btn)

    def add(self, layout):
        """
        :param Layout layout:
        """
        layout.add_widget(self.lbl)
        layout.add_widget(self.txt_in)
        layout.add_widget(self.edit_btn)


# write only list view of bits for a dual Register
#  that represents the non-dual part
class WriteView(object):
    def __init__(self, reg):
        """
        :param Register reg:
        """
        self.reg = reg
        dualBit = reg.dualBit
        lstDualBits = None
        if dualBit == -1:
            pass
        elif isinstance(dualBit, slice):
            lstDualBits = xrange(*reg.dualBit.indices(reg.word.bitLen))
        else:
            lstDualBits = [dualBit]
        self.dualMask = 0 if lstDualBits is None else reduce(lambda tot, cur: tot | (1 << cur), lstDualBits, 0)

    def __setitem__(self, index, val):
        opt = 0
        if isinstance(index, slice): opt |= 1
        if isinstance(self.reg.dualBit, slice): opt |= 2
        if opt == 0:
            if index == self.reg.dualBit:
                raise KeyError("cannot set the dual bit on both words")
        elif opt == 1:
            lstBits = sorted(xrange(*index.indices(len(self.reg))))
            if self.reg.dualBit in lstBits:
                raise KeyError("cannot set the dual bit on both words")
        elif opt == 3:
            lstBits = sorted(xrange(*index.indices(len(self.reg))))
            lstDual = sorted(xrange(*index.indices(len(self.reg))))
            for c in lstBits:
                if c in lstDual:
                    raise KeyError("cannot set the dual bit on both words")
        elif opt == 2:
            lstDual = sorted(xrange(*index.indices(len(self.reg))))
            if index in lstDual:
                raise KeyError("cannot set the dual bit on both words")
        orig = self.reg.curReg
        for c in xrange(self.reg.numDual):
            self.reg.curReg = c
            self.reg[index] = val
        self.reg.curReg = orig

    def __getitem__(self, index):
        opt = 0
        if isinstance(index, slice): opt |= 1
        if isinstance(self.reg.dualBit, slice): opt |= 2
        if opt == 0:
            if index == self.reg.dualBit:
                raise KeyError("cannot get the dual bit on both words")
        elif opt == 1:
            lstBits = sorted(xrange(*index.indices(len(self.reg))))
            if self.reg.dualBit in lstBits:
                raise KeyError("cannot get the dual bit on both words")
        elif opt == 3:
            lstBits = sorted(xrange(*index.indices(len(self.reg))))
            lstDual = sorted(xrange(*index.indices(len(self.reg))))
            for c in lstBits:
                if c in lstDual:
                    raise KeyError("cannot get the dual bit on both words")
        elif opt == 2:
            lstDual = sorted(xrange(*index.indices(len(self.reg))))
            if index in lstDual:
                raise KeyError("cannot get the dual bit on both words")
        orig = self.reg.curReg
        val = 0
        origVal = None
        for c in xrange(self.reg.numDual):
            self.reg.curReg = c
            val = self.reg[index]
            if origVal is not None and origVal != val:
                globalVars.log.Log("different values were stored in register: %u and %u" % (origVal, val), 2)
            origVal = val
        self.reg.curReg = orig
        return val

    def __len__(self):
        return len(self.reg)

    def getNonDualNumberData(self):
        """
        :rtype: (int|long, int|long)
        :returns: (number data, anti-mask)
        """
        num = self.reg.word.num
        dualMask = self.dualMask
        if dualMask:
            num = num | dualMask
            num ^= dualMask
        return num, dualMask

    def changedNumberData(self, origData):
        """
        :param (int|long, int|long) origData:
        :rtype: bool
        """
        return self.reg.changedNumberData(origData)


# Base(and instantiable) Register class for parts with serial latched registers coded by address
class Register(object):
    wordSz = 32
    ctlSz = 3

    def getCurReg(self):
        return self._curReg

    def setCurReg(self, val):
        if val > 0 and not self.isDual:
            raise ValueError("cannot switch to unavailable register (not dual)")
        if hasattr(self, "lbl"): self.dualLstLbl[self._curReg] = self.lbl
        if hasattr(self, "built"): self.dualLstBlt[self._curReg] = self.built
        if self._curReg == val: return
        self._curReg = val
        self.lbl = self.dualLstLbl[self._curReg]
        self.built = self.dualLstBlt[self._curReg]
        self.word = self.dualLst[self._curReg]

    curReg = property(getCurReg, setCurReg)

    def __init__(self, regNum, dualBit=-1):
        """
        :param int|long regNum:
        :param int|slice dualBit:
        """
        self.ctlMsk = (1 << self.ctlSz) - 1
        # self.word
        # self.lbl
        self.numDual = 1
        if dualBit != -1:
            if isinstance(dualBit, slice):
                self.numDual <<= len(xrange(*dualBit.indices(self.wordSz)))
            else:
                self.numDual = 2
        self.isDual = dualBit != -1
        self.dualBit = dualBit
        self.regNum = regNum
        self._curReg = 0
        self.built = False
        self.dualLst = [
            bitMapInt(regNum, self.wordSz) for _ in xrange(self.numDual)]
        self.dualLstLbl = [
            ValuePack({"text": ""}) for _ in xrange(self.numDual)]
        self.dualLstBlt = [
            False for _ in xrange(self.numDual)]
        if self.numDual < 1:
            raise ValueError("Cannot have less than 1 numDual")
        for c in xrange(self.numDual):
            self.lbl = self.dualLstLbl[c]
            self.built = self.dualLstBlt[c]
            self.word = self.dualLst[c]
            self.curReg = c
            for i, v in numShiftSlice(dualBit, c, self.wordSz):
                self.word[i] = v
            self.lbl.text = self.calcText()
        self.curReg = 0
        self.both = WriteView(self)

    def build(self, cls=Label, **kwargs):
        if self.built: return self.lbl
        self.lbl = cls(text=self.lbl.text, **kwargs)
        self.built = True
        return self.lbl

    def calcText(self):
        return "R%u: 0x%s" % (self.regNum, self.toStr((self.wordSz + 3) / 4, False, 16))

    def flipNum(self):
        return self.word.flipNum()

    def progNum(self, num):
        if self.ctlMsk > 0 and (num & self.ctlMsk != self.regNum): return False
        origReg = self.curReg
        if self.isDual:
            if isinstance(self.dualBit, slice):
                lstDualBits = xrange(*self.dualBit.indices(self.word.bitLen))
                tmp = 0
                for pwr, index in enumerate(lstDualBits):
                    tmp |= ((num >> index) & 1) << pwr
                self.curReg = tmp
            else:
                self.curReg = int(bool(num & (1 << self.dualBit)))
        # orig = self.word.num
        self.word.num = num
        self.lbl.text = self.calcText()
        # diffWord = bitMapInt(orig ^ num, len(self.word))
        if self.isDual:
            self.curReg = origReg
        return True

    def progNumBits(self, num, bitRange):
        lstVals = [
            (bitRange[x], bool(num & 2 ** x)) for x in xrange(len(bitRange))]
        lstVals = sorted(lstVals, key=lambda x1: x1[0])
        lstDualBits = [self.dualBit]
        if isinstance(self.dualBit, slice):
            lstDualBits = xrange(*self.dualBit.indices(self.word.bitLen))
        doFirst = [bit for bit in lstVals if bit[0] in lstDualBits]
        doFirst = dict(doFirst)
        tmp = self.curReg
        pwr = 0
        for c in lstDualBits:
            if c in doFirst:
                if bool(doFirst[c]) ^ bool(tmp & (1 << pwr)):
                    tmp ^= 1 << pwr
            pwr += 1
        self.curReg = tmp
        orig = self.word.num
        for bit, val in lstVals:
            if bit not in doFirst: self.word[bit] = val
        if self.word.num ^ orig != 0: self.edited()

    def written(self):
        self.lbl.color = kWHITE

    def edited(self):
        self.lbl.color = kRED

    def __setitem__(self, index, val):
        if not self.isDual:
            orig = self.word.num
            self.word[index] = val
            if orig ^ self.word.num: self.edited()
        elif self.dualBit == index:
            self.curReg = val
            return
        elif not isinstance(index, slice):
            if self.word[index] == val: return
            lstDualBits = [self.dualBit]
            if isinstance(self.dualBit, slice):
                lstDualBits = xrange(*self.dualBit.indices(self.word.bitLen))
            if index in lstDualBits:
                bitVal = 1 << lstDualBits.index(index)
                # use xor because we know that self.word[index] != val
                self.curReg = self.curReg ^ bitVal
            else:
                self.word[index] = val
                self.edited()
        else:
            self.progNumBits(val, xrange(*index.indices(self.word.bitLen)))
        self.lbl.text = self.calcText()

    def __getitem__(self, index):
        return self.word[index]

    def __iter__(self):
        return self.word.__iter__()

    def toStr(self, padZero=0, bigEnd=False, radix=10):
        return self.word.toStr(padZero, bigEnd, radix)

    def flipToStr(self, padZero=0, bigEnd=False, radix=10):
        return self.word.flipToStr(padZero, bigEnd, radix)

    def __len__(self):
        return len(self.word)

    def spiWriteData(self, spiBus):
        spiWriteNum(self.word.num, self.wordSz, spiBus)
        self.written()

    def getNonDualNumberData(self):
        """
        :rtype: (int|long, int|long)
        :returns: (number data, anti-mask)
        """
        return self.word.num, 0

    def changedNumberData(self, origData):
        """
        :param (int|long, int|long) origData:
        :rtype: bool
        """
        numData, antiMask = origData
        curData = self.word.num
        if antiMask:
            curData = (curData | antiMask) ^ antiMask
        return numData != curData

class RegWithoutAddr(Register): # has a R/W bit default 0
    wordSz = 16
    ctlSz = 0
    addrSz = 7
    def __init__(self, addr, dualBit=-1):
        super(RegWithoutAddr, self).__init__(addr, dualBit)
        self[:] = 0
        self.addr = addr
        self.rwBit = 0
    def calcText(self):
        if self.dualBit == -1:
            return "(0x%02X): 0x%s" % (self.regNum, self.toStr((self.wordSz + 3) / 4, False, 16))
        else:
            return "(0x%02X ws %u): 0x%s" % (self.regNum, self.curReg, self.toStr((self.wordSz + 3) / 4, False, 16))

    def spiWriteData(self, spiBus):
        nBits = self.addrSz + 1 + self.wordSz
        num = SimpleBitComp(
            [(0, 16), (16, 17), (17,)], nBits,
            self.word.num, self.rwBit, self.addr)
        spiWriteNum(num, nBits, spiBus)
        self.written()

class RegWithoutAddr1(Register): # has a R/W bit default 0
    wordSz = 16
    ctlSz = 0
    addrSz = 7
    def __init__(self, addr, dualBit=-1):
        super(RegWithoutAddr1, self).__init__(addr, dualBit)
        self[:] = 0
        self.addr = addr
        self.rwBit = 0
    def calcText(self):
        if self.dualBit == -1:
            return "(R%u): 0x%s" % (self.regNum, self.toStr((self.wordSz + 3) / 4, False, 16))
        else:
            return "(R%u ws %u): 0x%s" % (self.regNum, self.curReg, self.toStr((self.wordSz + 3) / 4, False, 16))

    def spiWriteData(self, spiBus):
        nBits = self.addrSz + 1 + self.wordSz
        num = SimpleBitComp(
            [(0, 16), (16, 23), (23,)], nBits,
            self.word.num, self.addr, self.rwBit)
        spiWriteNum(num, nBits, spiBus)
        self.written()


class Reg32_4(Register):
    ctlSz = 4
class Reg24(Register):
    wordSz = 24
    ctlSz = 2
class Reg24_3(Register):
    wordSz = 24
    ctlSz = 3


# Base Hittite Register
class HMCReg(Register):
    wordSz = 24
    ctlSz = 5

    def progNum(self, num):
        origReg = self.curReg
        if self.isDual: self.curReg = int(bool(num & (1 << self.dualBit)))
        # orig = self.word.num
        self.word.num = num
        self.lbl.text = self.calcText()
        # diffWord = bitMapInt(orig ^ num, len(self.word))
        """toCallBack = set()
        for c in xrange(self.wordSz):
            if diffWord[c]: toCallBack |= self.lstBitCbs[c]
        for cb in toCallBack: cb(self, True)"""
        if self.isDual: self.curReg = origReg
        return True

    def spiWriteData(self, spiBus, addr=0):
        nBits = self.wordSz + self.ctlSz + 3 + 7
        spiWriteNum(self.word.num, nBits, spiBus)
        self.written()


def invBitCh(bit):
    return bitCh(int(not bit))


def bitCh(x):
    return chr(x + ord('0'))


def initFunc(moduleName):
    loadData = globalVars.loadData
    if globalVars.useLoadData:
        loadData.curModule = moduleName
        loadData.allModules[moduleName] = {"log": None, "parts": {}}

    def decor(func):
        modLog = globalVars.log.SpawnLog(moduleName)
        if globalVars.useLoadData: loadData.allModules[moduleName]["log"] = modLog
        func(modLog)
        return func

    return decor


def defCb(v):
    return v


def initPartClsBase(cls):
    if globalVars.dctParts[cls.XM_partName] is not cls:
        raise ValueError("XM_partName must equal key name for part class")
    partManufs = globalVars.partManufs
    globalVars.partCats[cls.catId].parts[cls.XM_partName] = cls
    cls.initCls(1)
    if globalVars.useLoadData:
        loadData = globalVars.loadData
        loadData.allModules[loadData.curModule]["parts"][cls.XM_partName] = cls
    if cls.mfgName not in partManufs:
        partManufs[cls.mfgName] = {}
    partManufs[cls.mfgName][cls.mfgPartName] = cls
    globalVars.mfgPartNames[cls.mfgPartName] = cls
    return cls


globalVars.initPartCls = initPartClsBase


def DctSetItem(self, k, value, callOn=defCb):
    dict.__setitem__(self, k, value)
    return callOn(value)


def partClass(partName):
    return Bind(DctSetItem, globalVars.dctParts, partName, callOn=globalVars.initPartCls)


def lstBinToVal(lst):
    rtn = 0
    for elem in lst[::-1]:
        rtn *= 2
        rtn += elem
    return rtn


def valToBin(val, padZero=0, bigEnd=False):
    rtn = ""
    while val > 0:
        rtn += DIGITS[val % 2]
        val /= 2
    if padZero > len(rtn): rtn += "0" * (padZero - len(rtn))
    if not bigEnd: rtn = rtn[::-1]
    return rtn
def isReserv(opt):
    return opt != "Reserved", opt
def isReservStr(strFmt, obj):
    opt = strFmt % obj
    return opt != "Reserved", opt
internedSignRegValFns = {}
#returns a 2-tuple of functions to be used in obj.regDataFuncs
#  where obj is and instance of BaseProgADF_PLL (by inheritance)
#  these functions are used for signed %{bitlen}-bit integers
#  uses a twos complement sign system
def signRegValFunc(bitlen):
    rtn = internedSignRegValFns.get(bitlen, None)
    if rtn is None:
        mod = 1 << bitlen
        mask = mod >> 1
        rtn = (
            lambda data: data%mod,
            lambda regNum: regNum if regNum & mask == 0 else (regNum - mod))
        internedSignRegValFns[bitlen] = rtn
    return rtn
internedSignBitRegValFns = {}
#returns a 2-tuple of functions to be used in obj.regDataFuncs
#  where obj is and instance of BaseProgADF_PLL (by inheritance)
#  these functions are used for signed %{bitlen}-bit integers
#  uses a simple sign bit and absolute magnitude (allows for possible -0)
def signBitRegValFunc(bitlen):
    rtn = internedSignBitRegValFns.get(bitlen, None)
    if rtn is None:
        signBitMask = 1 << (bitlen - 1)
        numMask = signBitMask - 1
        rtn = (
            lambda data: (signBitMask if data < 0 else 0) | abs(data),
            lambda regNum: (regNum & numMask) * -1 if regNum & signBitMask else 1)
        internedSignBitRegValFns[bitlen] = rtn
    return rtn

#returns a 2-tuple of functions to be used in obj.regDataFuncs
#  where obj is and instance of BaseProgADF_PLL (by inheritance)
#  these functions are used for unsigned %{bitlen}-bit numbers that
#    are 2 ** bitlen when regNum == 0 and are never zero
def nonZeroRegValFunc(bitlen):
    mod = 1 << bitlen
    return (
        lambda data: data%mod,
        lambda regNum: regNum if regNum != 0 else mod)
lstUnitFmt = [
    "%s T%s", "%s G%s", "%s M%s", "%s k%s",
    "%s %s", "%s m%s", u"%s \u03BC%s", "%s n%s"]
def unitConv(num, unit, e10=0, fmtBase="%.15g"):
    num *= 10 ** e10
    siPos = 4
    if num < 1.0:
        while num < 1.0 and siPos + 1 < len(lstUnitFmt):
            siPos += 1
            num *= 1e3
    else:
        while num >= 1e3 and siPos > 0:
            siPos -= 1
            num /= 1e3
    return (lstUnitFmt[siPos] % (fmtBase, unit)) % num


def unitSeconds(num, e10=0, fmtBase="%.15g"):
    return unitConv(num, "s", e10, fmtBase)


def unitHertz(num, e10=0, fmtBase="%.15g"):
    return unitConv(num, "Hz", e10, fmtBase)


def unitAmps(num, e10=0, fmtBase="%.15g"):
    return unitConv(num, "A", e10, fmtBase)


def unitVolts(num, e10=0, fmtBase="%.15g"):
    return unitConv(num, "V", e10, fmtBase)


def unitFarads(num, e10=0, fmtBase="%.15g"):
    return unitConv(num, "F", e10, fmtBase)


def unitWatts(num, e10=0, fmtBase="%.15g"):
    return unitConv(num, "W", e10, fmtBase)


def unitOhms(num, e10=0, fmtBase="%.15g"):
    return unitConv(num, u"\u2126", e10, fmtBase)


def unitJoules(num, e10=0, fmtBase="%.15g"):
    return unitConv(num, "J", e10, fmtBase)

dBmCnv = "%.15gdBm"
def dBmify(x):
    return isReservStr(dBmCnv, x)

globalVars.log.Log("Importing opengl_utils")
from kivy.graphics import opengl_utils
globalVars.log.Log("Importing opengl")
from kivy.graphics import opengl

globalVars.log.Log("getting MAX_TEXTURE_SIZE")
# 0x0D33 is the constant for max texture size
opengl_utils.gl_register_get_size(0x0D33, 1)
# noinspection PyArgumentList
MAX_TEXTURE_SIZE = opengl.glGetIntegerv(0x0D33)[0]
globalVars.log.Log("got MAX_TEXTURE_SIZE")

from .FileManagement import FileChooser
from .LibIndAppData import preSaveCheck, preLoadCheck
from .Logging import GetTmFmt


# used by the ADI PLLs to show a log of the register spi-output order
# acts like a general log output medium
class UserLog(object):
    MAX_LINES = None
    FONT_HEIGHT = None
    kwargsLbl = {
        "valign": "top", "halign": "left", "size_hint": (None, None)}

    def __init__(self, **kwargs):
        if "size_hint" not in kwargs: kwargs["size_hint"] = (1.0, 1.0)
        self.lbl = Label(text="", **UserLog.kwargsLbl)
        if UserLog.FONT_HEIGHT is None:
            self.lbl.text = "".join([chr(ord(' ') + c) for c in xrange(95)])
            self.lbl.texture_update()
            UserLog.FONT_HEIGHT = self.lbl.texture.height
            self.lbl.text = ""
            UserLog.MAX_LINES = int(MAX_TEXTURE_SIZE / UserLog.FONT_HEIGHT)
            self.MAX_LINES = UserLog.MAX_LINES
            self.FONT_HEIGHT = UserLog.FONT_HEIGHT
            globalVars.log.Log("UserLog.MAX_LINES = %u" % UserLog.MAX_LINES, 3)
            globalVars.log.Log("UserLog.FONT_HEIGHT = %u" % UserLog.FONT_HEIGHT, 3)
        self.lbl.bind(texture_size=self.lbl.setter("size"))
        self.widget = ScrollView(**kwargs)
        self.widget.bind(on_scroll_start=ScrollAllower)
        self.widget.add_widget(self.lbl)
        self.logName = ""
        self.popup = None
        self.saverCB = lambda fName: self.save(fName)
        self.loaderCB = lambda fName: self.load(fName)

    def log(self, Str):
        txt = self.lbl.text + GetTmFmt() + " %s\n" % Str
        txts = txt.split("\n")
        # limit to  227 lines
        txts = txts[max(0, len(txts) - UserLog.MAX_LINES):]
        self.lbl.text = "\n".join(txts)
        self.lbl.texture_update()

    def save(self, fName):
        with open(fName, "wb") as fl:
            fl.write(self.lbl.text.encode("UTF-8"))
            fl.flush()
        if self.popup is not None:
            self.popup.dismiss()
            self.popup = None

    def load(self, fName):
        with open(fName, "rb") as fl:
            self.lbl.text = fl.read().decode("UTF-8")
        if self.popup is not None:
            self.popup.dismiss()
            self.popup = None

    def saveDialog(self):
        a = FileChooser(
            "Save Log To Internal SD Card", "Save Log", self.saverCB,
            globalVars.logCtx.baseDir if len(globalVars.logCtx.baseDir) else "./", self.logName,
            Bind(setattr, self, "logName"), preSaveCheck, "xmlog")
        self.popup = Popup(size_hint=(.5, .8), content=a.layout, title="Save Log")
        self.popup.open()

    def loadDialog(self):
        a = FileChooser(
            "Load Log From Internal SD Card", "Load Log", self.loaderCB,
            globalVars.logCtx.baseDir if len(globalVars.logCtx.baseDir) else "./", self.logName,
            Bind(setattr, self, "logName"), preLoadCheck, "xmlog")
        self.popup = Popup(size_hint=(.5, .8), content=a.layout, title="Load Log")
        self.popup.open()

    def clear(self):
        self.lbl.text = ""


# used by the ADI PLLs to show a log of the register spi-output order
# acts like a general log output medium
# unlike UserLog this one supports [M]any more [L]ines
class UserLogML(object):
    MAX_LINES = None
    MAX_COLS = None
    FONT_HEIGHT = None
    kwargsLbl = {
        "valign": "top", "halign": "left", "size_hint": (None, None)}

    def __init__(self, **kwargs):
        if "size_hint" not in kwargs: kwargs["size_hint"] = (1.0, 1.0)
        self.lbls = []
        self.box_layout = BoxLayout(orientation="vertical", size_hint=(None, None))
        if UserLogML.FONT_HEIGHT is None:
            lbl = Label(text="", **self.kwargsLbl)
            lbl.text = "".join([chr(ord(' ') + c) for c in xrange(95)])
            lbl.texture_update()
            UserLogML.FONT_HEIGHT = lbl.texture.height
            UserLogML.MAX_LINES = int(MAX_TEXTURE_SIZE / UserLogML.FONT_HEIGHT)
            self.MAX_LINES = UserLogML.MAX_LINES
            self.FONT_HEIGHT = UserLogML.FONT_HEIGHT
            lbl.text = "\n".join(lbl.text)
            globalVars.log.Log("UserLog.MAX_LINES = %u" % UserLogML.MAX_LINES, 3)
            globalVars.log.Log("UserLog.FONT_HEIGHT = %u" % UserLogML.FONT_HEIGHT, 3)
        # TODO: note
        # TODO:   whenever you create a new label to add to the log ScrollView box_layout
        # TODO:   call it `lbl` one must do lbl.bind(texture_size=lbl.setter("size"))
        # TODO:   load, save, and clear methods are finished

        self.widget = ScrollView(**kwargs)
        self.widget.bind(on_scroll_start=ScrollAllower)
        self.widget.add_widget(self.box_layout)
        self.logName = ""
        self.popup = None
        self.saverCB = lambda fName: self.save(fName)
        self.loaderCB = lambda fName: self.load(fName)

    def log(self, Str):
        txt = "%s %s\n" % (GetTmFmt(), Str)
        lbl = None
        ln = UserLogML.MAX_LINES - 1
        start = 0
        if len(self.lbls):
            lbl = self.lbls[-1]
            txt = lbl.text + txt
        txts = txt.split("\n")
        # limit to  227 lines
        if lbl is not None:
            lbl.text = "\n".join(txts[:ln])
            lbl.texture_update()
            start = ln
            print "changing existing label"
        for c in xrange(start, len(txts), ln):
            print "adding new label"
            lbl = Label(text="\n".join(txts[c:c + ln]), **self.kwargsLbl)
            lbl.bind(texture_size=lbl.setter("size"))
            lbl.texture_update()
            self.lbls.append(lbl)
            self.box_layout.add_widget(lbl)
        self.update_size()

    def save(self, fName):
        with open(fName, "wb") as fl:
            fl.write("\n".join([lbl.text for lbl in self.lbls]).encode("UTF-8"))
            fl.flush()
        if self.popup is not None:
            self.popup.dismiss()
            self.popup = None

    def load(self, fName):
        self.clear()
        with open(fName, "rb") as fl:
            text = fl.read().decode("UTF-8")
        ln = UserLogML.MAX_LINES
        lines = text.split("\n")
        lbls = [Label(text="\n".join(lines[c: c + ln]), **self.kwargsLbl) for c in xrange(0, len(lines), ln)]
        for lbl in lbls:
            lbl.bind(texture_size=lbl.setter("size"))
            self.box_layout.add_widget(lbl)
        self.lbls = lbls
        self.update_size()
        if self.popup is not None:
            self.popup.dismiss()
            self.popup = None

    def saveDialog(self):
        a = FileChooser(
            "Save Log To Internal SD Card", "Save Log", self.saverCB,
            globalVars.logCtx.baseDir if len(globalVars.logCtx.baseDir) else "./", self.logName,
            Bind(setattr, self, "logName"), preSaveCheck, "xmlog")
        self.popup = Popup(size_hint=(.5, .8), content=a.layout, title="Save Log")
        self.popup.open()

    def loadDialog(self):
        a = FileChooser(
            "Load Log From Internal SD Card", "Load Log", self.loaderCB,
            globalVars.logCtx.baseDir if len(globalVars.logCtx.baseDir) else "./", self.logName,
            Bind(setattr, self, "logName"), preLoadCheck, "xmlog")
        self.popup = Popup(size_hint=(.5, .8), content=a.layout, title="Load Log")
        self.popup.open()

    def clear(self):
        self.box_layout.clear_widgets()
        self.lbls = []
        self.update_size()

    def update_size(self):
        height = 0
        width = 0
        for lbl in self.lbls:
            w, h = lbl.size
            width = max(w, width)
            height += h
        self.box_layout.size = width, height