try:
    from .TinyLib import *
except ValueError:
    from TinyLib import *
import os
import threading

isByt = lambda x: type(x) == int
#SPI Mock reads out a timing diagram, MSB first
#  if it prints reading ... 00000001 then that byte was 1
#  if it prints reading ... 10000000 then that byte was 128
class MOCK_spiobj(object):
    spidev = None
    def __init__(self, port=None,cs=None):
        assert self.spidev is not None
        print "CTOR_MOCK_SPIOBJ"
        self.port = port
        self.cs = cs
        self.doPrint = True
        self._lsbfirst = False
        self._mode = 0
        self._cshigh = False
        self._threewire = False
        self._no_cs = False
        self._loop = False
        self._max_speed_hz = 3600000
        self._bits_per_word = 8
        self.data = []
        if self.port is not None and self.cs is not None:
            self.open(self.port, self.cs)
    def open(self, port, cs):
        self.port = port
        self.cs = cs
        self.spidev.ports[(port, cs)] = list()
        self.data = self.spidev.ports[(port, cs)]
    def close(self):
        print "DTOR_MOCK_SPIOBJ"
    def devread(self):
        #ch is char
        Prn = "reading: %u, %u\n  " % (self.port, self.cs)
        if self._lsbfirst:
            for ch in self.data:
                Add = ""
                Val = ord(ch)
                for c in xrange(8):
                    Add += str(Val % 2)
                    Val >>= 1
                Prn += Add
        else:
            for ch in self.data:
                Val = ord(ch)
                Add = ""
                for c in xrange(8):
                    Add += str(Val % 2)
                    Val >>= 1
                Prn += Add[::-1]
        self.data[:] = []
        if self.doPrint: print Prn
    @property
    def cshigh(self): return self._cshigh
    @cshigh.setter
    def cshigh(self, val):
        val = bool(val)
        self._cshigh = val
        if val: self._mode |= self.spidev.SPI_CS_HIGH
        else: self._mode &= 0xFF ^ self.spidev.SPI_CS_HIGH
        if self.doPrint:
            print "cshigh: bus=%u, cs=%u" % (self.port, self.cs), bool(val)
    @property
    def mode(self): return self._mode & (self.spidev.SPI_CPHA|self.spidev.SPI_CPOL)
    @mode.setter
    def mode(self, val):
        val = int(val)
        self._mode = 0xFF ^ (self._mode & (self.spidev.SPI_CPHA|self.spidev.SPI_CPOL))
        self._mode |= val & (self.spidev.SPI_CPHA|self.spidev.SPI_CPOL)
    @property
    def lsbfirst(self): return self._lsbfirst
    @lsbfirst.setter
    def lsbfirst(self, val):
        val = bool(val)
        self._lsbfirst = val
        if val: self._mode |= self.spidev.SPI_LSB_FIRST
        else: self._mode &= 0xFF ^ self.spidev.SPI_LSB_FIRST
    @property
    def threewire(self): return self._threewire
    @threewire.setter
    def threewire(self, val):
        val = bool(val)
        self._threewire = val
        if val: self._mode |= self.spidev.SPI_3WIRE
        else: self._mode &= 0xFF ^ self.spidev.SPI_3WIRE
    @property
    def no_cs(self): return self._no_cs
    @no_cs.setter
    def no_cs(self, val):
        val = bool(val)
        self._no_cs = val
        if val: self._mode |= self.spidev.SPI_NO_CS
        else: self._mode &= 0xFF ^ self.spidev.SPI_NO_CS
    @property
    def loop(self): return self._loop
    @loop.setter
    def loop(self, val):
        val = bool(val)
        self._loop = val
        if val: self._mode |= self.spidev.SPI_LOOP
        else: self._mode &= 0xFF ^ self.spidev.SPI_LOOP
    @property
    def max_speed_hz(self): return self._max_speed_hz
    @max_speed_hz.setter
    def max_speed_hz(self, val):
        if type(val) != int:
            raise TypeError("max_speed_hz must be of type int")
        if val < 0 or val >= 2 ** 32:
            raise ValueError(
                "Cannot set max_speed_hz to %u, must be uint32"%val)
        self._max_speed_hz = val
    @property
    def bits_per_word(self): return self._bits_per_word
    @bits_per_word.setter
    def bits_per_word(self, val):
        if type(val) != int:
            raise TypeError("max_speed_hz must be of type int")
        if val < 8 or val > 16:
            raise ValueError("Cannot set bits_per_word out of interval [8, 16]")
        self._bits_per_word = val
    def writebytes(self, Bytes):
        if not all(map(isByt,Bytes)):
            raise TypeError(
                "Elements of Bytes must be of type int (not long or str)")
        self.data.extend(map(chr, Bytes))
        if self.doPrint: print "WRITE BYTES"
        self.devread()
    def xfer(self, Bytes):
        if not all(map(isByt,Bytes)):
            raise TypeError(
                "Elements of Bytes must be of type int (not long or str)")
        self.data.extend(map(chr, Bytes))
        self.devread()
    def xfer2(self, Bytes):
        if not all(map(isByt,Bytes)):
            raise TypeError(
                "Elements of Bytes must be of type int (not long or str)")
        self.data.extend(map(chr, Bytes))
        self.devread()
    def __enter__(self): pass
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return False

class MOCK_spidev(object):
    SPI_CPHA = 0x01
    SPI_CPOL = 0x02
    SPI_CS_HIGH = 0x04
    SPI_LSB_FIRST = 0x08
    SPI_3WIRE = 0x10
    SPI_LOOP = 0x20
    SPI_NO_CS = 0x40
    SPI_READY = 0x80
    SpiDev = MOCK_spiobj
    ports = {}

MOCK_spiobj.spidev = MOCK_spidev

class MOCK_gpio(object):
    BCM = 11
    UNKNOWN = -1
    HARD_PWM = 43
    BOARD = 10
    I2C = 42
    FALLING = 32
    BOTH = 33
    RPI_REVISION = 3
    OUT = 0
    HIGH = 1
    VERSION = "0.6.3"
    RISING = 31
    PUD_OFF = 20
    PUD_DOWN = 21
    PUD_UP = 22
    LOW = 0
    SPI = 41
    IN = 1
    SERIAL = 40
    _mode = None
    _warnings = True
    RPI_INFO = {
        "P1_REVISION": 3, "RAM": "1024M", "REVISION": "a22082", "TYPE": "Pi 3 Model B",
        "PROCESSOR": "BCM2837", "MANUFACTURER": "Embest"}
    _BCM_MAP = { # maps BCM->BOARD
        2: 3, 3: 5, 4: 7, 14: 8,
        15: 10, 17: 11, 18: 12, 27: 13,
        22: 15, 23: 16, 24: 18, 10: 19,
        9: 21, 25: 22, 11: 23, 8: 24,
        7: 26, 0: 27, 1: 28, 5: 29,
        6: 31, 12: 32, 13: 33, 19: 35,
        16: 36, 26: 37, 20: 38, 21: 40}
    _BOARD_MAP = { # maps BOARD->BOARD
        3: 3, 5: 5, 7: 7, 8: 8,
        10: 10, 11: 11, 12: 12, 13: 13,
        15: 15, 16: 16, 18: 18, 19: 19,
        21: 21, 22: 22, 23: 23, 24: 24,
        26: 26, 27: 27, 28: 28, 29: 29,
        31: 31, 32: 32, 33: 33, 35: 35,
        36: 36, 37: 37, 38: 38, 40: 40}
    _pin_states = {
        3: None, 5: None, 7: None, 8: None,
        10: None, 11: None, 12: None, 13: None,
        15: None, 16: None, 18: None, 19: None,
        21: None, 22: None, 23: None, 24: None,
        26: None, 27: None, 28: None, 29: None,
        31: None, 32: None, 33: None, 35: None,
        36: None, 37: None, 38: None, 40: None}
    _DO_PRINT_OUTPUT = True
    _channels = {}
    @classmethod
    def _setup(cls, chan, md, init, pud):
        pin = cls._channels[chan]
        if md == cls.IN:
            assert init is None, "initial must not be specified for input"
        else:
            assert pud is None, "pull_up_down must not be specified for output"
        if init is None: init = cls.LOW
        if pud is None: pud = cls.PUD_OFF
        cls._pin_states[pin] = init
    @classmethod
    def setup(cls, channel, ioMode, initial=None, pull_up_down=None):
        _channel_match = lambda x: x in cls._channels
        _ioMode_match = lambda x: x in [cls.IN, cls.OUT]
        _inital_match = lambda x: x in [None, cls.LOW, cls.HIGH]
        _pud_match = lambda x: x in [None, cls.PUD_UP, cls.PUD_DOWN, cls.PUD_OFF]
        if isinstance(channel, (list, tuple)):
            assert all(map(_channel_match, channel))
            if isinstance(ioMode, (list, tuple)):
                assert len(ioMode) == len(channel)
                assert all(map(_ioMode_match, ioMode))
            else:
                assert _ioMode_match(ioMode)
                ioMode = [ioMode] * len(channel)
            if isinstance(initial, (list, tuple)):
                assert len(initial) == len(channel)
                assert all(map(_inital_match, initial))
            else:
                assert _inital_match(initial)
                initial = [initial] * len(channel)
            if isinstance(pull_up_down, (list, tuple)):
                assert len(pull_up_down) == len(channel)
                assert all(map(_pud_match, pull_up_down))
            else:
                assert _pud_match(pull_up_down)
                pull_up_down = [pull_up_down] * len(channel)
            map(cls._setup, channel, ioMode, initial, pull_up_down)
        else:
            assert _channel_match(channel)
            assert _ioMode_match(ioMode)
            assert _inital_match(initial)
            assert _pud_match(pull_up_down)
            cls._setup(channel, ioMode, initial, pull_up_down)
    @classmethod
    def cleanup(cls, channel=None):
        if channel is None:
            for k in cls._pin_states:
                cls._pin_states[k] = None
            return
        _channel_match = lambda x: x in cls._channels
        if isinstance(channel, (list, tuple)):
            assert all(map(_channel_match, channel))
            for chan in channel:
                pin = cls._channels[chan]
                cls._pin_states[pin] = None
        else:
            assert _channel_match(channel)
            pin = cls._channels[channel]
            cls._pin_states[pin] = None
    @classmethod
    def getmode(cls):
        return cls._mode
    @classmethod
    def setmode(cls, mode):
        assert mode in [cls.BCM, cls.BOARD]
        cls._mode = mode
        cls._channels = cls._BCM_MAP if mode == cls.BCM else cls._BOARD_MAP
    @classmethod
    def input(cls, channel):
        if cls._mode is None: raise RuntimeError(
            "Please set pin numbering mode using GPIO.setmode(GPIO.BOARD) or GPIO.setmode(GPIO.BCM)")
        assert channel in cls._channels
        pin = cls._channels[channel]
        if cls._pin_states[pin] is None:
            raise RuntimeError("The GPIO channel has not been set up as an INPUT")
        return cls._pin_states[pin]
    @classmethod
    def output(cls, channel, state):
        if cls._mode is None: raise RuntimeError(
            "Please set pin numbering mode using GPIO.setmode(GPIO.BOARD) or GPIO.setmode(GPIO.BCM)")
        assert channel in cls._channels
        assert isinstance(state, (bool, int, long))
        pin = cls._channels[channel]
        if cls._pin_states[pin] is None:
            raise RuntimeError("The GPIO channel has not been set up as an OUTPUT")
        if cls._DO_PRINT_OUTPUT:
            old = cls._pin_states[pin]
            new = bool(state)
            cls._pin_states[pin] = new
            print "MOCK_gpio(%u): %r->%r" % (pin, old, new)
        else:
            cls._pin_states[pin] = bool(state)
    @classmethod
    def setwarnings(cls, b):
        assert isinstance(b, (long, int, bool))
        cls._warnings = b
    @classmethod
    def add_event_callback(cls): pass
    @classmethod
    def event_detected(cls): pass
    @classmethod
    def remove_event_detect(cls): pass
    @classmethod
    def add_event_detect(cls): pass
    @classmethod
    def wait_for_edge(cls): pass
    @classmethod
    def gpio_function(cls): pass
    class PWM(object): pass
#spidev = None
#gpio = None
def InitLibIO(spiModule, gpioModule):
    """
    :param MOCK_spidev spiModule:
    :param MOCK_gpio gpioModule:
    :return:
    """
    global spidev
    global gpio
    spidev = spiModule
    gpio = gpioModule
    gpio.setmode(gpio.BCM)
IO_NONE = 0
IO_SPI = 1
IO_PIN = 2
IO_STATIC_P = 3
IO_I2C = 4
PIN_MODE_SYS = 0
PIN_MODE_USR = 1
ioTypenames = ["IO_NONE", "IO_SPI", "IO_PIN", "IO_STATIC_P", "IO_I2C"]
def getNBitsLstCS(lstCS):
    return (len(lstCS) - 1).bit_length()
class BaseIO_Object(object):
    ioType = IO_NONE
    pinTypes = {
        0: set(), # reserved GPIO 0 or SDA 0 VC (rPi pin 27)
        1: set(), # reserved GPIO 1 or SCL 0 VC (rPi pin 28)
        2: {IO_PIN,IO_I2C},
        3: {IO_PIN,IO_I2C},
        4: {IO_PIN},
        5: {IO_PIN},
        6: {IO_PIN},
        7: {IO_PIN, IO_SPI}, # SPI 0 CE1
        8: {IO_PIN, IO_SPI}, # SPI 0 CE0
        9: {IO_PIN, IO_SPI}, # SPI 0 MISO
        10:{IO_PIN, IO_SPI}, # SPI 0 MOSI
        11:{IO_PIN, IO_SPI}, # SPI 0 SCLK
        12:{IO_PIN},
        13:{IO_PIN},
        14:{IO_PIN}, # UART 0 Tx
        15:{IO_PIN}, # UART 0 Rx
        16:{IO_PIN, IO_SPI}, # SPI 1 CE2
        17:{IO_PIN, IO_SPI}, # SPI 1 CE1
        18:{IO_PIN, IO_SPI}, # SPI 1 CE0
        19:{IO_PIN, IO_SPI}, # SPI 1 MISO
        20:{IO_PIN, IO_SPI}, # SPI 1 MOSI
        21:{IO_PIN, IO_SPI}, # SPI 1 SCLK
        22:{IO_PIN},
        23:{IO_PIN},
        24:{IO_PIN},
        25:{IO_PIN},
        26:{IO_PIN},
        27:{IO_PIN}}
    pinAllocTable = {}
    spiPins = {
        0: ((9, 10, 11), {0: 8, 1: 7}),
        1: ((19, 20, 21), {0: 18, 1: 17, 2: 16})}
    ioObjLk = threading.Lock()
    @classmethod
    def deserialize(cls, fl):
        raise NotImplementedError("serialize was not implemented")
    @classmethod
    def freeIO(cls, ioInf):
        ioInf = tuple(ioInf)
        with cls.ioObjLk:
            if ioInf[0] == IO_PIN:
                if len(ioInf) != 2: raise ValueError(
                    "io alloc functions for IO_PIN only accept a 2-tuple")
                if cls.pinAllocTable.get(ioInf[1], (IO_NONE,)) == ioInf:
                    del cls.pinAllocTable[ioInf[1]]
                    return
            elif ioInf[0] == IO_SPI:
                if len(ioInf) not in [2, 3]: raise ValueError(
                    "io alloc functions for IO_SPI only accept (2 or 3)-tuples")
                if len(ioInf) == 2:
                    typ, bus = ioInf
                    pins, chipSelects = cls.spiPins[bus]
                    for pin in pins:
                        if cls.pinAllocTable.get(pin, (IO_NONE,))[0] == IO_NONE: break
                    else:
                        for cs in chipSelects:
                            pin = chipSelects[cs]
                            if cls.pinAllocTable.get(pin, (IO_NONE,))[0:3] == ioInf + (cs,):
                                raise ValueError(
                                    "Cannot free spi bus when chip selects are allocated with that spi bus")
                        for pin in pins:
                            del cls.pinAllocTable[pin]
                        return
                else:
                    typ, bus, cs = ioInf
                    pins, chipSelects = cls.spiPins[bus]
                    for pin in pins:
                        if cls.pinAllocTable.get(pin, (IO_NONE,)) != ioInf[:2]:
                            break
                    else:
                        pin = chipSelects[cs]
                        if cls.pinAllocTable.get(pin, (IO_NONE,)) == ioInf:
                            del cls.pinAllocTable[pin]
                            return
            else:
                raise ValueError("IO must have been previously allocated")
            raise KeyError("no IO of type %s was allocated with those parameters"%ioTypenames[ioInf[0]])
    @classmethod
    def acquireIO(cls, ioInf):
        ioInf = tuple(ioInf)
        with cls.ioObjLk:
            if ioInf[0] == IO_PIN:
                assert len(ioInf) == 3, "io alloc functions for IO_PIN only accept a 3-tuple"
                cur_alloc = cls.pinAllocTable.get(ioInf[1], (IO_NONE,))
                if cur_alloc[0] == IO_NONE:
                    cls.pinAllocTable[ioInf[1]] = ioInf
                    return
                # elif cur_alloc[0] == IO_PIN:
                #     if cur_alloc[2] == PIN_MODE_USR and ioInf[2] == PIN_MODE_USR:
                #         cls.pinAllocTable[ioInf[1]] = ioInf
                #         return
            elif ioInf[0] == IO_SPI:
                if len(ioInf) not in [2, 3]: raise ValueError(
                    "io alloc functions for IO_SPI only accept (2 or 3)-tuples")
                if len(ioInf) == 2:
                    typ, bus = ioInf
                    pins = cls.spiPins[bus][0]
                    for pin in pins:
                        if cls.pinAllocTable.get(pin, (IO_NONE,))[0] != IO_NONE: break
                    else:
                        for pin in pins:
                            cls.pinAllocTable[pin] = ioInf
                        return
                else:
                    typ, bus, cs = ioInf
                    pins, chipSelects = cls.spiPins[bus]
                    for pin in pins:
                        typPin = cls.pinAllocTable.get(pin, (IO_NONE,))[0]
                        if typPin == IO_NONE: raise ValueError(
                            "spi bus must be allocated before allocating a specific chip select")
                        elif typPin != IO_SPI: break
                    else:
                        pin = chipSelects[cs]
                        typPin = cls.pinAllocTable.get(pin, (IO_NONE,))[0]
                        if typPin == IO_NONE:
                            cls.pinAllocTable[pin] = ioInf
                            return
            else:
                raise ValueError("IO type must be one of IO_SPI or IO_PIN")
            raise KeyError("no pins available for %s"%ioTypenames[ioInf[0]])
    def serialize(self, fl):
        raise NotImplementedError("serialize was not implemented")
    def getDispCols(self): return ["", ""]
class SpiBusInstanceStorage(object):
    __slots__ = ["spi", "allInst"]
    def __init__(self, spi, allInst):
        """
        :param MOCK_spiobj spi:
        :param set[spiIO_Object] allInst:
        """
        self.spi = spi
        self.allInst = allInst
class SpiMultiInstanceStorage(object):
    __slots__ = ["allInst"]
    def __init__(self, allInst):
        """
        :param set[spiIO_Object] allInst:
        """
        self.allInst = allInst
def setMuxSpi(muxPins, v):
    """
    :param tuple[int] muxPins:
    :param int v:
    """
    for pin in muxPins:
        gpio.output(pin, bool(v & 1))
        v >>= 1

def getMaskNBits(n):
    return (1 << n) - 1
multi_packer = DataPacker(DataArray(DataInt(1)), DataArray(DataInt(1)))
class spiIO_Object(BaseIO_Object):
    """
    :type spiObjects: dict[int, dict[int,SpiBusInstanceStorage]]
    :type spiMultiObjects: dict[tuple, SpiMultiInstanceStorage]
    """
    ioType = IO_SPI
    spiBusPins = [(10,9,11), (20,19,21)]
    spiObjects = {}
    spiMultiObjects = {}
    spiObjLk = threading.Lock()
    #ensureInit is NOT thread safe 
    @classmethod
    def ensureInit(cls, bus, cs, multi=None):
        """
        multi can be None
        multi can be tuple (0, list of chip selects, list of chip select mux bits)
        :param int bus:
        :param int cs:
        :param (int, tuple[int], tuple[int]) multi:
        :return:
        """
        if multi is None:
            if bus not in cls.spiObjects:
                cls.acquireIO((IO_SPI, bus))
                cls.spiObjects[bus] = {}
            if cs not in cls.spiObjects[bus]:
                cls.acquireIO((IO_SPI, bus, cs))
                spiObj = spidev.SpiDev()
                spiObj.open(bus, cs)
                cls.spiObjects[bus][cs] = SpiBusInstanceStorage(spiObj, set())
            return
        assert multi[0] == 0
        if multi in cls.spiMultiObjects:
            return
        lstCS = multi[1]
        muxPins = multi[2]
        csIndex = cs >> len(muxPins)
        # muxBits = cs & getMaskNBits(len(muxPins))
        real_cs = lstCS[csIndex]
        cls.ensureInit(bus, real_cs)
        c = 0
        try:
            while c < len(muxPins):
                cls.acquireIO((IO_PIN, muxPins[c], PIN_MODE_SYS))
                gpio.setup(muxPins[c], gpio.OUT)
                c += 1
        except ValueError:
            for pin in muxPins[:c]:
                cls.freeIO((IO_PIN, pin, PIN_MODE_SYS))
                gpio.cleanup(muxPins[c])
            raise
        cls.spiMultiObjects[multi] = SpiMultiInstanceStorage(set())
    @classmethod
    def alloc(cls, inst):
        """
        :param spiIO_Object inst:
        :return: MOCK_spiobj
        """
        with cls.spiObjLk:
            cls.ensureInit(inst.bus, inst.cs, inst.multi)
            cs = inst.cs
            if inst.multi is not None:
                lstCS = inst.multi[1]
                muxPins = inst.multi[2]
                csIndex = cs >> len(muxPins)
                cs = lstCS[csIndex]
            spiLock = cls.spiObjects[inst.bus][cs]
            spiLock.allInst.add(inst)
            if inst.multi is not None:
                cls.spiMultiObjects[inst.multi].allInst.add(inst)
            return spiLock.spi
    @classmethod
    def dealloc(cls, inst):
        """
        :param spiIO_Object inst:
        """
        bus, cs = inst.bus, inst.cs
        if inst.multi is not None:
            lstCS = inst.multi[1]
            muxPins = inst.multi[2]
            csIndex = cs >> len(muxPins)
            cs = lstCS[csIndex]
        with cls.spiObjLk:
            if bus not in cls.spiObjects: return
            if cs not in cls.spiObjects[bus]: return
            cls.spiObjects[bus][cs].allInst.discard(inst)
            if len(cls.spiObjects[bus][cs].allInst) == 0:
                cls.spiObjects[bus][cs].spi.close()
                del cls.spiObjects[bus][cs]
                cls.freeIO((IO_SPI, bus, cs))
                if len(cls.spiObjects[bus]) == 0:
                    del cls.spiObjects[bus]
                    cls.freeIO((IO_SPI, bus))
        if inst.multi is None:
            return
        smo = cls.spiMultiObjects.get(inst.multi[2], None)
        if smo is None:
            return
        smo.allInst.discard(inst)
        if len(smo.allInst) != 0:
            return
        for pin in inst.multi[2]:
            cls.freeIO((IO_PIN, pin, PIN_MODE_SYS))

    @classmethod
    def deserialize(cls, fl):
        bus = ord(fl.read(1))
        cs = ord(fl.read(1))
        if bus == 255:
            bus = ord(fl.read(1))
            f_multi = ord(fl.read(1))
            if f_multi == 255:
                multi = None
            elif f_multi == 0:
                lstCS, lstMux = multi_packer.Unpack(fl)
                multi = (f_multi, tuple(lstCS), tuple(lstMux))
            else:
                print "expected multi to be 0"
                multi = None
            return cls(bus, cs, multi)
        else:
            return cls(bus, cs)
    def __init__(self, bus, cs, multi=None):
        """
        :param int bus:
        :param int cs:
        :param None|(int, tuple[int], tuple[int]) multi:
        """
        self.bus, self.cs = bus, cs
        self.multi = multi
        self.spiBus = self.alloc(self)
    def serialize(self, fl):
        if self.multi is None and self.bus < 255:
            fl.write(chr(self.bus) + chr(self.cs))
        else:
            fl.write(chr(255) + chr(self.cs) + chr(self.bus) + chr(self.multi[0]))
            multi_packer.Pack([self.multi[1],self.multi[2]], fl)
        fl.flush()
    def close(self):
        self.dealloc(self)
    def getDispCols(self): return ["SPI %u" % self.bus, "LE %u" % self.cs]
    @property
    def cshigh(self): return self.spiBus.cshigh
    @cshigh.setter
    def cshigh(self, val): self.spiBus.cshigh = val
    @property
    def mode(self): return self.spiBus.mode
    @mode.setter
    def mode(self, val): self.spiBus.mode = val
    @property
    def lsbfirst(self): return self.spiBus.lsbfirst
    @lsbfirst.setter
    def lsbfirst(self, val): self.spiBus.lsbfirst = val
    @property
    def threewire(self): return self.spiBus.threewire
    @threewire.setter
    def threewire(self, val): self.spiBus.threewire = val
    @property
    def no_cs(self): return self.spiBus.no_cs
    @no_cs.setter
    def no_cs(self, val): self.spiBus.no_cs = val
    @property
    def loop(self): return self.spiBus.loop
    @loop.setter
    def loop(self, val): self.spiBus.loop = val
    @property
    def max_speed_hz(self): return self.spiBus.max_speed_hz
    @max_speed_hz.setter
    def max_speed_hz(self, val):
        if type(val) != int:
            raise TypeError("max_speed_hz must be of type int")
        if val <= 0 or val >= 2 ** 32:
            raise ValueError(
                "Cannot set max_speed_hz to %u, must be uint32" % val)
        self.spiBus.max_speed_hz = val
    @property
    def bits_per_word(self): return self.spiBus.bits_per_word
    @bits_per_word.setter
    def bits_per_word(self, val):
        if type(val) != int:
            raise TypeError("max_speed_hz must be of type int")
        if val < 8 or val > 16:
            raise ValueError("Cannot set bits_per_word out of interval [8, 16]")
        self.spiBus.bits_per_word = val
    def writebytes(self, Bytes):
        if not all(map(isByt,Bytes)):
            raise TypeError(
                "Elements of Bytes must be of type int (not long or str)")
        if self.multi is not None:
            lstCS = self.multi[1]
            muxPins = self.multi[2]
            muxBits = self.cs & getMaskNBits(len(muxPins))
            setMuxSpi(muxPins, muxBits)
        self.spiBus.writebytes(Bytes)
    def xfer(self, Bytes):
        if not all(map(isByt,Bytes)):
            raise TypeError(
                "Elements of Bytes must be of type int (not long or str)")
        if self.multi is not None:
            lstCS = self.multi[1]
            muxPins = self.multi[2]
            muxBits = self.cs & getMaskNBits(len(muxPins))
            setMuxSpi(muxPins, muxBits)
        self.spiBus.xfer(Bytes)
    def xfer2(self, Bytes):
        if not all(map(isByt,Bytes)):
            raise TypeError(
                "Elements of Bytes must be of type int (not long or str)")
        if self.multi is not None:
            lstCS = self.multi[1]
            muxPins = self.multi[2]
            muxBits = self.cs & getMaskNBits(len(muxPins))
            setMuxSpi(muxPins, muxBits)
        self.spiBus.xfer2(Bytes)
    def __enter__(self): return self
    def __exit__(self, exc_type, exc_value, tb):
        self.close()
        return False
class PinIO_Object(BaseIO_Object):
    ioType = IO_PIN
    pinObjects = {}
    pinObjLk = threading.Lock()
    # TODO: mock the rPi.GPIO module to make the allocator work
    @classmethod
    def deserialize(cls, fl):
        return cls(ord(fl.read(1)) | (ord(fl.read(1)) << 8))
    def __init__(self, pinNum):
        """
        :param int|long pinNum:
        """
        self.pinNum = pinNum
        with self.pinObjLk:
            if pinNum not in self.pinObjects:
                self.pinObjects[pinNum] = ValuePack(
                    {"allInst":set(),"value":False})
                gpio.setup(pinNum, gpio.OUT, initial=gpio.LOW)
            self.pinObjects[pinNum].allInst.add(self)
            self._value = self.pinObjects[pinNum].value
    def serialize(self, fl):
        fl.write(chr(self.pinNum&0xFF) + chr((self.pinNum>>8)&0xFF))
        fl.flush()
    @property
    def value(self): return self._value
    @value.setter
    def value(self, val):
        val = bool(val)
        print "PinIO_Object(%u) - value: %s -> %s" % (
            self.pinNum, str(self._value), str(val))
        with self.pinObjLk:
            cur = self.pinObjects[self.pinNum]
            if cur.value != val:
                cur.value = val
                for inst in cur.allInst: inst._value = cur.value
                gpio.output(self.pinNum, val)
    def close(self):
        with self.pinObjLk:
            cur = self.pinObjects[self.pinNum]
            cur.allInst.discard(self)
            if len(cur.allInst) == 0:
                gpio.cleanup(self.pinNum)
                del self.pinObjects[self.pinNum]
                self.pinNum = None
    def __del__(self):
        if self.pinNum is not None:
            self.close()
dctClsIO = {
    IO_SPI:spiIO_Object,
    IO_PIN:PinIO_Object}
def serializeIO(ioType, obj, fl, inclType=True):
    if isinstance(obj, BaseIO_Object):
        assert ioType == obj.ioType, "ioType must be correct"
        if inclType: fl.write(chr(ioType))
        obj.serialize(fl)
        return
    obj = list(obj)
    if inclType: fl.write(chr(ioType|0x80))
    fl.write(chr(len(obj)))
    for inst in obj:
        assert isinstance(inst, BaseIO_Object), "only lists of BaseIO_Object are allowed"
        serializeIO(ioType, inst, fl, False)
def deserializeIO(fl):
    ioType = ord(fl.read(1))
    isList = bool(ioType & 0x80)
    ioType &= 0x7F
    cls = dctClsIO[ioType]
    if isList:
        Len = ord(fl.read(1))
        return [cls.deserialize(fl) for _ in xrange(Len)]
    else:
        return cls.deserialize(fl)

def hasSpiBusCs(bus, cs):
    if os.name == "posix":
        return os.path.exists("/dev/spidev%u.%u"%(bus,cs))
    if bus == 0:
        return cs in {0, 1}
    elif bus == 1:
        return cs in {0, 1, 2}
    else:
        return False
def enumerateSpiBus():
    if os.name == "posix":
        lst = filter(lambda x: x.startswith("spidev"), os.listdir("/dev/"))
        return sorted(set(map(lambda x: int(x[6:x.rfind('.')]), lst)))
    else:
        return [0, 1]
def enumerateSpiBusCs(bus):
    if os.name == "posix":
        lst = filter(lambda x: x.startswith("spidev%u."%bus), os.listdir("/dev/"))
        try:
            return sorted(set(map(lambda x: int(x[x.rfind('.')+1:]), lst)))
        except ValueError:
            print "ERROR: Bad SPI Filename format"
            return []
    else:
        return [[0, 1], [0, 1, 2]][bus]
def enumerateAllSpiBusCs():
    return [(bus, cs) for bus in enumerateSpiBus() for cs in enumerateSpiBusCs(bus)]
def enumeratePins(key=None):
    return sorted(filter(
        lambda x: len(BaseIO_Object.pinTypes[x]) == 1 and IO_PIN in BaseIO_Object.pinTypes[x] and x not in disabledPins,
        BaseIO_Object.pinTypes),key=key)
bcm_to_pin = {
    2: 3,
    3: 5,
    4: 7,
    17: 11,
    27: 13,
    22: 15,
    10: 19,
    9: 21,
    11: 23,
    0: 27,
    5: 29,
    6: 31,
    13: 33,
    19: 35,
    26: 37,
    14: 8,
    15: 10,
    18: 12,
    23: 16,
    24: 18,
    25: 22,
    8: 24,
    7: 26,
    1: 28,
    12: 32,
    16: 36,
    20: 38,
    21: 40}
pin_to_bcm = {bcm_to_pin[k]: k for k in bcm_to_pin}
def getPinNumFromBCM_ID(bcm):
    return bcm_to_pin[bcm]
def getBCM_ID_FromPinNum(pin):
    return pin_to_bcm[pin]
def isMultiIO(ioType):
    return ioType in {IO_STATIC_P, IO_PIN}
MUX_PINS_TO_USE = (pin_to_bcm[11], pin_to_bcm[12])
disabledPins = set()
multiEnabled = False
def getMulti(bus):
    if not multiEnabled:
        return None
    lstCS = tuple(enumerateSpiBusCs(bus))
    nBits = len(MUX_PINS_TO_USE)# getNBitsLstCS(lstCS)
    multi = (0, lstCS, MUX_PINS_TO_USE)
    return [
        ((cs << nBits) | muxVal, muxVal, multi)
        for cs in lstCS
        for muxVal in xrange(1 << len(MUX_PINS_TO_USE))
    ]

def setMultiEnabled(b):
    global multiEnabled
    multiEnabled = bool(b)


def setDisabledPins(pins):
    global disabledPins
    disabledPins = set(pins)


def getSpiBusInfo(spiObj):
    if isinstance(spiObj, MOCK_spiobj):
        return spiObj.port, spiObj.cs
    elif isinstance(spiObj, spiIO_Object):
        return spiObj.bus, spiObj.cs
    elif isinstance(spiObj, PinIO_Object):
        return -1, spiObj.pinNum
    elif os.name == "posix":
        name = os.readlink("/proc/self/fd/" + str(spiObj.fileno()))
        return map(int, name.lower().replace("/dev/spidev", "").split("."))
    return -1, -1

def initializeMuxedSpi(mux_pins):
    pass