from os.path import exists, isfile


def preLoadCheck(fName, text):
    if not isfile(fName):
        return False, "cannot load from \"%s\"; not a file" % text
    return True, ""


def preSaveCheck(fName, text):
    if exists(fName):
        if not isfile(fName):
            return False, "cannot save to \"%s\"; must not exist or must be a file" % text
    return True, ""


class ProgStyleStorage(object):
    __slots__ = [
        "lstW",
        "lstH",
        "btnH",
        "fntSz",
        "scroll_dist_product",
        "brdImgW",
        "relYBtns"
    ]

    def __init__(self):
        self.lstW = None
        self.lstH = None
        self.btnH = None
        self.fntSz = None
        self.scroll_dist_product = None
        self.brdImgW = None
        self.relYBtns = None


# a command list that represents the users scene history
#  (not the history in that scene)
# used by the forward and back buttons
class CmdList(object):
    def __init__(self):
        self.data = list()
        self.index = 0

    def push(self, cmd, name, noCpy=True):
        if noCpy and len(self.data) > 0 and self.data[self.index][0] == cmd:
            if self.index + 1 < len(self.data):
                self.data[:] = self.data[:self.index + 1]
                return True
            return False
        self.data[:] = self.data[:self.index + 1]
        self.data.append((cmd, name))
        self.index = len(self.data) - 1
        return True

    def unPop(self):
        if len(self.data) == 0: return None
        self.index += 1
        if self.index >= len(self.data):
            self.index = len(self.data)
            return None
        else:
            return self.data[self.index]

    def pop(self):
        if len(self.data) == 0: return None
        self.index -= 1
        if self.index <= -1:
            self.index = -1
            return None
        else:
            return self.data[self.index]

    def clear(self):
        self.data[:] = []
        self.index = 0

    def atNext(self):
        if len(self.data) == 0 or self.index + 1 == len(self.data): return None
        return self.data[self.index + 1]

    def atPrev(self):
        if len(self.data) == 0 or self.index == 0: return None
        return self.data[self.index - 1]

    def __len__(self):
        return len(self.data)

    def atCur(self):
        if len(self.data) == 0: return None
        return self.data[self.index]