from .DataEntryHelpers import *


MB_HELP = 0x00004000L
MB_OK = 0x00000000L
MB_OKCANCEL = 0x00000001L
MB_ABORTRETRYIGNORE = 0x00000002L
MB_YESNOCANCEL = 0x00000003L
MB_YESNO = 0x00000004L
MB_RETRYCANCEL = 0x00000005L
MB_CANCELTRYCONTINUE = 0x00000006L
IDOK = 1
IDCANCEL = 2
IDABORT = 3
IDRETRY = 4
IDIGNORE = 5
IDYES = 6
IDNO = 7
IDTRYAGAIN = 10
IDCONTINUE = 11
# MessageBox models the windows api function MessageBox
#   except that it passes the return value to a callback
class MessageBox(object):
    btnNames = {
        1: "Ok",
        2: "Cancel",
        3: "Abort",
        4: "Retry",
        5: "Ignore",
        6: "Yes",
        7: "No",
        10: "Try Again",
        11: "Continue"
    }
    btnCodeMap = {
        0: [1],
        1: [1, 2],
        2: [3, 4, 5],
        3: [6, 7, 2],
        4: [6, 7],
        5: [4, 2],
        6: [2, 10, 11]
    }
    def __init__(self, uType, kwargsContent=None, **kwargsPopup):
        btnCode = uType & 0xF
        btnRtn = self.btnCodeMap[btnCode]
        if kwargsContent is None: kwargsContent = {}
        fillDictDefs(kwargsContent, dict(size_hint=(1.,1.)))
        self.layout = FloatLayout(**kwargsContent)
        self.btnGrid = GridLayout(
            size_hint=(1.,.2), cols=len(btnRtn), pos_hint={"center_x": .5, "top": .2})
        for btnId in btnRtn:
            btn = Button(text=self.btnNames[btnId], on_release=Bind(self.action, btnId))
            self.btnGrid.add_widget(btn)
        self.layout.add_widget(self.btnGrid)
        self.popup = Popup(**kwargsPopup)
        self.popup.content = self.layout
        self.textLbl = Label(
            size_hint=(1.,.8), pos_hint={"center_x": .5, "top":1.0},
            valign="middle", halign="left")
        self.textLbl.bind(size=lambda obj, val: setattr(obj, "text_size", val))
        self.layout.add_widget(self.textLbl)
        self.cbDone = None
    def action(self, rtn, obj):
        del obj
        self.popup.dismiss()
        if self.cbDone is not None:
            self.cbDone(rtn)
    def show(self, title, text, cbDone):
        self.cbDone = cbDone
        self.popup.title = title
        self.textLbl.text = text
        self.popup.open()