from .DataEntryHelpers import *
from .LibIO import IO_SPI, enumerateSpiBus, enumerateSpiBusCs, getSpiBusInfo, spiIO_Object, getMulti
from .LibIO import IO_PIN, enumeratePins, getPinNumFromBCM_ID, PinIO_Object
from .LibIO import IO_NONE, setMultiEnabled, setDisabledPins
from .LibBaseScreen import BaseScreen
from .LibAppData import AppData


MUX_PINS = (4, 14)


class IOConfig_Spi(object):
    indexMapper = None
    progMapper = None
    adapter = None
    progIndex = None
    prog = None
    SelectChainPos = None
    spiInf = None
    spiChain = None
    def __init__(self, ioConfig, app):
        """
        :param IOConfigHelpers ioConfig:
        :param AppData app:
        """
        assert isinstance(app, AppData)
        self.fmtBus = "Change\nBus"
        self.fmtLE = "Change\nLE"
        self.fmtChain = "Change\nChain Position"
        self.app = app
        self.cfg = ioConfig
        self.panel = GridLayout(cols=2, spacing=[5])
        self.theList = BorderedList(bdr=8)
        scroll = filter(lambda x: isinstance(x, ScrollView), self.theList.children)[0]
        scroll.bind(on_scroll_start=ScrollAllower)
        self.panel.add_widget(self.theList)
        self.SelectSpiBus = SelectPopupMaker(
            {x: "Spi Bus %u" % x for x in enumerateSpiBus()}, 0)
        self.leOptions = [
            map(lambda x: (x, "LE/CS %u" % x), enumerateSpiBusCs(bus))
            for bus in enumerateSpiBus()
        ]
        for bus in enumerateSpiBus():
            options = self.leOptions[bus]
            multiData = getMulti(bus)
            if multiData is None or len(multiData) == 0:
                continue
            muxValMax = max(multiData,key=lambda x: x[1])[1]
            if multiData is None or len(multiData) == 0:
                continue
            addOptions = [
                (cs, "LE%u [PCB 915]" % cs, multi)
                for cs, muxVal, multi in multiData
            ]
            options.extend(addOptions)
        self.SelectSpiLE = {
            x: SelectPopupMaker({x0: elem[1] for x0, elem in enumerate(options)}, 0)
            for x, options in enumerate(self.leOptions)}
        self.btnLayout = GridLayout(rows=3, size_hint_x=.2, spacing=[5])
        self.btnBus = XMLoneButton(text=self.fmtBus, disabled=True)
        self.btnLE = XMLoneButton(text=self.fmtLE, disabled=True)
        self.btnChain = XMLoneButton(text=self.fmtChain, disabled=True)
        for btn in [self.btnBus, self.btnLE, self.btnChain]:
            btn.disabled_background_color = self.cfg.dis[0]
            btn.background_normal_color = self.cfg.off[0]
            btn.background_down_color = self.cfg.on[0]
            btn.disabled_color = self.cfg.dis[1]
            btn.normal_color = self.cfg.off[1]
            btn.down_color = self.cfg.on[1]
            self.btnLayout.add_widget(btn)
        self.btnBus.bind(on_release=self.showSelSpiBus)
        self.btnLE.bind(on_release=self.showSelSpiLE)
        self.btnChain.bind(on_release=self.showSelChainPos)
        self.panel.add_widget(self.btnLayout)
        self.SpiChains = app.SpiChains

    def updateSpiLE(self):
        self.leOptions = [
            map(lambda x: (x, "LE/CS %u" % x), enumerateSpiBusCs(bus))
            for bus in enumerateSpiBus()
        ]
        for bus in enumerateSpiBus():
            options = self.leOptions[bus]
            multiData = getMulti(bus)
            if multiData is None or len(multiData) == 0:
                continue
            muxValMax = max(multiData,key=lambda x: x[1])[1]
            if multiData is None or len(multiData) == 0:
                continue
            addOptions = [
                (cs, "LE%u [PCB 915]" % cs, multi)
                for cs, muxVal, multi in multiData
            ]
            options.extend(addOptions)
        self.SelectSpiLE = {
            x: SelectPopupMaker({x0: elem[1] for x0, elem in enumerate(options)}, 0)
            for x, options in enumerate(self.leOptions)}
        if self.cfg.dctIO_Cfg[IO_NONE].selected >= 2:
            for x in self.SelectSpiLE:
                selectSpiLE = self.SelectSpiLE[x]
                for x0, elem in enumerate(self.leOptions[x]):
                    if len(elem) == 2:
                        selectSpiLE.adapter.get_view(x0).disabled = True

    def convProgLstItem1(self, row_index, prog):
        colors0 = (kXM_LIST_BG[row_index % 2], k3F_F_GREEN)
        colors1 = (k5_GREY, k3F_F_RED)
        color0, color1 = (kBLACK, kXM_GREEN1)
        lstDisp = prog.getDispCols()
        clsDicts = []
        colors = (kXM_LIST_BG[row_index % 2], k5_GREY)
        for disp, colW in zip(lstDisp, self.app.lstColW[:len(lstDisp)]):
            clsDicts.append({
                "cls": Label, "kwargs": {
                    "color": kBLACK,
                    "size_hint_x": colW, "text": disp}})
        # clsDicts[1]["kwargs"]["color"] = kXM_GREEN1
        clsDicts[1]["kwargs"]["color"] = kXM_RED
        clsDicts[1]["id"] = "XM_name"
        clsDicts[1]["colors"] = ((kXM_RED, None), (None, None))
        leftOver = 1.0
        elems = 0
        for part in clsDicts:
            kwargs = part["kwargs"]
            kwargs["font_size"] = self.app.progStyle.fntSz
            szX = kwargs["size_hint_x"]
            if szX is not None:
                leftOver -= szX
            else:
                elems += 1
        if elems > 0:
            leftOver /= elems
            for part in clsDicts:
                kwargs = part["kwargs"]
                if kwargs["size_hint_x"] is None: kwargs["size_hint_x"] = leftOver
        # txt = prog.dispName()
        return {
            "size_hint_y": None, "height": szYpx, "cls_dicts": clsDicts,
            "deselected_color": colors[0], "on_is_selected": self.showIO,
            "selected_color": colors[1],  # "on_state": onState,
            "deselected_attrs": dict(
                normal_color=color0,
                down_color=color0,
                background_normal_color=colors0[0],
                background_down_color=colors0[1]),
            "selected_attrs": dict(
                normal_color=color1,
                down_color=color1,
                background_normal_color=colors1[0],
                background_down_color=colors1[1])}

    def showPanel(self, app, parent, wid_index):
        """
        :param AppData app:
        :param parent:
        :param wid_index:
        """
        assert isinstance(app, AppData)
        parent.add_widget(self.panel, wid_index)
        data = filter(lambda (c, x): x.ioType == IO_SPI, enumerate(app.lstProgs))
        self.indexMapper = DynamicMap(lambda x: x[0], data)
        self.progMapper = DynamicMap(lambda x: x[1], data)
        self.adapter = ListAdapter(
            cls=XMOptListItem1, allow_empty_selection=False, selection_mode='single',
            data=self.progMapper, args_converter=self.convProgLstItem1)
        if len(self.indexMapper) == 0:
            self.btnChain.disabled = True
            self.btnBus.disabled = True
            self.btnLE.disabled = True
        else:
            if len(self.adapter.selection) == 0:
                self.adapter.select_list([self.adapter.get_view(0)], False)
        self.theList.adapter = self.adapter
        self.theList.populate()

    def hidePanel(self, app, parent):
        """
        :param AppData app:
        :param parent:
        """
        assert isinstance(app, AppData)
        del app
        parent.remove_widget(self.panel)

    def showIO(self, obj, val=None):
        del val
        if not obj.is_selected: return
        index = self.progIndex = obj.index
        prog = self.prog = self.progMapper[index]
        self.btnBus.disabled = False
        self.btnLE.disabled = False
        self.spiInf = list(getSpiBusInfo(prog.spiBus))
        if prog.spiBus.multi is not None:
            self.spiInf.append(prog.spiBus.multi)
        spiChain = self.spiChain = self.SpiChains[tuple(self.spiInf)]
        self.btnChain.disabled = not prog.allowChain
        if prog.allowChain:
            self.SelectChainPos = SelectPopupMaker(
                self.cfg.progToOptList(spiChain, prog), 0)
        else:
            self.SelectChainPos = None
        """self.btnChain.text = self.fmtChain % (
            self.cfg.progIndexToTxt(lstIndexOf(spiChain, prog)))
        self.btnBus.text = "Set Bus\n%u" % self.spiInf[0]
        self.btnLE.text = "Set LE\n%u" % self.spiInf[1]"""

    def showSelChainPos(self, obj):
        del obj
        prog, spiChain = self.prog, self.spiChain
        self.SelectChainPos.showDialog(self.setChainPos, self.cfg.progToSelIndex(lstIndexOf(spiChain, prog)))

    def showSelSpiBus(self, obj):
        del obj
        spiInf = self.spiInf
        self.SelectSpiBus.showDialog(self.setSpiBus, spiInf[0])

    def setSpiBus(self, useLst, adapter, popup, obj):
        del obj
        popup.dismiss()
        self.spiInf[0] = useLst[adapter.selection[0].index][1]
        self.updateSpiBusLE()

    def showSelSpiLE(self, obj):
        del obj
        spiInf = self.spiInf
        if len(self.spiInf) == 2:
            self.SelectSpiLE[spiInf[0]].showDialog(self.setSpiLE, spiInf[1])
        elif len(self.spiInf) == 3:
            self.SelectSpiLE[spiInf[0]].showDialog(self.setSpiLE, spiInf[1] + len(spiInf[2][1]))

    def setSpiLE(self, useLst, adapter, popup, obj):
        del obj
        popup.dismiss()
        optSelected = self.leOptions[self.spiInf[0]][useLst[adapter.selection[0].index][1]]
        self.spiInf[1] = optSelected[0]
        if len(optSelected) == 3:
            self.spiInf[2:] = [optSelected[2]]
        else:
            self.spiInf[2:] = ""
        self.updateSpiBusLE()

    def updateSpiBusLE(self):
        spiChain = self.spiChain
        prog = self.prog
        if prog in spiChain: spiChain.remove(prog)
        prog.setSpiBus(spiIO_Object(*self.spiInf))
        del self.adapter.cached_views[self.progIndex]
        self.theList.populate()
        self.adapter.select_list([self.adapter.get_view(self.progIndex)], False)

    def setChainPos(self, useLst, adapter, popup, obj):
        del useLst, obj
        popup.dismiss()
        prog, spiChain = self.prog, self.spiChain
        index = self.cfg.selLstToProgIndex(adapter.selection)
        if prog in spiChain: spiChain.remove(prog)
        if index != -1: spiChain.insert(index, prog)
        self.adapter.update_for_new_data()
        self.theList.populate()
        self.adapter.select_list([self.adapter.get_view(self.progIndex)], False)
        """self.btnChain.text = "Set Chain Position\n%s" % (
            self.cfg.progIndexToTxt(lstIndexOf(spiChain, prog)))"""


class IOConfig_Pin(object):
    indexMapper = None
    progMapper = None
    adapter = None
    progIndex = None
    prog = None
    def __init__(self, ioConfig, app):
        """
        :param IOConfigHelpers ioConfig:
        :param AppData app:
        """
        assert isinstance(app, AppData)
        self.fmtPin = "Change\nPin"
        self.app = app
        self.cfg = ioConfig
        self.panel = GridLayout(cols=2, spacing=[5])
        self.theList = BorderedList(bdr=8)
        scroll = filter(lambda x: isinstance(x, ScrollView), self.theList.children)[0]
        scroll.bind(on_scroll_start=ScrollAllower)
        self.panel.add_widget(self.theList)
        self.ordering = enumeratePins(getPinNumFromBCM_ID)
        self.dctOpts = {x: "Pin %u" % getPinNumFromBCM_ID(x) for x in self.ordering}
        # self.orderMap = {x: c for c, x in enumerate(self.ordering)} # redundant: generated in SelectPopupMaker
        self.options = OrderedOptMap(self.dctOpts, self.ordering)
        self.SelectPin = SelectPopupMaker(
            self.options, 0)
        self.btnLayout = GridLayout(cols=1, size_hint_x=.2, spacing=[5])
        self.thePinList = ListView()
        self.btnLayout.add_widget(self.thePinList)
        self.panel.add_widget(self.btnLayout)

    def updatePins(self):
        self.ordering = enumeratePins(getPinNumFromBCM_ID)
        self.dctOpts = {x: "Pin %u" % getPinNumFromBCM_ID(x) for x in self.ordering}
        self.options = OrderedOptMap(self.dctOpts, self.ordering)
        self.SelectPin = SelectPopupMaker(
            self.options, 0)

    def ioConvLstPin(self, row_index, rec):
        name, pin = rec
        assert isinstance(pin, PinIO_Object)
        Rtn = convLstOpts(
            row_index, (
                "%s: pin %u" % (name, getPinNumFromBCM_ID(pin.pinNum)),))
        Rtn.update({
            "height": szYpx,
            "on_release": lambda obj: self.SelectPin.showDialog(Bind(self.setPin, row_index), pin.pinNum),
            "normal_color": self.cfg.off[1],
            "down_color": self.cfg.on[1],
            "background_normal_color": Rtn.pop("background_color", self.cfg.off[0]),
            "background_down_color": self.cfg.on[0]})
        Rtn.pop("color", None)
        return Rtn

    def convProgLstItem1(self, row_index, prog):
        colors0 = (kXM_LIST_BG[row_index % 2], k3F_F_GREEN)
        colors1 = (k5_GREY, k3F_F_RED)
        color0, color1 = (kBLACK, kXM_GREEN1)
        lstDisp = prog.getDispCols()
        clsDicts = []
        colors = (kXM_LIST_BG[row_index % 2], k5_GREY)
        for disp, colW in zip(lstDisp, self.app.lstColW[:len(lstDisp)]):
            clsDicts.append({
                "cls": Label, "kwargs": {
                    "color": kBLACK,
                    "size_hint_x": colW, "text": disp}})
        # clsDicts[1]["kwargs"]["color"] = kXM_GREEN1
        clsDicts[1]["kwargs"]["color"] = kXM_RED
        clsDicts[1]["id"] = "XM_name"
        clsDicts[1]["colors"] = ((kXM_RED, None), (None, None))
        leftOver = 1.0
        elems = 0
        for part in clsDicts:
            kwargs = part["kwargs"]
            kwargs["font_size"] = self.app.progStyle.fntSz
            szX = kwargs["size_hint_x"]
            if szX is not None:
                leftOver -= szX
            else:
                elems += 1
        if elems > 0:
            leftOver /= elems
            for part in clsDicts:
                kwargs = part["kwargs"]
                if kwargs["size_hint_x"] is None: kwargs["size_hint_x"] = leftOver
        # txt = prog.dispName()
        return {
            "size_hint_y": None, "height": szYpx, "cls_dicts": clsDicts,
            "deselected_color": colors[0], "on_is_selected": self.showIO,
            "selected_color": colors[1],  # "on_state": onState,
            "deselected_attrs": dict(
                normal_color=color0,
                down_color=color0,
                background_normal_color=colors0[0],
                background_down_color=colors0[1]),
            "selected_attrs": dict(
                normal_color=color1,
                down_color=color1,
                background_normal_color=colors1[0],
                background_down_color=colors1[1])}

    def showPanel(self, app, parent, wid_index):
        """
        :param AppData app:
        :param parent:
        :param wid_index:
        """
        assert isinstance(app, AppData)
        parent.add_widget(self.panel, wid_index)
        data = filter(lambda (c, x): x.ioType == IO_PIN, enumerate(app.lstProgs))
        self.indexMapper = DynamicMap(lambda x: x[0], data)
        self.progMapper = DynamicMap(lambda x: x[1], data)
        self.adapter = ListAdapter(
            cls=XMOptListItem1, allow_empty_selection=False, selection_mode='single',
            data=self.progMapper, args_converter=self.convProgLstItem1)
        if len(self.indexMapper) > 0:
            if len(self.adapter.selection) == 0:
                self.adapter.select_list([self.adapter.get_view(0)], False)
        self.theList.adapter = self.adapter
        self.theList.populate()

    def hidePanel(self, app, parent):
        """
        :param AppData app:
        :param parent:
        """
        assert isinstance(app, AppData)
        del app
        parent.remove_widget(self.panel)

    def showIO(self, obj, val=None):
        del val
        if not obj.is_selected: return
        index = self.progIndex = obj.index
        prog = self.prog = self.progMapper[index]
        thePinList = self.thePinList
        thePinList.adapter = ListAdapter(
            cls=XMButton, allow_empty_selection=True, selection_mode='single',
            data=zip(prog.getPinNames(), prog.pins), args_converter=self.ioConvLstPin)
        thePinList.populate()

    def setPin(self, pinIndex, useLst, adapter, popup, obj):
        del obj
        popup.dismiss()
        prog = self.prog
        prog.setSinglePin(pinIndex, PinIO_Object(useLst[adapter.selection[0].index][1]))
        del self.adapter.cached_views[self.progIndex]
        self.theList.populate()
        self.adapter.select_list([self.adapter.get_view(self.progIndex)], False)


class IOConfig_Board(object):
    def __init__(self, ioConfig, app):
        """
        :param IOConfigHelpers ioConfig:
        :param AppData app:
        """
        self.app = app
        self.cfg = ioConfig
        panel = self.panel = GridLayout(rows=3, spacing=[5])
        callbacks = [
            lambda obj: self.select40PinPassthru(),
            lambda obj: self.select26PinPassthru(),
            lambda obj: self.select26PinMuxed()
        ]
        self.widgets = [
            (
                ClkTxtIn(text="40 Pin Connector\n  32x32 plate (Passthrough)"),
                ClkTxtIn(text="Board # 0393")
            ),
            (
                ClkTxtIn(text="26 Pin Connector\n  16x16 plate (Passthrough)"),
                ClkTxtIn(text="Board # 0655")
            ),
            (
                ClkTxtIn(text="26 Pin Connector\n  32x32 plate (Multiplexed SPI LE)"),
                ClkTxtIn(text="Board # 0915")
            )
        ]
        for c, (a, b) in enumerate(self.widgets):
            panel.add_widget(a)
            panel.add_widget(b)
            for x in (a, b):
                x.bdw = 0
                x.bdr = 0
                x.bgc = (1, 1, 1, 1)
                # x.background_color = (1, 1, 1)
                # x.color = (0, 0, 0)
                # x.background_disabled_normal = ""
                # x.background_disabled_down = ""
                # x.background_down = ""
                # x.background_normal = ""
            a.bind(on_release=callbacks[c])
            b.bind(on_release=callbacks[c])

    def setColor(self, index):
        self.selected = index
        for c, (a, b) in enumerate(self.widgets):
            color = (0.5, 1, 0.5, 1) if c == index else (1, 1, 1, 1)
            Button().background_disabled_down = ""
            a.bgc = color
            b.bgc = color

    def select40PinPassthru(self):
        self.setColor(0)
        keys = [k for k in self.app.SpiChains if len(k) == 3]
        for k in keys:
            del self.app.SpiChains[k]
        for prog in self.app.lstProgs:
            if prog.spiBus.multi is None:
                continue
            bus = prog.spiBus.bus
            prog.setSpiBus(spiIO_Object(bus, enumerateSpiBusCs(bus)))
        setMultiEnabled(False)
        setDisabledPins(set())
        self.cfg.dctIO_Cfg[IO_SPI].updateSpiLE()
        self.cfg.dctIO_Cfg[IO_PIN].updatePins()


    def select26PinPassthru(self):
        self.setColor(1)
        keys = [k for k in self.app.SpiChains if len(k) == 3]
        for k in keys:
            del self.app.SpiChains[k]
        for prog in self.app.lstProgs:
            if prog.spiBus.multi is None:
                continue
            bus = prog.spiBus.bus
            prog.setSpiBus(spiIO_Object(bus, enumerateSpiBusCs(bus)))
        setMultiEnabled(False)
        setDisabledPins({5, 6, 12, 13, 16, 19, 20, 21, 26})
        self.cfg.dctIO_Cfg[IO_SPI].updateSpiLE()
        self.cfg.dctIO_Cfg[IO_PIN].updatePins()

    def select26PinMuxed(self):
        self.setColor(2)
        setMultiEnabled(True)
        setDisabledPins({5, 6, 12, 13, 16, 19, 20, 21, 26})
        for bus in enumerateSpiBus():
            multiData = getMulti(bus)
            if multiData is None:
                continue
            for (cs, muxVal, multi) in multiData:
                self.app.SpiChains[(bus, cs, multi)] = []
        self.cfg.dctIO_Cfg[IO_SPI].updateSpiLE()
        self.cfg.dctIO_Cfg[IO_PIN].updatePins()

    def showPanel(self, app, parent, wid_index):
        """
        :param AppData app:
        :param parent:
        :param wid_index:
        """
        assert isinstance(app, AppData)
        parent.add_widget(self.panel, wid_index)

    def hidePanel(self, app, parent):
        """
        :param AppData app:
        :param parent:
        """
        assert isinstance(app, AppData)
        del app
        parent.remove_widget(self.panel)


class IOConfigHelpers(BaseScreen):
    on = (k5_GREY, kXM_GREEN1, k3F_F_RED)
    off = (kWHITE, kBLACK, k3F_F_GREEN)
    dis = (kA_GREY, k4_GREY)

    @staticmethod
    def selToProgIndex(selIndex):
        return selIndex - 1

    @staticmethod
    def progToSelIndex(progIndex):
        return progIndex + 1

    noSelIndex = -1

    @staticmethod
    def progIndexToTxt(progIndex):
        return str(progIndex) if progIndex != -1 else "Not Set"

    @classmethod
    def progSpiChainToTxt(cls, spiChain, prog):
        return cls.progIndexToTxt(lstIndexOf(spiChain, prog))

    @classmethod
    def progToOptList(cls, spiChain, prog):
        rtn = list(xrange(-1, len(spiChain) + int(prog not in spiChain)))
        for c in xrange(len(rtn)):
            rtn[c] = (True, cls.progIndexToTxt(rtn[c]))
        return rtn

    @classmethod
    def selLstToProgIndex(cls, selection):
        return cls.selToProgIndex(selection[0].index if len(selection) else cls.noSelIndex)

    def showIO(self, index):
        if self.curSelect is not None:
            ioType = self.lstIOTypes[self.curSelect]
            self.dctIO_Cfg[ioType].hidePanel(self.app, self.panel)
        else:
            self.panel.remove_widget(self.dummy)
        if self.curSelect == index:
            index = 0
        self.curSelect = index
        ioType = self.lstIOTypes[index]
        self.colorize()
        self.dctIO_Cfg[ioType].showPanel(self.app, self.panel, 0)

    def unShowIO(self):
        if self.curSelect is not None:
            ioType = self.lstIOTypes[self.curSelect]
            self.dctIO_Cfg[ioType].hidePanel(self.app, self.panel)
            self.curSelect = None
            self.panel.add_widget(self.dummy, 0)
            self.colorize()

    def getName(self):
        return "I/O Config"

    def build(self, app, ext_data=None):
        """
        :param AppData app:
        :param None ext_data:
        """
        assert isinstance(app, AppData)
        assert isinstance(app.wnd, FloatLayout)
        super(IOConfigHelpers, self).build(app)
        app.wnd.add_widget(self.panel)

    def hide(self, app):
        """
        :param AppData app:
        """
        assert isinstance(app, AppData)
        assert isinstance(app.wnd, FloatLayout)
        super(IOConfigHelpers, self).hide(app)
        app.wnd.remove_widget(self.panel)
        self.unShowIO()

    def getHelp(self, app):
        """
        :param AppData app:
        """
        return None

    def __init__(self, app):
        """
        :param AppData app:
        """
        assert isinstance(app, AppData)
        self.logObj = app.logObj
        self.app = app
        style = dict(app.defStyle)
        style["size_hint"] = list(style["size_hint"])
        style["pos_hint"] = dict(style["pos_hint"])
        style["pos_hint"]['x'] += (1 - .99) * style["size_hint"][0] / 2
        style["size_hint"][0] *= .99
        style["size_hint"][1] *= .98
        self.panel = GridLayout(cols=2, spacing=[5], **style)
        self.btnPanel = GridLayout(cols=1, size_hint_x=.1)
        self.panel.add_widget(self.btnPanel)
        self.lstBtns = [
            XMLoneButton(text="SPI /\n3 Wire"),
            XMLoneButton(text="PIN /\nParallel"),
            XMLoneButton(text="IFACE /\nInterface\nBoard"),
        ]
        self.curSelect = None

        self.lstIOTypes = [
            IO_SPI,
            IO_PIN,
            IO_NONE
        ]
        self.dctIO_Cfg = {
            IO_SPI: IOConfig_Spi(self, app),
            IO_PIN: IOConfig_Pin(self, app),
            IO_NONE: IOConfig_Board(self, app)
        }
        for c, btn in enumerate(self.lstBtns):
            btn.on_release = Bind(self.showIO, c)
            btn.disabled_color = self.dis[1]
            btn.disabled_background_color = self.dis[0]
            self.btnPanel.add_widget(btn)
        self.dummy = Label(text="Select Interface to Configure", font_size=32, size_hint_x=.9)
        self.panel.add_widget(self.dummy)
        self.colorize()

    def colorize(self):
        for c, btn in enumerate(self.lstBtns):
            if c == self.curSelect:
                btn.down_color = btn.normal_color = self.on[1]
                btn.background_normal_color = self.on[0]
                btn.background_down_color = self.on[2]
            else:
                btn.down_color = btn.normal_color = self.off[1]
                btn.background_normal_color = self.off[0]
                btn.background_down_color = self.off[2]