"""
TinyLib: Tiny and much more standalone than the other modules in this project
"""
import os
import codecs
import struct
from DataPackers import *
from decimal import Decimal
import traceback

class BaseContext(object):
    def __add__(self, other): return AddContext(self, other)
    def __mul__(self, other): return MulContext(self, other)
    def __enter__(self): pass
    def __exit__(self, exc_t, exc_v, tb): return False


class OpContext(BaseContext):
    def __init__(self, Cntxt0, Cntxt1):
        self.Cntxt0 = Cntxt0
        self.Cntxt1 = Cntxt1
    def __enter__(self):
        self.Cntxt0.__enter__()
        self.Cntxt1.__enter__()
    def __exit__(self, exc_t, exc_v, tb):
        Res0 = self.Cntxt0.__exit__(exc_t, exc_v, tb)
        Res1 = self.Cntxt1.__exit__(exc_t, exc_v, tb)
        return self.OpFunc(Res0, Res1)
    def OpFunc(self, Res0, Res1): return False


class AddContext(OpContext):
    def OpFunc(self, Res0, Res1): return Res0 or Res1


class MulContext(OpContext):
    def OpFunc(self, Res0, Res1): return Res0 and Res1


class VarContext(BaseContext):
    def __init__(self, Var, DctAttr):
        self.DctAttr = DctAttr
        self.Var = Var
    def __enter__(self):
        self.DctOrig = {k:getattr(self.Var, k) for k in self.DctAttr}
        for k in self.DctAttr: setattr(self.Var, k, self.DctAttr[k])
    def __exit__(self, exc_t, exc_v, tb):
        for k in self.DctOrig: setattr(self.Var, k, self.DctOrig[k])


class FuncContext(BaseContext):
    #e1 means pass exception data to f1
    def __init__(self,f0,f1,e1,args0=None,args1=None,kwargs0=None,kwargs1=None):
        self.f0 = f0
        self.f1 = f1
        self.e1 = e1
        self.args0 = list() if args0 is None else args0
        self.args1 = list() if args1 is None else args1
        self.kwargs0 = dict() if kwargs0 is None else kwargs0
        self.kwargs1 = dict() if kwargs1 is None else kwargs1
    def __enter__(self): self.f0(*self.args0, **self.kwargs0)
    def __exit__(self, exc_t, exc_v, tb):
        if self.e1: return self.f1(exc_t, exc_v, tb, *self.args1, **self.kwargs1)
        else: return self.f1(*self.args1, **self.kwargs1)


class GetSetContext(BaseContext):
    def __init__(self, Var, AttrGetter, AttrSetter, SetVal, *args):
        self.FuncAttrs = (AttrGetter, AttrSetter)
        self.Var = Var
        self.SetVal = SetVal
        self.args = args
    def __enter__(self):
        self.OrigVal = getattr(self.Var, self.FuncAttrs[0])(*self.args)
        getattr(self.Var, self.FuncAttrs[1])(*(list(self.args)+[self.SetVal]))
    def __exit__(self, exc_t, exc_v, tb):
        getattr(self.Var, self.FuncAttrs[1])(*(list(self.args)+[self.OrigVal]))


class GetSetContext1(BaseContext):
    def __init__(self, Getter, Setter, k, SetVal, ArgsGet, ArgsSet):
        self.Getter = Getter
        self.Setter = Setter
        self.k = k
        self.ArgsGet = ArgsGet
        self.ArgsSet = ArgsSet
        self.SetVal = SetVal
    def __enter__(self):
        self.OrigVal = self.Getter(self.k, *self.ArgsGet)
        self.Setter(self.k, self.SetVal, *self.ArgsSet)
    def __exit__(self, exc_t, exc_v, tb):
        self.Setter(self.k, self.OrigVal, *self.ArgsSet)

"""
# get time string for use in log files or text data
def GetTmFmt():
    Tm = time.localtime()
    return "%u-%02u-%02u %02u:%02u:%02u" % (Tm.tm_year, Tm.tm_mon, Tm.tm_mday, Tm.tm_hour, Tm.tm_min, Tm.tm_sec)


# get time string for use in filenames
def fNameSafeGetTmFmt():
    Tm = time.localtime()
    return "%u-%02u-%02u_%02u.%02u.%02u" % (Tm.tm_year, Tm.tm_mon, Tm.tm_mday, Tm.tm_hour, Tm.tm_min, Tm.tm_sec)


class Logger:
    Lvls = ["SEVERE", "ERROR", "WARNING", "INFO", "FINE", "FINER", "FINEST"]
    def __init__(self, Module, PrnLvl, FlObj, FlLvl, ErrLvl=2, LogLk=None):
        self.Mod = Module
        self.Fl = FlObj
        self.PrnLvl = PrnLvl
        self.FlLvl = FlLvl
        self.LogLk = LogLk if LogLk is not None else threading.Lock()
        self.ErrLvl = ErrLvl
    def Log(self, Str, Lvl):
        if Lvl <= self.PrnLvl or Lvl <= self.FlLvl:
            StrLog = GetTmFmt() + " [" + Logger.Lvls[Lvl] + "][" + self.Mod + "] " + Str + "\n"
            with self.LogLk:
                if Lvl <= self.PrnLvl:
                    if Lvl <= self.ErrLvl:sys.stderr.write(StrLog)
                    else:sys.stdout.write(StrLog)
                if Lvl <= self.FlLvl:
                    self.Fl.write(StrLog)
                    self.Fl.flush()
    def SpawnLog(self, Module, PrnLvl=None, FlLvl=None, ErrLvl=2):
        if PrnLvl is None: PrnLvl = self.PrnLvl
        if FlLvl is None: FlLvl = self.FlLvl
        return Logger(Module, PrnLvl, self.Fl, FlLvl, ErrLvl, self.LogLk)
"""


class AliasedDict(dict):
    def __init__(self, *args, **kwargs):
        super(AliasedDict, self).__init__(*args, **kwargs)
        self.Aliases = {}
    #makes self[k1] represent self[k0]
    def addAlias(self, k0, k1):
        supObj = super(AliasedDict, self)
        self.Aliases[k0] = self.Aliases.get(k0, {k0}) | {k1}
        self.Aliases[k1] = set(self.Aliases[k0])
        for k in self.Aliases[k1]:
            self.Aliases[k] = self.Aliases.get(k, {k}) | self.Aliases[k1]
        for k in self.Aliases[k0]:
            supObj.__setitem__(k, supObj.__getitem__(k0))
    def remAlias(self, k):
        if k in self.Aliases:
            for k0 in self.Aliases[k]:
                self.Aliases[k0].discard(k)
            self.Aliases[k] = {k}
    def __setitem__(self, k, v):
        supObj = super(AliasedDict, self)
        if k not in self.Aliases:
            self.Aliases[k] = {k}
        for k0 in self.Aliases[k]:
            supObj.__setitem__(k0, v)
    def __delitem__(self, k):
        if k in self.Aliases:
            for k0 in self.Aliases[k]: self.Aliases[k0].discard(k)
            del self.Aliases[k]
        super(AliasedDict, self).__delitem__(k)


#Acts like a javascript object (except when accessing undefined properties)
#  NOTE: emulation of a default (undefined) value for keys not in inst
#    is possible by using NoCpy=True and passing in a dictionary that
#    has been given a default value of None.
#x <==> y, means x is interchangeable with y.
#  In other words x does the same thing as y.
#inst[attrName] = n <==> setattr(inst,attrName,n)
#n = inst[attrName] <==> n = getattr(inst,attrName)
#del inst[attrName] <==> delattr(inst,attrName
#k in inst <==> hasattr(inst, k)
#for k in inst <==> for k in dir(inst)
class ValuePack(object):
    def __init__(self, DctAttr = None, NoCpy=False):
        if DctAttr is None: self.__dict__ = {}
        else: self.__dict__ = DctAttr if NoCpy else dict(DctAttr)
    def __getitem__(self, Attr):
        return self.__dict__[Attr]
    def __setitem__(self, Attr, Val):
        self.__dict__[Attr] = Val
    def __delitem__(self, Attr):
        del self.__dict__[Attr]
    def __contains__(self, Attr):
        return Attr in self.__dict__
    def __iter__(self):
        return self.__dict__.__iter__()
    def __repr__(self):
        return "ValuePack(%s)" % repr(self.__dict__)
    def __str__(self):
        return str(self.__dict__)


def DefFltrFunc(Attr):
    return not Attr.startswith("__")


def ValuePackFromObj(Obj, FltrFunc=DefFltrFunc):
    return ValuePack({k:getattr(Obj, k) for k in dir(Obj) if FltrFunc(k)})


class SlowerValuePack(ValuePack):
    def __setattr__(self, Attr, Val):
        if Attr == "__dict__":
            super(SlowerValuePack, self).__setattr__(Attr, Val)
        else: self[Attr] = Val
    def __getattr__(self, Attr):
        if Attr == "__dict__":
            return getattr(super(SlowerValuePack, self), Attr)
        else: return self[Attr]
    def __delattr__(self, Attr):
        if Attr == "__dict__":
            super(SlowerValuePack, self).__delattr__(Attr)
        else: del self[Attr]
    def __repr__(self):
        return "SlowerValuePack(%s)" % repr(self.__dict__)


#StrictDict: Strict Dictionary that will throw exceptions if invalid data is
#  assigned to a key depending on the result of x, Msg = DctVerify[key](data)
#  x == 0: valid data
#  x == 1: TypeError(Msg)
#  x == 2: ValueError(Msg)
#  else: throws Msg
class StrictDict(dict):
    def __init__(self, DctVerify, *args, **kwargs):
        super(StrictDict, self).__init__(*args, **kwargs)
        self.DctVerify = dict(DctVerify)
    #adds a verification function that, when called as Func(v)
    #  will return 0 to allow the assignment of v to self[k]
    #  will return x,Msg to prevent self[k] being set to v
    #    x=1 for TypeError
    #    x=2 for ValueError
    def AddVerify(self, k, Func):
        if k not in self.DctVerify:
            self.DctVerify[k] = set()
        self.DctVerify[k].add(Func)
    def RemVerify(self, k, Func):
        if k not in self.DctVerify or Func not in self.DctVerify[k]:return
        self.DctVerify[k].remove(Func)
    def __setitem__(self, k, v):
        for Func in self.DctVerify.get(k, set()):
            x, Msg = Func(v)
            if x == 0: continue
            elif x == 1: Msg = TypeError(Msg)
            elif x == 2: Msg = ValueError(Msg)
            Txt = Msg.message
            Txt = "Failed set on %s ERROR: " % repr(k) + Txt
            Msg.args = (Txt,)
            raise Msg
        super(StrictDict, self).__setitem__(k, v)


def mkAliased(inst):
    if isinstance(inst.__dict__, AliasedDict): return inst
    return SlowerValuePack(AliasedDict(inst.__dict__), NoCpy=True)


class FnCallGetItem(object):
    def __init__(self, fn):
        self.fn = fn
    def __getitem__(self, k):
        return self.fn(k)


# indexed version of python3 map
class DynamicMap(object):
    def __init__(self, func, lstObj):
        self.lstObj = lstObj
        self.func = func
    def __getitem__(self, index):
        if isinstance(index, slice): return map(self.func, self.lstObj[index])
        else: return self.func(self.lstObj[index])
    def __setitem__(self, index, val):
        raise NotImplementedError("DynamicMap does not support item assignment")
    def __iter__(self):
        for item in self.lstObj: yield self.func(item)
    def __len__(self):
        return len(self.lstObj)


#a class where an instance acts like a list where each index represents a key
#  in the dictionary(dct) object that has been assigned to it
class DctListRepr(object):
    def __init__(self, dct, size, default):
        self.dct = dct
        self.size = size
        self.default = default
    def __getitem__(self, index):
        if index < 0: index += self.size
        return self.dct.get(index, self.default)
    def __setitem__(self, index, val):
        raise NotImplementedError(
            "DctListRepr does not support item assignment")
    def __iter__(self):
        for c in xrange(self.size): yield self[c]
    def __len__(self):
        return self.size


# like Bind from java-script
def Bind(Func, *args, **kwargs):
    def Bound(*CurArgs, **CurKwargs):
        return Func(*(args + CurArgs), **dict(CurKwargs, **kwargs))
    return Bound


def restrict(num, a, b):
    if num < a: return a
    elif num > b: return b
    return num


def tryExceptWrap(log, reRaise=False, logLvl=2):
    def Rtn(fn):
        def Function(*args, **kwargs):
            # noinspection PyBroadException
            try:
                return fn(*args, **kwargs)
            except:
                log.Log("result from function %s:\n"%fn.__name__ + traceback.format_exc()[:-1], logLvl)
                if reRaise: raise
        Function.__name__ = fn.__name__
        Function.__doc__ = fn.__doc__
        Function.__module__ = fn.__module__
        return Function
    return Rtn


#returns the index where Val could be inserted in List for List to remain sorted
def BinaryApprox(List, Val):
    Len = len(List)
    CurLen = Len
    Pos = Len >> 1
    while CurLen > 0:
        if Val <= List[Pos] and (Pos == 0 or Val >= List[Pos - 1]):
            return Pos
        elif Val < List[Pos]:
            CurLen >>= 1
            Pos -= (CurLen + 1) >> 1
        else:
            CurLen = (CurLen - 1) >> 1
            Pos += (CurLen >> 1) + 1
    return Len


def BinaryApprox1(List, Val):
    Len = len(List)
    CurLen = Len
    Pos = Len >> 1
    while CurLen > 0:
        if (Pos+1 >= Len or Val < List[Pos+1]) and  Val >= List[Pos]:
            return Pos
        elif Val < List[Pos]:
            CurLen >>= 1
            Pos -= (CurLen + 1) >> 1
        else:
            CurLen = (CurLen - 1) >> 1
            Pos += (CurLen >> 1) + 1
    return Len


def dummyKeyFn(lst, i):
    return lst[i]


def BisectSearchBase(Lst, Val, KeyFn=dummyKeyFn):
    End = len(Lst)
    Begin = 0
    while End - 2 > Begin:
        Pivot = (End + Begin) >> 1
        v = KeyFn(Lst, Pivot)
        if Val > v: Begin = Pivot + 1
        else: End = Pivot + 1
    while End - Begin and KeyFn(Lst, Begin) < Val:
        Begin += 1
    while End - Begin and Val < KeyFn(Lst, End - 1):
        End -= 1
    return Begin, End


#iterative version of GCD
def FastGcd(a, b):
    while b != 0:
        r = a % b
        a = b
        b = r
    return a


#Simplify Fraction n/d
def sFrac(n, d):
    a = FastGcd(n, d)
    return n/a, d/a


#Tries multiplying 'n' and 'd' by 'mulBy' a maximum of 'maxMul' times
#  or when 'sFrac(n, d)' are less than 'n, d'
#  before simplifying fraction n/d
def sFracBetter(n, d, maxMul=10, mulBy=10):
    while n % 1 > 0 or d % 1 > 0:
        n *= mulBy
        d *= mulBy
        maxMul -= 1
        if maxMul <= 0: break
    return sFrac(int(n), int(d))


def adder(amt=1):
    return lambda x:x + amt


#ranges are used to stop the user from entering in bad data
#  str(BaseRange instance) should return a relatively human readable string
#    of what 'satisfies' that instance
#  userInfo will return a more human readable string than str(BaseRange instance)
#  maxLen will return a positive definite maximum length that a string could be
#    and still satisfy the instance
#  maxLen will return -1 if there is not a maximum length
#ranges may also be used for non numerical types to validate the data
class BaseRange(object):
    wantsExtData = False
    def __init__(self, isNeg=True, val=True, noneable=False):
        self.val, self.isNeg,self.noneable = val, isNeg, noneable
        self.verifyOnChg = False
    def __repr__(self): return "BaseRange" + repr((self.isNeg, self.val))
    def __str__(self):
        if not self.val: return "False"
        if self.isNeg: return "True"
        return "x >= 0"
    def userInfo(self, title, conv=None, extData=None): return str(self).replace("x", title)
    def satisfies(self, x, extData=None): return self.val and (self.isNeg or x >= 0)
    def maxLen(self, conv, extData=None): return -1
    def getMinMax(self, extData=None):
        """
        :param any extData:
        :rtype: (any, any)|None
        """
        return None
    def __call__(self, x, extData=None): return self.satisfies(x, extData)

class BaseNumRange(BaseRange):
    def __str__(self):
        minMax = self.getMinMax()
        if minMax is None: minMax = ('?', '?')
        return "x >= %s and x <= %s" % (minMax[0], minMax[1])
    def maxLen(self, conv, extData=None):
        minMax = self.getMinMax(extData)
        if minMax is None:
            return -1
        return max(map(lambda x: len(conv[1](x).replace("-", "")), minMax))
    def userInfo(self, title, conv=None, extData=None):
        minMax = self.getMinMax(extData)
        Min = '?'
        Max = '?'
        if minMax is not None:
            cnv = str if conv is None else conv[1]
            Min = cnv(minMax[0])
            Max = cnv(minMax[1])
        return "%s must be between %s and %s" % (title, Min, Max)
    def satisfies(self, x, extData=None):
        minMax = self.getMinMax(extData)
        if minMax is None:
            return super(BaseNumRange, self).satisfies(x, extData)
        else:
            return minMax[0] <= x <= minMax[1]
class NumRange(BaseNumRange):
    def __init__(self, Min, Max, noneable=False):
        super(NumRange, self).__init__(Min < 0, True, noneable)
        self.Min = Min
        self.Max = Max
        self.verifyOnChg = all(map(isinstance, [Min, Max], ((int,long),)*2))
    def getMinMax(self, extData=None):
        return self.Min, self.Max
    def __repr__(self):
        args = (self.Min,self.Max)
        if self.noneable is not False:
            args = args + (self.noneable,)
        return "NumRange" + repr(args)


class InListRange(BaseRange):
    def __init__(self, lst, inv=0):
        super(InListRange, self).__init__()
        self.lst = lst
        self.inv = bool(inv)
    def __str__(self):
        return "x %s " % ["in", "not in"][self.inv] + repr(self.lst)
    def __repr__(self):
        return "InListRange" + repr((self.lst, self.inv))
    def userInfo(self, title, conv=None, extData=None):
        rtn = " must%s be one of: %s"
        rtn = rtn % (" not" if self.inv else "", ", ".join(map(str, self.lst)))
        return title + rtn
    def satisfies(self, x, extData=None):
        return (x in self.lst) ^ self.inv
    def maxLen(self, conv, extData=None): return -1


class NumListRange(InListRange):
    def __init__(self, lst, inv=0):
        lst = sorted(lst)
        super(NumListRange, self).__init__(lst, inv)
        self.Min = lst[0]
        self.Max = lst[-1]
        self.verifyOnChg = all(map(lambda x: isinstance(x, (int,long)), lst))
        self.noneable = False
        self.isNeg = self.Min < 0
    def getMinMax(self, extData=None):
        return self.Min, self.Max
    def maxLen(self, conv, extData=None):
        MinStr = conv[1](self.Min).replace("-", "")
        MaxStr = conv[1](self.Max).replace("-", "")
        return max(map(len, [MinStr, MaxStr]))


def Filter(Iter, func):
    if isinstance(Iter, dict): return {k:Iter[k] for k in Iter if func(k)}
    elif isinstance(Iter, (list, tuple)): return filter(func, Iter)
    elif isinstance(Iter, ValuePack):
        return ValuePack(Filter(Iter.__dict__, func))
    else:
        raise TypeError(
            "Iter must satisfy isinstance(Iter, (dict,ValuePack,list,tuple))")


def makeDctKeys(keys, valueFn, fnArgs, passKey=0):
    if not passKey:
        return {k: valueFn(*fnArgs) for k in keys}
    elif passKey == 1:
        return {k: valueFn(k, *fnArgs) for k in keys}
    else:
        return {k: valueFn(*(fnArgs + [k])) for k in keys}

def makeListLen(n):
    return [None] * n

#stacks functions in args
#args[0](args[1](args[2](x)))
def Stack(*args):
    if len(args) == 1: return args[0]
    elif len(args) == 2: return lambda x: args[0](args[1](x))
    else:
        def RtnFunc(x):
            for Func in args[::-1]: x = Func(x)
            return x
        return RtnFunc


#C/C++ style of string repr
def C_Repr(Str):
    return "\"%s\"" % Str.encode("string_escape").replace("\"", "\\\"")


def MyRepr(obj):
    if isinstance(obj, (str, unicode)): return C_Repr(obj)
    elif isinstance(obj, (tuple, list, set, dict)):
        MidStr = ", ".join(
            ["%s: %s" % (MyRepr(k), MyRepr(obj[k])) for k in obj]
            if isinstance(obj, dict) else
            map(C_Repr, obj)
        )
        if isinstance(obj, tuple): return "(%s)" % MidStr
        elif isinstance(obj, list): return "[%s]" % MidStr
        elif len(obj) > 0 or isinstance(obj, dict): return "{%s}" % MidStr
        else: return "set()"
    else: return repr(obj)


def Repr80Ch(obj, OpenNewLine=True, SpaceAdd=0):
    del OpenNewLine
    # Fixme: maybe OpenNewLine means something
    lstLines = [""]
    if isinstance(obj, dict):
        lstLines[0] = "{"
        SpaceAdd += 4
        lstLines.append(" " *SpaceAdd)
        lstK = sorted(obj)
        for k, c in zip(lstK, xrange(len(lstK))):
            StrAdd = "%s: %s"
            if c + 1 < len(lstK): StrAdd += ","
            else: StrAdd += "}"
            StrAdd %= (MyRepr(k), MyRepr(obj[k]))
            lenCur = len(lstLines[-1])
            if lenCur > SpaceAdd: lenCur += 1
            if lenCur + len(StrAdd) >= 80:
                lstLines.append(" "*SpaceAdd)
            else:
                if lenCur > SpaceAdd: StrAdd = " " + StrAdd
            lstLines[-1] += StrAdd
    elif isinstance(obj, list):
        lstLines[0] = "["
        SpaceAdd += 4
        lstLines.append(" " * SpaceAdd)
        for c in xrange(len(obj)):
            StrAdd = "%s"
            if c + 1 < len(obj): StrAdd += ","
            else: StrAdd += "]"
            StrAdd %= MyRepr(obj[c])
            lenCur = len(lstLines[-1])
            if lenCur > SpaceAdd: lenCur += 1
            if lenCur + len(StrAdd) >= 80:
                lstLines.append(" "*SpaceAdd)
            else:
                if lenCur > SpaceAdd: StrAdd = " " + StrAdd
            lstLines[-1] += StrAdd
    else: return repr(obj)
    return "\n".join(lstLines)


def closeEnough(a, b, rel_tol=2**-52, abs_tol=2**-52):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


# data sheet helper
def lstOptsFromDS_Str(Str):
    return Repr80Ch(map(
        lambda x: x.title().replace("Rf", "RF").replace("If", "IF"),
        Str.split("\n")))


#updates the dictionary getattr(obj,dctMain) with the dictionary dctUpdate with
#  keys not present in dctMain
#if not hasattr(obj, dctMain) then do nothing
def updDctIfExist(obj, dctMain, dctUpdate):
    if not hasattr(obj, dctMain): return
    setattr(obj, dctMain, dict(getattr(obj, dctMain)))
    dctMain = getattr(obj, dctMain)
    for k in dctUpdate:
        if k in dctMain: continue
        dctMain[k] = dctUpdate[k]


# copies data at keys from dctDefs that are absent from dctMain to dctMain
def fillDictDefs(dctMain, dctDefs):
    for k in dctDefs:
        if k not in dctMain:
            dctMain[k] = dctDefs[k]


def dctDefaultI(obj, default):
    fillDictDefs(obj, default)
    return obj


def dctDefault(obj, default):
    rtn = dict(obj)
    return dctDefaultI(rtn, default)


def dctAssignI(obj, assign):
    obj.update(assign)
    return obj


def dctAssign(obj, assign):
    rtn = dict(obj)
    return dctAssignI(rtn, assign)


def lstGetDefault(lst, index, default=None, key=None):
    return (
        default
        if index >= len(lst) else (
            lst[index]
            if key is None else
            key(lst[index])
        )
    )


def lstFind(lst, value, key=None):
    nLst = len(lst)
    c = 0
    if key is None:
        while c < nLst:
            if lst[c] == value: break
            c += 1
    else:
        while c < nLst:
            if key(lst[c]) == value: break
            c += 1
    return c


def getInsertionSlice(index):
    return slice(index, index)


def mkNormPath(path):
    s = path.startswith("/")
    path = "/".join([x for x in path.split("/") if len(x)])
    if s: path = "/" + path
    return path


# expects both paths to only contain '/' component separators
def is_path_equal(path0, path1):
    if "//" not in path0 and "//" not in path1:
        a = path0.endswith("/")
        b = path1.endswith("/")
        if a == b:
            return path0 == path1
        elif a:
            return path0[:-1] == path1
        elif b:
            return path0 == path1[:-1]
    return filter(len, path0.split("/")) == filter(len, path1.split("/"))


def callOnce(fn):
    def Rtn(*args, **kwargs):
        if Rtn.calling: return
        Rtn.calling = True
        try:
            return fn(*args, **kwargs)
        finally:
            Rtn.calling = False
    Rtn.calling = False
    Rtn.__name__ = fn.__name__ + "@callOnce"
    Rtn.__module = fn.__module__
    Rtn.__doc__ = fn.__doc__
    return Rtn


def map2D(func, lst2D):
    return [[func(x) for x in y] for y in lst2D]


def ReprMaxLenCh(obj, maxLen):
    Str = repr(obj)
    Lst = Str.split(" ")
    Lines = [""]
    for c in xrange(len(Lst)):
        Elem = Lst[c] + (" " if c + 1 < len(Lst) else "")
        if len(Lines[-1]) + len(Elem) < maxLen: Lines[-1] += Elem
        else:
            if len(Elem) <= maxLen:
                Lines.append(Elem)
            elif len(Elem) > maxLen:
                End = maxLen-len(Lines[-1])
                Lines[-1] += Elem[:End]
                Elem = Elem[End:]
                while len(Elem) > maxLen:
                    Lines.append(Elem[:maxLen])
                    Elem = Elem[maxLen:]
    return "\n".join(Lines)

# noinspection SpellCheckingInspection
AlNums = "123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"


def numToXM_PN(num):
    Rtn = chr(ord('1') + num % 9)
    num /= 9
    Rtn = AlNums[num % 35] + Rtn
    num /= 35
    Rtn = chr(ord('1') + num % 9) + Rtn
    num /= 9
    Rtn = chr(ord('A') + num % 26) + Rtn
    return Rtn


def XM_PNtoNum(pn):
    Rtn = ord(pn[0]) - ord('A')
    Rtn *= 9
    Rtn += ord(pn[1]) - ord('1')
    Rtn *= 35
    Rtn += ord(pn[2]) - (ord('1') if pn[2].isdigit() else (ord('A') - 9))
    Rtn *= 9
    Rtn += ord(pn[3]) - ord('1')
    return Rtn


def IsXM_PN(pn):
    """
    :param str|unicode pn:
    :rtype: bool
    """
    if len(pn) != 4: return False
    if not pn[0].isalpha(): return False
    if not pn[1].isdigit() or pn[1] == '0': return False
    if not pn[2].isalnum() or pn[2] == '0': return False
    if not pn[3].isdigit() or pn[3] == '0': return False
    return True

def _getCodecName(codecName):
    return codecs.lookup(codecName).name
T_INTEGER = 0
T_FLOAT = 1
T_DECIMAL = 2
T_STRING = 3
STR_ENC_LATIN1 = 0
STR_ENC_UTF_8 = 1
STR_ENC_UTF_16 = 2
STR_ENC_UTF_16BE = 3
STR_ENC_UTF_32 = 4
STR_ENC_UTF_32BE = 5
STR_ENC = map(_getCodecName, [
    'latin-1', # STR_ENC_LATIN1
    'utf-8', # STR_ENC_UTF_8
    'utf-16', # STR_ENC_UTF_16
    'utf-16-be', # STR_ENC_UTF_16BE
    'utf-32', # STR_ENC_UTF_32
    'utf-32-be' # STR_ENC_UTF_32BE
])

def SaveWithT_Info(fl, data, **kwargs):
    if isinstance(data, (int, long)):
        headLen = kwargs.pop("headLen", 2)
        fl.write(chr(T_INTEGER))
        if data < 0:
            data *= -1
            fl.write("\x01")
        else: fl.write("\0")
        Str = GetStrBytesLong(data)
        fl.write(AlignStrBytesLong(len(Str)-1, headLen))
        fl.write(Str)
    elif isinstance(data, float):
        fl.write(chr(T_FLOAT) + struct.pack("!d", data))
    elif isinstance(data, Decimal):
        headLen = kwargs.pop("headLen", 2)
        sign, digits, exponent = data.as_tuple()
        Num = int("".join(map(str, digits)))
        Str = GetStrBytesLong(Num)
        HeadBytes = struct.pack("!?i", sign, exponent) + AlignStrBytesLong(len(Str)-1, headLen)
        fl.write(chr(T_DECIMAL) + HeadBytes + Str)
    elif isinstance(data, (str, unicode)):
        headLen = kwargs.pop("headLen", 2)
        encoding = codecs.lookup(kwargs.pop("encoding", 'utf-8')).name
        code = STR_ENC.index(encoding)
        Str = data.encode(encoding)
        fl.write(chr(T_STRING) + chr(code) + AlignStrBytesLong(len(Str), headLen) + Str)
    else:
        raise TypeError("Unrecognized object type: %s" % data.__class__.__name__)


def LoadWithT_Info(fl, **kwargs):
    Byt = fl.read(1)
    if len(Byt) == 0:
        raise ValueError("EOF encountered while reading next type code")
    Type = ord(Byt)
    if Type == T_INTEGER:
        headLen = kwargs.pop("headLen", 2)
        sign = [1, -1][ord(fl.read(1))]
        Len = GetLongStrBytes(fl.read(headLen)) + 1
        return GetLongStrBytes(fl.read(Len)) * sign
    elif Type == T_FLOAT:
        return struct.unpack("!d", fl.read(8))[0]
    elif Type == T_DECIMAL:
        headLen = kwargs.pop("headLen", 2)
        sign, exponent = struct.unpack("!?i", fl.read(5))
        Len = GetLongStrBytes(fl.read(headLen)) + 1
        Num = GetLongStrBytes(fl.read(Len))
        digits = tuple(map(int, str(Num)))
        return Decimal((sign, digits, exponent))
    elif Type == T_STRING:
        headLen = kwargs.pop("headLen", 2)
        code = ord(fl.read(1))
        Len = GetLongStrBytes(fl.read(headLen))
        return fl.read(Len).decode(STR_ENC[code])
    else:
        fl.seek(-1, os.SEEK_CUR)
        raise ValueError("Unrecognized type code encountered at byte %u (corrupt file?)"%fl.tell())


# list view of bits mapped to a positive integer word
# slices result in bits combining to make numbers
class bitMapInt(object):
    @staticmethod
    def numToStr(num, padZero=0, bigEnd=False, radix=10):
        assert 1 <= radix <= 16
        rtn = ""
        while num > 0:
            rtn += DIGITS[num % radix]
            num /= radix
        if padZero > len(rtn): rtn += (padZero - len(rtn)) * "0"
        if not bigEnd: return rtn[::-1]
        return rtn

    def __init__(self, num, bitLen):
        """
        :param int|long num:
        :param int|long bitLen:
        """
        self.bitLen = bitLen
        self.wholeMask = (1 << bitLen) - 1
        self.num = num & self.wholeMask

    def flipNum(self):
        rtn = 0
        mask = 1
        for c in xrange(self.bitLen):
            rtn <<= 1
            rtn |= int(bool(self.num & mask))
            mask <<= 1
        return rtn

    def setitemLI(self, index, val, ignoreEnd=0):
        """
        :param int|long|slice index:
        :param int|long val:
        :param int|long ignoreEnd:
        :rtype: int|long
        """
        if isinstance(index, slice):
            rng = xrange(*index.indices(self.bitLen))
            num = self.num
            if ignoreEnd == 0:
                for c in rng:
                    mask = 1 << c
                    if val & 1:
                        num |= mask
                    elif num & mask:
                        num ^= mask
                    val >>= 1
            elif ignoreEnd < len(rng):
                val >>= ignoreEnd
                for c0 in xrange(ignoreEnd, len(rng)):
                    c = rng[c0]
                    mask = 1 << c
                    if val & 1:
                        num |= mask
                    elif num & mask:
                        num ^= mask
                    val >>= 1
            self.num = num
            return len(rng)
        if ignoreEnd > 0:
            return 1
        if index < 0:
            index += self.bitLen
        mask = 1 << index
        if val:
            self.num |= mask
        elif self.num & mask:
            self.num ^= mask
        return 1

    def setitemL(self, index, val):
        """
        :param int|long|slice index:
        :param int|long val:
        :rtype: int|long
        """
        if isinstance(index, slice):
            rng = xrange(*index.indices(self.bitLen))
            num = self.num
            for c in rng:
                mask = 1 << c
                if val & 1:
                    num |= mask
                elif num & mask:
                    num ^= mask
                val >>= 1
            self.num = num
            return len(rng)
        if index < 0:
            index += self.bitLen
        mask = 1 << index
        if val:
            self.num |= mask
        elif self.num & mask:
            self.num ^= mask
        return 1

    def getitemL(self, index):
        """
        :param int|long|slice index:
        :rtype: (int|long, int|long)
        """
        if isinstance(index, slice):
            rng = xrange(*index.indices(self.bitLen))
            rtn = 0
            maskR = 1
            num = self.num
            for c in rng:
                if (1 << c) & num:
                    rtn |= maskR
                maskR <<= 1
            return rtn, len(rng)
        if index < 0:
            index += self.bitLen
        mask = 1 << index
        return int(self.num & mask > 0), 1

    def __setitem__(self, index, val):
        """
        :param int|long|slice index:
        :param int|long val:
        """
        if isinstance(index, slice):
            num = self.num
            for c in xrange(*index.indices(self.bitLen)):
                mask = 1 << c
                if val & 1:
                    num |= mask
                elif num & mask:
                    num ^= mask
                val >>= 1
            self.num = num
        else:
            if index < 0:
                index += self.bitLen
            mask = 1 << index
            if val:
                self.num |= mask
            elif self.num & mask:
                self.num ^= mask
        '''
        if isinstance(index, slice):
            for c in xrange(*index.indices(self.bitLen)):
                mask = 1 << c
                if val & 1:
                    self.num |= mask
                else:
                    mask ^= self.wholeMask
                    self.num &= mask
                val >>= 1
        else:
            if index < 0: index += self.bitLen
            mask = 1 << index
            if val:
                self.num |= mask
            else:
                mask ^= self.wholeMask
                self.num &= mask'''

    def __getitem__(self, index):
        """
        :param int|long|slice index:
        :rtype: int|long
        """
        if isinstance(index, slice):
            rng = xrange(*index.indices(self.bitLen))
            rtn = 0
            maskR = 1
            num = self.num
            for c in rng:
                if (1 << c) & num:
                    rtn |= maskR
                maskR <<= 1
            return rtn
        '''if isinstance(index, slice):
            lstMask = map(lambda x: 1 << x, xrange(*index.indices(self.bitLen)))
            rtn = 0
            for mask in lstMask[::-1]:
                rtn <<= 1
                if mask & self.num:
                    rtn |= 1
            return rtn
        if index < 0: index += self.bitLen
        mask = 1 << index
        return int(self.num & mask > 0)'''
        if index < 0:
            index += self.bitLen
        mask = 1 << index
        return int(self.num & mask > 0)

    def __iter__(self):
        for c in xrange(self.bitLen):
            yield self[c]

    def toStr(self, padZero=0, bigEnd=False, radix=10):
        return bitMapInt.numToStr(self.num, padZero, bigEnd, radix)

    def flipToStr(self, padZero=0, bigEnd=False, radix=10):
        return bitMapInt.numToStr(self.flipNum(), padZero, bigEnd, radix)

    def __len__(self):
        return self.bitLen

    def __add__(self, other):
        num, bitLen = other if isinstance(other, tuple) else (other.num, other.bitLen)
        return bitMapInt(self.num | (num << self.bitLen), self.bitLen + bitLen)

    def __iadd__(self, other):
        num, bitLen = other if isinstance(other, tuple) else (other.num, other.bitLen)
        self.num |= num << self.bitLen
        self.bitLen += bitLen
        return self

    def offsetView(self, offset):
        if offset == 0: return self
        return bitMapIntView(self, xrange(offset, self.bitLen))


# a view of a bitMapInt instance, 'inst'
# bit n corresponds to lstBitView[n] in inst
# lstBitView must be read_index-able, iterable, read_slice-able, and support len()
class bitMapIntView(object):
    def __init__(self, inst, lstBitView):
        self.inst = inst
        self.lstBitView = lstBitView
        self.bitLen = len(self.lstBitView)

    def __setitem__(self, index, val):
        if isinstance(index, slice):
            for c in xrange(*index.indices(self.bitLen)):
                mask = 1 << self.lstBitView[c]
                if val & 1:
                    self.inst.num |= mask
                else:
                    mask ^= self.inst.wholeMask
                    self.inst.num &= mask
                val >>= 1
        else:
            if index < 0: index += self.bitLen
            mask = 1 << self.lstBitView[index]
            if val:
                self.inst.num |= mask
            else:
                mask ^= self.inst.wholeMask
                self.inst.num &= mask

    def __getitem__(self, index):
        if isinstance(index, slice):
            lstMask = map(
                lambda x: 1 << self.lstBitView[x],
                xrange(*index.indices(self.bitLen)))
            rtn = 0
            for mask in lstMask[::-1]:
                rtn <<= 1
                if mask & self.inst.num: rtn |= 1
            return rtn
        if index < 0: index += self.bitLen
        mask = 1 << self.lstBitView[index]
        return int(self.inst.num & mask > 0)

    def __iter__(self):
        for c in xrange(self.bitLen):
            yield self[c]

    def __len__(self):
        return self.bitLen


def addCol(data, rows, cols, fill=None):
    data.extend([None] * rows)
    for y1 in xrange(rows):
        y1 = rows - y1 - 1
        base = y1 * cols
        for x1 in xrange(cols):
            x1 = cols - x1
            data[base + y1 + x1 - 1] = data[base + x1 - 1]
        data[base + y1 + cols] = fill

def testAddCol(in_data, rows, cols, fill, output):
    title = "testAddCol%s" % repr((in_data, rows, cols, fill, output))
    print "BEGIN " + title
    data = list(in_data)
    addCol(data, rows, cols, fill)
    assert data == output, title
    print "END " + title
def testAddColAll():
    lst9 = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    testAddCol(lst9, 3, 3, None, [1, 2, 3, None, 4, 5, 6, None, 7, 8, 9, None])
    testAddCol(lst9, 9, 1, None, [1, None, 2, None, 3, None, 4, None, 5, None, 6, None, 7, None, 8, None, 9, None])
    testAddCol(lst9, 1, 9, None, [1, 2, 3, 4, 5, 6, 7, 8, 9, None])
    lst6 = [1, 2, 3, 4, 5, 6]
    testAddCol(lst6, 2, 3, None, [1, 2, 3, None, 4, 5, 6, None])
    testAddCol(lst6, 3, 2, None, [1, 2, None, 3, 4, None, 5, 6, None])
    testAddCol(lst6, 6, 1, None, [1, None, 2, None, 3, None, 4, None, 5, None, 6, None])
    testAddCol(lst6, 1, 6, None, [1, 2, 3, 4, 5, 6, None])


# converts a html formatted color #RRGGBB (optional '#' character) to float (r,b,g)
def StrRGB(Str):
    if Str.startswith("#"):
        Str = Str[1:]
    lenColor = len(Str) / 3
    num = 16 ** lenColor - 1
    Triple = [Str[c:c + lenColor] for c in xrange(0, len(Str), lenColor)]
    for c in xrange(len(Triple)):
        Triple[c] = int(Triple[c], 16)
    return tuple(map(lambda x: float(x) / num, Triple))


# Inverse of StrRGB
def RGBStr(rgb, nDigitC=2):
    num = 16 ** nDigitC - 1
    fmt = "%%0%uX" % nDigitC
    Triple = [fmt % int(rgb[x] * num + .5) for x in xrange(3)]
    return "".join(Triple)


# debug helper prints out lines of repr
# if isinstance(obj, (list, tuple))
#  for each element in the list prints a repr(obj[n]) line
# elif isinstance(obj, basestring)
#  prints one line that is the repr(obj)
# elif isinstance(obj, dict)
#  prints on each sorted line repr(k): repr(v)
def dbg_PrnLstRepr(obj):
    if isinstance(obj, (str, unicode)):
        print repr(obj)
    elif isinstance(obj, (list, dict, tuple)):
        return "\n".join(sorted(
            map(repr, obj)
            if isinstance(obj, (list, tuple))else
            [repr(k) + ": " + repr(obj[k]) for k in obj]))


def lstIndexOf(lst, obj):
    try:
        return lst.index(obj)
    except ValueError:
        return -1


def toHexBasic(num):
    return "%X" % num


def fromHexStr(Str):
    return int(Str, 16)


def fromFloat(num):
    return "%.15g" % num


def fromDecimal(num):
    return str(num)[:26]


internedFormatFns = {
    "%.15g": fromFloat,
    "%X": toHexBasic
}
def getFormatFn(fmt):
    rtn = internedFormatFns.get(fmt, None)
    if rtn is None:
        assert isinstance(fmt, (str, unicode))
        rtn = lambda x: fmt % x
        internedFormatFns[fmt] = rtn
    return rtn

# noinspection SpellCheckingInspection
DIGITS = "0123456789ABCDEF"


def leftPad(s, n, ch):
    return min(n - len(s), 0) * ch + s


def rightPad(s, n, ch):
    return s + min(n - len(s), 0) * ch