from .LibBaseProgPart import *
from kivy.adapters.simplelistadapter import SimpleListAdapter
from .WidgetHelpers import *
from kivy.utils import escape_markup


class RegisterContext(object):
    """
    :type regs: list[Register]|dict[int|long,Register]|None
    :type regCtlSz: int|long
    :type regWordSz: int|long
    :type regCtlMask: int|long
    """
    regs = None
    # if regCtlSz == 0
    #   then the address is equal to `num >> regWordSz`
    regCtlSz = 3
    regWordSz = 32
    regCtlMask = 0x7
    def initRegs(self, regSpecs, regMaker):
        """
        :param list[slice|int|None]|dict[int,slice|int|None] regSpecs:
        :param (int, int|slice) -> Register regMaker:
        """
        # regSpecs key is register number
        #   value is a slice with indices being dual bit numbers or None
        if self.regCtlSz > 0:
            self.regCtlMask = (1 << self.regCtlSz) - 1
        else:
            self.regCtlMask = (1 << self.regWordSz) - 1
        if isinstance(regSpecs, list):
            self.regs = [None] * len(regSpecs)
            for c in xrange(len(self.regs)):
                cur = regSpecs[c]
                if cur is None: cur = -1
                self.regs[c] = regMaker(c, cur)
        else:
            self.regs = {k: None for k in regSpecs}
            for k in regSpecs:
                cur = regSpecs[k]
                if cur is None: cur = -1
                self.regs[k] = regMaker(k, cur)

    def mapRegs(self, fn):
        """
        :param (Register) -> T fn:
        :rtype: list[T]
        """
        indices = xrange(len(self.regs)) if isinstance(self.regs, list) else sorted(self.regs)
        szRes = 0
        regs = self.regs
        for k in indices:
            szRes += len(regs[k].dualLst)
        rtn = [None] * szRes
        c = 0
        for k in indices:
            reg = regs[k]
            for c1 in xrange(len(reg.dualLst)):
                reg.curReg = c1
                rtn[c] = fn(reg)
                c += 1
        return rtn

    def mapRegsWithKey(self, fn):
        """
        :param (Register, int|(int, int)) -> T fn:
        :rtype: list[T]
        """
        indices = xrange(len(self.regs)) if isinstance(self.regs, list) else sorted(self.regs)
        szRes = 0
        regs = self.regs
        for k in indices:
            szRes += len(regs[k].dualLst)
        rtn = [None] * szRes
        c = 0
        for k in indices:
            reg = regs[k]
            if len(reg.dualLst) <= 1:
                rtn[c] = fn(reg, k)
                c += 1
                continue
            for c1 in xrange(len(reg.dualLst)):
                reg.curReg = c1
                rtn[c] = fn(reg, (k, c1))
                c += 1
        return rtn

    def progReg(self, code):
        if self.regCtlSz > 0:
            reg = self.regs[code & self.regCtlMask]
            assert isinstance(reg, Register)
            reg.progNum(code)
        else:
            reg = self.regs[code >> self.regWordSz]
            assert isinstance(reg, Register)
            reg.progNum(code & self.regCtlMask)

    def getReg(self, regId):
        """
        :param int|(int, int) regId:
        :rtype: Register
        """
        dual = None
        if isinstance(regId, (tuple, list)):
            dual = regId[1]
            regId = regId[0]
        reg = self.regs[regId]
        """ :type: Register"""
        if dual is not None:
            reg.curReg = dual
        elif reg.isDual:
            reg = reg.both
        return reg


DC_DATA_ONLY = 0 # change data
DC_SYNCH = 1 # change data -> synchronize user widget
DC_VERIFY = 2 # verify -> change data
DC_RECALC = 4 # change data -> run calcs
DC_VERIFY_SYNCH = 3 # verify -> change data -> synchronize user widget
DC_SYNCH_RECALC = 5 # change data -> synchronize user widget -> run calcs
DC_VERIFY_RECALC = 6 # verify -> change data -> run calcs
DC_VERIFY_SYNCH_RECALC = 7 # verify -> change data -> synchronize user widget -> run calcs


class DataContext(object):
    def getData(self, k):
        """
        :param str|unicode k:
        """
        raise NotImplementedError("Not Implemented")
    def getDataDefault(self, k, v=None):
        """
        :param str|unicode k:
        :param v:
        """
        raise NotImplementedError("Not Implemented")
    def setData(self, k, v, updateDisp=DC_DATA_ONLY):
        """
        :param str|unicode k:
        :param any v:
        :param int updateDisp:
        """
        raise NotImplementedError("Not Implemented")

class BaseUpstream(object):
    def pull(self, context):
        raise NotImplementedError("Not Implemented")

    def push(self, context, v):
        """
        :param context:
        :param v:
        :rtype: bool
        """
        raise NotImplementedError("Not Implemented")


class UnionUpstream(BaseUpstream):
    def __init__(self, lstUpstream, pullAll=False):
        """
        :param list[BaseUpstream] lstUpstream:
        :param bool pullAll:
        """
        assert len(lstUpstream) >= 2
        self.lstUpstream = lstUpstream
        self.pullAll = pullAll
    def pull(self, context):
        lstUpstream = self.lstUpstream
        v = lstUpstream[0].pull(context)
        if self.pullAll:
            for c in xrange(1, len(lstUpstream)):
                if v != lstUpstream[c].pull(context):
                    raise ValueError("Inconsistent values in list of upstreams")
        return v
    def push(self, context, v):
        """
        :param context:
        :param v:
        :rtype: bool
        """
        rtn = False
        for upstream in self.lstUpstream:
            rtn |= upstream.push(context, v)
        return rtn


class BitsetContext(object):
    """
    :type bitlen: int|long
    :type word: bitMapInt|None
    """
    bitlen = 0
    word = None
    def initBitset(self):
        self.word = bitMapInt(0, self.bitlen)



class BitsetUpstream(BaseUpstream):
    def __init__(self, lstWrite, toFn=None, fromFn=None, passContext=False):
        """
        :param list[(int|None, int|None, int|None)|(int|None, int|None)|int] lstWrite:
        :param None|((any, BitsetContext) -> int|long|(int|long, int)) toFn:
        :param None|((int|long, BitsetContext) -> any) fromFn:
        :param bool passContext:
        """
        self.lstWrite = lstWrite
        self.toFn = toFn
        self.fromFn = fromFn
        self.passContext = passContext
    def pull(self, context):
        """
        :param BitsetContext context:
        :rtype: any
        """
        v = 0
        numShift = 0
        word = context.word
        assert word is not None
        for write in self.lstWrite:
            index = write
            if not isinstance(write, (int, long)):
                index = slice(*write)
            v1, shiftBy = word.getitemL(index)
            v |= v1 << numShift
            numShift += shiftBy
        if self.fromFn is not None:
            v = self.fromFn(v, context) if self.passContext else self.fromFn(v)
        return v
    def push(self, context, v):
        """
        :param BitsetContext context:
        :param any v:
        :rtype: bool
        """
        if self.toFn is not None:
            v = self.toFn(v, context) if self.passContext else self.toFn(v)
        ignoreEnd = 0
        if isinstance(v, (list, tuple)):
            ignoreEnd = v[1]
            v = v[0]
        assert isinstance(ignoreEnd, (int, long))
        assert isinstance(v, (int, long))
        word = context.word
        origNum = word.num
        for write in self.lstWrite:
            index = write
            if not isinstance(write, (int, long)):
                index = slice(*write)
            sz = word.setitemLI(index, v, ignoreEnd)
            if ignoreEnd:
                ignoreEnd = max(0, ignoreEnd - sz)
            v >>= sz
        return word.num != origNum


class RegisterDownstream(BaseUpstream):
    def __init__(self, lstWrite, toFn=None, fromFn=None, passContext=False):
        """
        most significant `lstWrite` elements are first`
        :param list[(int|(int, int), int, int, int)|(int|(int, int), int, int)|(int|(int, int), int)] lstWrite:
        :param None|((any, RegisterContext) -> int|long|(int|long, int)) toFn:
        :param None|((int|long, RegisterContext) -> any) fromFn:
        :param bool passContext:
        """
        self.lstWrite = lstWrite
        self.toFn = toFn
        self.fromFn = fromFn
        self.passContext = passContext
    def pull(self, context):
        """
        :param RegisterContext context:
        :rtype: any
        """
        v = 0
        numShift = 0
        for write in self.lstWrite:
            reg = context.getReg(write[0])
            if len(write) == 2:
                v |= reg[write[1]] << numShift
                numShift += 1
            else:
                slice_ = slice(*write[1:])
                v |= reg[slice_] << numShift
                numShift += len(xrange(*slice_.indices(len(reg))))
        if self.fromFn is not None:
            v = self.fromFn(v, context) if self.passContext else self.fromFn(v)
        return v
    def push(self, context, v):
        """
        :param RegisterContext context:
        :param any v:
        :rtype: bool
        """
        if self.toFn is not None:
            v = self.toFn(v, context) if self.passContext else self.toFn(v)
        ignoreEnd = 0
        if isinstance(v, (list, tuple)):
            ignoreEnd = v[1]
            v = v[0]
        assert isinstance(ignoreEnd, (int, long))
        assert isinstance(v, (int, long))
        rtn = False
        for write in self.lstWrite:
            reg = context.getReg(write[0])
            origData = reg.getNonDualNumberData()
            if len(write) == 2:
                if ignoreEnd:
                    ignoreEnd -= 1
                    v >>= 1
                else:
                    reg[write[1]] = v & 1
                    v >>= 1
            else:
                slice_ = slice(*write[1:])
                indices = xrange(*slice_.indices(len(reg)))
                shiftBy = len(indices)
                if ignoreEnd >= shiftBy:
                    v >>= shiftBy
                    ignoreEnd -= shiftBy
                elif ignoreEnd:
                    start = slice_.start
                    if start is None: start = 0
                    step = slice_.step
                    if step is None: step = 1
                    stop = slice_.stop
                    slice_ = slice(start + ignoreEnd * step, step, stop)
                    v >>= ignoreEnd
                    shiftBy -= ignoreEnd
                    reg[slice_] = v
                    v >>= shiftBy
                else:
                    reg[slice_] = v
                    v >>= shiftBy
            if reg.changedNumberData(origData):
                rtn = True
        return rtn


DVT_BASE = 0
DVT_NUMBER = 1
DVT_LIST = 2
DVT_LIST_INF = 3
DVT_CHK = 4
DVT_STATIC = 5

class BaseUserDataView(object):
    """
    :type display: bool
    :type style: dict[str|unicode,any]|None
    :type defVal: int|long|float|Decimal|bool
    :type widgetIndex: int|long|None
    :type name: str|unicode
    :type shortTitle: str|unicode
    :type readOnly: bool
    :type upstream: BaseUpstream|None
    :type dataViewType: int|long
    """
    display = True
    titleMarkup = False
    dataViewType = DVT_BASE
    defVal = 0
    widgetIndex = None
    style = None
    def __init__(self, name, shortTitle="", readOnly=False):
        """
        :param str|unicode name:
        :param str|unicode shortTitle:
        :param bool readOnly:
        """
        self.name = name
        self.shortTitle = shortTitle
        self.readOnly = readOnly
        self.upstream = None

    def setWidgetIndex(self, index):
        self.widgetIndex = index
        return self

    def setMarkup(self, markup):
        """
        :param bool markup:
        :rtype: BaseUserDataView
        """
        self.titleMarkup = markup
        return self

    def updateWidget(self, v, wid):
        """
        :param any v:
        :param Widget wid:
        """
        raise NotImplementedError("Not Implemented")
    def getStyle(self):
        """
        :rtype: dict[str|unicode,any]
        """
        if self.style is None:
            return {}
        else:
            return self.style
    def setUpstream(self, upstream):
        """
        :param BaseUpstream upstream:
        :rtype: BaseUserDataView
        """
        self.upstream = upstream
        return self
    def build(self, context):
        """
        :param context:
        :rtype: (tuple[Widget], int)
        """
        raise NotImplementedError("Not Implemented")

    def buildAsOne(self, context):
        """
        :param context:
        :rtype: (GridLayout, int)
        """
        rtn = GridLayout(rows=1, spacing=[10, 0])
        widgets, index = self.build(context)
        widIndex = self.widgetIndex
        if widIndex is None:
            for c, wid in enumerate(widgets):
                rtn.add_widget(wid)
            widIndex = len(widgets) - index - 1
        else:
            for c, wid in enumerate(widgets):
                if c == index:
                    continue
                rtn.add_widget(wid)
            rtn.add_widget(widgets[index], widIndex)
        return rtn, widIndex

    def setDisplay(self, display):
        """
        :param bool display:
        :rtype: BaseUserDataView
        """
        self.display = display
        return self


class BasicPopupMaker(object):
    def __init__(self, kwargsPopup=None, kwargsTxtIn=None, addContent=(), kwargsDone=None):
        """
        :param dict[str|unicode,any]|None kwargsPopup:
        :param dict[str|unicode,any]|None kwargsTxtIn:
        :param tuple|list addContent:
        :param dict[str|unicode,any]|None kwargsDone:
        """
        if kwargsDone is None:
            kwargsDone = defDone
        else:
            kwargsDone = dict(defDone, **kwargsDone)
        if kwargsTxtIn is None:
            kwargsTxtIn = defTxtIn
        else:
            kwargsTxtIn = dict(defTxtIn, **kwargsTxtIn)
        if kwargsPopup is None:
            kwargsPopup = defPopup
        else:
            kwargsPopup = dict(defPopup, **kwargsPopup)
        popup = PopupWithClose(on_dismiss=self.onPopupDismiss, on_open=self.onPopupOpen, **kwargsPopup)
        done = Button(**kwargsDone)
        self.done = done
        txtIn = TextInput(**kwargsTxtIn)
        self.txtIn = txtIn
        content = FloatLayout(size_hint=(1.0, 1.0), pos_hint={'x': 0.0, "top": 1.0})
        content.add_widget(txtIn)
        content.add_widget(done)
        popup.content = content
        calcTop = txtIn.pos_hint.get('y', None)
        if calcTop is None:
            calcTop = txtIn.pos_hint.get("top", 0.8 + txtIn.size_hint_y) - txtIn.size_hint_y
        doneTop = done.pos_hint.get("top", None)
        if doneTop is None:
            doneTop = done.pos_hint.get('y', None)
            if doneTop is None:
                doneTop = .1
            else:
                doneTop += done.size_hint_y
        calcSz = (1.0, calcTop - doneTop)
        kwargsInner = {"size_hint": calcSz, "pos_hint": {'x': 0.0, "top": calcTop}}
        for part in addContent:
            content.add_widget(part)
        self.addContent = addContent
        content.add_widget(self.getContent(done, txtIn, popup, kwargsInner))
        self.popup = popup
        self.ranInitContent = False
        self.initContent = None
        self.freeContent = None
        self.extArgs = None

    def getContent(self, done, txtIn, popup, kwargsInner):
        return Label(text="UNNAMED POPUP", **kwargsInner)

    def onPopupOpen(self, popup):
        if self.initContent is not None:
            assert self.extArgs is not None
            args = self.initContent(self, self.done, self.txtIn, popup, *self.extArgs)
            self.ranInitContent = True
            if args is not None:
                self.extArgs = args

    def onPopupDismiss(self, popup):
        if self.freeContent is not None:
            if self.ranInitContent:
                assert self.extArgs is not None
                self.freeContent(self, self.done, self.txtIn, popup, *self.extArgs)
            self.ranInitContent = False
            self.initContent = None
            self.freeContent = None
            self.extArgs = None

    def showDialog(self, initContent, freeContent, extArgs=(), title=""):
        self.initContent = initContent
        self.freeContent = freeContent
        self.extArgs = extArgs
        self.popup.title = title
        self.popup.open()


class BasicSelectPopupMaker(BasicPopupMaker):
    defAdapter = SimpleListAdapter(data=[], cls=Label)

    """
    :type lst: ListView|None
    """
    def __init__(self, kwargsPopup=None, kwargsTxtIn=None, addContent=(), kwargsDone=None):
        """
        :param dict[str|unicode,any]|None kwargsPopup:
        :param dict[str|unicode,any]|None kwargsTxtIn:
        :param tuple|list addContent:
        :param dict[str|unicode,any]|None kwargsDone:
        """
        self.lst = None
        self.scrollView = None
        super(BasicSelectPopupMaker, self).__init__(kwargsPopup, kwargsTxtIn, addContent, kwargsDone)

    def getContent(self, done, txtIn, popup, kwargsContent):
        self.lst = ListView(adapter=self.defAdapter, **kwargsContent)
        for child in self.lst.children:
            if isinstance(child, ScrollView):
                self.scrollView = child
                child.bind(on_scroll_start=ScrollAllower)
                break
        txtIn.font_size = 24
        scroll = filter(lambda x: isinstance(x, ScrollView), self.lst.children)[0]
        assert isinstance(scroll, ScrollView)
        # scroll.scroll_timeout = int(scroll.scroll_timeout * 30 / 48.)
        return self.lst


defPopup1 = dict(defPopup, size_hint=(defPopup["size_hint"][0] + 0.2, defPopup["size_hint"][1]))
class InfoSelectPopupMaker(BasicPopupMaker):
    grid = None
    scrl = None
    def __init__(self, kwargsPopup=None, kwargsTxtIn=None, addContent=(), kwargsDone=None):
        kwargsPopup = defPopup1 if kwargsPopup is None else dict(defPopup1, **kwargsPopup)
        super(InfoSelectPopupMaker, self).__init__(kwargsPopup, kwargsTxtIn, addContent, kwargsDone)
    def getContent(self, done, txtIn, popup, kwargsContent):

        """def onEach(cur):
            curLst = cur.children
            curLst[0].bind(on_press=lambda *args: curLst[1].on_press(curLst[1]))
            curLst[1].bind(
                on_state=lambda obj, val: (
                    cur.select() if fromState[val] else cur.deselect()))"""
        txtIn.font_size = 24
        self.grid = grid = GridLayout(cols=2, spacing=[10, 0], **kwargsContent)
        self.scrl = scrl = ScrollView(size_hint=(1.0, 1.0))
        scrl.bind(on_scroll_start=ScrollAllower)
        # scrl.scroll_timeout = int(scrl.scroll_timeout * lblHeight / 48.)
        grid.add_widget(scrl)
        # grid.add_widget(grid1)
        return grid


def buildNumpad(lst, grid, done, txtIn, content, nCols=None):
    """

    :param list[str|unicode|(str|unicode, str|unicode)|Widget|dict] lst:
    :param GridLayout grid:
    :param Widget done:
    :param TextInput txtIn:
    :param Layout content:
    :param int|long|None nCols:
    """
    if nCols is None: nCols = grid.cols
    lstWidNumpad = [None] * len(lst)
    lstWidTxtIn = []
    lstWidDone = []
    c = 0
    for txt in lst:
        wid = None
        if isinstance(txt, Widget):
            wid = txt
        elif isinstance(txt, (str, unicode)):
            wid = Button(text=txt, on_release=Bind(addTxt, txtIn, txt))
        elif isinstance(txt, tuple):
            txtCmd, txt = txt
            wid = Button(text=txt, on_release=Bind(addTxt, txtIn, txtCmd))
        elif isinstance(txt, dict):
            cls = txt.get("cls", Button)
            wid = cls(**txt["kwargs"])
            txtCmd = txt.get("txtCmd", None)
            if txtCmd is not None:
                wid.bind(on_release=Bind(addTxt, txtIn, txtCmd))
            tgt = txt.get("tgt", "numpad")
            szPush = txt.get("szPush", 0.25)
            if tgt == "txtIn":
                lstWidTxtIn.append((szPush, wid))
                wid = None
            elif txt == "done":
                lstWidDone.append((szPush, wid))
                wid = None
        if wid is not None:
            lstWidNumpad[c] = wid
            c += 1
    szLst = c
    rem = szLst % nCols
    gridLst = lstWidNumpad[:szLst - rem]
    lstWidDone.extend(map(lambda btn1: (0.25, btn1), lstWidNumpad[szLst - rem:szLst]))
    for btn in gridLst:
        grid.add_widget(btn)
    if len(lstWidDone):
        pushWidgets(done, lstWidDone)
        for szPush, btn in lstWidDone:
            content.add_widget(btn)
    if len(lstWidTxtIn):
        pushWidgets(txtIn, lstWidTxtIn)
        for szPush, btn in lstWidTxtIn:
            content.add_widget(btn)
    return lstWidTxtIn


class BaseNumpadPopupMaker(BasicPopupMaker):
    """
    :type numpadSpec: list[str|unicode|(str|unicode, str|unicode)|Widget|dict]
    :type grid: GridLayout|None
    :type alignedWids: list[(int|float|long, Widget)]|None
    :type kwargsGrid: dict
    """
    numpadSpec = []
    grid = None
    alignedWids = None
    kwargsGrid = {"cols": 3}
    def getContent(self, done, txtIn, popup, kwargsContent):
        kwargs = dict(self.kwargsGrid)
        kwargs.update(kwargsContent)
        self.grid = grid = GridLayout(**kwargs)
        self.alignedWids = buildNumpad(self.numpadSpec, grid, done, txtIn, popup.content)
        return grid


DEFAULT_CLEAR_BUTTON_SPEC = {
    "cls": XMLoneButton, "txtCmd": "\0", "tgt": "txtIn", "szPush": 0.2,
    "kwargs": {
        "text": "Clear", "color": kBLACK,
        "background_normal_color": kWHITE,
        "background_down_color": kD_GREY
    },
}


class DecimalUIntNumpadPopupMaker(BaseNumpadPopupMaker):
    numpadSpec = list("123456789") + [
        Widget(), "0", ("\b", "Backspace"), DEFAULT_CLEAR_BUTTON_SPEC]


class DecimalSIntNumpadPopupMaker(BaseNumpadPopupMaker):
    numpadSpec = list("123456789") + [
        ("\1-", "-"), "0", ("\b", "Backspace"), DEFAULT_CLEAR_BUTTON_SPEC]


class DecimalUFloatNumpadPopupMaker(BaseNumpadPopupMaker):
    numpadSpec = list("123456789") + [
        ".", "0", ("\b", "Backspace"), DEFAULT_CLEAR_BUTTON_SPEC]


class DecimalSFloatNumpadPopupMaker(BaseNumpadPopupMaker):
    numpadSpec = list("123456789") + [
        ("\1-", "-"), "0", ".", ("\b", "Backspace"), DEFAULT_CLEAR_BUTTON_SPEC]



class HexUIntNumpadPopupMaker(BaseNumpadPopupMaker):
    # noinspection SpellCheckingInspection
    numpadSpec = list("0123456789ABCDEF") + [
        ("\b", "Backspace"), DEFAULT_CLEAR_BUTTON_SPEC]
    kwargsGrid = {"cols": 4}


class HexSIntNumpadPopupMaker(BaseNumpadPopupMaker):
    # noinspection SpellCheckingInspection
    numpadSpec = list("0123456789ABCDEF") + [
        ("\1-", "-"), ("\b", "Backspace"), DEFAULT_CLEAR_BUTTON_SPEC]
    kwargsGrid = {"cols": 4}


class StaticUserDataView(BaseUserDataView):
    dataViewType = DVT_STATIC
    def __init__(self, name, shortTitle=""):
        super(StaticUserDataView, self).__init__(name, shortTitle, True)

    def updateWidget(self, v, wid):
        """
        :param str|unicode v:
        :param Label wid:
        """
        assert isinstance(wid, Label)
        if not isinstance(v, (str,unicode)):
            globalVars.log.Log(
                "StaticUserDataView.updateWidget(v, wid) expected v"
                " to be string-type, got %s %r" % (v.__class__.__name__, v), 2)
            v = str(v)
        wid.text = v

    def build(self, context):
        """
        :param DataContext context:
        :rtype: ((Label, Label), int)
        """
        lbl = sizedLabel(text=self.shortTitle)
        style = self.getStyle()
        wid = Label(**style)
        return (lbl, wid), 1


class NumberUserDataView(BaseUserDataView):
    """
    :type popupMaker: BasicPopupMaker|None
    :type conv: (((str|unicode) -> int|float|Decimal)|None, (int|float|Decimal) -> str|unicode)
    :type varRange: BaseRange|tuple|None
    :type longTitle: str|unicode
    :type unit: str|unicode|None
    """
    dataViewType = DVT_NUMBER
    unitMarkup = False
    unitLblSz = 0.2
    style = dict(size_hint_y=None, height=32)
    def __init__(self, name, popupMaker, conv, varRange, longTitle="", shortTitle="", unit=None, readOnly=False):
        """
        :param str|unicode name:
        :param BaseNumpadPopupMaker|None popupMaker:
        :param (((str|unicode) -> int|float|Decimal)|None, (int|float|Decimal) -> str|unicode) conv:
        :param BaseRange|tuple|None varRange:
        :param str|unicode longTitle:
        :param str|unicode shortTitle:
        :param str|unicode|None unit:
        :param bool readOnly:
        """
        super(NumberUserDataView, self).__init__(name, shortTitle, readOnly)
        self.longTitle = longTitle
        self.unit = unit
        self.unitLbl = None if unit is None else Label(text=unit)
        # if unit is not None: self.setWidgetIndex(1)
        assert conv[1] is not None
        if not self.readOnly:
            assert varRange is not None
            assert conv[0] is not None
            if not isinstance(varRange, BaseRange):
                varRange = NumRange(*varRange)
            assert isinstance(varRange, BaseRange)
            assert popupMaker is not None
        self.varRange = varRange
        self.popupMaker = popupMaker
        self.conv = conv

    def setUnitMarkup(self, markup):
        """
        :param bool markup:
        :rtype: BaseUserDataView
        """
        self.unitMarkup = markup
        return self

    def updateWidget(self, v, wid):
        """
        :param any v:
        :param BrdLabel wid:
        """
        assert isinstance(wid, (ClkTxtIn, BrdLabel))
        wid.text = self.conv[1](v)

    @staticmethod
    def funcDone(context, name, conv, txtIn, popup, obj=None):
        """
        :param DataContext context:
        :param str|unicode name:
        :param ((str|unicode) -> int|float|Decimal, (int|float|Decimal) -> str|unicode))|None conv:
        :param TextInput txtIn:
        :param Popup popup:
        :param Widget|None obj:
        """
        del obj
        try:
            val = conv[0](txtIn.text)
            context.setData(name, val, DC_VERIFY_SYNCH_RECALC)
        except Exception as Exc:
            txtMsg = str(Exc)
            bubble = ErrBubble(
                pos_hint={'center_x': 0.5, 'y': 1.2}, text=txtMsg,
                background_color=(1.75, 1, 1, 1.25),
                size_hint=(None, .2), width=int(8 * len(txtMsg)))
            popup.content.add_widget(bubble)
            fn = lambda obj1: obj1.content.remove_widget(bubble)
            popup.bind(on_dismiss=fn)
            bubble.addUnbindAtRemove(popup, "on_dismiss", fn)
            globalVars.log.Log("User Error: " + txtMsg, 2)
        else:
            popup.dismiss()

    @staticmethod
    def initContent(popupMaker, done, txtIn, popup, self, val, context):
        """
        :param BaseNumpadPopupMaker popupMaker:
        :param Button done:
        :param TextInput txtIn:
        :param Popup popup:
        :param NumberUserDataView self:
        :param int|float|Decimal val:
        :param DataContext context:
        :rtype: (NumberUserDataView, (Widget) -> any, ((ListAdapter, any) -> any)|None, ((int|float|None, int|float|None), dict)|None)
        """
        if val is None:
            val = self.defVal
        txtIn.text = self.conv[1](val)
        onDone = Bind(self.funcDone, context, self.name, self.conv, txtIn, popup)
        done.bind(on_release=onDone)
        onTxtChg = None
        if self.varRange.verifyOnChg:
            onTxtChg = (
                Bind(txtChanger, self.varRange, self.conv, extData=context)
                if self.varRange.wantsExtData else
                Bind(txtChanger, self.varRange, self.conv))
            txtIn.bind(text=onTxtChg)
        origTxtInSz = None
        if self.unitLbl is not None:
            origTxtInSz = injectPushWidgets(txtIn, popupMaker.alignedWids, [(self.unitLblSz, self.unitLbl)])
            popup.content.add_widget(self.unitLbl)
        return self, onDone, onTxtChg, origTxtInSz

    @staticmethod
    def freeContent(popupMaker, done, txtIn, popup, self, onDone, onTxtChg, origTxtInSz):
        """
        :param BaseNumpadPopupMaker popupMaker:
        :param Button done:
        :param TextInput txtIn:
        :param Popup popup:
        :param NumberUserDataView self:
        :param (Widget) -> any onDone:
        :param ((ListAdapter, any) -> any)|None onTxtChg:
        :param ((int|float|None, int|float|None), dict)|None origTxtInSz:
        """
        done.unbind(on_release=onDone)
        if onTxtChg is not None:
            txtIn.unbind(text=onTxtChg)
        if self.unitLbl is not None:
            popup.content.remove_widget(self.unitLbl)
        if origTxtInSz is not None:
            space, pos_hint = origTxtInSz
            pushWidgetsBase(txtIn, popupMaker.alignedWids, space, pos_hint)

    def build(self, context):
        """
        :param DataContext context:
        :rtype: ((Label, BrdLabel), int)
        """
        titleTxt = self.shortTitle
        if self.unit is not None and len(self.unit):
            if self.titleMarkup and self.unitMarkup:
                titleTxt += " (%s)" % self.unit
            elif self.titleMarkup:
                titleTxt += escape_markup(" (%s)" % self.unit)
            elif self.unitMarkup:
                titleTxt += escape_markup(titleTxt) + self.unit
            else:
                titleTxt += " (%s)" % self.unit
        lbl = sizedLabel(text=titleTxt, markup=self.titleMarkup or self.unitMarkup)
        style = self.getStyle()
        assert isinstance(self.popupMaker, BaseNumpadPopupMaker) or self.popupMaker is None
        wid = (
            BrdLabel(**style)
            if self.readOnly else
            ClkTxtIn(
                on_release=lambda obj: self.popupMaker.showDialog(
                    self.initContent, self.freeContent,
                    (self, context.getDataDefault(self.name), context),
                    self.longTitle), **style))
        return (lbl, wid), 1


class ChkBoxUserDataView(BaseUserDataView):
    dataViewType = DVT_CHK
    defVal = False
    style = dict(size_hint_y=None, height=32)
    initiated_change = False

    def updateWidget(self, v, wid):
        """
        :param bool v:
        :param CheckBox wid:
        """
        assert isinstance(wid, CheckBox)
        self.initiated_change = True
        wid.active = v
        self.initiated_change = False

    @staticmethod
    def onChange(self, context, obj, val):
        """
        :param ChkBoxUserDataView self:
        :param DataContext context:
        :param CheckBox obj:
        :param bool val:
        """
        del obj
        if not self.initiated_change:
            context.setData(self.name, val, DC_VERIFY_SYNCH_RECALC)

    def build(self, context):
        """
        :param DataContext context:
        :rtype: ((Label, CheckBox), int)
        """
        lbl = sizedLabel(text=self.shortTitle)
        style = self.getStyle()
        wid = CheckBox(**style)
        if self.readOnly:
            wid.disabled = True
        else:
            wid.bind(active=Bind(self.onChange, self, context))
        return (lbl, wid), 1


class ListUserDataView(BaseUserDataView):
    """
    :type popupMaker: BasicPopupMaker
    :type allowEmptySel: bool
    :type longTitle: str|unicode
    :type orderMap: dict[int|long,int|long]|None
    :type lstOpts: tuple[(bool,str|unicode)]|list[(bool,str|unicode)]|dict[int|long,str|unicode]|OrderedOptMap
    :type useLst: list[(str|unicode,int|long)]
    :type adapter: ListAdapter
    :type addSizeHintX: int|long|float
    """
    dataViewType = DVT_LIST
    defVal = -1
    style = dict(size_hint_y=None, height=32)
    row_height = 30

    def __init__(self, name, lstOpts, popupMaker, allowEmptySel=False, longTitle="", shortTitle="", readOnly=False):
        """
        :param str|unicode name:
        :param tuple[(bool,str|unicode)]|list[(bool,str|unicode)]|dict[int|long,str|unicode]|OrderedOptMap lstOpts:
        :param BasicPopupMaker popupMaker:
        :param bool allowEmptySel:
        :param str|unicode longTitle:
        :param str|unicode shortTitle:
        :param bool readOnly:
        """
        super(ListUserDataView, self).__init__(name, shortTitle, readOnly)
        self.longTitle = longTitle
        self.allowEmptySel = allowEmptySel
        self.popupMaker = popupMaker
        useLst = None
        self.orderMap = None
        self.lstOpts = lstOpts
        if isinstance(lstOpts, (tuple, list)):
            useLst = [None] * reduce(lambda total, x: total + int(x[0]), lstOpts, 0)
            cur = 0
            for c in xrange(len(useLst)):
                while not lstOpts[cur][0]: cur += 1
                useLst[c] = (lstOpts[cur][1], cur)
                cur += 1
        elif isinstance(lstOpts, dict):
            iterLst = sorted(lstOpts)
            useLst = [(lstOpts[k], k) for k in iterLst]
        elif isinstance(lstOpts, OrderedOptMap):
            useLst = lstOpts.useLst
            if lstOpts.ordering is not None:
                self.orderMap = {x: c for c, x in enumerate(lstOpts.ordering)}
        if useLst is None: raise TypeError("lstOpts must be instance of dict, tuple or list")
        self.useLst = useLst
        if not allowEmptySel and len(self.useLst):
            self.defVal = self.useLst[0][1]
        self.adapter = ListAdapter(
            cls=XMButton1, allow_empty_selection=self.allowEmptySel, selection_mode="single",
            data=self.useLst, args_converter=convLstOpts)
        maxLen = max(map(lambda x: len(x[0]), self.useLst))
        maxLen -= minLenResize
        self.addSizeHintX = 0
        if maxLen > 0:
            self.addSizeHintX = .015 * maxLen

    def updateWidget(self, v, wid):
        """
        :param any v:
        :param BrdLabel wid:
        """
        assert isinstance(wid, (ClkTxtIn, BrdLabel))
        c = self.getIndexFromVal(v)
        wid.text = "" if c == -1 else self.useLst[c][0]

    def getIndexFromVal(self, val):
        curSel = -1 if self.allowEmptySel else 0
        if val >= 0:
            Tmp = len(self.useLst)
            if self.orderMap is None:
                Tmp = BinaryApprox1(DynamicMap(lambda x: x[1], self.useLst), val)
            else:
                Tmp = self.orderMap.get(val, Tmp)
            if Tmp < len(self.useLst):
                curSel = Tmp
        return curSel

    @staticmethod
    def funcDone(context, name, useLst, adapter, popup, obj=None):
        """
        :param DataContext context:
        :param str|unicode name:
        :param list[(str|unicode, int)] useLst:
        :param ListAdapter adapter:
        :param Popup popup:
        :param Widget|None obj:
        """
        del obj
        val = useLst[adapter.selection[0].index][1]
        try:
            context.setData(name, val, DC_VERIFY_SYNCH_RECALC)
        except ValueError as Exc:
            txtMsg = str(Exc)
            bubble = ErrBubble(
                pos_hint={'center_x': 0.5, 'y': 1.2}, text=txtMsg,
                background_color=(1.75, 1, 1, 1.25),
                size_hint=(None, .2), width=int(8 * len(txtMsg)))
            popup.content.add_widget(bubble)
            fn = lambda obj1: obj1.content.remove_widget(bubble)
            popup.bind(on_dismiss=fn)
            bubble.addUnbindAtRemove(popup, "on_dismiss", fn)
            globalVars.log.Log("User Error: " + txtMsg, 2)
        else:
            popup.dismiss()

    @staticmethod
    def initContent(popupMaker, done, txtIn, popup, self, val, context):
        """
        :param BasicSelectPopupMaker popupMaker:
        :param Button done:
        :param TextInput txtIn:
        :param Popup popup:
        :param ListUserDataView self:
        :param int|None val:
        :param DataContext context:
        :rtype: (ListUserDataView, (Widget) -> any, (ListAdapter, any) -> any, float|int|None)
        """
        adapter = self.adapter
        popupMaker.lst.adapter = adapter
        if self.row_height is not None:
            nh = self.row_height * len(self.useLst)
            popupMaker.lst.container.height = nh
        if val is None:
            val = self.defVal
        curSel = self.getIndexFromVal(val)
        map(adapter.deselect_item_view, adapter.selection)
        popupMaker.scrollView.scroll_y = 0
        if curSel != -1:
            adapter.select_item_view(adapter.get_view(curSel))
            if len(self.useLst) <= 5:
                popupMaker.scrollView.scroll_y = 1
            else:
                # 30 is the height of a normal entry in the ListView
                wndSize = popupMaker.scrollView.height / 30.0
                popupMaker.scrollView.scroll_y =  1 - restrict((curSel - round(wndSize / 2 + 0.1)) / (len(self.useLst) - wndSize), 0, 1)
        else:
            # reset scroll to make sure options are not invisible
            # (due to bug that makes them invisible)
            popupMaker.scrollView.scroll_y = 1
        updTxtPopup(self.useLst, txtIn, adapter, val)
        onDone = Bind(self.funcDone, context, self.name, self.useLst, self.adapter, popup)
        done.bind(on_release=onDone)
        onSelChg = Bind(updTxtPopup, self.useLst, txtIn)
        # noinspection PyArgumentList
        adapter.bind(on_selection_change=onSelChg)
        origPopupSzX = None
        if self.addSizeHintX > 0:
            origPopupSzX = popup.size_hint_x
            popup.size_hint_x = origPopupSzX + self.addSizeHintX
        return self, onDone, onSelChg, origPopupSzX

    @staticmethod
    def freeContent(popupMaker, done, txtIn, popup, self, onDone, onSelChg, origPopupSzX):
        """
        :param BasicSelectPopupMaker popupMaker:
        :param Button done:
        :param TextInput txtIn:
        :param Popup popup:
        :param ListUserDataView self:
        :param (Widget) -> any onDone:
        :param (ListAdapter, any) -> any onSelChg:
        :param float|int|long|None origPopupSzX:
        """
        del txtIn
        del self
        done.unbind(on_release=onDone)
        lst = popupMaker.lst
        assert isinstance(lst, ListView)
        lst.adapter.unbind(on_selection_change=onSelChg)
        lst.adapter = popupMaker.defAdapter
        if origPopupSzX is not None:
            popup.size_hint_x = origPopupSzX

    def build(self, context):
        """
        :param DataContext context:
        :rtype: ((Label, BrdLabel), int)
        """
        lbl = sizedLabel(text=self.shortTitle)
        style = self.getStyle()
        assert isinstance(self.popupMaker, BasicPopupMaker)
        wid = (
            BrdLabel(**style)
            if self.readOnly else
            ClkTxtIn(on_release=lambda obj: self.popupMaker.showDialog(
                self.initContent, self.freeContent,
                (self, context.getDataDefault(self.name), context),
                self.longTitle), **style))
        return (lbl, wid), 1


class InfoListUserDataView(BaseUserDataView):
    """
    :type popupMaker: BasicPopupMaker
    :type allowEmptySel: bool
    :type longTitle: str|unicode
    :type orderMap: dict[int|long,int|long]|None
    :type useLst: list[(str|unicode,int|long)]
    :type lstOpts: tuple[(bool,str|unicode)]|list[(bool,str|unicode)]|dict[int|long,str|unicode]|OrderedOptMap
    :type infLst: list[str|unicode]|dict[int|long,str|unicode]
    :type addSizeHintX: int|long|float
    :type tmpTxtIn: TextInput|None
    """
    dataViewType = DVT_LIST_INF
    defVal = -1
    style = dict(size_hint_y=None, height=32)

    def __init__(self, name, lstOpts, infLst, popupMaker, allowEmptySel=False, longTitle="", shortTitle="", readOnly=False):
        """
        :param str|unicode name:
        :param tuple|list|dict|OrderedOptMap lstOpts:
        :param list[str|unicode]|dict[int|long,str|unicode] infLst:
        :param InfoSelectPopupMaker popupMaker:
        :param bool allowEmptySel:
        :param str|unicode longTitle:
        :param str|unicode shortTitle:
        :param bool readOnly:
        """
        assert not readOnly, "It is wasteful to use an InfoListUserDataView for readOnly data"
        super(InfoListUserDataView, self).__init__(name, shortTitle, readOnly)
        self.longTitle = longTitle
        self.allowEmptySel = allowEmptySel
        self.popupMaker = popupMaker
        self.tmpTxtIn = None
        useLst = None
        self.orderMap = None
        self.lstOpts = lstOpts
        if isinstance(lstOpts, (tuple, list)):
            useLst = [None] * reduce(lambda total, x: total + int(x[0]), lstOpts, 0)
            cur = 0
            for c in xrange(len(useLst)):
                while not lstOpts[cur][0]: cur += 1
                useLst[c] = (lstOpts[cur][1], cur)
                cur += 1
        elif isinstance(lstOpts, dict):
            iterLst = sorted(lstOpts)
            useLst = [(lstOpts[k], k) for k in iterLst]
        elif isinstance(lstOpts, OrderedOptMap):
            useLst = lstOpts.useLst
            self.orderMap = {x: c for c, x in enumerate(lstOpts.ordering)}
        if useLst is None: raise TypeError("lstOpts must be instance of dict, tuple or list")
        self.useLst = useLst
        if not allowEmptySel and len(self.useLst):
            self.defVal = self.useLst[0][1]
        # self.adapter = ListAdapter(cls=XMButton1, allow_empty_selection=self.allowEmptySel, selection_mode="single", data=self.useLst, args_converter=convLstOpts)
        maxLen = max(map(lambda x: len(x[0]), self.useLst))
        maxLen -= minLenResize
        self.addSizeHintX = 0
        if maxLen > 0:
            self.addSizeHintX = .015 * maxLen
        lstInfo = self.infLst = infLst
        if isinstance(lstInfo, dict):
            lstInfo = DctListRepr(lstInfo, max(map(lambda x: x[1], useLst)) + 1, "NONE")
        self.lstInfo = lstInfo
        self.grid1 = grid1 = GridLayout(cols=2, size_hint=(1., 1.), spacing=[0, 0])
        self.lstChk = lstChk = [None] * len(useLst)
        self.lblInf = sizedLabel(text="", size_hint=(1., 1.,))
        self.group = id(self)
        lblHeight = 30
        cbChkState = self.onChkBoxState
        for c in xrange(len(useLst)):
            lstChk[c] = CheckBox(
                group=self.group, state=toState[False],
                size_hint_x=None, width=30, allow_no_selection=allowEmptySel)
            lbl = sizedCls(ClkLabel, text=useLst[c][0])
            lbl.height += 20
            lstChk[c].bind(state=cbChkState)
            lbl.bind(on_press=Bind(togChkBox, lstChk[c]))
            grid1.add_widget(lstChk[c])
            grid1.add_widget(lbl)
            lblHeight = max(lblHeight, lbl.height)
        self.curSel = -1
    def onChkBoxState(self, obj, val):
        index = self.lstChk.index(obj)
        if fromState[val]:
            self.tmpTxtIn.text = self.useLst[index][0]
            self.lblInf.text = self.lstInfo[index]
            self.curSel = index

    def updateWidget(self, v, wid):
        """
        :param any v:
        :param BrdLabel wid:
        """
        assert isinstance(wid, (ClkTxtIn, BrdLabel)), "bad type: %s" % wid.__class__.__name__
        c = self.getIndexFromVal(v)
        wid.text = "" if c == -1 else self.useLst[c][0]

    def getIndexFromVal(self, val):
        curSel = -1 if self.allowEmptySel else 0
        if val >= 0:
            Tmp = len(self.useLst)
            if self.orderMap is None:
                Tmp = BisectSearchBase(self.useLst, val, lambda useLst, i: useLst[i][1])[0]
            else:
                Tmp = self.orderMap.get(val, Tmp)
            if Tmp < len(self.useLst):
                curSel = Tmp
        return curSel

    @staticmethod
    def funcDone(context, self, popup, obj=None):
        """
        :param DataContext context:
        :param InfoListUserDataView self:
        :param Popup popup:
        :param Widget|None obj:
        """
        del obj
        val = self.useLst[self.curSel][1]
        try:
            context.setData(self.name, val, DC_VERIFY_SYNCH_RECALC)
        except ValueError as Exc:
            txtMsg = str(Exc)
            bubble = ErrBubble(
                pos_hint={'center_x': 0.5, 'y': 1.2}, text=txtMsg,
                background_color=(1.75, 1, 1, 1.25),
                size_hint=(None, .2), width=int(8 * len(txtMsg)))
            popup.content.add_widget(bubble)
            fn = lambda obj1: obj1.content.remove_widget(bubble)
            popup.bind(on_dismiss=fn)
            bubble.addUnbindAtRemove(popup, "on_dismiss", fn)
            globalVars.log.Log("User Error: " + txtMsg, 2)
        else:
            popup.dismiss()

    @staticmethod
    def initContent(popupMaker, done, txtIn, popup, self, val, context):
        """
        :param InfoSelectPopupMaker popupMaker:
        :param Button done:
        :param TextInput txtIn:
        :param Popup popup:
        :param InfoListUserDataView self:
        :param int|None val:
        :param DataContext context:
        :rtype: (InfoListUserDataView, (Widget) -> any, float|int|None)
        """
        self.tmpTxtIn = txtIn
        popupMaker.scrl.add_widget(self.grid1)
        popupMaker.grid.add_widget(self.lblInf)
        if val is None:
            val = self.defVal
        curSel = self.curSel = self.getIndexFromVal(val)
        # if self.curSel == -1 then all checkboxes will be unchecked
        for c, chk in enumerate(self.lstChk):
            chk.state = toState[c == curSel]
        if curSel != -1:
            txtIn.text = self.useLst[curSel][0]
        onDone = Bind(self.funcDone, context, self, popup)
        done.bind(on_release=onDone)
        origPopupSzX = None
        if self.addSizeHintX > 0:
            origPopupSzX = popup.size_hint_x
            popup.size_hint_x = origPopupSzX + self.addSizeHintX
        return self, onDone, origPopupSzX

    @staticmethod
    def freeContent(popupMaker, done, txtIn, popup, self, onDone, origPopupSzX):
        """
        :param InfoSelectPopupMaker popupMaker:
        :param Button done:
        :param TextInput txtIn:
        :param Popup popup:
        :param InfoListUserDataView self:
        :param (Widget) -> any onDone:
        :param float|int|long|None origPopupSzX:
        """
        del txtIn
        popupMaker.grid.remove_widget(self.lblInf)
        popupMaker.scrl.remove_widget(self.grid1)
        self.tmpTxtIn = None
        done.unbind(on_release=onDone)
        if origPopupSzX is not None:
            popup.size_hint_x = origPopupSzX

    def build(self, context):
        """
        :param DataContext context:
        :rtype: ((Label, BrdLabel), int)
        """
        lbl = sizedLabel(text=self.shortTitle)
        style = self.getStyle()
        assert isinstance(self.popupMaker, BasicPopupMaker)
        wid = (
            BrdLabel(**style)
            if self.readOnly else
            ClkTxtIn(on_release=lambda obj: self.popupMaker.showDialog(
                self.initContent, self.freeContent,
                (self, context.getDataDefault(self.name), context),
                self.longTitle), **style))
        return (lbl, wid), 1

def ChangeAlignedWids((szFrac, wid)):
    return szFrac / 2, wid

theListMaker = BasicSelectPopupMaker()
theInfoListMaker = InfoSelectPopupMaker()
theDecUIntMaker = DecimalUIntNumpadPopupMaker()
theDecSIntMaker = DecimalSIntNumpadPopupMaker()
theDecUFloatMaker = DecimalUFloatNumpadPopupMaker()
theDecSFloatMaker = DecimalSFloatNumpadPopupMaker()
theDecUFloatMakerBig = DecimalUFloatNumpadPopupMaker({"size_hint": (0.8, defPopup["size_hint"][1])})
theDecUFloatMakerBig.alignedWids = map(ChangeAlignedWids, theDecUFloatMakerBig.alignedWids)
theDecSFloatMakerBig = DecimalSFloatNumpadPopupMaker({"size_hint": (0.8, defPopup["size_hint"][1])})
theDecSFloatMakerBig.alignedWids = map(ChangeAlignedWids, theDecSFloatMakerBig.alignedWids)
theHexUIntMaker = HexUIntNumpadPopupMaker()
theHexSIntMaker = HexSIntNumpadPopupMaker()