from . import globalVars


pos_hint_attr_names = [['x', "center_x", "right"], ['y', "center_y", "top"]]
def get_iattr(pos_hint, attrs):
    """
    :param dict[str|unicode, int|long|float] pos_hint:
    :param list[str|unicode] attrs:
    """
    for c, attr in enumerate(attrs):
        if attr in pos_hint:
            return c
    raise ValueError("expected one of %s to be present" % str(attrs))


def swapWidgets(wid0, wid1):
    par0 = wid0.parent
    par1 = wid1.parent
    if par0 is None and par1 is None: return
    if par0 is None:
        tmp = wid0
        wid0 = wid1
        wid1 = tmp
        par0 = wid0.parent
        par1 = wid1.parent
    if par1 is None:
        index0 = par0.children.index(wid0)
        par0.remove_widget(wid0)
        par0.add_widget(wid1, index0)
    elif par0 is not None and par1 is not None:
        index0 = par0.children.index(wid0)
        index1 = par1.children.index(wid1)
        par0.remove_widget(wid0)
        par1.remove_widget(wid1)
        par0.add_widget(wid1, index0)
        par1.add_widget(wid0, index1)



def inferPosSizeHint(origWidget, lstOthers, pushHoriz=True):
    """

    :param Widget origWidget:
    :param list[(int|float, Widget)] lstOthers:
    :param bool pushHoriz:
    :rtype: ((int|long|float|None, int|long|float|None), dict[str|unicode, int|long|float], int|long)
    """
    space = list(origWidget.size_hint)
    pos_hint = dict(origWidget.pos_hint)
    iVec = int(not pushHoriz)
    szOff = 0
    for szFrac, wid in lstOthers:
        szOff += wid.size_hint[iVec]
    space[iVec] += szOff
    space = tuple(space)
    attrs = pos_hint_attr_names[iVec]
    iattr = get_iattr(pos_hint, attrs)
    if iattr == 1:
        pos_hint[attrs[1]] += szOff / 2.0
    elif iattr == 2:
        pos_hint[attrs[2]] += szOff
    return space, pos_hint, iattr


def pushWidgetsBase(origWidget, lstOthers, space, pos_hint, pushHoriz=True, iattr=None):
    """
    :param Widget origWidget:
    :param list[(int|float, Widget)] lstOthers:
    :param (int|long|float|None, int|long|float|None) space:
    :param dict[str|unicode, int|long|float] pos_hint:
    :param bool pushHoriz:
    :param int|long|None iattr:
    """
    iVec = int(not pushHoriz)
    attrs = pos_hint_attr_names[iVec]
    if iattr is None:
        iattr = get_iattr(pos_hint, attrs)
    szFracLeft = 1.0
    for szFrac, wid in lstOthers:
        theNewSz = list(space)
        theNewPos = dict(pos_hint)
        theNewSz[iVec] *= szFrac
        if iattr == 2:
            theNewPos[attrs[2]] = szFracLeft * space[iVec]
        szFracLeft -= szFrac
        if iattr == 0:
            theNewPos[attrs[0]] = szFracLeft * space[iVec]
        elif iattr == 1:
            theNewPos[attrs[1]] = (szFracLeft + szFrac / 2.0) * space[iVec]
        wid.size_hint = theNewSz
        wid.pos_hint = theNewPos
    if szFracLeft < 0:
        globalVars.log.Log("szFracLeft < 0 means only negative space is left for origWidget", 2)
    elif szFracLeft * space[iVec] != origWidget.size_hint[iVec]:
        theNewSz = list(space)
        theNewSz[iVec] *= szFracLeft
        theNewPos = dict(pos_hint)
        if iattr == 1:
            theNewPos[attrs[1]] -= (1.0 - szFracLeft) / 2
        elif iattr == 2:
            theNewPos[attrs[1]] -= szFracLeft * space[iVec]
        origWidget.size_hint = theNewSz
        origWidget.pos_hint = theNewPos
    return space, pos_hint


def pushWidgets(origWidget, lstOthers, pushHoriz=True):
    """
    :param Widget origWidget:
    :param list[(int|float, Widget)] lstOthers:
    :param bool pushHoriz:
    """
    return pushWidgetsBase(
        origWidget, lstOthers, tuple(origWidget.size_hint), origWidget.pos_hint, pushHoriz)


def realignPushedWidgets(origWidget, lstOthers, pushHoriz=True):
    """
    :param Widget origWidget:
    :param list[(int|float, Widget)] lstOthers:
    :param bool pushHoriz:
    """
    space, pos_hint, iattr = inferPosSizeHint(origWidget, lstOthers, pushHoriz)
    # print "realign inferred original pos_hint=%r, size_hint=%r" % (pos_hint, space)
    return pushWidgetsBase(origWidget, lstOthers, space, pos_hint, pushHoriz, iattr)


def injectPushWidgets(origWidget, lstOthers, lstInject, pushHoriz=True):
    """
    :param Widget origWidget:
    :param list[(int|float, Widget)] lstOthers:
    :param list[(int|float, Widget)] lstInject:
    :param bool pushHoriz:
    """
    space, pos_hint, iattr = inferPosSizeHint(origWidget, lstOthers, pushHoriz)
    # print "inject inferred original pos_hint=%r, size_hint=%r" % (pos_hint, space)
    return pushWidgetsBase(origWidget, lstInject + lstOthers, space, pos_hint, pushHoriz, iattr)


def ejectPushWidgets(origWidget, lstOthers, lstEject, pushHoriz=True):
    """
    :param Widget origWidget:
    :param list[(int|float, Widget)] lstOthers:
    :param list[(int|float, Widget)] lstEject:
    :param bool pushHoriz:
    """
    space, pos_hint, iattr = inferPosSizeHint(origWidget, lstEject + lstOthers, pushHoriz)
    # print "eject inferred original pos_hint=%r, size_hint=%r" % (pos_hint, space)
    return pushWidgetsBase(origWidget, lstOthers, space, pos_hint, pushHoriz, iattr)