import os
import traceback
from .Logging import LoggingContext
from .TinyLib import ValuePack
from .GlobalStorages import GlobalStorage
import platform

__all__ = [
    "TinyLib",
    "partmods",
    "LibProgramIt",
    "LibIO",
    "images",
    "globalVars",
    "FileVersionCompat"
]
rPi = os.name != "nt" and (platform.uname()[0].lower() == "linux") and ("arm" in platform.uname()[4].lower())
# get the base directory (%PROJECT_DIR%/ProgramIt)
baseDir = os.path.abspath(__file__).replace("\\", "/")
_pos = baseDir.rfind("/")
if _pos == -1:
    baseDir = os.path.abspath(".").replace("\\", "/") + "/"
else:
    baseDir = baseDir[:_pos + 1]
del _pos
# start the Logger
logCtx = LoggingContext(baseDir, "ProgramIt.log")
FlLog, log = logCtx.createAndSwapLogFile("PROGRAM_IT")
log.Log("PROCESS ID: %u" % os.getpid())
globalVars = GlobalStorage(False, False, rPi, rPi, None, 20, 55, logCtx, log, FlLog)

#globalVars.canPowerOff = True
def fRequire(fName):
    if not os.path.isfile(fName):
        raise OSError("Required File not found: %s" % fName)
images = ValuePack()
images.fullLogoPi = "images/X-MW-logo.png"
images.smallLogo = "images/X-logo.png"
images.fullLogo = "images/X-MW-logo1.png"
images.help = "images/HelpImg1.png"
images.home = "images/HomeImg1.png"
images.forward = "images/ForwardImg1.png"
images.back = "images/BackImg1.png"
images.close = "images/CloseImg.png"
images.bubbleBG = "images/bubbleBG3.png"
images.bubbleArrow = "images/bubbleArrow1.png"
if rPi: images.fullLogo = images.fullLogoPi
for k in images:
    if k.startswith("_"): continue
    images[k] = baseDir + images[k]
    fRequire(images[k])
images.baseImg = baseDir + "images/"
from kivy.config import Config
Config.set("graphics","width","800")
Config.set("graphics","height","480")
if rPi:
    Config.set("graphics", "fullscreen", "1")
    Config.set("input", "mouse", "mouse")
    # noinspection SpellCheckingInspection
    Config.set("input", "mtdev_%(name)s", "probesysfs,provider=mtdev")
    # noinspection SpellCheckingInspection
    Config.set("input", "hid_%(name)s", "probesysfs,provider=hidinput")
else:
    Config.set("graphics", "fullscreen", "0")
    
Config.set("kivy","desktop", "0")
Config.set("kivy", "keyboard_mode", "dock")
Config.set("widgets", "scroll_distance", str(globalVars.scroll_distance))
Config.set("widgets", "scroll_timeout", str(globalVars.scroll_timeout))
Config.write()
import kivy
kivy.require("1.9.1")
import LibProgramIt
import FileVersionCompat
import partmods
TheApp = None
# noinspection PyBroadException
try:
    globalVars.app = TheApp = LibProgramIt.ProgramItApp(logObj=log)
    TheApp.run()
except:
    tb_data = traceback.format_exc()[:-1]
    if TheApp is not None:
        # noinspection PyBroadException
        try:TheApp.stopApp()
        except:pass
    logCtx.closeLog(FlLog, log, [
        (tb_data, 1),
        "ProgramIt Exited"
    ], True)
else:
    if globalVars.execExit is not None:
        # noinspection PyBroadException
        try:
            import subprocess
            subprocess.Popen(globalVars.execExit, close_fds=True)
        except:
            log.Log("Error while trying to do execExit: " + traceback.format_exc()[:-1], 2)
        else:
            log.Log("Started execExit successfully", 3)
    logCtx.closeLog(FlLog, log, ["ProgramIt Exited"], False)
    if globalVars.powerOff:
        if rPi: os.system("sudo poweroff")
        else: print ("POWER_OFF " * 8 + "\n") * 8
