from ..LibBaseProgPart import *

log = None

@initFunc(__name__.replace("ProgramIt.", ""))
def init(progLog):
    # noinspection PyGlobalUndefined
    global log
    log = progLog

stNames = ["Low", "High"]

class SwitchStorage(object):
    """
    :type val: int|long|None
    :type layout: BorderedGridLayout|None
    :type topLbl: Label|None
    :type grid: GridLayout|None
    :type chgVal: ((int|long) -> any)|None
    :type opts: list[(CheckBox, ClkLabel)]|None
    """
    __slots__ = ["val", "layout", "topLbl", "grid", "chgVal", "opts"]
    def __init__(self):
        self.val = None
        self.layout = None
        self.topLbl = None
        self.grid = None
        self.chgVal = None
        self.opts = None
    def init(self, val, title, chgSwitch, nThrow, swLbls, **kwargs):
        """
        :param int|long val:
        :param str|unicode title:
        :param (int|long) -> any chgSwitch:
        :param int|long nThrow:
        :param lst[str|unicode] swLbls:
        :param kwargs:
        """
        self.val = val
        self.layout = BorderedGridLayout(cols=1, **kwargs)
        self.topLbl = sizedLabel(
            size_hint=(1., None), text=title, halign="center")
        self.grid = GridLayout(cols=2)
        self.layout.add_widget(self.topLbl)
        self.layout.add_widget(self.grid)
        self.chgVal = chgSwitch
        self.opts = [
            (CheckBox(
                state=toState[self.val == c], size_hint_x=None,
                width=40, group=self, allow_no_selection=False),
             sizedCls(
                 ClkLabel, text=swLbls[c], size_hint=(1.0, 1.)))
            for c in xrange(nThrow)]
        for c in xrange(nThrow):
            btn, lbl = self.opts[c]
            btn.bind(state=Bind(radioBtnChg, self, c))
            lbl.bind(on_press=Bind(togChkBox, btn))
            lbl.color = stRGB[self.val == c]
            self.grid.add_widget(btn)
            self.grid.add_widget(lbl)
class BinaryStorage(object):
    __slots__ = ["layout", "topLbl", "grid", "lbls"]
    def __init__(self):
        self.layout = None
        self.topLbl = None
        self.grid = None
        self.lbls = None

class BaseSwitch(BaseProgPart):
    XM_partName = "Unknown"
    mfgPartName = "Unknown"
    catId = CAT_SWITCH
    mfgName = "Unknown"
    interface = "parallel"
    ioType = IO_PIN
    nThrow = 2
    nPins = 2
    swLbls = ["RF1  On", "RF2  On"]
    defSW_Val = 0
    optMaps = {0:[0, 1], 1:[1, 0]}
    txtTable = ["V1", "V2"]
    layout = None
    top_custom_name = None
    def __init__(self, fl=None, ver=0):
        self.states = [0] * self.nPins
        if not hasattr(self, "oldStates"):
            try: self.oldStates = list(self.optMaps[self.defSW_Val])
            except KeyError:
                log.Log(
                    "index %u was not found in the optMaps. class %s" %(
                    self.defSW_Val, self.__class__.__name__), 2)
                self.oldStates = list(self.states)
        ErrMsg = "Error with optMaps[%s]=%s in class " + self.__class__.__name__
        for k in self.optMaps:
            Val = len(self.optMaps[k]) == self.nPins
            assert Val, ErrMsg % (repr(k), repr(self.optMaps[k]))
        super(BaseSwitch, self).__init__(fl, ver)
        if fl is not None:
            data = GetLongStrBytes(fl.read(((self.nThrow - 1).bit_length() + 7) // 8))
            self.states[:] = self.optMaps[data]
            self.defSW_Val = data
        self.switch = SwitchStorage()
        self.binary = BinaryStorage()
        self.txtTbl = [
            "[color=#FFF]%s  [/color][color=#%%s]%%s[/color]" % txt
            for txt in self.txtTable]
        self.buildFirst()
    def serialize(self, fl, ver):
        super(BaseSwitch, self).serialize(fl, ver)
        data = self.defSW_Val
        if hasattr(self, "switch"):
            data = self.switch.val
        fl.write(AlignStrBytesLong(data, ((self.nThrow - 1).bit_length() + 7) // 8))
        fl.flush()
    def onhide(self, app):
        self.layout.remove_widget(self.switch.layout)
        self.layout.remove_widget(self.binary.layout)
        self.top_custom_name.remove(self.layout)
        super(BaseSwitch, self).onhide(app)
    def build(self, app):
        layout = super(BaseSwitch, self).build(app)
        if self.binary.layout is None:
            self.buildFirst()
            self.chgSwitch(self.switch.val)
        sizeY = .61
        posY = .93
        posY1 = posY - sizeY - .03
        self.top_custom_name.add(layout)
        layout.add_widget(self.switch.layout)
        layout.add_widget(self.binary.layout)
        #layout.add_widget(self.binLayout)
        self.progBtn.pos_hint = {"top": posY1, 'x': .1}
        self.layout = layout
    def updBinLbls(self):
        for c in xrange(len(self.binary.lbls)):
            lbl = self.binary.lbls[c]
            color = cmpColors[self.states[c]!=self.oldStates[c]]
            lbl.text = self.txtTbl[c]%(color,stNames[self.states[c]])
    def setSwitch(self, index):
        if index not in self.optMaps:
            log.Log(
                "index %u was not found in the optMaps. class %s" %(
                index, self.__class__.__name__), 2)
            return
        self.states[:] = self.optMaps[index]
        self.updBinLbls()
    def chgSwitch(self, index):
        self.setSwitch(index)
        self.onProgChg()
    def on_chg_custom_name(self, v):
        if self.top_custom_name is not None:
            self.top_custom_name.txt_in.text = v
    def buildFirst(self):
        sizeX = .4
        padX = .015
        #posX = .5 - sizeX/2
        posX = .1
        sizeY = .61
        posY = .93
        # posY1 = posY - sizeY - .03
        self.top_custom_name = TopCustomNameStorage(self)
        self.switch.init(
            self.defSW_Val, "Switch Setting",
            lambda i: self.chgSwitch(i), self.nThrow, self.swLbls,
            size_hint=(sizeX-padX,sizeY),pos_hint={"top":posY,'x':posX+padX/2})
        '''self.switch.val = self.defSW_Val
        self.switch.layout = BorderedGridLayout(
            size_hint=(sizeX-padX,sizeY),pos_hint={"top":posY,'x':posX+padX/2},
            cols=1)
        self.switch.topLbl = sizedLabel(
            size_hint=(1.,None),text="Switch Setting",halign="center")
        self.switch.grid = GridLayout(cols=2)
        self.switch.layout.add_widget(self.switch.topLbl)
        self.switch.layout.add_widget(self.switch.grid)
        self.switch.chgVal = self.chgSwitch
        self.switch.opts = [
            (CheckBox(
                state=toState[self.switch.val==c],size_hint_x=None,
                width=40,group=self,allow_no_selection=False),
             sizedCls(
                 ClkLabel,text=self.swLbls[c],size_hint=(1.0,1.)))
            for c in xrange(self.nThrow)]
        for c in xrange(self.nThrow):
            btn, lbl = self.switch.opts[c]
            btn.bind(state=Bind(radioBtnChg, self.switch, c))
            lbl.bind(on_press=Bind(togChkBox, btn))
            lbl.color = stRGB[self.switch.val==c]
            self.switch.grid.add_widget(btn)
            self.switch.grid.add_widget(lbl)'''
        self.binary.layout = BorderedGridLayout(
            size_hint=(sizeX-padX,sizeY),cols=1,
            pos_hint={"top":posY,'x':posX+sizeX+padX/2})
        self.binary.topLbl = sizedLabel(
            size_hint=(1.,None),text="Control Values",halign="center")
        self.binary.grid = GridLayout(cols=1, spacing=[2,2])
        self.binary.layout.add_widget(self.binary.topLbl)
        self.binary.layout.add_widget(self.binary.grid)
        self.binary.lbls = map(
            lambda txt, st:(sizedLabel(
                text=txt%("FFF", stNames[st]),
                size_hint=(1.,1.),padding_x=20,markup=True)),
            self.txtTbl, self.states)
        for lbl in self.binary.lbls:
            self.binary.grid.add_widget(lbl)
        self.setSwitch(self.defSW_Val)
    def getPinNames(self):
        return self.txtTable
    def programIt(self, obj=None):
        for st, pin in zip(self.states, self.pins):
            pin.value = st
        self.oldStates = list(self.states)
        super(BaseSwitch, self).programIt(obj)
        self.updBinLbls()
class BaseDualSwitch(BaseProgPart):
    XM_partName = "Unknown"
    mfgPartName = "Unknown"
    catId = CAT_SWITCH
    mfgName = "Unknown"
    interface = "parallel"
    ioType = IO_PIN
    nThrow = 2
    nPins = 2
    swLbls0 = ["RFA1  On", "RFA2  On"]
    swLbls1 = ["RFB1  On", "RFB2  On"]
    defSW_Val0, defSW_Val1 = 0, 0
    optMaps0 = {0: [1], 1: [0]}
    optMaps1 = {0: [1], 1: [0]}
    txtTable = ["A", "B"]
    layout = None
    swGrid = None
    def __init__(self, fl=None, ver=0):
        self.states = [0] * self.nPins
        if not hasattr(self, "oldStates"):
            try: self.oldStates = list(self.optMaps0[self.defSW_Val0]) + list(self.optMaps1[self.defSW_Val1])
            except KeyError:
                log.Log(
                    "index key0 (%u) or key1 (%u) was not found in the optMaps. class %s" %(
                    self.defSW_Val0, self.defSW_Val1, self.__class__.__name__), 2)
                self.oldStates = list(self.states)
        else: assert len(self.oldStates) == len(self.states)
        # ErrMsg = "Error with optMaps[%s]=%s in class " + self.__class__.__name__
        allLens0, allLens1 = map(len, self.optMaps0.values()), map(len, self.optMaps1.values())
        assert map(lambda x: x == allLens0[0], allLens0), "All of OptMaps0 must have same len"
        assert map(lambda x: x == allLens1[0], allLens1), "All of OptMaps1 must have same len"
        self.lenBin0 = allLens0[0]
        self.lenBin1 = allLens1[0]
        assert self.lenBin0 + self.lenBin1 == self.nPins, "optMaps0 and optMaps1 must both fit in nPins"
        super(BaseDualSwitch, self).__init__(fl, ver)
        if fl is not None:
            data = GetLongStrBytes(fl.read(((self.nThrow - 1).bit_length() + 7) // 8))
            self.states[:self.lenBin0] = self.optMaps0[data]
            self.defSW_Val0 = data
            data = GetLongStrBytes(fl.read(((self.nThrow - 1).bit_length() + 7) // 8))
            self.states[self.lenBin0:] = self.optMaps1[data]
            self.defSW_Val1 = data
        self.switch0 = SwitchStorage()
        self.switch1 = SwitchStorage()
        self.binary = BinaryStorage()
        self.txtTbl = [
            "[color=#FFF]%s  [/color][color=#%%s]%%s[/color]" % txt
            for txt in self.txtTable]
        self.buildFirst()
    def serialize(self, fl, ver):
        super(BaseDualSwitch, self).serialize(fl, ver)
        data = self.defSW_Val0
        if hasattr(self, "switch0"):
            data = self.switch0.val
        fl.write(AlignStrBytesLong(data, ((self.nThrow - 1).bit_length() + 7) // 8))
        data = self.defSW_Val1
        if hasattr(self, "switch1"):
            data = self.switch1.val
        fl.write(AlignStrBytesLong(data, ((self.nThrow - 1).bit_length() + 7) // 8))
        fl.flush()
    def onhide(self, app):
        self.layout.remove_widget(self.swGrid)
        self.layout.remove_widget(self.binary.layout)
        super(BaseDualSwitch, self).onhide(app)
    def build(self, app):
        layout = super(BaseDualSwitch, self).build(app)
        if self.binary.layout is None:
            self.buildFirst()
            self.chgSwitch0(self.switch0.val)
            self.chgSwitch1(self.switch1.val)
        sizeY = .61
        posY = .93
        posY1 = posY - sizeY - .03
        layout.add_widget(self.swGrid)
        layout.add_widget(self.binary.layout)
        #layout.add_widget(self.binLayout)
        self.progBtn.pos_hint = {"top": posY1, 'x': .1}
        self.layout = layout
    def updBinLbls(self):
        for c in xrange(len(self.binary.lbls)):
            lbl = self.binary.lbls[c]
            color = cmpColors[self.states[c]!=self.oldStates[c]]
            lbl.text = self.txtTbl[c]%(color,stNames[self.states[c]])
    def setSwitch0(self, index):
        if index not in self.optMaps0:
            log.Log(
                "index %u was not found in the optMaps. class %s" %(
                index, self.__class__.__name__), 2)
            return
        self.states[:self.lenBin0] = self.optMaps0[index]
        self.updBinLbls()
    def setSwitch1(self, index):
        if index not in self.optMaps1:
            log.Log(
                "index %u was not found in the optMaps. class %s" %(
                index, self.__class__.__name__), 2)
            return
        self.states[self.lenBin0:] = self.optMaps1[index]
        self.updBinLbls()
    def chgSwitch0(self, index):
        self.setSwitch0(index)
        self.onProgChg()
    def chgSwitch1(self, index):
        self.setSwitch1(index)
        self.onProgChg()
    def buildFirst(self):
        sizeX = .4
        padX = .015
        #posX = .5 - sizeX/2
        posX = .1
        sizeY = .61
        posY = .93
        # posY1 = posY - sizeY - .03
        # SWITCH 0
        self.switch0.init(
            self.defSW_Val0, "Switch Setting A",
            lambda i: self.chgSwitch0(i), self.nThrow, self.swLbls0)
        '''self.switch0.val = self.defSW_Val0
        self.switch0.layout = BorderedGridLayout(cols=1)
        self.switch0.topLbl = sizedLabel(
            size_hint=(1.,None),text="Switch Setting A",halign="center")
        self.switch0.grid = GridLayout(cols=2)
        self.switch0.layout.add_widget(self.switch0.topLbl)
        self.switch0.layout.add_widget(self.switch0.grid)
        self.switch0.chgVal = self.chgSwitch0
        self.switch0.opts = [
            (CheckBox(
                state=toState[self.switch0.val==c],size_hint_x=None,
                width=40,group=self.switch0,allow_no_selection=False),
             sizedCls(
                 ClkLabel,text=self.swLbls0[c],size_hint=(1.0,1.)))
            for c in xrange(self.nThrow)]
        for c in xrange(self.nThrow):
            btn, lbl = self.switch0.opts[c]
            btn.bind(state=Bind(radioBtnChg, self.switch0, c))
            lbl.bind(on_press=Bind(togChkBox, btn))
            lbl.color = stRGB[self.switch0.val==c]
            self.switch0.grid.add_widget(btn)
            self.switch0.grid.add_widget(lbl)'''
        # SWITCH 1
        self.switch1.init(
            self.defSW_Val1, "Switch Setting B",
            lambda i: self.chgSwitch1(i), self.nThrow, self.swLbls1)
        '''self.switch1.val = self.defSW_Val1
        self.switch1.layout = BorderedGridLayout(cols=1)
        self.switch1.topLbl = sizedLabel(
            size_hint=(1.,None),text="Switch Setting B",halign="center")
        self.switch1.grid = GridLayout(cols=2)
        self.switch1.layout.add_widget(self.switch1.topLbl)
        self.switch1.layout.add_widget(self.switch1.grid)
        self.switch1.chgVal = self.chgSwitch1
        self.switch1.opts = [
            (CheckBox(
                state=toState[self.switch1.val==c],size_hint_x=None,
                width=40,group=self.switch1,allow_no_selection=False),
             sizedCls(
                 ClkLabel,text=self.swLbls1[c],size_hint=(1.0,1.)))
            for c in xrange(self.nThrow)]
        for c in xrange(self.nThrow):
            btn, lbl = self.switch1.opts[c]
            btn.bind(state=Bind(radioBtnChg, self.switch1, c))
            lbl.bind(on_press=Bind(togChkBox, btn))
            lbl.color = stRGB[self.switch1.val==c]
            self.switch1.grid.add_widget(btn)
            self.switch1.grid.add_widget(lbl)'''
        self.swGrid = BorderedGridLayout(
            size_hint=(sizeX-padX,sizeY),pos_hint={"top":posY,'x':posX+padX/2},
            cols=2, spacing=[10,10], padding=[10,10])
        self.swGrid.add_widget(self.switch0.layout)
        self.swGrid.add_widget(self.switch1.layout)
        # BINARY
        self.binary.layout = BorderedGridLayout(
            size_hint=(sizeX-padX,sizeY),cols=1,
            pos_hint={"top":posY,'x':posX+sizeX+padX/2})
        self.binary.topLbl = sizedLabel(
            size_hint=(1.,None),text="Control Values",halign="center")
        self.binary.grid = GridLayout(cols=1, spacing=[2,2])
        self.binary.layout.add_widget(self.binary.topLbl)
        self.binary.layout.add_widget(self.binary.grid)
        self.binary.lbls = map(
            lambda txt, st:(sizedLabel(
                text=txt%("FFF", stNames[st]),
                size_hint=(1.,1.),padding_x=20,markup=True)),
            self.txtTbl, self.states)
        for lbl in self.binary.lbls: self.binary.grid.add_widget(lbl)
        self.setSwitch0(self.defSW_Val0)
        self.setSwitch1(self.defSW_Val1)
    def getPinNames(self):
        return self.txtTable
    def programIt(self, obj=None):
        for st, pin in zip(self.states, self.pins):
            pin.value = st
        self.oldStates = list(self.states)
        super(BaseDualSwitch, self).programIt(obj)
        self.updBinLbls()
