from ..UserDataViewHelpers import *
from ..LibGetTabPanel import *

log = None

@initFunc(__name__.replace("ProgramIt.", ""))
def init(progLog):
    # noinspection PyGlobalUndefined
    global log
    log = progLog


def getNameOfEntry(entry):
    """
    :param (BaseUserDataView, str|unicode|None) entry:
    :rtype: str|unicode
    """
    return entry[0].name


def getCommonMinMax(lstMinMax):
    rtn = [0, 0]
    nMinMax = len(lstMinMax)
    c = 0
    while c < nMinMax:
        minMax = lstMinMax[c]
        c += 1
        if minMax is not None:
            rtn[0] = minMax[0]
            rtn[1] = minMax[1]
            break
    if c >= nMinMax:
        return None
    while c < nMinMax:
        minMax = lstMinMax[c]
        if minMax is not None:
            rtn[0] = max(minMax[0], rtn[0])
            rtn[1] = min(minMax[1], rtn[1])
        c += 1
    return rtn[0], rtn[1]

def sumMinMax(lstMinMax):
    rtn = [0, 0]
    nMinMax = len(lstMinMax)
    c = 0
    while c < nMinMax:
        minMax = lstMinMax[c]
        c += 1
        if minMax is not None:
            rtn[0] = minMax[0]
            rtn[1] = minMax[1]
            break
    if c >= nMinMax:
        return None
    while c < nMinMax:
        minMax = lstMinMax[c]
        if minMax is not None:
            rtn[0] += minMax[0]
            rtn[1] += minMax[1]
        c += 1
    return rtn[0], rtn[1]

def mulMinMax(mul, minMax):
    """
    :param int|long|float mul:
    :param (int|long, int|long)|None minMax:
    :rtype: (int|long, int|long)|None
    """
    return None if minMax is None else (minMax[0] * mul, minMax[1] * mul)


def getMinMaxCtx(context, k):
    """
    :param BaseProgWithRegs context:
    :param str|unicode k:
    :rtype: (int|long|float|Decimal, int|long|float|Decimal)|None
    """
    entry = context.getEntry(k)
    assert isinstance(entry, NumberUserDataView)
    varRange = entry.varRange
    assert varRange is not None
    if isinstance(varRange, BaseRange):
        return varRange.getMinMax(context)
    if len(varRange) == 2:
        return varRange
    else:
        raise ValueError("Unexpected varRange=%s" % repr(varRange))


class BaseCalc(object):
    """
    :type requires: list[str|unicode]|None
    """
    requires = None
    def updateMode(self, context):
        """
        :param BaseProgWithRegs context:
        """
        raise NotImplementedError("Not Implemented")
    def update(self, context):
        """
        :param BaseProgWithRegs context:
        """
        raise NotImplementedError("Not Implemented")

class BasePFD_Calc(BaseCalc):
    """
    :type pfdFreqShift: int|long
    :type lstAutoROpts: list[(bool, str|unicode)]|dict[int|long, str|unicode]
    :type lstAutoRInfs: list[str|unicode]|dict[int|long, str|unicode]
    :type names: list[str|unicode]
    """
    pfdFreqShift = 0
    lstAutoROpts = map(isReserv, [
        "R Counter", "Desired PFD Frequency"])
    lstAutoRInfs = [
        "R counter is manually entered.",
        "R counter is calculated from desired PFD frequency."]
    names = ["selAutoR", "numR_Count", "numUsrPFD_Freq", "numPFD_Freq", "numR_Freq"]
    def __init__(self, names):
        self.names = names
        self.requires = self.names
    def getPFD_Freq(self, context):
        return context.getData(self.names[3]) * 10 ** self.pfdFreqShift
    def updateMode(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        selAutoR = context.getData(names[0])
        txtR_Count = context.getWidget(names[1])
        txtUsrPFD_Freq = context.getWidget(names[2])
        txtR_Count.disabled = selAutoR
        txtUsrPFD_Freq.disabled = not selAutoR
    def getR_Freq(self, context):
        """
        :param BaseProgWithRegs context:
        """
        return context.getData(self.names[4])
    def update(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        rFreq = self.getR_Freq(context)
        selAutoR = context.getData(names[0])
        val = None
        if selAutoR:
            minMax = getMinMaxCtx(context, names[1])
            pfdFreq = context.getData(names[2]) * 10 ** self.pfdFreqShift
            val = int(rFreq / pfdFreq + .5)
            val = restrict(val, minMax[0], minMax[1])
            context.setData(names[1], val, DC_SYNCH)
        if val is None:
            val = context.getData(names[1])
        context.setData(names[3], rFreq / (val * 10.0 ** self.pfdFreqShift), DC_SYNCH)

class Mul2PFD_Calc(BasePFD_Calc):
    def getR_Freq(self, context):
        rFreq = super(Mul2PFD_Calc, self).getR_Freq(context)
        if context.getData(self.names[5]):
            rFreq *= 2
        return rFreq

class MulDivPFD_Calc(BasePFD_Calc):
    def getR_Freq(self, context):
        rFreq = super(MulDivPFD_Calc, self).getR_Freq(context)
        names = self.names
        if context.getData(names[5]):
            rFreq *= 2
        if context.getData(names[6]):
            rFreq /= 2
        return rFreq

def getNamesPFD(mul=False, div=False, dual=-1):
    rtn = [
        "selAutoR",
        "numR_Count",
        "numUsrPFD_Freq",
        "numPFD_Freq",
        "numR_Freq",
    ]
    if mul:
        rtn.append("bRefDbl")
    if div:
        rtn.append("bRefDiv")
    if dual >= 0:
        for c in xrange(len(rtn)):
            if c == 4: continue
            rtn[c] += str(dual)
    return rtn


class BaseNVal_Calc(BaseCalc):
    """
    :type lstAutoNOpts: list[(bool, str|unicode)]|dict[int|long, str|unicode]
    :type lstAutoNInfs: list[str|unicode]|dict[int|long, str|unicode]
    :type names: list[str|unicode]
    :type otherNames: list[str|unicode]
    :type fracIntNames: list[str|unicode]
    :type nameAutoR: str|unicode
    :type numberConverter: (int|long|float|Decimal|bool) -> float|int|long|Decimal|bool
    """
    lstAutoNOpts = []
    lstAutoNInfs = []
    names = [
        "selAutoN",
        "numNVal",
        "numRF_Out",
        "numOutFreq"]
    otherNames = []
    fracIntNames = []
    nameAutoR = "AutoR"
    def __init__(self, names, otherNames, fracIntNames, nameAutoR):
        """
        :param list[str|unicode] names:
        :param list[str|unicode] otherNames:
        :param list[str|unicode] fracIntNames:
        :param str|unicode nameAutoR:
        """
        self.names = names
        self.otherNames = otherNames
        self.fracIntNames = fracIntNames
        self.nameAutoR = nameAutoR
        self.requires = self.names[:3] + self.names[4:] + self.otherNames + self.fracIntNames
        self.numberConverter = lambda x: float(x)
    def updateMode(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        fracIntNames = self.fracIntNames
        selAutoN = context.getData(names[0])
        txtRF_Out = context.getWidget(names[2])
        txtNVal = context.getWidget(names[1])
        fracIntWids = [context.getWidget(x) for x in fracIntNames]
        txtRF_Out.disabled = selAutoN != 0
        txtNVal.disabled = selAutoN != 1
        dfi = selAutoN != 2 # disable frac int
        for wid in fracIntWids:
            wid.disabled = dfi
    def update(self, context):
        """
        :param BaseProgWithRegs context:
        """
        selAutoN = context.getData(self.names[0])
        assert selAutoN in [0, 1, 2]
        if selAutoN == 2:
            self.updFracInt(context)
            self.updNVal(context)
        elif selAutoN == 1:
            self.updNVal(context)
            self.updFracInt(context)
        elif selAutoN == 0:
            self.updNVal(context)
            self.updFracInt(context)
        self.updOutFreq(context)
    # translates to and from internal frequency
    def getIntFreq(self, freq, context):
        """
        :param int|long|float|Decimal freq:
        :param BaseProgWithRegs context:
        :rtype: int|long|float|Decimal
        """
        return freq

    def getExtFreq(self, freq, context):
        """
        :param int|long|float|Decimal freq:
        :param BaseProgWithRegs context:
        :rtype: int|long|float|Decimal
        """
        return freq

    def updFracInt(self, context):
        """
        :param BaseProgWithRegs context:
        """
        raise NotImplementedError("Not Implemented")

    def updNVal(self, context):
        """
        :param BaseProgWithRegs context:
        """
        raise NotImplementedError("Not Implemented")

    def updOutFreq(self, context):
        """
        :param BaseProgWithRegs context:
        """
        pfdCalc = context.getCalc(self.nameAutoR)
        assert isinstance(pfdCalc, BasePFD_Calc)
        pfdFreq = pfdCalc.getPFD_Freq(context)
        names = self.names
        toNum = self.numberConverter
        pfdFreq = toNum(pfdFreq)
        nVal = toNum(context.getData(names[1]))
        context.setData(names[3], self.getExtFreq(nVal * pfdFreq, context), DC_SYNCH)


def getNamesNVal(dual=-1):
    rtn = [
        "selAutoN",
        "numNVal",
        "numRF_Out",
        "numOutFreq"
    ]
    if dual >= 0:
        addStr = str(dual)
        for c in xrange(len(rtn)):
            rtn[c] += addStr
    return rtn


class IntAB_Calc(BaseNVal_Calc):
    fracIntNameFmt = ["numACntr%u", "numBCntr%u"]
    otherNameFmt = ["selPreScl%u"]
    useAB_Range_NVal = True
    lstAutoNOpts = map(isReserv,[
        "Desired Frequency", "N Value", "A & B Counters"])
    lstAutoNInfs = [
        " ".join([
            "The N, A and B Counter values",
            "are calculated from the desired Frequency",
            "and the actual Output frequency is",
            "calculated based on the calculated N value."]),
        " ".join([
            "The N value is entered and",
            "the A and B Counter values",
            "are calculated to match it. The actual",
            "N value is placed in the entry box."]),
        " ".join([
            "The A and B Counter values are manually entered",
            "and the N value and output frequency are calculated."])]

    def __init__(self, names, otherNames, fracIntNames, nameAutoR, preSclShift=8, nMul=1, minPreScl=64):
        """
        :param list[str|unicode] names:
        :param list[str|unicode] otherNames:
        :param list[str|unicode] fracIntNames:
        :param str|unicode nameAutoR:
        :param int|long preSclShift:
        :param int|long nMul:
        """
        self.preSclShift = preSclShift
        self.minPreScl = minPreScl
        self.nMul = nMul
        super(IntAB_Calc, self).__init__(names, otherNames, fracIntNames, nameAutoR)
    def getPreSclDiv(self, context):
        """
        :param BaseProgWithRegs context:
        """
        # Str = "WARNING: default getPreSclDiv function called. CLASS %s"
        # log.Log(Str % self.__class__.__name__, 3)
        return self.preSclShift << context.getData(self.otherNames[0])
    def updFracInt(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        fracIntNames = self.fracIntNames
        selAutoN = context.getData(names[0])
        if selAutoN in [1, 0]:
            numNVal = context.getData(names[1])
            b, a = divmod(int(numNVal / float(self.nMul) + .5), self.getPreSclDiv(context))
            minMaxA = getMinMaxCtx(context, fracIntNames[0])
            minMaxB = getMinMaxCtx(context, fracIntNames[1])
            if minMaxA is not None: a = restrict(a, minMaxA[0], minMaxA[1])
            if minMaxB is not None: b = restrict(b, minMaxB[0], minMaxB[1])
            context.setData(fracIntNames[0], a, DC_SYNCH)
            context.setData(fracIntNames[1], b, DC_SYNCH)
        # registers update automatically
    def updNVal(self, context):
        """
        :param BaseProgWithRegs context:
        """
        pMod = self.getPreSclDiv(context)
        nMul = self.nMul
        names = self.names
        fracIntNames = self.fracIntNames
        calcAutoR = context.getCalc(self.nameAutoR)
        assert isinstance(calcAutoR, BasePFD_Calc)
        selAutoN = context.getData(names[0])
        nVal = None
        if selAutoN == 0:
            pfdFreq = calcAutoR.getPFD_Freq(context)
            numRF_Out = context.getData(names[2])
            nVal = int(self.getIntFreq(numRF_Out, context) / (pfdFreq * nMul) + .5) * nMul
        elif selAutoN == 2:
            numACntr = context.getData(fracIntNames[0])
            numBCntr = context.getData(fracIntNames[1])
            nVal = (numACntr + numBCntr * pMod) * nMul
        if nVal is None:
            nVal = context.getData(names[1])
        nVal = int(float(nVal) / nMul + .5) * nMul
        if selAutoN != 2:
            minMax = getMinMaxCtx(context, names[1])
            if self.useAB_Range_NVal:
                minMaxA = getMinMaxCtx(context, fracIntNames[0])
                minMaxB = getMinMaxCtx(context, fracIntNames[1])
                minMaxAB = mulMinMax(nMul, sumMinMax([minMaxA, mulMinMax(pMod, minMaxB)]))
                minMax = getCommonMinMax([minMax, minMaxAB])
            if minMax is not None:
                nVal = restrict(nVal, minMax[0], minMax[1])
        context.setData(names[1], nVal, DC_SYNCH)


def getNamesAB(dual=-1):
    rtn = [
        "numACntr",
        "numBCntr"
    ]
    if dual >= 0:
        for c in xrange(len(rtn)):
            rtn[c] += str(dual)
    return rtn

class IntB_Calc(BaseNVal_Calc):
    lstAutoNOpts = map(isReserv,[
        "Desired Frequency", "N Value", "B Counter"])
    lstAutoNInfs = [
        " ".join([
            "The N and B Counter values are",
            "calculated from the desired Frequency",
            "and the actual Output frequency is",
            "calculated based on the calculated N value."]),
        " ".join([
            "The N value is entered and the B",
            "Counter value is calculated to match it.",
            "The actual N value is placed in the entry box."]),
        " ".join([
            "The B Counter value is manually entered",
            "and the N value and output frequency are calculated."])]
    useB_Range_NVal = True
    def __init__(self, names, otherNames, fracIntNames, nameAutoR, preSclShift=8, nMul=1, minPreScl=64):
        """
        :param list[str|unicode] names:
        :param list[str|unicode] otherNames:
        :param list[str|unicode] fracIntNames:
        :param str|unicode nameAutoR:
        :param int|long preSclShift:
        :param int|long nMul:
        """
        self.preSclShift = preSclShift
        self.minPreScl = minPreScl
        self.nMul = nMul
        super(IntB_Calc, self).__init__(names, otherNames, fracIntNames, nameAutoR)
    def getPreSclDiv(self, context):
        """
        :param BaseProgWithRegs context:
        """
        rtn = self.preSclShift
        if len(self.otherNames):
            rtn <<= context.getData(self.otherNames[0])
        return rtn
    def updFracInt(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        fracIntNames = self.fracIntNames
        selAutoN = context.getData(names[0])
        if selAutoN in [1, 0]:
            numNVal = context.getData(names[1])
            mulMod = self.getPreSclDiv(context) * self.nMul
            b = int(numNVal / float(mulMod) + .5)
            minMaxB = getMinMaxCtx(context, fracIntNames[0])
            if minMaxB is not None: b = restrict(b, minMaxB[0], minMaxB[1])
            context.setData(fracIntNames[0], b, DC_SYNCH)
        # registers update automatically
    def updNVal(self, context):
        """
        :param BaseProgWithRegs context:
        """
        pMod = self.getPreSclDiv(context)
        nMul = self.nMul
        mulMod = pMod * nMul
        names = self.names
        fracIntNames = self.fracIntNames
        calcAutoR = context.getCalc(self.nameAutoR)
        assert isinstance(calcAutoR, BasePFD_Calc)
        selAutoN = context.getData(names[0])
        nVal = None
        if selAutoN == 0:
            pfdFreq = calcAutoR.getPFD_Freq(context)
            numRF_Out = context.getData(names[2])
            nVal = int(self.getIntFreq(numRF_Out, context) / (pfdFreq * mulMod) + .5) * mulMod
        elif selAutoN == 2:
            numBCntr = context.getData(fracIntNames[0])
            nVal = numBCntr * mulMod
        if nVal is None:
            nVal = context.getData(names[1])
        nVal = int(float(nVal) / mulMod + .5) * mulMod
        if selAutoN != 2:
            minMax = getMinMaxCtx(context, names[1])
            if self.useB_Range_NVal:
                minMaxB = getMinMaxCtx(context, fracIntNames[0])
                minMax = getCommonMinMax([minMax, mulMinMax(mulMod, minMaxB)])
            if minMax is not None:
                nVal = restrict(nVal, minMax[0], minMax[1])
        context.setData(names[1], nVal, DC_SYNCH)


def getNamesB(dual=-1):
    return ["numBCntr" + ("" if dual < 0 else str(dual))]

class IntFrac_Calc(BaseNVal_Calc):
    lstAutoNOpts = map(isReserv, [
        "Desired Frequency", "N Value", "Fraction & Integer"])
    lstAutoNInfs = [
        " ".join([
            "The N, Fraction and Integer values",
            "are calculated from the desired Frequency",
            "and the actual Output frequency is",
            "calculated based on the N value."]),
        " ".join([
            "The N value is entered and",
            "the Fraction and Integer values",
            "are calculated to match it. The actual",
            "N value is placed in the entry box."]),
        " ".join([
            "The Fraction and Integer values are manually entered",
            "and the N value and output frequency are calculated."])]
    def updFracInt(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        fracIntNames = self.fracIntNames
        otherNames = self.otherNames
        selAutoN = context.getData(names[0])
        if selAutoN in [1, 0]:
            mod = context.getData(otherNames[0])
            if mod == 0:
                Int = int(context.getData(names[1]) + 0.5)
                frac = 0
            else:
                Int, frac = divmod(int(context.getData(names[1]) * mod + 0.5), mod)
            minMaxI = getMinMaxCtx(context, fracIntNames[0])
            minMaxF = getMinMaxCtx(context, fracIntNames[1])
            if minMaxI is not None: Int = restrict(Int, minMaxI[0], minMaxI[1])
            if minMaxF is not None: frac = restrict(frac, minMaxF[0], minMaxF[1])
            # preInt = int(Int)
            # Int = restrict(int(Int), iRange.Min, iRange.Max)
            # if preInt > iRange.Max: frac += (preInt - Int) * mod
            # frac = int(restrict(frac, fRange.Min, min(fRange.Max, mod)))
            if frac > mod: frac = mod
            context.setData(fracIntNames[0], Int, DC_SYNCH)
            context.setData(fracIntNames[1], frac, DC_SYNCH)
        # registers should automatically update
    def updNVal(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        fracIntNames = self.fracIntNames
        otherNames = self.otherNames
        calcAutoR = context.getCalc(self.nameAutoR)
        assert isinstance(calcAutoR, BasePFD_Calc)
        selAutoN = context.getData(names[0])
        toNum = self.numberConverter
        mod = toNum(context.getData(otherNames[0]))
        nVal = context.getData(names[1])
        if selAutoN == 0:
            pfdFreq = calcAutoR.getPFD_Freq(context)
            nVal = self.getIntFreq(context.getData(names[2]), context) / pfdFreq
        elif selAutoN == 2:
            nVal = context.getData(fracIntNames[0]) + context.getData(fracIntNames[1]) / mod
        if mod != 0:
            nVal = int(nVal * mod + .5) / mod
        else:
            nVal = int(nVal + 0.5)
        minMax = getMinMaxCtx(context, names[1])
        if selAutoN != 2 and minMax is not None:
            nVal = toNum(restrict(nVal, minMax[0], minMax[1]))
        context.setData(names[1], nVal, DC_SYNCH)


def getNamesIntFrac(dual=-1):
    addStr = "" if dual < 0 else str(dual)
    return ["numIntVal" + addStr, "numFracVal" + addStr]

def getNamesIntFracO(dual=-1):
    addStr = "" if dual < 0 else str(dual)
    return ["numModVal" + addStr]


class IntFrac2_Calc(BaseNVal_Calc):
    lstAutoNOpts = map(isReserv, [
        "Desired Frequency", "N Value", "Integer & Fraction"])
    lstAutoNInfs = [
        " ".join([
            "The N, Fraction and Integer values",
            "are calculated from the desired Frequency",
            "and the actual Output frequency is",
            "calculated based on the N value."]),
        " ".join([
            "The N value is entered and",
            "the Fraction and Integer values",
            "are calculated to match it. The actual",
            "N value is placed in the entry box."]),
        " ".join([
            "The Fraction and Integer values are manually entered",
            "and the N value and output frequency are calculated."])]
    def __init__(self, names, otherNames, fracIntNames, nameAutoR):
        """
        :param list[str|unicode] names:
        :param list[str|unicode] otherNames:
        :param list[str|unicode] fracIntNames:
        :param str|unicode nameAutoR:
        """
        super(IntFrac2_Calc, self).__init__(names, otherNames, fracIntNames, nameAutoR)
    def updFracInt(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        fracIntNames = self.fracIntNames
        otherNames = self.otherNames
        selAutoN = context.getData(names[0])
        if selAutoN in [1, 0]:
            toNum = self.numberConverter
            mod1 = toNum(context.getData(otherNames[0]))
            mod2 = toNum(context.getData(otherNames[1]))
            Int, frac = divmod(context.getData(names[1]) * mod1 + toNum(.5) / mod2, mod1)
            minMaxI = getMinMaxCtx(context, fracIntNames[0])
            minMaxF1 = getMinMaxCtx(context, fracIntNames[1])
            minMaxF2 = getMinMaxCtx(context, fracIntNames[2])
            frac1, frac2 = divmod(int(frac * mod2), mod2)
            Int = int(Int)
            if minMaxI is not None: Int = restrict(Int, minMaxI[0], minMaxI[1])
            if minMaxF1 is not None: frac1 = restrict(frac1, minMaxF1[0], minMaxF1[1])
            if minMaxF2 is not None: frac2 = restrict(frac2, minMaxF2[0], minMaxF2[1])
            #preInt = int(Int)
            #Int = restrict(int(Int), iRange.Min, iRange.Max)
            #if preInt > iRange.Max: frac += (preInt - Int) * mod1
            # preFrac1 = int(frac1)
            # frac1 = int(restrict(frac1, f1Range.Min, min(f1Range.Max, mod1)))
            # if preFrac1 > f1Range.Max: frac2 += (preFrac1 - frac1) * mod2
            # frac2 = int(restrict(frac2, f2Range.Min, min(f2Range.Max, mod2)))
            if frac1 > mod1: frac1 = mod1
            if frac2 > mod2: frac2 = mod2
            context.setData(fracIntNames[0], int(Int), DC_SYNCH)
            context.setData(fracIntNames[1], int(frac1), DC_SYNCH)
            context.setData(fracIntNames[2], int(frac2), DC_SYNCH)
        # registers should automatically update
    def updNVal(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        fracIntNames = self.fracIntNames
        otherNames = self.otherNames
        calcAutoR = context.getCalc(self.nameAutoR)
        assert isinstance(calcAutoR, BasePFD_Calc)
        selAutoN = context.getData(names[0])
        minMax = getMinMaxCtx(context, names[1])
        toNum = self.numberConverter
        mod1 = toNum(context.getData(otherNames[0]))
        mod2 = toNum(context.getData(otherNames[1]))
        nVal = toNum(context.getData(names[1]))
        if selAutoN == 0:
            pfdFreq = toNum(calcAutoR.getPFD_Freq(context))
            nVal = toNum(self.getIntFreq(context.getData(names[2]), context)) / pfdFreq
        elif selAutoN == 2:
            nVal = toNum(context.getData(fracIntNames[0]))
            nVal += (toNum(context.getData(fracIntNames[1])) + toNum(context.getData(fracIntNames[2])) / mod2) / mod1
        nVal = int(nVal * mod1 * mod2 + toNum(.5)) / (mod1 * mod2)
        if selAutoN != 2 and minMax is not None:
            nVal = toNum(restrict(nVal, minMax[0], minMax[1]))
        context.setData(names[1], nVal, DC_SYNCH)


def getNamesIntFrac2(dual=-1):
    addStr = "" if dual < 0 else str(dual)
    return ["numIntVal" + addStr, "numFrac1Val" + addStr, "numFrac2Val" + addStr]

def getNamesIntFrac2O(dual=-1):
    addStr = "" if dual < 0 else str(dual)
    return ["numMod1Val" + addStr, "numMod2Val" + addStr]


class IntFrac2Div_Calc(IntFrac2_Calc):
    # translates to and from internal frequency
    def getIntFreq(self, freq, context):
        """
        :param int|long|float|Decimal freq:
        :param BaseProgWithRegs context:
        :rtype: int|long|float|Decimal
        """
        otherNames = self.otherNames
        selFdBkSel = context.getData(otherNames[2])
        freq = self.numberConverter(freq)
        if selFdBkSel == 1:
            selDivRF = context.getData(otherNames[3])
            freq *= self.numberConverter(2 ** selDivRF)
        return freq

    def getExtFreq(self, freq, context):
        """
        :param int|long|float|Decimal freq:
        :param BaseProgWithRegs context:
        :rtype: int|long|float|Decimal
        """
        otherNames = self.otherNames
        selFdBkSel = context.getData(otherNames[2])
        if selFdBkSel == 1:
            selDivRF = context.getData(otherNames[3])
            freq /= self.numberConverter(2 ** selDivRF)
        return freq


def getNamesIntFrac2DivO(dual=-1):
    rtn = ["numMod1Val", "numMod2Val", "selFdBkSel", "selDivRF"]
    if dual >= 0:
        addStr = str(dual)
        for c in xrange(len(rtn)):
            rtn[c] += addStr
    return rtn


class IntFracDiv_Calc(IntFrac_Calc):
    # translates to and from internal frequency
    def getIntFreq(self, freq, context):
        """
        :param int|long|float|Decimal freq:
        :param BaseProgWithRegs context:
        :rtype: int|long|float|Decimal
        """
        otherNames = self.otherNames
        selFdBkSel = context.getData(otherNames[1])
        if selFdBkSel == 1:
            selDivRF = context.getData(otherNames[2])
            freq *= self.numberConverter(2 ** selDivRF)
        return freq

    def getExtFreq(self, freq, context):
        """
        :param int|long|float|Decimal freq:
        :param BaseProgWithRegs context:
        :rtype: int|long|float|Decimal
        """
        otherNames = self.otherNames
        selFdBkSel = context.getData(otherNames[1])
        if selFdBkSel == 1:
            selDivRF = context.getData(otherNames[2])
            freq /= self.numberConverter(2 ** selDivRF)
        return freq


def getNamesIntFracDivO(dual=-1):
    rtn = ["numModVal", "selFdBkSel", "selDivRF"]
    if dual >= 0:
        addStr = str(dual)
        for c in xrange(len(rtn)):
            rtn[c] += addStr
    return rtn



class BaseModSpace_Calc(BaseCalc):
    """
    :type lstAutoModOpts: list[(bool, str|unicode)]|dict[int|long, str|unicode]
    :type lstAutoModInfs: list[str|unicode]|dict[int|long, str|unicode]
    :type names: list[str|unicode]
    :type otherNames: list[str|unicode]
    :type modNames: list[str|unicode]
    :type nameAutoR: str|unicode
    """
    lstAutoModOpts = []
    lstAutoModInfs = []
    names = [
        "selAutoMod",
        "numUsrSpaceCh",
        "numSpaceCh"]
    otherNames = []
    modNames = []
    nameAutoR = "AutoR"
    def __init__(self, names, otherNames, modNames, nameAutoR):
        """
        :param list[str|unicode] names:
        :param list[str|unicode] otherNames:
        :param list[str|unicode] modNames:
        :param str|unicode nameAutoR:
        """
        self.names = names
        self.otherNames = otherNames
        self.modNames = modNames
        self.nameAutoR = nameAutoR
        self.requires = self.names[:2] + self.names[3:] + self.otherNames + self.modNames
    def updateMode(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        modNames = self.modNames
        selAutoMod = context.getData(names[0])
        txtUsrSpaceCh = context.getWidget(names[1])
        txtUsrSpaceCh.disabled = not selAutoMod
        modWids = [context.getWidget(k) for k in  modNames]
        dMod = selAutoMod  # disable modulus
        for wid in modWids:
            wid.disabled = dMod
    def update(self, context):
        """
        :param BaseProgWithRegs context:
        """
        raise NotImplementedError("Not Implemented")


def getNamesModSpace(dual=-1):
    rtn = ["selAutoMod", "numUsrSpaceCh", "numSpaceCh"]
    if dual >= 0:
        addStr = str(dual)
        for c in xrange(len(rtn)):
            rtn[c] += addStr
    return rtn


class AuxMod_Calc(BaseModSpace_Calc):
    lstAutoModOpts = map(isReserv, [
        "Auxiliary Modulus Value", "Desired Frequency Step"])
    lstAutoModInfs = [
        "Auxiliary modulus value is manually entered",
        " ".join([
            "Auxiliary modulus value is calculated based on",
            "desired frequency step."])]
    def update(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        otherNames = self.otherNames
        modNames = self.modNames
        calcAutoR = context.getCalc(self.nameAutoR)
        assert isinstance(calcAutoR, BasePFD_Calc)
        pfdFreq = calcAutoR.getPFD_Freq(context) * 1e3
        selAutoMod = context.getData(names[0])
        if selAutoMod:
            minMax = getMinMaxCtx(context, modNames[0])
            val = int(sFracBetter(pfdFreq, context.getData(names[1]))[0])
            if minMax is not None: val = restrict(val, minMax[0], minMax[1])
            context.setData(modNames[0], val, DC_SYNCH)
        mod1 = context.getData(otherNames[0])
        mod2 = context.getData(modNames[0])
        context.setData(names[2], pfdFreq / (mod1 * mod2), DC_SYNCH)


class MainMod_Calc(BaseModSpace_Calc):
    lstAutoModOpts = map(isReserv, [
        "Modulus Value", "Desired Channel Spacing"])
    lstAutoModInfs = [
        "Modulus value is manually entered.",
        "Modulus value is calculated based on desired channel spacing."]
    def update(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        modNames = self.modNames
        calcAutoR = context.getCalc(self.nameAutoR)
        assert isinstance(calcAutoR, BasePFD_Calc)
        pfdFreq = calcAutoR.getPFD_Freq(context) * 1e3
        selAutoMod = context.getData(names[0])
        numModVal = None
        if selAutoMod:
            minMax = getMinMaxCtx(context, modNames[0])
            numUsrSpaceCh = context.getData(names[1])
            numModVal = int(pfdFreq / numUsrSpaceCh + .5)
            if minMax is not None:
                numModVal = restrict(numModVal, minMax[0], minMax[1])
            context.setData(modNames[0], numModVal, DC_SYNCH)
        if numModVal is None:
            numModVal = context.getData(modNames[0])
        if numModVal == 0:
            context.setData(names[2], 0, DC_SYNCH)
        else:
            context.setData(names[2], pfdFreq / numModVal, DC_SYNCH)

# theMulDivDualPFD_Calc1 = MulDivPFD_Calc(["selAutoR1", "numR_Count1", "numUsrPFD_Freq1", "numPFD_Freq1", "numR_Freq", "bRefDbl1", "bRefDiv1"])

thePFD_Calc = BasePFD_Calc(getNamesPFD())
theMilliPFD_Calc = BasePFD_Calc(getNamesPFD())
theMilliPFD_Calc.pfdFreqShift = -3
theMulDivPFD_Calc = MulDivPFD_Calc(getNamesPFD(True, True))
theMul2PFD_Calc = Mul2PFD_Calc(getNamesPFD(True))
theAuxModCalc = AuxMod_Calc(getNamesModSpace(), ["numMod1Val"], ["numMod2Val"], "AutoR")
theMainModCalc = MainMod_Calc(getNamesModSpace(), [], ["numModVal"], "AutoR")
theIntFracCalc = IntFrac_Calc(getNamesNVal(), getNamesIntFracO(), getNamesIntFrac(), "AutoR")
theIntFrac2Calc = IntFrac2_Calc(getNamesNVal(), getNamesIntFrac2O(), getNamesIntFrac2(), "AutoR")
theIntFrac2DivCalc = IntFrac2Div_Calc(getNamesNVal(), getNamesIntFrac2DivO(), getNamesIntFrac2(), "AutoR")
theIntFracDivCalc = IntFracDiv_Calc(getNamesNVal(), getNamesIntFracDivO(), getNamesIntFrac(), "AutoR")
theIntB_Calc = IntB_Calc(getNamesNVal(), [], getNamesB(), "AutoR", 1, 1, 1)


def lstEntriesToDctWidgets(lstEntries, defData):
    """
    :param list[(BaseUserDataView, str|unicode)] lstEntries:
    :param dict[str|unicode, T] defData:
    :rtype: (dict[str|unicode, int|long], list[T])
    """
    rtn = {lstEntries[c][0].name: c for c in xrange(len(lstEntries))}
    i = len(lstEntries)
    extData = []
    for k in defData:
        if rtn.setdefault(k, i) == i:
            extData.append(defData[k])
            i += 1
    return rtn, extData

# changes value
#   acts like a command
#   can be string command:
#     "delete": deletes the data entry and defData
#     "del-entryOnly": deletes only the data entry
#     "del-dataOnly": deletes only the data in defData
#   can be dictionary command:
#     the values in the dictionary indicate the new values for attributes named by the key
#     Keys for Base (all DataViews):
#       "shortTitle", "readOnly", "name", "upstream"
#     Keys for NumberUserDataView (dataViewType == DVT_NUMBER):
#       "popupMaker", "conv", "varRange", "longTitle"
#     Keys for ListUserDataView (dataViewType == DVT_LIST):
#       "lstOpts", "popupMaker", "allowEmptySel", "longTitle"
#     Keys for ChkBoxUserDataView (dataViewType == DVT_CHK):
#       Only Base Keys
#     Keys for InfoListUserDataView (dataViewType == DVT_LIST_INF):
#       "lstOpts", "infLst", "popupMaker", "allowEmptySel", "longTitle"
#     Keys for StaticUserDataView (dataViewType == DVT_STATIC):
#       Only Base Keys except "readOnly"
#   can be tuple command:
#     tuple of (int, dictionary command)
#     int represents the dataViewType attribute of the target DataView class
def changeLstEntries(lstEntries, defData, changes):
    """
    :param list[(BaseUserDataView, str|unicode)] lstEntries:
    :param dict[str|unicode,int|long|float|Decimal] defData:
    :param dict[str|unicode,None|str|unicode|dict[str|unicode,any]|(int|long, dict[str|unicode,any])] changes:
    :rtype: (list[(BaseUserDataView, str|unicode)], dict[str|unicode,int|long|float|Decimal])
    """
    lstResEntries = [None] * len(lstEntries)
    defData = dict(defData)
    resC = 0
    for c, (entry, tgt) in enumerate(lstEntries):
        curChanges = changes.get(entry.name, None)
        if curChanges is not None:
            if isinstance(curChanges, (str, unicode)):
                if curChanges == "delete":
                    defData.pop(entry.name, None)
                    continue
                elif curChanges == "del-entryOnly":
                    continue
                elif curChanges == "del-dataOnly":
                    defData.pop(entry.name, None)
                else:
                    raise ValueError("Unrecognized command: k = %r, command = %r" % (entry.name, curChanges))
            else:
                dataViewType = entry.dataViewType
                if not isinstance(curChanges, dict):
                    dataViewType, curChanges = curChanges
                assert isinstance(curChanges, dict)
                args = dict(curChanges)
                name = args.setdefault("name", entry.name)
                shortTitle = args.setdefault("shortTitle", entry.shortTitle)
                readOnly = args.setdefault("readOnly", entry.readOnly)
                upstream = args.setdefault("upstream", entry.upstream)
                display = args.setdefault("display", entry.display)
                if entry.dataViewType == DVT_NUMBER:
                    assert isinstance(entry, NumberUserDataView)
                    args.setdefault("popupMaker", entry.popupMaker)
                    args.setdefault("conv", entry.conv)
                    args.setdefault("varRange", entry.varRange)
                    args.setdefault("longTitle", entry.longTitle)
                    args.setdefault("unit", entry.unit)
                elif entry.dataViewType == DVT_LIST:
                    assert isinstance(entry, ListUserDataView)
                    args.setdefault("lstOpts", entry.lstOpts)
                    args.setdefault("popupMaker", entry.popupMaker)
                    args.setdefault("allowEmptySel", entry.allowEmptySel)
                    args.setdefault("longTitle", entry.longTitle)
                elif entry.dataViewType == DVT_LIST_INF:
                    assert isinstance(entry, InfoListUserDataView)
                    args.setdefault("lstOpts", entry.lstOpts)
                    args.setdefault("infLst", entry.infLst)
                    args.setdefault("popupMaker", entry.popupMaker)
                    args.setdefault("allowEmptySel", entry.allowEmptySel)
                    args.setdefault("longTitle", entry.longTitle)
                elif entry.dataViewType == DVT_CHK:
                    assert isinstance(entry, ChkBoxUserDataView)
                elif entry.dataViewType == DVT_STATIC:
                    assert isinstance(entry, StaticUserDataView)
                    assert readOnly
                else: raise TypeError(
                    "Unrecognized UserDataView dataViewType = %u, clsName = %s" % (
                        entry.dataViewType, entry.__class__.__name__))
                if dataViewType == DVT_NUMBER:
                    args.setdefault("longTitle", "")
                    args.setdefault("unit", None)
                    entry = NumberUserDataView(
                        name, args["popupMaker"], args["conv"], args["varRange"],
                        args["longTitle"], shortTitle, args["unit"], readOnly)
                elif dataViewType == DVT_LIST:
                    args.setdefault("longTitle", "")
                    args.setdefault("allowEmptySel", False)
                    entry = ListUserDataView(
                        name, args["lstOpts"], args["popupMaker"], args["allowEmptySel"],
                        args["longTitle"], shortTitle, readOnly)
                elif dataViewType == DVT_LIST_INF:
                    args.setdefault("longTitle", "")
                    args.setdefault("allowEmptySel", False)
                    entry = InfoListUserDataView(
                        name, args["lstOpts"], args["infLst"], args["popupMaker"], args["allowEmptySel"],
                        args["longTitle"], shortTitle, readOnly)
                elif dataViewType == DVT_CHK:
                    entry = ChkBoxUserDataView(
                        name,
                        shortTitle, readOnly)
                elif dataViewType == DVT_STATIC:
                    entry = StaticUserDataView(name, shortTitle)
                else: raise TypeError(
                    "Unrecognized UserDataView dataViewType = %u" % dataViewType)
                entry.setUpstream(upstream).setDisplay(display)
                tgt = args.get("tgt", tgt)
        lstResEntries[resC] = (entry, tgt)
        resC += 1
    if resC < len(lstResEntries):
        lstResEntries[resC:] = []
    return lstResEntries, defData

def remWidCalcSys(calcSystems, dctCalcSystemUpd, remWid, remCalcs):
    """
    :param dict[str|unicode,BaseCalc|(BaseCalc|None,str|unicode|list[str|unicode])] calcSystems:
    :param dict[str|unicode,(str|unicode, bool)] dctCalcSystemUpd:
    :param remWid:
    :param remCalcs:
    :rtype: (dict[str|unicode,BaseCalc|(BaseCalc|None,str|unicode|list[str|unicode])], dict[str|unicode,(str|unicode, bool)])
    """
    calcSystemsOut = {}
    for k in calcSystems:
        v = calcSystems[k]
        if k in remCalcs:
            if not isinstance(v, BaseCalc) and v is not None:
                v = (None, v[1])
                calcSystemsOut[k] = v
        else:
            calcSystemsOut[k] = v
    dctCalcSystemUpdOut = {
        k: v
        for k, v in dctCalcSystemUpd.iteritems()
        if k not in remWid and v[0] in calcSystemsOut}
    return calcSystemsOut, dctCalcSystemUpdOut

def addAfterCalcSys(calcSystems, afterName, addName):
    """
    :param dict[str|unicode,BaseCalc|(BaseCalc|None,str|unicode|list[str|unicode])] calcSystems:
    :param str|unicode afterName:
    :param str|unicode addName:
    :rtype: dict[str|unicode,BaseCalc|(BaseCalc|None,str|unicode|list[str|unicode])]
    """
    entry = calcSystems.get(afterName, None)
    if entry is None:
        entry = (None, addName)
    elif isinstance(entry, BaseCalc):
        entry = (entry, addName)
    else:
        assert not isinstance(entry, BaseCalc)
        calc, after = entry
        if isinstance(after, (str, unicode)):
            after = [after, addName]
        else:
            after = after + [addName]
        entry = calc, after
    calcSystems[afterName] = entry
    return calcSystems

def remAfterCalcSys(calcSystems, afterName, remName):
    """
    :param dict[str|unicode,BaseCalc|(BaseCalc|None,str|unicode|list[str|unicode])] calcSystems:
    :param str|unicode afterName:
    :param str|unicode remName:
    :rtype: dict[str|unicode,BaseCalc|(BaseCalc|None,str|unicode|list[str|unicode])]
    """
    entry = calcSystems.get(afterName, None)
    if entry is None:
        return calcSystems
    elif isinstance(entry, BaseCalc):
        return calcSystems
    assert not isinstance(entry, BaseCalc)
    calc, after = entry
    if isinstance(after, (str, unicode)):
        if after == remName:
            entry = calc
    else:
        try:
            pos = after.index(remName)
            after = after[:pos] + after[pos + 1:]
        except ValueError:
            pass
        if len(after) == 0:
            entry = calc
        elif len(after) == 1:
            entry = calc, after[0]
        else:
            entry = calc, after
    calcSystems[afterName] = entry
    return calcSystems

def setCalc(calcSystems, calcName, newCalcObj):
    """
    :param dict[str|unicode,BaseCalc|(BaseCalc|None,str|unicode|list[str|unicode])] calcSystems:
    :param str|unicode calcName:
    :param BaseCalc|None newCalcObj:
    :rtype: dict[str|unicode,BaseCalc|(BaseCalc|None,str|unicode|list[str|unicode])]
    """
    entry = calcSystems.get(calcName, None)
    if entry is None:
        entry = newCalcObj
    elif isinstance(entry, BaseCalc):
        entry = newCalcObj
    else:
        assert not isinstance(entry, BaseCalc)
        entry = newCalcObj, entry[1]
    calcSystems[calcName] = entry
    return calcSystems

def insertDataView(lstDataEntries, entryInsert, beforeName):
    """
    :param list[(BaseUserDataView, str|unicode|None)] lstDataEntries:
    :param (BaseUserDataView, str|unicode|None) entryInsert:
    :param str|unicode beforeName:
    """
    index = len(lstDataEntries)
    for c, entry in enumerate(lstDataEntries):
        if entry[0].name == beforeName:
            index = c
            break
    lstDataEntries.insert(index, entryInsert)

def consolidateRegs(defRegsOrig, ctlSz, addDefRegs):
    """
    :param list[int|long] defRegsOrig:
    :param int|long ctlSz:
    :param list[int|long] addDefRegs:
    :rtype: list[int|long]
    """
    if ctlSz > 8:
        raise ValueError("ctrlSz is a bit too large (number of regs > 256) (number of ctl bits > 8)")
    ctlMask = (1 << ctlSz)
    lstRtn = [0] * ctlMask
    ctlMask -= 1
    for code in defRegsOrig:
        lstRtn[code & ctlMask] = code
    for code in addDefRegs:
        lstRtn[code & ctlMask] = code
    return filter(lambda x: bool(x), lstRtn)

def fltrDefRegs1(ctlSz, wordSz, origNums, chgNums):
    mask = (1 << (ctlSz if ctlSz > 0 else wordSz)) - 1
    dctRegs = {}
    if ctlSz > 0:
        for num in origNums + chgNums:
            dctRegs[num & mask] = num
    else:
        for num in origNums + chgNums:
            dctRegs[num >> wordSz] = num
    return list(dctRegs.values())

simplePllInfoLabels = {
    "gridCfgMd": (
        "gridCfgInf",
        {
            "selAutoN": "Select the method of entering the N Counter value",
            "selAutoMod": "Select the method of entering modulus",
            "selAutoR": "Select the method of entering the R Counter value"
        },
        Label,
        {"size_hint": (1.0, None), "height": 32}
    )
}

def hideWriteButtonOnTab(prog, tabName):
    """
    :param BaseProgWithRegs prog:
    :param str|unicode tabName:
    """
    tab = prog.dctNamed[tabName]
    def hideWriteButton(obj, val):
        if val == tab:
            if prog.progBtn.parent is not None:
                prog.mainParent = prog.progBtn.parent
                prog.mainParent.remove_widget(prog.progBtn)
    prog.tabPanel.bind(current_tab=hideWriteButton)

class BaseProgWithRegs(BaseProgPart, DataContext, RegisterContext):
    """
    :type lstDataEntries: list[(BaseUserDataView, str|unicode|None)]
    :type defData: dict[str|unicode,int|float|Decimal|long|bool]
    :type dctWidgets: dict[str|unicode, int|long]
    :type extData: list[int|float|Decimal|long|bool]
    :type defRegs: list[int|long]
    :type regDuals: list[int|long]
    :type dctCalcSystemUpd: dict[str|unicode,(str|unicode, bool)]
    :type calcSystems: dict[str|unicode,BaseCalc|(BaseCalc|None,str|unicode|list[str|unicode])]
    :type regMaker: (int|long, int|long|slice) -> Register
    :type widgets: list[(tuple[Widget], int)]|None
    :type data: list[int|float|Decimal|long|bool]|None
    :type tabLayout: list[(str|unicode, str|unicode, ((dict[str|unicode,any]) -> Widget, str|unicode|None, dict[str|unicode,any])|((dict[str|unicode,any]) -> Widget, str|unicode|None, dict[str|unicode,any], any)|((dict[str|unicode,any]) -> Widget, str|unicode|None, dict[str|unicode,any],any,int|long))]
    :type tabPanel: TabbedPanel|None
    :type dctNamed: dict[str|unicode,Widget]|None
    :type layout: Widget|None
    :type binLayoutStyle: dict[str|unicode,any]
    :type binLayoutClsName: str|unicode
    :type binLayout: Layout|None
    :type addUserLog: str|unicode|None
    :type infoLabels: dict[str|unicode,(str|unicode, dict[str|unicode,str|unicode], any, dict[str|unicode,any])]
    :type tabNameNoWrite: set[str|unicode]
    :type tabNoWrite: set[Widget]|None
    :type mainParent: Widget|None
    :type custom_name_txt_in: TextInput|None
    """
    # all the attributes below are private-write, public-readonly
    #   in this case reading/writing is equivalent to direct getattr/setattr
    dctWidgets = {}
    extData = []
    widgets = None
    data = None
    tabPanel = None
    dctNamed = None
    layout = None
    binLayout = None
    cntrRstReg = None
    cntrRstIndex = None
    autoCntrRstIndex = None
    defDataPacker = DataPacker(DataArray(
        DataObj(
            DataVarBytes(1, 1),
            DataDirect(SaveWithT_Info, LoadWithT_Info, headLen=2, encoding='utf-8'))))
    defRegsPacker = DataPacker(DataArray(
        DataInt()))
    # all the attributes below are protected-write, public-readonly
    #   in this case reading/writing is equivalent to direct getattr/setattr
    lstDataEntries = []
    defData = {}
    defRegs = []
    regDuals = []
    dctCalcSystemUpd = {}
    calcSystems = {}
    regMaker = Register
    regCtlSz = -1
    regWordSz = -1
    tabLayout = []
    tabPanelStyle = {}
    progBtnPosHint = None
    binLayoutStyle = dict(pos_hint={'y': .02, "right": .99}, size_hint=(.79, .125), cols=5)
    binLayoutClsName = "BorderedGridLayout"
    showBinaryOnMain = True
    addUserLog = None
    cntrRstName = None
    autoCntrRstName = None
    infoLabels = {}
    tabNameNoWrite = set()
    tabNoWrite = None
    mainParent = None
    custom_name_txt_in = None
    @classmethod
    def initCls(cls, numBuses):
        cls.dctWidgets, cls.extData = lstEntriesToDctWidgets(cls.lstDataEntries, cls.defData)
        if cls.cntrRstName is not None:
            i = cls.dctWidgets[cls.cntrRstName]
            entry, tgt = cls.lstDataEntries[i]
            assert isinstance(entry.upstream, RegisterDownstream)
            assert len(entry.upstream.lstWrite) == 1
            cls.cntrRstReg = entry.upstream.lstWrite[0][0]
            cls.cntrRstIndex = i
        if cls.autoCntrRstName is not None:
            cls.autoCntrRstIndex = cls.dctWidgets[cls.autoCntrRstName]
        cName = cls.__name__
        assert cls.regCtlSz != -1, "Expected subclass ('%s') to override regCtlSz" % cName
        assert cls.regWordSz != -1, "Expected subclass ('%s') to override regWordSz" % cName
        runTest = True
        if runTest:
            dataEntries = cls.lstDataEntries
            def canChange(name):
                return lstGetDefault(dataEntries, cls.dctWidgets[name], False, lambda x: x[0].display and not x[0].readOnly)
            results = {}
            for k in cls.calcSystems:
                v = cls.calcSystems[k]
                if not isinstance(v, BaseCalc):
                    v = v[0]
                if v is None:
                    continue
                assert isinstance(v, BaseCalc)
                if v.requires is None:
                    continue
                result = filter(canChange, sorted(set(v.requires).difference(cls.dctCalcSystemUpd)))
                if len(result):
                    results[k] = result
            if len(results):
                StrLst = [""] * len(results)
                for c, k in enumerate(sorted(results)):
                    result = results[k]
                    StrLst[c] = "Expected calcSystem[%r] to be referred to by: %s" % (k, ", ".join(map(repr, result)))
                log.Log("For class=%s:\n  %s" % (cls.__name__, "\n  ".join(StrLst)), 2)



    def __init__(self, fl=None, ver=0):
        self.userLog = UserLogML(size_hint=(1.0, .9))
        self.regOutFmt = "Writing 0x%%0%uX to " % ((self.regWordSz + 3) / 4)
        super(BaseProgWithRegs, self).__init__(fl, ver)
        self.initRegs(self.regDuals, self.regMaker)
        self.widgets = None
        self.data = None
        self.tabPanel = None
        self.layout = None
        self.dctNamed = None
        self.binLayout = None
        # self.dctGridSync = None
        defRegs = self.defRegs
        defData = self.defData
        if fl is not None:
            defData = {k: v for k, v in self.defDataPacker.Unpack(fl)[0]}
            defRegs = self.defRegsPacker.Unpack(fl)[0]
        self.buildData(defRegs, defData)
        self.buildFirst()

    def writeSingleReg(self, obj=None):
        num = self.getData("numWriteData")
        spiWriteNum(num, 24, self.spiBus)

    def serialize(self, fl, ver):
        # TODO: deal with data outside len(lstDataEntries)
        super(BaseProgWithRegs, self).serialize(fl, ver)
        defRegs = self.dumpRegs()
        lstDataEntries = self.lstDataEntries
        data = self.data
        lstPairs = [
            (entry.name, data[c])
            for c, (entry, tgt) in enumerate(lstDataEntries)
            if entry.upstream is None]
        self.defDataPacker.Pack([lstPairs], fl)
        self.defRegsPacker.Pack([defRegs], fl)

    def dumpRegs(self):
        return self.mapRegs(lambda reg: reg.word.num)

    def buildData(self, defRegs, defData):
        """
        :param list[int|long] defRegs:
        :param dict[str|unicode,int|long|float|Decimal] defData:
        """
        lstDataEntries = self.lstDataEntries
        nDataEntries = len(lstDataEntries)
        self.data = data = [0] * nDataEntries
        data.extend(self.extData)
        for code in defRegs:
            self.progReg(code)
        for c in xrange(nDataEntries):
            entry, txt = lstDataEntries[c]
            if entry.upstream is None:
                data[c] = defData.get(entry.name, entry.defVal)
            else:
                try:
                    data[c] = defData[entry.name]
                except KeyError:
                    data[c] = entry.upstream.pull(self)
                else:
                    entry.upstream.push(self, data[c])

    def build(self, app):
        assert isinstance(app, AppData)
        self.layout = super(BaseProgWithRegs, self).build(app)
        if self.tabPanel is None: self.buildFirst()
        add_widget = self.layout.add_widget
        add_widget(self.tabPanel)
        if self.showBinaryOnMain:
            add_widget(self.binLayout)
        if self.progBtnPosHint is None:
            self.progBtn.pos_hint = {
                'y': 0.045 * app.defStyle["size_hint"][1], 'x': .01}
        else:
            self.progBtn.pos_hint = self.progBtnPosHint

    def postBuildFirst(self):
        for calcName in self.calcSystems:
            self.runCalc(calcName, True)
        lstCalcs = ["AutoR"]
        for calcName in lstCalcs:
            self.runCalc(calcName, False)

    def buildFirst(self, add_custom_name_changer="gridOther"):
        """
        :param str|unicode add_custom_name_changer:
        """
        try:
            tabPanel, dctNamed = getTabPanel(self.tabLayout, **self.tabPanelStyle)
        except:
            log.Log("Error with XM_partName=%r, mfgPN=%r" % (self.XM_partName, self.mfgPartName))
            raise
        self.tabPanel = tabPanel
        """:type: TabbedPanel"""
        self.dctNamed = dctNamed
        """:type: dict[str|unicode,Widget]"""
        #self.dctGridSync = {}
        #""":type: dict[str|unicode,AdvGridSync]"""
        lstDataEntries = self.lstDataEntries
        nDataEntries = len(lstDataEntries)
        self.widgets = widgets = [None] * nDataEntries
        """:type: list[(tuple[Widget], int)]"""
        data = self.data
        assert data is not None
        clsName = self.binLayoutClsName
        if clsName == "BorderedGridLayout":
            self.binLayout = BorderedGridLayout(**self.binLayoutStyle)
        elif clsName == "GridLayout":
            self.binLayout = GridLayout(**self.binLayoutStyle)
        elif clsName.startswith("named-"):
            self.binLayout = self.dctNamed[clsName[6:]]
            self.showBinaryOnMain = False
        if self.binLayout is None:
            raise TypeError("Expected binLayoutClsName to name a layout (got %s)" % clsName)
        self.addAllRegs(self.binLayout)
        for c in xrange(nDataEntries):
            entry, tgt = lstDataEntries[c]
            if not entry.display: continue
            group, index = widgets[c] = entry.build(self)
            if tgt is not None:
                tgtLayout = dctNamed[tgt]
                for wid in group:
                    tgtLayout.add_widget(wid)
            entry.updateWidget(data[c], group[index])
            labelInf = self.infoLabels.get(tgt, None)
            if labelInf is None: continue
            tgtInf, dctText, cls, style = labelInf
            dctNamed[tgtInf].add_widget(sizedCls1(
                cls, text=dctText.get(entry.name, ""), **style))
        addUserLog = self.addUserLog
        gridLog = None if addUserLog is None else dctNamed.get(addUserLog, None)
        if gridLog is not None:
            assert addUserLog is not None
            assert isinstance(gridLog, GridLayout)
            gridLog.add_widget(self.userLog.widget)
            gridLogBtns = GridLayout(spacing=[5, 5], rows=1, size_hint=(1.0, 0.1))
            dctNamed["gridLogBtns@" + addUserLog] = gridLogBtns
            gridLogBtns.add_widget(Button(
                text="Clear", on_release=lambda obj: self.userLog.clear()))
            gridLogBtns.add_widget(Button(
                text="Save", on_release=lambda obj: self.userLog.saveDialog()))
            """self.gridHistBtns.add_widget(Button(
                text="Load", on_release=lambda obj: self.userLog.loadDialog()))"""
            gridLog.add_widget(gridLogBtns)
        self.postBuildFirst()
        if add_custom_name_changer is not None and add_custom_name_changer in dctNamed:
            grid = dctNamed[add_custom_name_changer]
            txt_in = TextInput(text=self.custom_name, multiline=False)
            self.custom_name_txt_in = txt_in
            index = len(grid.children)
            grid.add_widget(Label(text="Custom Name"), index)
            grid.add_widget(txt_in, index)
            if getattr(grid, "cols", None) is not None:
                for c in xrange(2, grid.cols):
                    grid.add_widget(Label(), index)
            txt_in.bind(text=lambda obj, v: self.set_custom_name(v))
        self.tabNoWrite = {dctNamed[k] for k in self.tabNameNoWrite}
        if len(self.tabNoWrite):
            self.tabPanel.bind(current_tab=self.curTabEvt)

    def on_chg_custom_name(self, v):
        txt_in = self.custom_name_txt_in
        if txt_in is not None and txt_in.text != v:
            txt_in.text = v

    def curTabEvt(self, obj, val):
        if val in self.tabNoWrite:
            if self.progBtn.parent is not None:
                self.mainParent = self.progBtn.parent
                self.mainParent.remove_widget(self.progBtn)
        elif self.mainParent is not None and self.progBtn.parent is None:
            self.mainParent.add_widget(self.progBtn)
            self.mainParent = None

    def addAllRegs(self, binLayout):
        """
        :param Layout binLayout:
        """
        raise NotImplementedError("Not Implemented")

    def getCalc(self, calcName):
        """
        :param str|unicode calcName:
        :rtype: BaseCalc|None
        """
        calcEntry = self.calcSystems.get(calcName, None)
        if calcEntry is None:
            return None
        if isinstance(calcEntry, BaseCalc):
            return calcEntry
        else:
            return calcEntry[0]

    def runCalc(self, calcName, isMode=False):
        """
        :param str|unicode calcName:
        :param bool isMode:
        """
        calcEntry = self.calcSystems.get(calcName, None)
        if calcEntry is None:
            return
        calc = calcEntry if isinstance(calcEntry, BaseCalc) else calcEntry[0]
        """ :type: BaseCalc """
        calcNext = (
            []
            if isinstance(calcEntry, BaseCalc) else (
                [calcEntry[1]]
                if isinstance(calcEntry[1], (str, unicode)) else
                calcEntry[1]
            )
        )
        """ :type: list[str|unicode] """
        if isMode:
            if calc is not None:
                calc.updateMode(self)
        else:
            if calc is not None:
                calc.update(self)
            for curCalcNext in calcNext:
                self.runCalc(curCalcNext, isMode)

    def getWidget(self, k):
        """
        :param str|unicode k:
        :rtype: Widget
        """
        assert self.widgets is not None
        i = self.dctWidgets[k]
        group, index = self.widgets[i]
        return group[index]

    def getDataAndWidget(self, k):
        """
        :param str|unicode k:
        :rtype: (int|long|float|Decimal|bool, Widget)
        """
        assert self.widgets is not None
        assert self.data is not None
        i = self.dctWidgets[k]
        group, index = self.widgets[i]
        return self.data[i], group[index]

    def getEntry(self, k):
        """
        :param str|unicode k:
        :rtype: BaseUserDataView
        """
        i = self.dctWidgets[k]
        entry, tgt = self.lstDataEntries[i]
        return entry

    def setData(self, k, v, updateDisp=DC_DATA_ONLY):
        err0 = "user entry widget must exist when updateDisp includes DC_VERIFY"
        err1 = "user entry widget must exist when updateDisp includes DC_SYNCH"
        i = self.dctWidgets[k]
        entry = self.lstDataEntries[i][0] if i < len(self.lstDataEntries) else None
        upd = None
        if updateDisp & DC_VERIFY:
            assert entry is not None, err0
            if isinstance(entry, NumberUserDataView):
                if not entry.varRange.satisfies(v, self):
                    raise ValueError(entry.varRange.userInfo(entry.shortTitle, entry.conv, self))
        if updateDisp & DC_RECALC:
            upd = self.dctCalcSystemUpd.get(k, None)
        self.data[i] = v
        if updateDisp & DC_SYNCH:
            assert entry is not None, err1
            if entry.upstream is not None:
                assert isinstance(entry.upstream, BaseUpstream)
                if entry.upstream.push(self, v):
                    self.onProgChg()
            group, index = self.widgets[i]
            entry.updateWidget(v, group[index])
        if upd is not None:
            calcName, isMode = upd
            self.runCalc(calcName, isMode)

    def getData(self, k):
        i = self.dctWidgets[k]
        return self.data[i]

    def getDataDefault(self, k, v=None):
        try:
            c = self.dctWidgets[k]
            v = self.data[c]
        except KeyError:
            pass
        return v

    def programIt(self, obj=None):
        self.userLog.log("Sequencing ...")
        seq = self.getWriteRegSequence()
        lstStrSeq = map(lambda x: self.fmtReg(x), seq)
        max_line_size = 80
        line_size = 4
        prevBreak = 0
        strSeq = ""
        for c, s in enumerate(lstStrSeq):
            if line_size + len(s) >= max_line_size:
                strSeq += ", ".join(lstStrSeq[prevBreak:c]) + "\n    "
                line_size = 4 + len(s)
                prevBreak = c
            else:
                line_size += len(s) + 2
        strSeq += ", ".join(lstStrSeq[prevBreak:])
        self.userLog.log("Sequence order: \n    " + strSeq)
        for regId in seq:
            self.outputReg(regId)
        super(BaseProgWithRegs, self).programIt(obj)

    def onProgChg(self):
        super(BaseProgWithRegs, self).onProgChg()

    def onhide(self, app):
        remove_widget = self.layout.remove_widget
        if self.showBinaryOnMain:
            remove_widget(self.binLayout)
        remove_widget(self.tabPanel)
        super(BaseProgWithRegs, self).onhide(app)

    @staticmethod
    def fmtReg(regId, short=True):
        """
        :param int|long|(int|long, int|long) regId:
        :param bool short:
        :rtype: str|unicode
        """
        if isinstance(regId, (int, long)):
            return "R%u" % regId
        elif short:
            return "R%u ws %u" % (regId[0], regId[1])
        else:
            return "R%u word select %u" % (regId[0], regId[1])

    def getWriteRegSequence(self):
        """
        :rtype: list[int|long|(int|long,int|long)]
        """
        raise NotImplementedError("Not Implemented")

    def outputReg(self, regId):
        logStr = self.fmtReg(regId, False)
        reg = self.getReg(regId)
        userLog = self.userLog
        if self.cntrRstReg == regId and self.autoCntrRstIndex is not None and self.data[self.autoCntrRstIndex]:
            i = self.cntrRstIndex
            userLog.log("Counter Reset: 1")
            upstream = self.lstDataEntries[i][0].upstream
            self.data[i] = 1
            upstream.push(self, self.data[i])
            userLog.log(self.regOutFmt % reg[:] + logStr)
            reg.spiWriteData(self.spiBus)
            userLog.log("Counter Reset: 0")
            self.data[i] = 0
            upstream.push(self, self.data[i])
            userLog.log(self.regOutFmt % reg[:] + logStr)
            reg.spiWriteData(self.spiBus)
        else:
            userLog.log(self.regOutFmt % reg[:] + logStr)
            reg.spiWriteData(self.spiBus)
        return reg
