from ..LibBaseProgPart import *

log = None

@initFunc(__name__.replace("ProgramIt.", ""))
def init(progLog):
    # noinspection PyGlobalUndefined
    global log
    log = progLog

class BaseProg_P_DSA(BaseProgPart):
    XM_partName = "Unknown"
    mfgPartName = "Unknown"
    mfgName = "Unknown"
    interface = "parallel"
    ioType = IO_PIN
    catId = CAT_DSA
    lsbVal = .5
    nPins = 6
    invBits = False
    @staticmethod
    def binTitleFmt(pinIndex):
        return "V%u" % (pinIndex+1)
    layout = None
    attnPanel = None
    binLayout = None
    binLbl = None
    top_custom_name = None
    @classmethod
    def initCls(cls, numBuses):
        binTitleFunc = cls.binTitleFmt
        if isinstance(cls.binTitleFmt, (str, unicode)):
            # use __mod__ for string format
            binTitleFunc = cls.binTitleFmt.__mod__
        cls.binTitles = map(binTitleFunc, xrange(cls.nPins))
        cls.wholeMask = (1 << cls.nPins) - 1
        return super(BaseProg_P_DSA, cls).initCls(numBuses)
    def __init__(self, fl=None, ver=0):
        super(BaseProg_P_DSA, self).__init__(fl, ver)
        self.attnWord = self.wholeMask if self.invBits else 0
        if fl is not None:
            numRead = (self.nPins + 7) // 8
            self.attnWord = GetLongStrBytes(fl.read(numRead))
        self.attn = [0] * self.nPins
        for c in xrange(self.nPins):
            self.attn[c] = (self.attnWord >> c) & 0x01
        self.oldBitsDisp = list(self.attn)
        self.cnvBitStr = invBitCh if self.invBits else "%u".__mod__
        self.buildFirst()
    def serialize(self, fl, ver):
        super(BaseProg_P_DSA, self).serialize(fl, ver)
        fl.write(AlignStrBytesLong(self.attnWord, (self.nPins + 7) // 8))
        fl.flush()
    def colorizeBinary(self):
        for cur, old, lbl in zip(self.attn,self.oldBitsDisp,self.binLbl):
            lbl.color = cmpRGB[int(cur!=old)]
    def onhide(self, app):
        self.layout.remove_widget(self.attnPanel.layout)
        self.layout.remove_widget(self.binLayout)
        self.top_custom_name.remove(self.layout)
        super(BaseProg_P_DSA, self).onhide(app)
    def build(self, app):
        layout = super(BaseProg_P_DSA, self).build(app)
        if self.binLayout is None: self.buildFirst()
        sizeY = .61
        posY = .93
        posY1 = posY - sizeY - .03
        self.top_custom_name.add(layout)
        layout.add_widget(self.attnPanel.layout)
        layout.add_widget(self.binLayout)
        self.layout = layout
        self.progBtn.pos_hint = {"top": posY1, 'x': .1}

    def on_chg_custom_name(self, v):
        if self.top_custom_name is not None:
            self.top_custom_name.txt_in.text = v

    def buildFirst(self):
        sizeX = .4
        padX = .015
        #posX = .5 - sizeX/2
        posX = .1
        sizeY = .61
        posY = .93
        posY1 = posY - sizeY - .03
        #sizeX = 1.0 - 2*posX
        self.top_custom_name = TopCustomNameStorage(self)
        self.attnPanel = mkChkBoxList(
            {"size_hint":(sizeX-padX,sizeY),"bdr":8,
             "pos_hint":{"top":posY, 'x':posX+padX/2}}, self.nPins, self.lsbVal,
            self.attn,"Attenuation: ","dB", "Attenuation Setting")
        self.attnPanel.onChgAny = lambda theObj: self.onProgChg()
        self.binLayout = BorderedGridLayout(
            cols=self.nPins+1,rows=2,pos_hint={'x':.3, "top":posY1},
            size_hint=(.595, .1),bdr=8)
        self.binLayout.add_widget(Label(text=""))
        self.binLbl = [
            Label(text=self.cnvBitStr(bit)) for bit in self.attn]
        for c in xrange(self.nPins):
            self.binLayout.add_widget(Label(text=self.binTitles[c]))
        self.binLayout.add_widget(Label(text="binary"))
        for bitLbl in self.binLbl: self.binLayout.add_widget(bitLbl)
    def getPinNames(self):
        return self.binTitles
    def programIt(self, obj=None):
        for pin, bit in zip(self.pins, bitMapInt(self.attnWord, self.nPins)):
            pin.value = bit
        self.oldBitsDisp = list(self.attn)
        self.colorizeBinary()
        super(BaseProg_P_DSA, self).programIt(obj)
    def updAttnWord(self):
        #map(lambda x, y: long(x << y), self.attn, xrange(self.nBits))
        #  is equivalent to the list comprehension given lst->reduce(fn, lst)
        tmp = reduce(
            lambda a, b: a|b,
            [self.attn[x] << x for x in xrange(self.nPins)])
        if self.invBits: tmp ^= self.wholeMask
        self.attnWord = tmp
    def onProgChg(self):
        self.updAttnWord()
        for lbl, bit in zip(self.binLbl, self.attn):
            lbl.text = self.cnvBitStr(bit)
        self.colorizeBinary()
        super(BaseProg_P_DSA, self).onProgChg()