from ..LibBaseProgPart import *

log = None

@initFunc(__name__.replace("ProgramIt.", ""))
def init(progLog):
    # noinspection PyGlobalUndefined
    global log
    log = progLog
class BaseProg_S_DPS(BaseProgPart):
    mfgName = "Unknown"
    interface = "3-wire"
    WordSz = 7
    catId = CAT_DPS
    XM_partName = "Unknown"
    mfgPartName = "Unknown"
    PhRes = 4
    binTitleFmt = "D%u"
    layout = None
    phasePanel = None
    binLayout = None
    binLbl = None
    top_custom_name = None

    @classmethod
    def initCls(cls, numBuses):
        fmt = cls.binTitleFmt.__mod__ if isinstance(cls.binTitleFmt, (str, unicode)) else cls.binTitleFmt
        cls.binTitles = [fmt(x) for x in xrange(cls.WordSz)]

    def __init__(self, fl=None, ver=0):
        super(BaseProg_S_DPS, self).__init__(fl, ver)
        self.data = bitMapInt(0, self.WordSz)
        if fl is not None:
            self.data.num = GetLongStrBytes(fl.read((self.WordSz + 7) // 8))
        self.oldBitsDisp = list(self.data)
        self.buildFirst()
        self.spiBus.cshigh = True

    def serialize(self, fl, ver):
        super(BaseProg_S_DPS, self).serialize(fl, ver)
        fl.write(AlignStrBytesLong(self.data.num, (self.WordSz + 7) // 8))

    def colorizeBinary(self):
        for c in xrange(len(self.binLbl)):
            self.binLbl[c].color = cmpRGB[
                int(self.data[c] != self.oldBitsDisp[c])]

    def onhide(self, app):
        self.layout.remove_widget(self.phasePanel.layout)
        self.layout.remove_widget(self.binLayout)
        self.top_custom_name.remove(self.layout)
        super(BaseProg_S_DPS, self).onhide(app)

    def build(self, app):
        layout = super(BaseProg_S_DPS, self).build(app)
        if not hasattr(self, "binLayout"): self.buildFirst()
        sizeY = .61
        posY = .93
        posY1 = posY - sizeY - .03
        self.top_custom_name.add(layout)
        layout.add_widget(self.phasePanel.layout)
        layout.add_widget(self.binLayout)
        self.progBtn.pos_hint = {"top": posY1, 'x': .1}
        self.layout = layout

    def on_chg_custom_name(self, v):
        if self.top_custom_name is not None:
            self.top_custom_name.txt_in.text = v

    def buildFirst(self):
        sizeX = .4
        padX = .015
        # posX = .5 - sizeX/2
        posX = .1
        sizeY = .61
        posY = .93
        posY1 = posY - sizeY - .03
        lastI = self.WordSz - 1
        self.top_custom_name = TopCustomNameStorage(self)
        self.phasePanel = CheckBitsPanel(
            "Phase Shift Setting", u"Phase Shift: %.15g\xb0", u"+%.15g\xb0",
            360.0 / 2 ** self.PhRes, self.PhRes,
            self.data[self.WordSz - self.PhRes:self.WordSz],
            lambda theObj: self.onProgChg(), size_hint=(sizeX - padX, sizeY), bdr=8,
            pos_hint={"top": posY, 'x': posX + padX / 2})
        self.binLayout = BorderedGridLayout(
            cols=self.WordSz + 1, rows=2, pos_hint={'x': .3, "top": posY1},
            size_hint=(.595, .1), bdr=8)
        self.binLayout.add_widget(Label(text=""))
        for c in xrange(self.WordSz):
            self.binLayout.add_widget(Label(text=self.binTitles[self.WordSz - (1 + c)]))
        self.binLayout.add_widget(Label(text="binary"))
        self.binLbl = [None] * self.WordSz
        for c in xrange(self.WordSz):
            self.binLbl[c] = Label(text=bitCh(self.data[lastI - c]))
            self.binLayout.add_widget(self.binLbl[c])
        self.binLbl[::-1] = self.binLbl

    def programIt(self, obj=None):
        self.oldBitsDisp = list(self.data)
        self.colorizeBinary()
        self.spiBus.cshigh = True
        spiPulseData(self.spiBus, [self.data.num])
        super(BaseProg_S_DPS, self).programIt(obj)

    def onProgChg(self):
        self.data[self.WordSz - self.PhRes:self.WordSz] = self.phasePanel.word.num
        for c in xrange(len(self.data)):
            self.binLbl[c].text = bitCh(self.data[c])
        self.colorizeBinary()
        super(BaseProg_S_DPS, self).onProgChg()