from ..LibBaseProgPart import *

log = None

@initFunc(__name__.replace("ProgramIt.", ""))
def init(progLog):
    # noinspection PyGlobalUndefined
    global log
    log = progLog

invBitCh = lambda bit: bitCh(int(not bit))
class BaseProg_S_DSA(BaseProgPart):
    XM_partName = "Unknown"
    mfgPartName = "Unknown"
    mfgName = "Unknown"
    interface = "3-wire"
    ioType = IO_SPI
    catId = CAT_DSA
    lsbVal = .5
    nBits = 4
    invBits = False
    binTitleFmt = "D%u"
    lsbFirst = False
    allowChain = True
    wordSz = None
    layout = None
    attnPanel = None
    binLayout = None
    binLbl = None
    top_custom_name = None
    @classmethod
    def initCls(cls, numBuses):
        #index Step
        iS = 1 if cls.lsbFirst else -1
        binTitleFunc = cls.binTitleFmt
        if isinstance(cls.binTitleFmt, (str, unicode)):
            # use __mod__ for string format
            binTitleFunc = cls.binTitleFmt.__mod__
        cls.binTitles = map(binTitleFunc, xrange(cls.nBits))[::iS]
        cls.wholeMask = (1 << cls.nBits) - 1
        if cls.wordSz is None: cls.wordSz = cls.nBits
        return super(BaseProg_S_DSA, cls).initCls(numBuses)
    def __init__(self, fl=None, ver=0):
        super(BaseProg_S_DSA, self).__init__(fl, ver)
        self.attnWord = self.wholeMask if self.invBits else 0
        if fl is not None:
            self.load(fl)
        self.attn = [0] * self.nBits
        tmp = self.attnWord
        if self.invBits: tmp ^= self.wholeMask
        for c in xrange(self.nBits):
            self.attn[c] = tmp & 0x01
            tmp >>= 1
        self.oldBitsDisp = list(self.attn)
        self.cnvBitStr = invBitCh if self.invBits else "%u".__mod__
        self.buildFirst()
    def load(self, fl):
        self.attnWord = GetLongStrBytes(fl.read((self.nBits + 7) // 8))
    def serialize(self, fl, ver):
        super(BaseProg_S_DSA, self).serialize(fl, ver)
        fl.write(AlignStrBytesLong(self.attnWord, (self.nBits + 7) // 8))
    def colorizeBinary(self):
        #index Step
        iS = 1 if self.lsbFirst else -1
        for cur, old, lbl in zip(self.attn,self.oldBitsDisp,self.binLbl[::iS]):
            lbl.color = cmpRGB[int(cur!=old)]
    def onhide(self, app):
        self.layout.remove_widget(self.attnPanel.layout)
        self.layout.remove_widget(self.binLayout)
        self.top_custom_name.remove(self.layout)
        super(BaseProg_S_DSA, self).onhide(app)
    def build(self, app):
        layout = super(BaseProg_S_DSA, self).build(app)
        if not hasattr(self, "binLayout"): self.buildFirst()
        sizeY = .61
        posY = .93
        posY1 = posY - sizeY - .03
        self.top_custom_name.add(layout)
        layout.add_widget(self.attnPanel.layout)
        layout.add_widget(self.binLayout)
        self.layout = layout
        self.progBtn.pos_hint = {"top": posY1, 'x': .1}
    def on_chg_custom_name(self, v):
        if self.top_custom_name is not None:
            self.top_custom_name.txt_in.text = v
    def buildFirst(self):
        #index Step
        iS = 1 if self.lsbFirst else -1
        sizeX = .4
        padX = .015
        #posX = .5 - sizeX/2
        posX = .1
        sizeY = .61
        posY = .93
        posY1 = posY - sizeY - .03
        #sizeX = 1.0 - 2*posX
        self.top_custom_name = TopCustomNameStorage(self)
        self.attnPanel = mkChkBoxList(
            {"size_hint":(sizeX-padX,sizeY),"bdr":8,
             "pos_hint":{"top":posY, 'x':posX+padX/2}}, self.nBits, self.lsbVal,
            self.attn,"Attenuation: ","dB", "Attenuation Setting")
        self.attnPanel.onChgAny = lambda theObj: self.onProgChg()
        self.binLayout = BorderedGridLayout(
            cols=self.nBits+1,rows=2,pos_hint={'x':.3, "top":posY1},
            size_hint=(.595, .1),bdr=8)
        self.binLayout.add_widget(Label(text=""))
        self.binLbl = [
            Label(text=self.cnvBitStr(bit)) for bit in self.attn[::iS]]
        for c in xrange(self.nBits):
            self.binLayout.add_widget(Label(text=self.binTitles[c]))
        self.binLayout.add_widget(Label(text="binary"))
        for bitLbl in self.binLbl: self.binLayout.add_widget(bitLbl)
    def programIt(self, obj=None):
        if not isinstance(self.spiBus, ChainSpiBus):
            bus, cs = getSpiBusInfo(self.spiBus)
            if (bus, cs) in self.app.SpiChains and self in self.app.SpiChains[(bus, cs)]:
                self.app.chainProgramIt(bus, cs)
                return
        self.oldBitsDisp = list(self.attn)
        self.updAttnWord()
        self.colorizeBinary()
        spiWriteNum(self.attnWord, self.wordSz, self.spiBus)
        super(BaseProg_S_DSA, self).programIt(obj)
    def updAttnWord(self):
        #map(lambda x, y: long(x << y), self.attn, xrange(self.nBits))
        #  is equivalent to the list comprehension given to reduce(fn, lst)
        if self.lsbFirst:
            tmp = reduce(
                lambda x, y: x|y,
                [self.attn[x0] << ((7-x0)%8) for x0 in xrange(self.nBits)])
        else:
            tmp = reduce(
                lambda x, y: x|y,
                [self.attn[x0] << x0 for x0 in xrange(self.nBits)])
        if self.invBits: tmp ^= self.wholeMask
        self.attnWord = tmp
    def onProgChg(self):
        #index Step
        iS = 1 if self.lsbFirst else -1
        self.updAttnWord()
        for lbl, bit in zip(self.binLbl[::iS], self.attn):
            lbl.text = self.cnvBitStr(bit)
        self.colorizeBinary()
        super(BaseProg_S_DSA, self).onProgChg()
    def executeCMD(self, s):
        if s.lower() == "help":
            return "S" + "\n".join([
                "list of commands",
                "  write  -  writes the data to the device",
                "  setAttn <float>  -  sets the attenuation value in dB",
                "  setBits <unsigned integer>  -  sets the bit values with a number",
                "  minAttn  -  sets the attenuation to the minimum",
                "  maxAttn  -  sets the attenuation to the maximum",
                "  help  -  shows this help"
            ])
        elif s.lower() == "write":
            self.programIt(None)
            return "S" + "wrote data"
        elif s.lower().startswith("setbits "):
            numStr = s[8:].strip()
            base = 10
            if numStr.startswith("0b"):
                base = 2
                numStr = numStr[2:]
            elif numStr.startswith("0x"):
                base = 16
                numStr = numStr[2:]
            elif numStr.startswith("0o"):
                base = 8
                numStr = numStr[2:]
            tmp = int(numStr, base)
            for c, (box, lbl) in enumerate(self.attnPanel.widgets[:-1]):
                box.active = tmp & 1
                tmp >>= 1
            return "S" + "set attenuation bits to %s" % "".join(map(str, self.attn))
        elif s.lower() in ["minattn", "maxattn"]:
            x = s.lower() == "maxattn"
            for c, (box, lbl) in enumerate(self.attnPanel.widgets[:-1]):
                box.active = x
            return "S" + "set attenuation bits to %s" % "".join(map(str, self.attn))
        else:
            return "E" + "unrecognized command, type help to see list of commands"