from BaseADF_PLL import *

"""
Implements:
    ADF4156
    ADF4157
    ADF4158
    ADF4159
    ADF4169
"""

log = None

@initFunc(__name__.replace("ProgramIt.", ""))
def init(progLog):
    global log
    log = progLog

#other functions may be added via the get_enterSetLst calls
#  after the definition of this class
class ProgADF41XX(PLL_MulDivPFD, PLL_IntFrac, BaseProgADF_PLL):
    XM_partName = "Unknown"
    mfgPartName = "ADF41**"
    catId = CAT_PLL
    mfgName = "Analog Devices"
    interface = "3-wire"
    DevOnly = False
    lstMuxoutOpts = map(isReserv,[
        "3-state Output","DVdd","GND","R-Divider Output","N-Divider Output",
        "Reserved", "Digital Lock Detect", "Serial Data Output", "Reserved",
        "Reserved", "Clk Divider Output", "Reserved", "Fast-lock switch",
        "R-Divider/2", "N-Divider/2", "Ramp Status"])
    lstClkDivOpts = map(isReserv,[
        "Clock Divider Off", "Fast Lock Divider", "Reserved", "Ramp Divider"])
    lstChrgPmpVals = [.3125*(x+1) for x in xrange(16)]
    lstChrgPmpOpts = map(lambda x: isReserv(str(x)+" mA"), lstChrgPmpVals)
    lstNegBldVals = [3.87, 11.6, 27, 58, 120, 243, 491, 986]
    lstNegBldOpts = map(lambda x: isReserv(str(x)+u" \u03BCA"), lstNegBldVals)
    lstPreSclOpts = map(isReserv,["4/5", "8/9"])
    lstNSelOpts = map(isReserv,[
        u"Load on \u03A3-\u0394CLK", "Load Delayed 4 Cycle"])
    #LocK DeTeCt Precision
    #Correction Lock Detect Precision
    lstLDPOpts = map(isReserv,["24 PFD Cycles","40 PFD Cycles"])
    lstPhDtcPolOpts = map(isReserv,["Negative", "Positive"])
    lstLESelOpts = map(isReserv,["LE From Pin", "LE Synch With RefB"])
    lstRmpStatOpts = {
        0:"Normal Operation", 2:"Readback To Muxout",
        3:"Ramp Complete To Muxout", 16:"Charge Pump Up",
        17:"Charge Pump Down"}
    lstSDModMdOpts = {
        0:"Normal Operation", 14: "Disabled When Frac = 0"}
    lstRmpMdOpts = map(isReserv,[
        "Continuous Sawtooth", "Continuous Triangular",
        "Single Sawtooth Burst", "Single Ramp Burst"])
    lstTXInterOpts = map(isReserv,[
        "Interrupt Off", "Load Channel - Continue Sweep",
        "Load Channel - Stop Sweep"])
    lstTXRmpClkOpts = map(isReserv,["CLK Div", "TX Data"])
    lstR0BitNumOpts = map(lambda x: isReserv(str(x)), range(32))
    lstDelClkOpts = map(isReserv,["PFD CLK", "PFD CLK * CLK1"])
    lstExtDivOpts = [isReservStr(u"\u00F7%u", 2 ** x) for x in xrange(4)]
    extraBindFuncs = {
        "mkSmRmpChk": ("addSmChkLine","gridMiscRamp"),
        "mkSmRmpLst": ("addSmallLstLine","gridMiscRamp"),
        #"addDelayLn": ("addSmallLine","gridDelay"),
        #"mkDelayLst": ("addSmallLstLine","gridDelay"),
        #"mkDelayChk": ("addSmChkLine","gridDelay"),
        "mkRmp1Ln": ("addSmallLine","gridRmp1"),
        "mkRmp2Ln": ("addSmallLine","gridRmp2"),
        "mkDelLn": ("addSmallLine","gridDel"),
        "mkDelChk": ("addSmChkLine","gridDel"),
        "mkDelLst": ("addSmallLstLine","gridDel"),
        "mkTXChk": ("addSmChkLine","gridTXData"),
        "mkTXLst":("addSmallLstLine","gridTXData"),
        "mkVCO_Chk": ("addChkLine", "gridVCO"),
        "mkCfgMdChk": ("addSmChkLine", "gridCfgMd"),
        "addSmOtherLn": ("addSmallLine", "gridOther")}
    lstBindFuncs = dict(BaseProgADF_PLL.lstBindFuncs, **extraBindFuncs)
    lstChkBigger = BaseProgADF_PLL.lstChkBigger + [10]
    #chk funcs 7, 8, 9, 10, 11
    lstChkFuncs = BaseProgADF_PLL.lstChkFuncs + [
        "mkSmRmpChk", "mkDelChk", "mkTXChk", "mkVCO_Chk", "mkCfgMdChk"]
    #line funcs 8, 9, 10, 11
    lstLineFuncs = BaseProgADF_PLL.lstLineFuncs + [
        "mkRmp1Ln", "mkRmp2Ln", "mkDelLn", "addSmOtherLn"]
    #lst func 4, 5, 6
    lstLstFuncs = BaseProgADF_PLL.lstLstFuncs + [
        "mkSmRmpLst", "mkDelLst", "mkTXLst"]
    useFuncs = {
        "R_Freq": (2,"MHz",1), "PFD_Freq": (2,"MHz",0),
        "UsrPFD_Freq": (2,"MHz",1), "NVal":(0,"",1),
        "RF_Out": (0,"MHz",1), "OutFreq": (0,"MHz",0),
        "IntVal":(0,"",1), "FracVal":(0,"",1),
        "Muxout":(1,), "rampOn":(7,"updRampMisc"),
        "PreScl":(0,), "R_Count": (2,"",1),
        #"ExtDiv":(0,),
        "selExtDiv":(11,"updAutoNOrder"),
        "phaseAdjust":(2,"updReg1Dat"), "Phase":(4,1), "Clk1Div":(4,1),
        "bRefDbl":(0,"","updPFD_Freq"), "bRefDiv":(0,"","updPFD_Freq"),
        "ChrgPmp":(2,), "r2CSR":(2,"updReg2Dat"),
        #"r3CntRst":(2,"updReg3Dat"),
        "bAutoCntrRst": (2,None),
        "r3ChrgPmp3St":(4,"updReg3Dat"),
        "r3PwrDown":(2,"updReg3Dat"), "PhDtcPol":(1,1),
        "LDP":(1,1), "bFSK":(7,"updReg3Dat"), "bPSK":(7,"updReg3Dat"),
        "RmpMd":(4,1), "r3NotSDRst": (2,"updReg3Dat"), "NSel":(1,1),
        "r3LOL":(2,"updReg3Dat"), "constNegBld":(4,"updReg3Dat"),
        "NegBld":(2,1), "ClkDiv":(4,1),
        "Clk2Div1":(8,1), "FDevStep1":(8,0),
        "Clk2Div2":(9,1), "FDevStep2":(9,0),
        "RmpStat":(4,1), "LESel":(4,1),
        "DevVal1":(8,1), "TotRamp1": (8,0),
        "DevVal2":(9,1), "TotRamp2": (9,0),
        "DevOff1":(8,1), "TStep1": (8,0),
        "DevOff2":(9,1), "TStep2": (9,0),
        "dualRamp":(7,"updRampMisc"), "bRmpFSK":(7,"updRampMisc"),
        "TXInter":(6,1),
        "TXRmpClk":(6,1),
        "bTXInv":(9,"updReg5Dat"),
        "Step1":(8,1), "TRamp1": (8,0),
        "Step2":(9,1), "TRamp2": (9,0),
        "DelWord":(10,1),
        "bDelStart":(8,"updReg7Dat"),
        "DelClk":(5,1),
        "bRmpDel":(8,"updReg7Dat"),
        "bFastRmp":(7,"updRampMisc"),
        "bTXTrig":(9,"updReg7Dat"),
        "bSFT":(7,"updReg7Dat"),
        "bTriDel":(8,"updReg7Dat"),
        "bTXTrigDel":(9,"updReg7Dat"),
        "SDModMd":(1,1),
        "calcNegBld": (4,"updNegBld"),
        "WriteData": (11,1)}
    titles = {
        "RF_Out": ("Desired VCO Freq","Desired VCO Output Frequency"),
        "OutFreq": ("Actual VCO Freq",),
        "DelWord": ("Delay Start Word",),
        "R_Freq": ("Reference Freq",),
        "R_Count": ("Ref Counter (R)",),
        "Phase": ("Phase",),
        "PreScl": ("Internal Prescaler (P/P+1)",),
        #"ExtDiv": ("VCO Divider",),
        "selExtDiv": ("VCO Divide by 2",),
        "Clk1Div": ("CLK1 Divider",),
        "Clk2Div1": ("Ramp 1 CLK2",), "FDevStep1": ("Fdev per step (kHz)",),
        "Clk2Div2": ("Ramp 2 CLK2",), "FDevStep2": ("Fdev per step (kHz)",),
        "DevVal1": ("Ramp 1 DEV",), "TotRamp1": ("Total Ramp (kHz)",),
        "DevVal2": ("Ramp 2 DEV",), "TotRamp2": ("Total Ramp (kHz)",),
        "DevOff1": ("Ramp 1 DEVoff",), "TStep1": (u"Time Per Step (\u03BCs)",),
        "DevOff2": ("Ramp 2 DEVoff",), "TStep2": (u"Time Per Step (\u03BCs)",),
        "Step1": ("Ramp 1 Steps",), "TRamp1": (u"Time Per Ramp (\u03BCs)",),
        "Step2": ("Ramp 2 Steps",), "TRamp2": (u"Time Per Ramp (\u03BCs)",),
        "PFD_Freq": ("Actual PFD Freq",), "IntVal": ("Integer Part of N",),
        "FracVal": ("Fractional Part of N",), "NVal": ("N Value",),
        "UsrPFD_Freq": ("Desired PFD Freq",),
        "Muxout": ("Muxout",),
        "rampOn": ("Ramp On",),
        "phaseAdjust": ("Phase Adjust","Phase Adjust Enable"),
        "bRefDbl": ("Ref x2 Doubler",),
        "bRefDiv": ("Ref /2 Divider",),
        "ChrgPmp": ("CP Current Setting","Charge Pump Current Setting"),
        "r2CSR": ("CSR Enable","Cycle Slip Reduction Enable"),
        #"r3CntRst": ("Counter Reset",),
        "bAutoCntrRst": ("Reset Counters on Write",),
        "r3ChrgPmp3St": ("CP 3-State",),
        "r3PwrDown": ("Power Down",),
        "PhDtcPol": ("PD Polarity","Phase Detector Polarity"),
        "LDP": ("LDP","Lock Detect Precision"),
        "bFSK": ("FSK Enable",),
        "bPSK": ("PSK Enable",),
        "RmpMd": ("Ramp Mode",),
        "r3NotSDRst": (u"\u03A3-\u0394 Reset Disable",),
        "NSel": ("N SEL",),
        "r3LOL": ("LOL",),
        "constNegBld": ("Const Neg Bleed","Constant Negative Bleed Enable"),
        "NegBld": ("Neg Bleed Current",),
        "ClkDiv": ("Clk Div Mode",),
        "RmpStat": ("Ramp Status",),
        "LESel": ("LE SEL",),
        "dualRamp": ("Dual Ramp",),
        "bRmpFSK": ("FSK Ramp Enable",),
        "TXInter": ("Interrupt",),
        "TXRmpClk": ("TX Data Ramp CLK",),
        "bTXInv": ("TX Data Invert",),
        "bDelStart": ("Delay Start","Delay Start Enable"),
        "DelClk": ("Delay CLK Select",),
        "bRmpDel": ("Delay Between Ramps",),
        "bFastRmp": ("Fast Ramp",),
        "bTXTrig": ("TX Data Trigger",),
        "bSFT": ("Single Full Triangle",),
        "bTriDel": ("Triangle Delay",),
        "bTXTrigDel": ("TX Data Trigger Delay",),
        "SDModMd": (u"\u03A3-\u0394 Mode",u"\u03A3-\u0394 Modulator Mode"),
        "calcNegBld": ("Autoset Neg Bleed",),
        "WriteData": ("Enter Register Hex",)}
    regDatas = {
        "numIntVal":[(0,15,27)], "numFracVal":[(1,15,28),(0,3,15)],
        "selMuxout":[(0,27,31)], "rampOn":[(0,31)], "selPreScl":[(2,22)],
        "numR_Count":[(2,15,20)], "phaseAdjust":[(1,28)], "numPhase":[(1,3,15)],
        "numClk1Div":[(2,3,15)], "bRefDbl":[(2,20)], "bRefDiv":[(2,21)],
        "selChrgPmp":[(2,24,28)], "r2CSR":[(2,28)],"r3CntRst":[(3,3)],
        "r3ChrgPmp3St":[(3,4)], "r3PwrDown":[(3,5)], "selPhDtcPol":[(3,6)],
        "selLDP":[(3,7)], "bFSK":[(3,8)], "bPSK":[(3,9)],
        "selRmpMd":[(3,10,12)], "r3NotSDRst":[(3,14)], "selNSel":[(3,15)],
        "r3LOL":[(3,16)], "constNegBld":[(3,21)], "selNegBld":[(3,22,25)],
        "selClkDiv":[(4,19,21)], "numClk2Div1":[((4,0),7,19)],
        "numClk2Div2":[((4,1),7,19)], "selRmpStat":[(4,21,26)],
        "selLESel":[(4,31)], "numDevVal1":[((5,0),3,19)],
        "numDevVal2":[((5,1),3,19)], "numDevOff1":[((5,0),19,23)],
        "numDevOff2":[((5,1),19,23)], "dualRamp":[(5,24)], "bRmpFSK":[(5,25)],
        "selTXInter":[(5,26,28)], "selTXRmpClk":[(5,29)], "bTXInv":[(5,30)],
        "numStep1":[((6,0),3,23)], "numStep2":[((6,1),3,23)],
        "numDelWord":[(7,3,15)], "bDelStart":[(7,15)], "selDelClk":[(7,16)],
        "bRmpDel":[(7,17)], "bFastRmp":[(7,19)], "bTXTrig":[(7,20)],
        "bSFT":[(7,21)], "bTriDel":[(7,22)], "bTXTrigDel":[(7, 23)],
        "selSDModMd":[(4,26,31)]}
    regDataFuncs = {
        "numPhase": signRegValFunc(12), "numR_Count": nonZeroRegValFunc(5),
        "numDevVal1": signRegValFunc(16), "numDevVal2": signRegValFunc(16)}
    lstTxtDatas = [
        ("RF_Out", "num", "%.15g"),
        ("PreScl", "lst"),
        #("ExtDiv", "lst"),
        ("selExtDiv", "chkExtDiv"),
        ("NVal", "num", "%.15g"),
        ("IntVal", "num"),
        ("FracVal", "num"),
        ("OutFreq", "num", "%.15g"),
        ("R_Freq", "num", "%.15g"),
        ("bRefDbl", "chkRefMul2"),
        ("R_Count", "num"),
        ("bRefDiv", "chkRefDiv2"),
        ("UsrPFD_Freq", "num", "%.15g"),
        ("PFD_Freq", "num", "%.15g"),
        ("Muxout", "lst"),
        ("Phase","num", "%i"),
        ("Clk1Div","num"),
        ("NSel", "lst"),
        ("LDP", "lst"),
        ("PhDtcPol", "lst"),
        ("SDModMd", "lst"),
        ("phaseAdjust", "chkPhaseAdjust"),
        ("r3NotSDRst", "chkR3NotSDRst"),
        ("r3LOL", "chkR3LOL"),
        #("r3CntRst", "chkR3CntRst"),
        ("bAutoCntrRst", "chkAutoCntrRst"),
        ("r3PwrDown", "chkR3PwrDown"),
        ("ChrgPmp", "lst"),
        ("NegBld", "lst"),
        ("calcNegBld","chkNegBld"),
        ("r3ChrgPmp3St","chkR3ChrgPmp3St"),
        ("r2CSR", "chkR2CSR"),
        ("constNegBld", "chkConstNegBld"),
        ("ClkDiv", "lst"),
        ("LESel", "lst"),
        ("RmpMd", "lst"),
        ("RmpStat", "lst"),
        ("bFSK", "chkFSK"),
        ("bPSK", "chkPSK"),
        ("bRmpFSK", "chkRmpFSK"),
        ("rampOn", "chkRampOn"),
        ("bFastRmp", "chkFastRmp"),
        ("dualRamp", "chkDualRamp"),
        ("bSFT", "chkSFT"),
        ("Clk2Div1", "num"), ("FDevStep1", "num", "%.6g"),
        ("DevVal1", "num", "%i"), ("TotRamp1", "num", "%.6g"),
        ("DevOff1", "num"), ("TStep1", "num", "%.6g"),
        ("Step1", "num"), ("TRamp1", "num", "%.6g"),
        ("Clk2Div2", "num"), ("FDevStep2", "num", "%.6g"),
        ("DevVal2", "num", "%i"), ("TotRamp2", "num", "%.6g"),
        ("DevOff2", "num"), ("TStep2", "num", "%.6g"),
        ("Step2", "num"), ("TRamp2", "num", "%.6g"),
        ("DelWord", "num"),
        ("bDelStart", "chkDelStart"),
        ("DelClk", "lst"),
        ("bRmpDel", "chkRmpDel"),
        ("bTriDel", "chkTriDel"),
        ("TXRmpClk", "lst"),
        ("bTXInv", "chkTXInv"),
        ("bTXTrig", "chkTXTrig"),
        ("bTXTrigDel", "chkTXTrigDel"),
        ("TXInter", "lst"),
        ("WriteData", "num", "%08X")]
    lstEnterSets = [
        ("lst", "Muxout", "updReg0Dat"),
        ("float", "UsrPFD_Freq", (.1,260.), "updPFD_Freq"),
        #("lst", "ExtDiv", "updFracIntDat"),
        ("lst", "ClkDiv", "updReg4Dat"),
        ("lst", "LESel", "updReg4Dat"),
        ("lst", "RmpStat", "updReg4Dat"),
        ("lst", "RmpMd", "updRampMisc"),
        ("lst", "ChrgPmp", "updChrgPmp"),
        ("lst", "NegBld", "updReg3Dat"),
        ("lst", "NSel", "updReg3Dat"),
        ("lst", "LDP", "updReg3Dat"),
        ("lst", "PhDtcPol", "updReg3Dat"),
        ("lst", "TXInter", "updReg5Dat"),
        ("lst", "SDModMd", "updReg4Dat"),
        ("lst", "TXRmpClk", "updReg5Dat"),
        ("lst", "PreScl", "updReg2Dat"),
        ("lst", "DelClk", "updReg7Dat"),
        ("float", "RF_Out", (34.0, 13000.0), "updAutoNOrder"),#MHz
        ("float", "R_Freq", (10.0, 260.0), "updPFD_Freq"),#MHz
        ("int", "DelWord", (0,4095), "updReg7Dat"),
        ("int", "R_Count", (1,32), "updPFD_Freq"),
        ("int", "IntVal", (23,4095), "updAutoNOrder"),
        ("int", "FracVal", (0,2 **25-1), "updAutoNOrder"),
        ("float", "NVal", (23,4096-.5**25), "updAutoNOrder"),
        ("int", "Phase", (-2048, 2047), "updReg1Dat"),
        ("int", "Clk1Div", (0, 4095), "updReg2Dat"),
        ("int", "Clk2Div1", (0, 4095), "updR4Ramp1"),
        ("int", "Clk2Div2", (0, 4095), "updR4Ramp2"),
        ("int", "DevVal1", (-32768, 32767), "updR5Ramp1"),
        ("int", "DevVal2", (-32768, 32767), "updR5Ramp2"),
        ("int", "DevOff1", (0,9), "updR5Ramp1"),
        ("int", "DevOff2", (0,9), "updR5Ramp2"),
        ("int", "Step1", (0, 1048575), "updR6Ramp1"),
        ("int", "Step2", (0, 1048575), "updR6Ramp2"),
        ("hex", "WriteData", (0, 2**32-1), None)]
    defRegs = [
        0x003C0000, 0x00000001, 0x0701000A, 0x00C300C3,
        0x38180104, 0x38180144, 0x00000005, 0x00800005,
        0x00000006, 0x00800006, 0x00000007]
    defData = {
        "numOutFreq":6000.0, "numR_Freq":100.0, "numRF_Out":6000.0,
        "numPFD_Freq":50., "calcNegBld":True, "numWriteData":0,
        "numFDevStep1":0, "numTotRamp1":0, "numTStep1":0.04, "numTRamp1":0.04,
        "numFDevStep2":0, "numTotRamp2":0, "numTStep2":0.04, "numTRamp2":0.04,
        "selExtDiv":0, "bAutoCntrRst":True, "numModVal":2**25,
        "numNVal":120.0,"numUsrPFD_Freq":50.}
    defBoxStyle = dict(pos_hint={'x': 0.0, "top": 1.0}, size_hint_y=.08)
    defInnerTop = defBoxStyle["pos_hint"]["top"] - defBoxStyle["size_hint_y"]
    tabLayout = [
        ("Config Mode", "mainTab", (
            BorderedGridLayout, None, dict(cols=2, spacing=[10, 10], padding=[6]), [
                (Bordered, "layoutCfgMd", {}, [
                    (Label, None, dict(defBoxStyle, text="Configuration Mode")),
                    (GridLayout, "gridCfgMd", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 2, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})]),
                (Bordered, "layoutCfgInf", {}, [
                    (Label, None, dict(defBoxStyle, text="Configuration Info")),
                    (GridLayout, "gridCfgInf", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 1, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})])])),
        ("Main", "vcoTab", (
            BorderedGridLayout, "gridMain", dict(cols=2, spacing=[10, 10], padding=[6]), [
                (Bordered, "layoutVCO", {"size_hint_x": 1.2}, [
                    (Label, None, dict(defBoxStyle, text="N Counter (VCO)")),
                    (GridLayout, "gridVCO", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 3, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})]),
                (Bordered, "layoutRef", {}, [
                    (Label, None, dict(defBoxStyle, text="Reference (PFD)")),
                    (GridLayout, "gridRef", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 3, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})])])),
        ("Config", "cfgTab", (
            BorderedGridLayout, "gridCfg", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("Charge Pump", "chrgTab", (
            BorderedGridLayout, "gridChrg", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("Ramp Main", "miscRampTab", (
            BorderedGridLayout, "gridMiscRamp", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("Ramp 1", "ramp1Tab", (
            BorderedGridLayout, "gridRmp1", dict(cols=4, spacing=[10, 10], padding=[5]*4))),
        ("Ramp 2", "ramp2Tab", (
            BorderedGridLayout, "gridRmp2", dict(cols=4, spacing=[10, 10], padding=[5]*4))),
        ("TXdata/Delay", "delTxTab", (
            BorderedGridLayout, "gridDelTx", dict(cols=2, spacing=[10, 10], padding=[5]*4), [
                (Bordered, "layoutTXData", {}, [
                    (Label, None, dict(defBoxStyle, text="TXData")),
                    (GridLayout, "gridTXData", {
                        "pos_hint":{'x':0.0,"top":defInnerTop},"padding":[6],
                        "size_hint":(1.0,defInnerTop),"spacing":[10,10],"cols":2})]),
                (Bordered, "layoutDel", {"size_hint_x":.9}, [
                    (Label, None, dict(defBoxStyle, text="Delays")),
                    (GridLayout, "gridDel", {
                        "pos_hint":{'x':0.0,"top":defInnerTop},"padding":[6],
                        "size_hint":(1.0,defInnerTop),"spacing":[10,10],"cols":2})])])),
        ("History", "histTab", (
            BorderedGridLayout, "gridHist", dict(cols=1, spacing=[10, 10], padding=[5, 5, 5, 5]))),
        ("Other", "otherTab", (
            BorderedGridLayout, "gridOther", dict(cols=3, spacing=[10, 10], padding=[20, 10, 5, 5])))
    ]
    mkDefaultTabs = True
    def __init__(self, fl=None, ver=0):
        if not hasattr(self, "regs"):
            duals = [-1,-1,-1,-1,6,23,23,-1]
            self.regs = [Register(x, duals[x]) for x in xrange(8)]
        super(ProgADF41XX, self).__init__(fl, ver)
        #set the only reserved bit of all the registers that needs to be 1
        #self.regs[3][17] = 1
        msg0 = "The ramp function must be enabled"
        msg1 = msg0+"\nRamp mode must be "
        self.needRmpOnBub = TouchDownMsg(text=msg0)
        self.needRmpMdBub = TouchDownMsg(text=msg1+"Continuous Triangular")
        self.needRmpMdBub1 = TouchDownMsg(text=msg1+"Single Ramp Burst")
        self.needRmpOn = Bind(tchDownBubble, self.needRmpOnBub)
        self.needRmpMdCT = Bind(tchDownBubble, self.needRmpMdBub)
        self.needRmpMdSRB = Bind(tchDownBubble, self.needRmpMdBub1)
        self.buildFirst()
        self.tabNoWrite.add(self.otherTab)
        self.tabPanel.bind(current_tab=self.curTabEvt)
    def setReg(self, regNum):
        regId = regNum & 7
        self.regs[regId][3:] = regNum >> 3
    def onRegFracIntChg(self, reg, isProg):
        #print isProg
        if not isProg: return
        self.data.numOutFreq = 1
        self.data.freqRF_Out = 1
        self.loadTxtAttr(*self.kTxtData.OutFreq)
        self.loadTxtAttr(*self.kTxtData.RF_Out)
    def onhide(self, app):
        self.curLayout.remove_widget(self.binLayout)
        self.curLayout.remove_widget(self.tabPanel)
        super(ProgADF41XX, self).onhide(app)
    def build(self, app):
        self.curLayout = super(ProgADF41XX, self).build(app)
        if not hasattr(self, "binLayout"):self.buildFirst()
        self.curLayout.add_widget(self.tabPanel)
        self.curLayout.add_widget(self.binLayout)
        self.progBtn.pos_hint={'y':0.045*app.defStyle["size_hint"][1],'x':.01}
        """if hasattr(self, "mainParent") and self.mainParent is not None:
            self.mainParent.remove_widget(self.progBtn)
            self.mainParent = self.progBtn.parent"""
    def getIntFreq(self, freq):
        return freq / 2. ** self.data.selExtDiv
    def getExtFreq(self, freq):
        return freq * 2 ** self.data.selExtDiv
    def buildFirst(self):
        self.tabPanel = self.getTabPanel()
        defStyle = {"size_hint": (1.0, None), "height": 32}
        self.addCfgMdLbls(self.gridCfgInf, defStyle)
        self.rampTabs = [self.ramp1Tab, self.ramp2Tab]
        self.gridRmp = [self.gridRmp1, self.gridRmp2]
        self.fillGridHist()
        self.buildTxtData(True)
        self.gridOther.add_widget(sizedButton(
            text=" Write Register ",on_release=self.writeReg))
        self.gridOther.add_widget(sizedLabel(text=""))
        self.gridOther.add_widget(Button(
            text="Reset To Defaults",
            on_release=self.resetToDefs,size_hint_y=None,height=36))
        self.gridOther.add_widget(sizedLabel(text=""))
        if hasattr(self, "chkFastRmp"):
            self.chkFastRmp.bind(on_touch_down=self.needRmpMdCT)
        if hasattr(self, "chkRmpFSK"):
            self.chkRmpFSK.bind(on_touch_down=self.needRmpOn)
        if hasattr(self, "chkFSK"):
            self.chkFSK.bind(on_touch_down=self.needRmpOn)
        if hasattr(self, "chkPSK"):
            self.chkPSK.bind(on_touch_down=self.needRmpOn)
        if hasattr(self, "inRmpMd"):
            self.inRmpMd.bind(on_touch_down=self.needRmpOn)
        if hasattr(self, "chkSFT"):
            self.chkSFT.bind(on_touch_down=self.needRmpMdSRB)
        if hasattr(self, "chkDualRamp"):
            self.chkDualRamp.bind(on_touch_down=self.needRmpOn)
        self.addAllRegs()
        if ("numClk2Div1" in self.data) and ("numClk2Div2" not in self.data):
            self.data = mkAliased(self.data)
            self.data.__dict__.addAlias("numClk2Div1", "numClk2Div2")
        self.updPFD_Freq()
        self.updRampMisc()
        self.updRampAll()
        self.loadTxtData()
        self.updAutoR()
        self.updAutoN()
    def updRampAll(self):
        self.updRampN(1)
        self.updRampN(2)
    def updRamp1(self): self.updRampN(1)
    def updRamp2(self): self.updRampN(2)
    #n is 1 or 2
    def updRampN(self,n):
        numClk2Div = self.data["numClk2Div%u"%n]
        numDevVal = self.data["numDevVal%u"%n]
        numDevOff = self.data["numDevOff%u"%n]
        numStep = self.data["numStep%u"%n]
        #initially multiply MHz by 1e6 to Hz but divide by 1e3
        #  to get kHz so in the end multiply by (1e6/1e3) or 1e3
        fDevMul = (numDevVal << numDevOff) * 1e3
        fDev = (self.data.numPFD_Freq * fDevMul) / (1<<25)
        #"%.6g"
        self.data["numFDevStep%u"%n] = fDev
        self.data["numTotRamp%u"%n] = fDev * numStep
        numTStep = self.data.numClk1Div*numClk2Div /(self.data.numPFD_Freq*1e6)
        self.data["numTStep%u"%n] = numTStep*1e6
        self.data["numTRamp%u"%n] = numTStep * (numStep + 1) * 1e6
        self.loadTxtAttr(*self.kTxtData["FDevStep%u"%n])
        self.loadTxtAttr(*self.kTxtData["TotRamp%u"%n])
        self.loadTxtAttr(*self.kTxtData["TStep%u"%n])
        self.loadTxtAttr(*self.kTxtData["TRamp%u"%n])
    #used in buildFirst
    #adds all the registers including word select expansions
    def addAllRegs(self):
        self.addedAllReg = True
        self.binLayout = BorderedGridLayout(
            pos_hint={'y':.02, "right":.99},size_hint=(.79,.125),cols=5)
        for reg in self.regs[:4] + self.regs[7:8]:
            self.binLayout.add_widget(reg.build())
        self.binLayout.add_widget(Label(text="Ramp 1"))
        for reg in self.regs[4:7]:
            reg.curReg = 0
            self.binLayout.add_widget(reg.build())
        self.binLayout.add_widget(Label(text=""))
        self.binLayout.add_widget(Label(text="Ramp 2"))
        for reg in self.regs[4:7]:
            if not reg.isDual:
                self.binLayout.add_widget(Label())
                continue
            reg.curReg = 1
            self.binLayout.add_widget(reg.build())
            reg.curReg = 0
    #can be used in buildFirst to replace addAllRegs
    #adds all the registers excluding word select expansions
    def add8Regs(self):
        self.addedAllReg = False
        self.binLayout = BorderedGridLayout(
            pos_hint={'y':.02, "right":.99},size_hint=(.79,.15),cols=4)
        for reg in self.regs:
            reg.curReg = 0
            self.binLayout.add_widget(reg.build())
    def loadReg(self, obj=None):
        try:
            self.setReg(self.data.numWriteData)
            self.onProgChg()
            self.loadDataFromRegs()
            self.loadTxtData()
        except:
            log.Log(traceback.format_exc()[:-1], 2)
    def writeReg(self, obj=None):
        tmp = self.data.numWriteData
        self.userLog.log("Directly writing 0x%08X to the part" % tmp)
        Len = (self.regs[0].wordSz + 7) / 8
        rtn = [0] * Len
        lsbfirst = self.spiBus.lsbfirst
        for c in xrange(Len):
            c = Len - c - 1
            rtn[c] = tmp & 0xFF
            if lsbfirst: rtn[c] = bitMapInt(rtn[c],8)[::-1]
            tmp >>= 8
        rtn = map(int, rtn)
        spiPulseData(self.spiBus, rtn)
    def updChrgPmp(self):
        self.updReg2Dat()
        self.updNegBld()
    def updNegBld(self):
        self.inNegBld.disabled = self.data.calcNegBld
        if self.data.calcNegBld:
            num = 4 * self.lstChrgPmpVals[self.data.selChrgPmp] * 1e3
            num /= (self.data.numIntVal + self.data.numFracVal / 2.0 ** 25)
            pos = BinaryApprox(self.lstNegBldVals, num)
            if pos == len(self.lstNegBldVals): pos -= 1
            elif pos != 0:
                dif0 = num - self.lstNegBldVals[pos]
                dif1 = num - self.lstNegBldVals[pos - 1]
                if abs(dif0) > abs(dif1): pos -= 1
            self.data.selNegBld = pos
            self.loadTxtAttr(*self.kTxtData.NegBld)
        self.updReg3Dat()
    def updAutoNOrder(self):
        super(ProgADF41XX, self).updAutoNOrder()
        if "calcNegBld" in self.useFuncs: self.updNegBld()
    #R_Count and bRefDbl and bRefDiv call this
    def updPFD_Freq(self):
        super(ProgADF41XX, self).updPFD_Freq()
        if hasattr(self, "gridRmp"): self.updRampAll()
    def updR4Ramp1(self):
        self.updReg4_0Dat()
        self.updRamp1()
    def updR4Ramp2(self):
        self.updReg4_1Dat()
        self.updRamp2()
    def updR5Ramp1(self):
        self.updReg5_0Dat()
        self.updRamp1()
    def updR5Ramp2(self):
        self.updReg5_1Dat()
        self.updRamp2()
    def updR6Ramp1(self):
        self.updReg6_0Dat()
        self.updRamp1()
    def updR6Ramp2(self):
        self.updReg6_1Dat()
        self.updRamp2()
    def updRampMisc(self):
        self.chkFastRmp.disabled = (
            self.data.selRmpMd != 1 or not self.data.rampOn)
        dualRmp = (
            self.data.dualRamp or
            (not self.chkFastRmp.disabled and self.data.bFastRmp) or
            self.data.bRmpFSK)
        if dualRmp and self.data.rampOn:
            if self.rampTabs[0].parent is None:
                self.tabPanel.add_widget(self.rampTabs[0],3)
            if self.rampTabs[1].parent is None:
                self.tabPanel.add_widget(self.rampTabs[1],3)
        elif self.data.rampOn:
            if self.rampTabs[0].parent is None:
                self.tabPanel.add_widget(self.rampTabs[0],3)
            if self.rampTabs[1].parent != None:
                self.tabPanel.remove_widget(self.rampTabs[1])
        else:
            if self.rampTabs[0].parent != None:
                self.tabPanel.remove_widget(self.rampTabs[0])
            if self.rampTabs[1].parent != None:
                self.tabPanel.remove_widget(self.rampTabs[1])
        self.chkSFT.disabled = not (self.data.selRmpMd == 3 and self.data.rampOn)
        self.chkDualRamp.disabled = not self.data.rampOn
        self.chkRmpFSK.disabled = not self.data.rampOn
        self.updReg0Dat()#for rampOn
        self.updReg3Dat()#for RmpMd
        self.updReg5Dat()#for dualRamp and bRmpFSK
        self.updReg7Dat()#for bFastRamp
    def outputReg(self, x):
        supObj = super(ProgADF41XX, self)
        doIt = ("r3CntRst" in self.data) and self.regHasAttr(x, "r3CntRst")
        if doIt and self.data.bAutoCntrRst:
            self.data.r3CntRst = True
            self.userLog.log("Counter Reset: 1")
            self.updRegAttrDat(["r3CntRst"])
            supObj.outputReg(x)
            self.data.r3CntRst = False
            self.userLog.log("Counter Reset: 0")
            self.updRegAttrDat(["r3CntRst"])
            supObj.outputReg(x)
        else: supObj.outputReg(x)
    def programIt(self, obj=None):
        self.userLog.log("Sequencing ...")
        seq = self.sequenceRegs([
            (0,0), (1,0), (2,0), (3,0), (4,0),
            (5,0), (6,0), (4,1), (5,1), (6,1), (7,0)])
        strSeq = ", ".join(map(self.fmtReg, seq))
        self.userLog.log("Sequence order: " + strSeq)
        for x in seq:self.outputReg(x)
        super(ProgADF41XX, self).programIt(obj)
    def onProgChg(self):
        try:super(ProgADF41XX, self).onProgChg()
        except:pass
    def sequenceRegs(self, regNums):
        midOrder = set(regNums)
        end = {(0,0), (1,0), (2,0)}
        if len(end) > 0:end.add((0,0))
        midOrder -= end
        end = sorted(end, key=lambda x: (x[0] << 1) | x[1])[::-1]
        midOrder = sorted(midOrder, key=lambda x: (x[0] << 1) | x[1])[::-1]
        return midOrder + end
#get_enterSetHex(ProgADF41XX, "WriteData", (0, 0xFFFFFFFF), "Enter Register Hex")
@partClass("XM-4158-0409")
class ProgXM_4158(ProgADF41XX):
    XM_partName = "XM-4158-0409"
    mfgPartName = "ADF4158"
    remLst = [
        "r3LOL", "constNegBld", "NegBld", "Clk2Div2", "RmpStat", "calcNegBld",
        "bTXInv", "bFastRmp", "bTXTrig", "bSFT", "bTriDel", "bTXTrigDel"]
    defData = dict(ProgADF41XX.defData, **{"Clk2Div2":1})
    del defData["calcNegBld"]
    lstTxtDatas, lstEnterSets, titles, useFuncs, regDatas = fltrFuncName(
        remLst, ProgADF41XX)
    defRegs = list(ProgADF41XX.defRegs) + [0x01030083]
    lstTxtDatas.insert(
        lstTxtDatas.index(("FDevStep2", "num", "%.6g")),
        ("gridRmp2", "blank_2"))
    postInitDiag = {"updateFuncs":lambda funcs: funcs.discard("updNegBld")}
    def __init__(self, fl=None, ver=0):
        duals = [-1,-1,-1,-1,-1,23,23,-1]
        self.regs = [Register(x, duals[x]) for x in xrange(8)]
        super(ProgXM_4158, self).__init__(fl, ver)
    def updRampMisc(self):
        dualRmp = self.data.dualRamp or self.data.bRmpFSK
        if dualRmp and self.data.rampOn:
            if self.rampTabs[0].parent is None:
                self.tabPanel.add_widget(self.rampTabs[0],3)
            if self.rampTabs[1].parent is None:
                self.tabPanel.add_widget(self.rampTabs[1],3)
        elif self.data.rampOn:
            if self.rampTabs[0].parent is None:
                self.tabPanel.add_widget(self.rampTabs[0],3)
            if self.rampTabs[1].parent is not None:
                self.tabPanel.remove_widget(self.rampTabs[1])
        else:
            if self.rampTabs[0].parent is not None:
                self.tabPanel.remove_widget(self.rampTabs[0])
            if self.rampTabs[1].parent is not None:
                self.tabPanel.remove_widget(self.rampTabs[1])
        self.chkDualRamp.disabled = not self.data.rampOn
        self.chkRmpFSK.disabled = not self.data.rampOn
        self.updReg0Dat()#for rampOn
        self.updReg3Dat()#for RmpMd
        self.updReg5Dat()#for dualRamp and bRmpFSK
    def updR4Ramp1(self):
        self.updReg4Dat()
        self.updRampAll()
    def programIt(self, obj=None):
        self.userLog.log("Sequencing ...")
        seq = self.sequenceRegs([
            (0,0), (1,0), (2,0), (3,0), (4,0),
            (5,0), (6,0), (5,1), (6,1), (7,0)])
        strSeq = ", ".join(map(self.fmtReg, seq))
        self.userLog.log("Sequence order: " + strSeq)
        for x in seq:self.outputReg(x)
        #NOTE: bypass the call to the direct super class of ProgXM_4158
        super(ProgADF41XX, self).programIt(obj)
@partClass("XM-A3P5-0708")
class ProgXM_A3P5(ProgADF41XX):
    DevOnly = True
    XM_partName = "XM-A3P5-0708"
    mfgPartName = "ADF4169"
    lstMuxoutOpts = map(isReserv,[
        "3-state Output", "DVdd", "GND", "R-Divider Output", "Reserved",
        "Reserved", "Digital Lock Detect", "Serial Data Output", "Reserved",
        "Reserved", "Clk Divider Output", "Reserved", "Reserved",
        "R-Divider/2", "N-Divider/2", "Ramp Status"])
@partClass("XM-A3X3-0409")
class ProgXM_A3X3(ProgADF41XX):
    boardImageName = images.baseImg+"0357A.png"
    XM_partName = "XM-A3X3-0409"
    mfgPartName = "ADF4169"
    lstMuxoutOpts = map(isReserv,[
        "3-state Output", "DVdd", "GND", "R-Divider Output", "Reserved",
        "Reserved", "Digital Lock Detect", "Serial Data Output", "Reserved",
        "Reserved", "Clk Divider Output", "Reserved", "Reserved",
        "R-Divider/2", "N-Divider/2", "Ramp Status"])
@partClass("XM-A5N5-0609")
class ProgXM_A5N5(ProgADF41XX):
    boardImageName = images.baseImg+"0357A.png"
    XM_partName = "XM-A5N5-0609"
    mfgPartName = "ADF4159"
    lstMuxoutOpts = map(isReserv,[
        "3-state Output", "DVdd", "GND", "R-Divider Output", "N-Divider Output",
        "Reserved", "Digital Lock Detect", "Serial Data Output", "Reserved",
        "Reserved", "Clk Divider Output", "Reserved", "Reserved",
        "R-Divider/2", "N-Divider/2", "Ramp Status"])
    regDatas = dict(ProgADF41XX.regDatas, **{
        "bParRmp":[(5,28)],
        "bRmpFstLk":[(7,18)]})
    lstTxtDatas = ProgADF41XX.lstTxtDatas + [
        ("bParRmp", "chkParRmp"),
        ("bRmpFstLk", "chkRmpFstLk")]
    titles = dict(ProgADF41XX.titles, **{
        "bParRmp":("Parabolic Ramp",),
        "bRmpFstLk":("Ramp Delay Fastlock",)})
    useFuncs = dict(ProgADF41XX.useFuncs, **{
        "bParRmp":(7,"updReg5Dat"),
        "bRmpFstLk":(8,"updReg7Dat")})
    def updRampMisc(self):
        self.chkParRmp.disabled = not self.data.rampOn
        super(ProgXM_A5N5, self).updRampMisc()
    def buildFirst(self):
        super(ProgXM_A5N5, self).buildFirst()
        self.chkParRmp.bind(on_touch_down=self.needRmpOn)
remUseFuncLst = [
    "rampOn", "phaseAdjust", "Phase", "bFSK", "bPSK", "NegBld", "RmpMd",
    "NSel", "r3LOL", "constNegBld", "Clk2Div1", "Clk2Div2", "RmpStat",
    "LESel", "DevVal1", "DevVal2", "DevOff1", "DevOff2", "dualRamp",
    "bRmpFSK", "TXInter", "TXRmpClk", "bTXInv", "Step1", "Step2", "DelWord",
    "bDelStart", "DelClk", "bRmpDel", "bFastRmp", "bTXTrig", "bSFT",
    "bTriDel", "bTXTrigDel", "SDModMd", "calcNegBld",
    "FDevStep1", "TotRamp1", "TStep1", "TRamp1",
    "FDevStep2", "TotRamp2", "TStep2", "TRamp2"]
remRegLst = [
    "rampOn", "phaseAdjust", "numPhase", "bFSK", "bPSK", "selNegBld",
    "selRmpMd", "selNSel", "r3LOL", "constNegBld", "numClk2Div1",
    "numClk2Div2", "selRmpStat", "selLESel", "numDevVal1", "numDevVal2",
    "numDevOff1", "numDevOff2", "dualRamp", "bRmpFSK", "selTXInter",
    "selTXRmpClk", "bTXInv", "numStep1", "numStep2", "numDelWord",
    "bDelStart", "selDelClk", "bRmpDel", "bFastRmp", "bTXTrig",
    "bSFT", "bTriDel", "bTXTrigDel", "selSDModMd"]
class ProgADF415X_67(ProgADF41XX):
    lstMuxoutOpts = map(isReserv,[
        "Three-State Output", "DVdd", "DGND", "R Divider Output", "N Divider Output",
        "Analog Lock Detect", "Digital Lock Detect", "Serial Data Output", "Reserved",
        "Reserved", "Clk Divider Output", "Reserved", "Fastlock Switch",
        "R Divider/2", "N Divider/2", "Reserved"])
    lstClkDivOpts = {
        0b00:"Clock Divider Off", 0b01:"Switched R Fastlock Enable"}
    lstEnterSets = [
        x for x in ProgADF41XX.lstEnterSets if x[1] not in remUseFuncLst]
    chgEnterSet(lstEnterSets, ("lst", "ChrgPmp", "updReg2Dat"))
    chgEnterSet(lstEnterSets, ("int", "Clk1Div", (0,4095), "updReg4Dat"))
    defData = dict(ProgADF41XX.defData, **{
        "numPFD_Freq":10.,"numR_Freq":10.0, "numNVal": 580.0,
        "numOutFreq":5800.0, "numRF_Out":5800.0,})
    defBoxStyle = dict(pos_hint={'x': 0.0, "top": 1.0}, size_hint_y=.08)
    defInnerTop = defBoxStyle["pos_hint"]["top"] - defBoxStyle["size_hint_y"]
    tabLayout = list(ProgADF41XX.tabLayout)
    tabLayout[2:] = [
        ("Config", "cfgTab", (
            BorderedGridLayout, "gridCfg", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("Charge Pump", "chrgTab", (
            BorderedGridLayout, "gridChrg", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("History", "histTab", (
            BorderedGridLayout, "gridHist", dict(cols=1, spacing=[10, 10], padding=[5, 5, 5, 5]))),
        ("Other", "otherTab", (
            BorderedGridLayout, "gridOther", dict(cols=3, spacing=[10, 10], padding=[20, 10, 5, 5])))
    ]
    def __init__(self, fl=None, ver=0):
        self.regs = [Register(x) for x in xrange(5)]
        super(ProgADF415X_67, self).__init__(fl, ver)
    def buildFirst(self):
        self.tabPanel = self.getTabPanel()
        defStyle = {"size_hint":(1.0,None), "height":32}
        self.addCfgMdLbls(self.gridCfgInf, defStyle)
        self.fillGridHist()
        self.buildTxtData(True)
        self.gridOther.add_widget(sizedButton(
            text=" Load Into Program ",on_release=self.loadReg))
        self.gridOther.add_widget(sizedButton(
            text=" Write Register ",on_release=self.writeReg))
        self.gridOther.add_widget(Button(
            text="     Reset To Defaults",
            on_release=self.resetToDefs,size_hint_y=None,height=24))
        self.gridOther.add_widget(sizedLabel(text=""))
        self.updPFD_Freq()
        self.loadTxtData()
        self.add5Regs()
        self.updAutoR()
        self.updAutoN()
    def add5Regs(self):
        self.binLayout = BorderedGridLayout(
            pos_hint={'y':.02, "right":.99},size_hint=(.79,.15),cols=5)
        for reg in self.regs:
            reg.curReg = 0
            self.binLayout.add_widget(reg.build())
    def sequenceRegs(self, regNums):
        rtn = set(regNums)
        rtn &= {(0,0), (1,0), (2,0), (3,0), (4,0)}
        return sorted(rtn, key=lambda x: (x[0] << 1) | x[1])[::-1]
pos0 = -1
pos1 = -1
for c in xrange(len(ProgADF415X_67.lstTxtDatas)):
    if ProgADF415X_67.lstTxtDatas[c][0:2] == ("ClkDiv", "lst"):
        if pos0 != -1:
            log.Log("Found duplicate (\"ClkDiv\", \"lst\")", 2)
        pos0 = c
    elif ProgADF415X_67.lstTxtDatas[c][0:2] == ("r3NotSDRst", "chkR3NotSDRst"):
        if pos1 != -1:
            log.Log("Found duplicate (\"r3NotSDRst\", \"chkR3NotSDRst\")", 2)
        pos1 = c
if pos0 != -1 and pos1 != -1:
    tmp = ProgADF415X_67.lstTxtDatas[pos0]
    ProgADF415X_67.lstTxtDatas[pos0] = ProgADF415X_67.lstTxtDatas[pos1]
    ProgADF415X_67.lstTxtDatas[pos1] = tmp
@partClass("XM-ADF4157-0000")
class ProgXM_ADF4157(ProgADF415X_67):
    XM_partName = "XM-ADF4157-0000"
    mfgPartName = "ADF4157"
    lstNegBldOpts = {0b00:"OFF",0b11:"ON"}
    lstEnterSets = list(ProgADF415X_67.lstEnterSets)
    lstEnterSets.append(("lst", "NegBld", "updReg4Dat"))
    regDatas = {
        k:ProgADF415X_67.regDatas[k] for k in ProgADF415X_67.regDatas
        if k not in remRegLst or k == "selNegBld"}
    regDatas.update({
        "selNegBld":[(4,23,25)], "selClkDiv":[(4,19,21)],
        "numClk1Div":[(4,7,19)]})
    lstTxtDatas = [
        x for x in ProgADF415X_67.lstTxtDatas
        if x[0] not in remUseFuncLst or x[0] == "NegBld"]
    useFuncs = {
        k:ProgADF415X_67.useFuncs[k] for k in ProgADF415X_67.useFuncs
        if k not in remUseFuncLst or k == "NegBld"}
    useFuncs.update({
        "ClkDiv": (1,), "Clk1Div": (4,1)})
    titles = {
        k:ProgADF415X_67.titles[k] for k in ProgADF415X_67.titles
        if k not in remUseFuncLst or k == "NegBld"}
    defRegs = [
        0x01220000, 0x00000001, 0x07408002, 0x00000043, 0x00000004]
    defData = dict(ProgADF415X_67.defData, **{
        "numWriteData":0,"selExtDiv":0,"bAutoCntrRst":False})
@partClass("XM-A5Y5-0409")
class ProgXM_A5Y5(PLL_MainMod, ProgADF415X_67):
    XM_partName = "XM-A5Y5-0409"
    mfgPartName = "ADF4156"
    lstNoiseMdOpts = map(isReserv,[
        "Low Noise Mode", "Reserved", "Reserved", "Low Spur Mode"])
    lstBindFuncs = dict(ProgADF415X_67.lstBindFuncs, **{
        "addModLine": ("addLine", "gridMod"),
        "addModLineM": ("addLineM", "gridMod"),
        "addModLst": ("addLstLine", "gridMod")})
    #line funcs 12, 13
    lstLineFuncs = ProgADF415X_67.lstLineFuncs + ["addModLine", "addModLineM"]
    lstLineBig = ProgADF415X_67.lstLineBig + [12, 13]
    #lst func 7
    lstLstFuncs = ProgADF415X_67.lstLstFuncs + ["addModLst"]
    lstLstBig = ProgADF415X_67.lstLstBig + [7]
    regDatas = {
        k:ProgADF415X_67.regDatas[k] for k in ProgADF415X_67.regDatas
        if k not in remRegLst or k == "numPhase"}
    regDatas.update({
        "selClkDiv":[(4,19,21)], "numFracVal":[(0,3,15)],
        "numClk1Div":[(4,7,19)], "numModVal":[(2,3,15)],
        "selNoiseMd":[(2,29,31)]})
    lstTxtDatas = [
        x for x in ProgADF415X_67.lstTxtDatas
        if x[0] not in remUseFuncLst or x[0] == "Phase"]
    lstTxtDatas.extend([
        ("ModVal", "num"), ("NoiseMd", "lst"),
        ("UsrSpaceCh", "num", "%.15g"), ("SpaceCh", "num", "%.15g")])
    useFuncs = {
        k:ProgADF415X_67.useFuncs[k] for k in ProgADF415X_67.useFuncs
        if k not in remUseFuncLst or k == "Phase"}
    useFuncs.update({
        "ModVal": (12,"",1), "IntVal": (0,"",1), "FracVal": (0,"",1),
        "SpaceCh": (12,"kHz",0), "NoiseMd": (1,), "ClkDiv": (1,),
        "UsrSpaceCh": (12,"kHz",1), "Clk1Div": (4,1), "PreScl": (7,1)})
    titles = {
        k:ProgADF415X_67.titles[k] for k in ProgADF415X_67.titles
        if k not in remUseFuncLst or k == "Phase"}
    titles.update({
        "SpaceCh": ("Actual Channel Spacing",),
        "ModVal": ("Modulus Part of N",),
        "NoiseMd": ("Noise Mode",),
        "UsrSpaceCh": ("Desired Channel Spacing",)})
    lstEnterSets = list(ProgADF415X_67.lstEnterSets)
    lstEnterSets.extend([
        ("int", "Phase", (-2048, 2047), "updReg1Dat"),
        ("float", "UsrSpaceCh", (1.0,10000.0), "updModSpace"),
        ("lst", "NoiseMd", "updReg2Dat"),
        ("int", "ModVal", (0, 4095), "updModSpace")])
    chgEnterSet(lstEnterSets, ("int", "FracVal", (0,4095), "updAutoNOrder"))
    chgEnterSet(
        lstEnterSets, ("float", "NVal", (23.,4096-.5**12), "updAutoNOrder"))
    chgEnterSet(lstEnterSets, ("int", "Phase", (0,4095), "updReg1Dat"))
    tabLayout = list(ProgADF415X_67.tabLayout)

    defBoxStyle = dict(pos_hint={'x': 0.0, "top": 1.0}, size_hint_y=.08)
    defInnerTop = defBoxStyle["pos_hint"]["top"] - defBoxStyle["size_hint_y"]
    tabLayout[1] = ("Main", "vcoTab", (
        GridSelFrameLayout, "gridMain", dict(
            rows=1, spacing=[10, 10], padding=[6], side_names=["VCO", "PFD", "Misc\n   /\nMOD"]), [
            (Bordered, "layoutMod", {}, [
                (Label, None, dict(defBoxStyle, text="Channel Spacing (Misc/Mod)")),
                (GridLayout, "gridMod", {
                    "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                    "cols": 3, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})], -4),
            (Bordered, "layoutRef", {}, [
                (Label, None, dict(defBoxStyle, text="Reference (PFD)")),
                (GridLayout, "gridRef", {
                    "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                    "cols": 3, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})], -4),
            (Bordered, "layoutVCO", {"size_hint_x": 1.2}, [
                (Label, None, dict(defBoxStyle, text="N Counter (VCO)")),
                (GridLayout, "gridVCO", {
                    "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                    "cols": 3, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})], -4)]))
    defRegs = [
        0x01220000, 0x00000009, 0x0040800A, 0x00000043, 0x00000004]
    defData = dict(ProgADF415X_67.defData, **{"numUsrSpaceCh":100.0, "numSpaceCh":100.0})
    del defData["numModVal"]
    def buildFirst(self):
        super(ProgXM_A5Y5, self).buildFirst()
        self.updAutoMod()
    def moveNCnt(self, obj):
        swapWidgets(self.gridSelFrame, self.layoutVCO)
    def moveMod(self, obj):
        swapWidgets(self.gridSelFrame, self.layoutMod)
    def moveRef(self, obj):
        swapWidgets(self.gridSelFrame, self.layoutRef)
