from BaseADF_PLL import *
from time import sleep

"""
Implements:
    ADF4360-0
    ADF4360-1
    ADF4360-2
    ADF4360-3
    ADF4360-4
    ADF4360-5
    ADF4360-6
    ADF4360-7
    ADF4360-8
    ADF4360-9
"""

log = None

@initFunc(__name__.replace("ProgramIt.", ""))
def init(progLog):
    global log
    log = progLog
sub10Fmt = "F[sub][size=10]%s[/size][/sub]"
class ProgADF4360_X(PLL_CalPFD_Freq, PLL_IntAB, BaseProgADF_PLL):
    DevOnly = True
    mfgName = "Analog Devices"
    interface = "3-wire"
    catId = CAT_PLL_VCO
    XM_partName = "Unknown"
    mfgPartName = "ADF4360-*"
    pfdFreqShift = -3
    lstBindFuncs = dict(BaseProgADF_PLL.lstBindFuncs, **{
        "addSmOtherLn": ("addSmallLine", "gridOther")})
    #8
    lstLineFuncs = BaseProgADF_PLL.lstLineFuncs + ["addSmOtherLn"]
    lstPreSclOpts = map(isReserv,["8/9", "16/17", "32/33", "Reserved"])
    lstMuxoutOpts = map(isReserv,[
        "Three-State Output", "Digital Lock Detect", "N Divider Output", "DVdd",
        "R Divider Output", "N-Channel Open-Drain Lock Detect",
        "Serial Data Output", "DGND"])
    lstCorePwrOpts = map(isReserv,["5mA", "10mA", "15mA", "20mA"])
    lstLkDtcPOpts = map(isReserv,["3 PFD Cycles","5 PFD Cycles"])
    lstPhDtcPolOpts = map(isReserv,["Negative", "Positive"])
    #CE PIN 0 means only Asynchronous Power Down is selected
    lstPwrDwnOpts = map(isReserv,[
        "Normal Operation", "Asynchronous Power Down",
        "Reserved", "Synchronous Power Down"])
    lstICPnVals = map(str, [.3125*(x+1) for x in xrange(8)])
    lstICP1Opts = map(isReserv, lstICPnVals)
    lstICP2Opts = map(isReserv, lstICPnVals)
    lstOutPwrOpts = map(isReserv,["3.5mA(-14dBm)", "5.0mA(-11dBm)", "7.5mA(-8dBm)", "11.0mA(-5dBm)"])
    lstABPWOpts = map(lambda x: isReserv("%.1fns" % x), [3.0, 1.3, 6.0, 3.0])
    lstBandClkOpts = map(isReserv, ["1", "2", "4", "8"])
    lstTxtDatas = [
        ("RF_Out", "num", "%.15g"),
        ("PreScl", "lst"),
        ("NVal", "num"),
        ("ACntr", "num"),
        ("BCntr", "num"),
        ("OutFreq", "num", "%.15g"),
        ("R_Freq", "num", "%.15g"),
        ("R_Count", "num"),
        ("UsrPFD_Freq", "num", "%.15g"),
        ("PFD_Freq", "num", "%.15g"),
        ("Muxout", "lst"),
        ("LkDtcP", "lst"),
        ("PhDtcPol", "lst"),
        ("CorePwr", "lst"),
        #("bCntRst", "chkCntRst"),
        ("bAutoCntrRst", "chkAutoCntrRst"),
        ("PwrDwn", "lst"),
        ("mtLkDtc", "chkMTLkDtc"),
        ("div2", "chkDiv2"),
        ("div2sel", "chkDiv2Sel"),
        ("ABPW", "lst"),
        ("BandClk", "lst"),
        ("OutPwr", "lst"),
        ("cpGain", "chkCPG"),
        ("chrgPmp3St","chkCP3St"),
        ("ICP1", "lst"),
        ("ICP2", "lst"),
        ("WriteData", "num", "%06X")]
    useFuncs = {
        "R_Freq": (3,"MHz",1),
        "PFD_Freq": (3,"kHz",0),
        "UsrPFD_Freq": (2,"kHz",1),
        "RF_Out": (0,"MHz",1),
        "R_Count": (3,"",1),
        "OutFreq": (1,"MHz",0),
        "CorePwr": (1,),
        #"bCntRst": (2,"updReg0Dat"),
        "bAutoCntrRst": (2,None),
        "Muxout": (1,),
        "PhDtcPol": (1,),
        "chrgPmp3St": (4,"updReg0Dat"),
        "cpGain": (4,"updCPGain"),
        "mtLkDtc": (2,"updReg0Dat"),
        "OutPwr": (1,),
        "ICP1": (2,),
        "ICP2": (2,),
        "PwrDwn": (1,),
        "PreScl": (0,),
        "div2": (2,"updReg2Dat"),
        "div2sel": (2,"updReg2Dat"), 
        "ABPW": (2,),
        "LkDtcP": (1,),
        "BandClk": (1,),
        "ACntr": (0,"",1),
        "BCntr": (0,"",1),
        "NVal":(0,"",1),
        "WriteData":(8,1)}
    #,("bTestMode", "chkTestMode")
    titles = {
        "R_Freq": (
            "Reference Freq (%s)"%(sub10Fmt%"REF"),
            "Reference Frequency"),
        "PFD_Freq": ("Actual PFD Freq (%s)"%(sub10Fmt%"PFD"),),
        "UsrPFD_Freq": ("Desired PFD Freq",),
        "RF_Out": ("Desired VCO Output Freq","RF Output Frequency"),
        "R_Count":(
            "REF Counter\n(R = %s/%s)"%(sub10Fmt%"REF",sub10Fmt%"PFD"),
            "Reference Counter (R)"),
        "CorePwr": ("Core Power Level",),
        #"bCntRst": ("Counter Reset",),
        "bAutoCntrRst": ("Reset Counters on Write",),
        "Muxout": ("Muxout",),
        "PhDtcPol": ("Phase Detector Polarity",),
        "chrgPmp3St": ("Charge Pump Tri-State",),
        "cpGain": ("Charge Pump Gain 1",),
        "mtLkDtc": ("Mute Until Lock Detect",),
        "OutPwr": ("Output Power Level",),
        "ICP1": ("Current Setting 1",),
        "ICP2": ("Current Setting 2",),
        "PwrDwn": ("Power Down",),
        "PreScl": ("Internal Prescaler (P/P+1)",),
        "div2": ("divide by 2",),
        "div2sel": ("divide by 2 select",), 
        "ABPW": ("Anti-Backlash Pulse Width",),
        "LkDtcP": ("Lock Detect Precision",),
        "BandClk": ("Band Select Clock",),
        "ACntr": ("A Counter",),
        "BCntr": ("B Counter",),
        "NVal": ("VCO Counter (N = BP+A)",),
        "OutFreq": ("VCO Output Freq (N * %s)"%(sub10Fmt%"PFD"),),
        "WriteData": ("Enter Register Hex",)}
    regDatas = {
        "selCorePwr":[(0,2,4)],
        "bCntRst":[(0,4)],
        "selMuxout":[(0,5,8)],
        "selPhDtcPol":[(0,8)],
        "chrgPmp3St":[(0,9)],
        #"cpGain":[(0,10)],
        #mt->Mute Till,
        "mtLkDtc":[(0,11)],
        "selOutPwr":[(0,12,14)],
        "selICP1":[(0,14,17)],
        "selICP2":[(0,17,20)],
        "selPwrDwn":[(0,20,22)],
        "selPreScl":[(0,22,24)],
        #reg 2
        "numACntr":[(2,2,7)],
        "numBCntr":[(2,8,21)],
        "cpGain":[(2,21)],
        "div2":[(2,22)],
        "div2sel":[(2,23)],
        "numR_Count":[(1,2,16)],
        "selABPW":[(1,16,18)],
        "selLkDtcP":[(1,18)],
        #"bTestMode":[(1,19)],
        "selBandClk":[(1,20,22)]}
    lstEnterSets = [
        ("float", "R_Freq", (10.,250.), "updPFD_Freq"),
        ("float", "UsrPFD_Freq", (10.,260000.), "updPFD_Freq"),
        ("float", "RF_Out", (350.,1800.), "updAutoNOrder"),
        ("int", "R_Count", (1, 2**14-1), "updPFD_Freq"),
        ("lst", "LkDtcP", "updReg1Dat"),
        ("lst", "PhDtcPol", "updReg0Dat"),
        ("lst", "PreScl", "updAutoNOrder"),
        ("lst", "Muxout", "updReg0Dat"),
        ("lst", "CorePwr", "updReg0Dat"),
        ("lst", "PwrDwn", "updReg0Dat"),
        ("lst", "OutPwr", "updReg0Dat"),
        ("lst", "ICP1", "updReg0Dat"),
        ("lst", "ICP2", "updReg0Dat"),
        ("lst", "ABPW", "updReg1Dat"),
        ("lst", "BandClk", "updReg1Dat"),
        ("int", "ACntr", (0, 31), "updAutoNOrder"),
        ("int", "BCntr", (3, 2**13-1), "updAutoNOrder"),
        ("int", "NVal", (24,2**18-1), "updAutoNOrder"),
        ("hex", "WriteData", (0, 2**24-1), None)]
    defRegs = [0x023212,0x0FF100, 0x1200C9]
    #MHz, MHz, kHz
    defData = {
        "numR_Freq":10.0, "numUsrPFD_Freq":200.0, "numPFD_Freq":200.0,
        "numWriteData":0,"numOutFreq":900.0,"numNVal":4500,"selAutoR":1,
        "selAutoN":0,"numRF_Out":900.0,"bAutoCntrRst":True}
    defBoxStyle = dict(pos_hint={'x': 0.0, "top": 1.0}, size_hint_y=.08)
    defInnerTop = defBoxStyle["pos_hint"]["top"] - defBoxStyle["size_hint_y"]
    tabLayout = [
        ("Config Mode", "mainTab", (
            BorderedGridLayout, None, dict(cols=2, spacing=[10, 10], padding=[6]), [
                (Bordered, "layoutCfgMd", {}, [
                    (Label, None, dict(defBoxStyle, text="Configuration Mode")),
                    (GridLayout, "gridCfgMd", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 2, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})]),
                (Bordered, "layoutCfgInf", {}, [
                    (Label, None, dict(defBoxStyle, text="Configuration Info")),
                    (GridLayout, "gridCfgInf", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 1, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})])])),
        ("Main", "vcoTab", (
            BorderedGridLayout, "gridMain", dict(cols=2, spacing=[10,10], padding=[6]), [
                (Bordered, "layoutVCO", {"size_hint_x": 1.2}, [
                    (Label, None, dict(defBoxStyle, text="N Counter (VCO)")),
                    (GridLayout, "gridVCO", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 3, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})]),
                (Bordered, "layoutRef", {}, [
                    (Label, None, dict(defBoxStyle, text="Reference (PFD)")),
                    (GridLayout, "gridRef", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 3, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})])])),
        ("Config", "cfgTab", (
            BorderedGridLayout, "gridCfg", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("Charge Pump", "chrgTab", (
            BorderedGridLayout, "gridChrg", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("History", "histTab", (
            BorderedGridLayout, "gridHist", dict(cols=1, spacing=[10, 10], padding=[5, 5, 5, 5]))),
        ("Other", "otherTab", (
            BorderedGridLayout, "gridOther", dict(cols=3, spacing=[10, 10], padding=[20, 10, 5, 5])))
    ]
    mkDefaultTabs = True
    def getPreSclDiv(self):
        return min(8 << self.data.selPreScl, 32)
    def updCPGain(self):
        self.updRegAttrDat(["cpGain"])
        self.regs[0][10] = self.regs[2][21]
    def __init__(self, fl=None, ver=0):
        self.regs = [Reg24(x) for x in xrange(3)]
        self.data = ValuePack()
        self.userLog = UserLog(size_hint=(1.0, .9))
        super(ProgADF4360_X, self).__init__(fl, ver)
        #"mkRefChk": ("addChkLine", "gridRef"),
        self.initData()
        self.buildFirst()
    def buildFirst(self):
        #defBoxStyle = dict(pos_hint={'x':0.0,"top":1.0}, size_hint_y=.08)
        #defInnerTop = defBoxStyle["pos_hint"]["top"]-defBoxStyle["size_hint_y"]
        self.tabPanel = self.getTabPanel()
        defStyle = {"size_hint":(1.0,None), "height":32}
        '''self.tabPanel.default_tab_text = "Config Mode"
        self.mainTab = self.tabPanel.tab_list[0]
        self.mainTab.content = BorderedGridLayout(
            cols=2,spacing=[10,10],padding=[6])
        self.layoutCfgMd, self.gridCfgMd = mkLblBorder(
            Label, GridLayout, {}, dict(defBoxStyle, text="Configuration Mode"),
            {"pos_hint":{'x':0.0,"top":defInnerTop},"cols":2,
             "size_hint":(1.0,defInnerTop),"spacing":[10,10],"padding":[6]})
        self.mainTab.content.add_widget(self.layoutCfgMd)
        self.layoutCfgInf, self.gridCfgInf = mkLblBorder(
            Label, GridLayout, {}, dict(defBoxStyle, text="Configuration Info"),
            {"pos_hint":{'x':0.0,"top":defInnerTop},"cols":1,
             "size_hint":(1.0,defInnerTop),"spacing":[10,10],"padding":[6]})'''
        self.gridCfgInf.add_widget(sizedCls1(
            Label,text="Select the method of entering the N Counter value",
            **defStyle))
        self.gridCfgInf.add_widget(sizedCls1(
            Label,text="Select the method of entering the R Counter value",
            **defStyle))
        '''self.mainTab.content.add_widget(self.layoutCfgInf)
        self.mkTab(
            "vcoTab", "gridMain", "Main VCO",
            cols=2, spacing=[10,10], padding=[6])
        self.layoutVCO, self.gridVCO = mkLblBorder(
            Label, GridLayout, {"size_hint_x":1.2},
            dict(defBoxStyle, text="N Counter"),
            {"pos_hint":{'x':0.0,"top":defInnerTop},"cols":3,
             "size_hint":(1.0,defInnerTop),"spacing":[10,10],"padding":[6]})
        self.layoutRef, self.gridRef = mkLblBorder(
            Label, GridLayout, {}, dict(defBoxStyle, text="Reference"),
            {"pos_hint":{'x':0.0,"top":defInnerTop},"cols":3,
             "size_hint":(1.0,defInnerTop),"spacing":[10,10],"padding":[6]})
        self.gridMain.add_widget(self.layoutRef, 0)
        self.gridMain.add_widget(self.layoutVCO, 1)
        """self.gridRef = GridLayout(
            cols=3,size_hint=(1.,1.),spacing=[10,10],padding=[6])
        self.gridPFD.add_widget(self.gridRef)"""
        #self.gridPFD.add_widget(Label())
        #self.buildMain()
        self.mkTab("cfgTab", "gridCfg", "Config")
        #self.mkTab("phLkTab", "gridPhLk", "Phase/Lock")
        self.mkTab("chrgTab", "gridChrg", "Charge Pump")
        #self.mkTab("outTab", "gridOut", "IN/OUT/PWR")
        self.mkTab("histTab", "gridHist", "History", cols=1, padding=[5,5,5,5])'''
        self.fillGridHist()
        #self.mkTab("otherTab", "gridOther", "Other", cols=3)
        """self.addSmallLine(
            self.gridOther, "Hex Value to Be Written",
            "%06X"%self.data.numWriteData, "txtWriteData",self.enterWriteData)"""
        #self.loadDataFromRegs()
        self.buildTxtData(True)
        self.gridOther.add_widget(sizedButton(
            text=" Write Register ",on_release=self.writeReg))
        self.gridOther.add_widget(sizedButton(
            text=" Set Register In Program ",on_release=self.setRegInProgram))
        self.loadTxtData()
        self.add3Regs()
        self.updAutoN()
        self.updAutoR()
        self.updPFD_Freq()
        self.updReg0Dat()
        self.updReg1Dat()
        self.updReg2Dat()
    def add3Regs(self):
        self.binLayout = BorderedGridLayout(
            pos_hint={'y':.02, "right":.99},size_hint=(.79,.15),cols=3)
        for reg in self.regs:
            reg.curReg = 0
            self.binLayout.add_widget(reg.build())
    def build(self, app):
        self.curLayout = super(ProgADF4360_X, self).build(app)
        if not hasattr(self, "binLayout"):self.buildFirst()
        self.curLayout.add_widget(self.tabPanel)
        self.curLayout.add_widget(self.binLayout)
        self.progBtn.pos_hint={'y':0.045*app.defStyle["size_hint"][1],'x':.01}
    def onhide(self, app):
        self.curLayout.remove_widget(self.binLayout)
        self.curLayout.remove_widget(self.tabPanel)
        super(ProgADF4360_X, self).onhide(app)
    def setReg(self, regNum):
        regId = regNum & 3
        self.regs[regId][2:] = regNum >> 2
    def setRegInProgram(self, obj=None):
        try:
            self.setReg(self.data.numWriteData)
            self.onProgChg()
            self.loadDataFromRegs()
            self.loadTxtData()
        except:
            log.Log(traceback.format_exc()[:-1], 2)
    def writeReg(self, obj=None):
        spiWriteNum(self.data.numWriteData, 24, self.spiBus)
        self.userLog.log("writing 0x%06X directly to part")

    def outputReg(self, x):
        supObj = super(ProgADF4360_X, self)
        if self.data.bAutoCntrRst and self.regHasAttr(x, "bCntRst"):
            self.data.bCntRst = True
            updArgs = (x[0],)
            if self.regs[x[0]].isDual: updArgs += x[1:]
            self.userLog.log("Counter Reset: 1")
            self.updRegNDat(*updArgs)
            supObj.outputReg(x)
            self.data.bCntRst = False
            self.userLog.log("Counter Reset: 0")
            self.updRegNDat(*updArgs)
            supObj.outputReg(x)
        else: supObj.outputReg(x)
    def programIt(self, obj=None):
        seq = [(1,0), (0,0), (2,0)]
        strSeq = ", ".join(map(self.fmtReg, seq))
        self.userLog.log("Sequence order: " + strSeq)
        for x in seq:
            if x[0] == 2:
                sleep(0.010)
            self.outputReg(x)
        super(ProgADF4360_X, self).programIt(obj)
    def onProgChg(self):
        try:super(ProgADF4360_X, self).onProgChg()
        except:pass
#get_enterSetHex(ProgADF4360_X, "WriteData", (0, 0xFFFFFF), "Enter Register Hex")
@partClass("XM-~A5P5-0609")
class ProgXM_A5P5(ProgADF4360_X):
    XM_partName = "XM-A5P5-0609"
    mfgPartName = "ADF4360-0"
    defData = dict(ProgADF4360_X.defData, **{
        "numRF_Out":2500.0, "numOutFreq": 2500.0, "numNVal": 12500})
    defRegs = [0x61A12, 0xFF104, 0x3000C9]
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13dBm)", "5.0mA(-11dBm)", "7.5mA(-8dBm)", "11.0mA(-5dBm)"])
    lstEnterSets = list(ProgADF4360_X.lstEnterSets)
    chgEnterSet(
        lstEnterSets, ("float", "RF_Out", (2400.,2725.), "updAutoNOrder"))
@partClass("XM-~A5P6-0609")
class ProgXM_A5P6(ProgADF4360_X):
    XM_partName = "XM-~A5P6-0609"
    mfgPartName = "ADF4360-1"
    defData = dict(ProgADF4360_X.defData, **{
        "numRF_Out":2400.0, "numOutFreq":2400.0, "numNVal": 12000})
    defRegs = [0x5DC02, 0xFF108, 0x3000C9]
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13dBm)", "5.0mA(-10.5dBm)", "7.5mA(-8dBm)", "11.0mA(-5dBm)"])
    lstEnterSets = list(ProgADF4360_X.lstEnterSets)
    chgEnterSet(
        lstEnterSets, ("float", "RF_Out", (2050.,2450.), "updAutoNOrder"))
@partClass("XM-A5P7-0609")
class ProgXM_A5P7(ProgADF4360_X):
    #REFERENCE COUNTER intervals are weird in data sheet
    XM_partName = "XM-A5P7-0609"
    mfgPartName = "ADF4360-2"
    defData = dict(ProgADF4360_X.defData, **{
        "numRF_Out":2200.0, "numOutFreq":2200.0, "numNVal": 11000})
    defRegs = [0x55F02, 0xFF108, 0x3000C9]
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13dBm)", "5.0mA(-11dBm)", "7.5mA(-8dBm)", "11.0mA(-5dBm)"])
    lstEnterSets = list(ProgADF4360_X.lstEnterSets)
    chgEnterSet(
        lstEnterSets, ("float", "RF_Out", (1850.,2170.), "updAutoNOrder"))
@partClass("XM-A5P8-0609")
class ProgXM_A5P8(ProgADF4360_X):
    #REFERENCE COUNTER intervals are weird in data sheet
    XM_partName = "XM-~A5P8-0609"
    mfgPartName = "ADF4360-3"
    defData = dict(ProgADF4360_X.defData, **{
        "numRF_Out":1800.0, "numOutFreq":1800.0, "numNVal": 9000})
    defRegs = [0x46502, 0xFF108, 0x3000C9]
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-12dBm)", "5.0mA(-9dBm)", "7.5mA(-6dBm)", "11.0mA(-3dBm)"])
    lstEnterSets = list(ProgADF4360_X.lstEnterSets)
    chgEnterSet(
        lstEnterSets, ("float", "RF_Out", (1600.,1950.), "updAutoNOrder"))
@partClass("XM-~A5P9-0609")
class ProgXM_A5P9(ProgADF4360_X):
    XM_partName = "XM-~A5P9-0609"
    mfgPartName = "ADF4360-4"
    defData = dict(ProgADF4360_X.defData, **{
        "numRF_Out":1600.0, "numOutFreq":1600.0, "numNVal": 8000})
    defRegs = [0x3E802, 0xFF108, 0x3000C9]
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13dBm)", "5.0mA(-10dBm)", "7.5mA(-7dBm)", "11.0mA(-4dBm)"])
    lstEnterSets = list(ProgADF4360_X.lstEnterSets)
    chgEnterSet(
        lstEnterSets, ("float", "RF_Out", (1450.,1750.), "updAutoNOrder"))
@partClass("XM-~A5R1-0609")
class ProgXM_A5R1(ProgADF4360_X):
    #REFERENCE COUNTER intervals are weird in data sheet
    XM_partName = "XM-~A5R1-0609"
    mfgPartName = "ADF4360-5"
    defData = dict(ProgADF4360_X.defData, **{
        "numRF_Out":1400.0, "numOutFreq":1400.0, "numNVal": 7000})
    defRegs = [0x56B02, 0xFF104, 0x3000C9]
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13.5dBm)", "5.0mA(-10.5dBm)", "7.5mA(-7.5dBm)", "11.0mA(-4.5dBm)"])
    lstEnterSets = list(ProgADF4360_X.lstEnterSets)
    chgEnterSet(
        lstEnterSets, ("float", "RF_Out", (1200.,1400.), "updAutoNOrder"))
@partClass("XM-~A5R2-0609")
class ProgXM_A5R2(ProgADF4360_X):
    #REFERENCE COUNTER intervals are weird in data sheet
    XM_partName = "XM-~A5R2-0609"
    mfgPartName = "ADF4360-6"
    defData = dict(ProgADF4360_X.defData, **{
        "numRF_Out":1200.0, "numOutFreq":1200.0, "numNVal": 6000})
    defRegs = [0x2EE02, 0xFF104, 0x3000C9]
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13.5dBm)", "5.0mA(-10.5dBm)", "7.5mA(-7.5dBm)", "11.0mA(-4.5dBm)"])
    lstEnterSets = list(ProgADF4360_X.lstEnterSets)
    chgEnterSet(
        lstEnterSets, ("float", "RF_Out", (1050.,1250.), "updAutoNOrder"))
@partClass("XM-~A5R3-0609")
class ProgXM_A5R3(ProgADF4360_X):
    DevOnly = True
    XM_partName = "XM-~A5R3-0609"
    mfgPartName = "ADF4360-7"
    defData = dict(ProgADF4360_X.defData, **{
        "numRF_Out":900.0, "numOutFreq":900.0, "numNVal": 4500})
    defRegs = [0x023212,0x0FF100, 0x1200C9]
class ProgADF4360_89(PLL_IntB, ProgADF4360_X):
    pfdFreqShift = -3
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13.5dBm)", "5.0mA(-10.5dBm)", "7.5mA(-7.5dBm)", "11.0mA(-4.5dBm)"])
    getPreSclDiv = lambda self: 1
    lstTxtDatas, lstEnterSets, titles, useFuncs, regDatas = fltrFuncName(
        ["PreScl", "div2", "div2sel"], ProgADF4360_X)
    titles["NVal"] = ("VCO Counter (N = B)",)
    chgEnterSet(
        lstEnterSets, ("float", "RF_Out", (65.,400.), "updAutoNOrder"))
@partClass("XM-~A5R4-0609")
class ProgXM_A5R4(ProgADF4360_89):
    XM_partName = "XM-~A5R4-0609"
    mfgPartName = "ADF4360-8"
    defData = dict(ProgADF4360_89.defData, **{
        "numRF_Out":120.0, "numOutFreq":120.0, "numNVal": 120,
        "numUsrPFD_Freq":1000.0, "numPFD_Freq": 1000.0})
    defRegs = [0x7802, 0xFF104, 0x3000C9]
    lstTxtDatas, lstEnterSets, titles, useFuncs, regDatas = fltrFuncName(
        ["ACntr"], ProgADF4360_89)
@partClass("XM-~A5R6-0609")
class ProgXM_ADF4360_9(ProgADF4360_89):
    XM_partName = "XM-~A5R6-0609"
    mfgPartName = "ADF4360-9"
    defData = dict(ProgADF4360_89.defData, **{
        "numRF_Out":360.0, "numOutFreq":360.0,"numR_Freq":19.2,
        "numUsrPFD_Freq":1600.0,"numDivOutFreq":90.0,
        "numNVal": 225, "numPFD_Freq": 1600.0})
    defRegs = [0xE10A, 0xFF1A4, 0x300031]
    lstMuxoutOpts = map(isReserv,[
        "DVdd", "Digital Lock Detect", "N Divider Output", "DVdd",
        "R Divider Output", "A Cntr/2 Out", "A Cntr Out", "DGND"])
    muxoutTitle = "Divout"
    lstTxtDatas, lstEnterSets, titles, useFuncs, regDatas = fltrFuncName(
        [], ProgADF4360_89)
    titles["Muxout"] = ("Divout Control",)
    titles["ACntr"] = ("Divout Divider",)
    lstTxtDatas.extend([
        ("DivOutFreq", "num", "%.15g")])
    titles["DivOutFreq"] = ("Freq at Divout",)
    useFuncs["DivOutFreq"] = (0,"MHz",0)
    chgEnterSet(lstEnterSets, ("int", "ACntr", (2, 31), "updDivout"))
    chgEnterSet(lstEnterSets, ("lst", "Muxout", "updDivout"))
    def updDivout(self):
        self.updRegAttrDat(["selMuxout", "numACntr"])
        if self.data.selMuxout not in [5, 6]: return
        self.data.numDivOutFreq = self.data.numOutFreq / self.data.numACntr
        if self.data.selMuxout == 5: self.data.numDivOutFreq /= 2
        self.loadTxtAttr(*self.kTxtData.DivOutFreq)
    def buildFirst(self):
        super(ProgXM_ADF4360_9, self).buildFirst()
        self.updDivout()
"""@partClass("XM-~ADF4360(7*)-0000")
class ProgXM_ADF4360_7(ProgADF4360_X):
    XM_partName = "XM-~ADF4360(7*)-0000"
    mfgPartName = "ADF4360-7"
    defData = dict(ProgADF4360_X.defData, **{
        "numRF_Out":900.0,"numR_Freq":10.0,"numUsrPFD_Freq":200.0})
    defRegs = [0x023212,0x0FF100, 0x1200C9]"""
