from BaseRegPart import *
from time import sleep

"""
Implements:
    ADF4360-0
    ADF4360-1
    ADF4360-2
    ADF4360-3
    ADF4360-4
    ADF4360-5
    ADF4360-6
    ADF4360-7
    ADF4360-8
    ADF4360-9
"""

log = None

@initFunc(__name__.replace("ProgramIt.", ""))
def init(progLog):
    # noinspection PyGlobalUndefined
    global log
    log = progLog

theIntAB_Calc_8_32 = IntAB_Calc(getNamesNVal(), ["selPreScl"], getNamesAB(), "AutoR", 8, 1, 32)
class ProgADF4360_X(BaseProgWithRegs):
    mfgName = "Analog Devices"
    interface = "3-wire"
    catId = CAT_PLL_VCO
    XM_partName = "Unknown"
    mfgPartName = "ADF4360-*"
    lstPreSclOpts = map(isReserv,["8/9", "16/17", "32/33", "Reserved"])
    lstMuxoutOpts = map(isReserv,[
        "Three-State Output", "Digital Lock Detect", "N Divider Output", "DVdd",
        "R Divider Output", "N-Channel Open-Drain Lock Detect",
        "Serial Data Output", "DGND"])
    lstCorePwrOpts = map(isReserv,["5mA", "10mA", "15mA", "20mA"])
    lstLkDtcPOpts = map(isReserv,["3 PFD Cycles","5 PFD Cycles"])
    lstPhDtcPolOpts = map(isReserv,["Negative", "Positive"])
    #CE PIN 0 means only Asynchronous Power Down is selected
    lstPwrDwnOpts = map(isReserv,[
        "Normal Operation", "Asynchronous Power Down",
        "Reserved", "Synchronous Power Down"])
    lstICPnVals = map(str, [.3125*(x+1) for x in xrange(8)])
    lstICP1Opts = map(isReserv, lstICPnVals)
    lstICP2Opts = map(isReserv, lstICPnVals)
    lstOutPwrOpts = map(isReserv,["3.5mA(-14dBm)", "5.0mA(-11dBm)", "7.5mA(-8dBm)", "11.0mA(-5dBm)"])
    lstABPWOpts = map(lambda x: isReserv("%.1fns" % x), [3.0, 1.3, 6.0, 3.0])
    lstBandClkOpts = map(isReserv, ["1", "2", "4", "8"])
    lstAutoROpts = thePFD_Calc.lstAutoROpts
    lstAutoRInfs = thePFD_Calc.lstAutoRInfs
    lstAutoNOpts = theIntAB_Calc_8_32.lstAutoNOpts
    lstAutoNInfs = theIntAB_Calc_8_32.lstAutoNInfs
    #,("bTestMode", "chkTestMode")
    defRegs = [0x023212,0x0FF100, 0x1200C9]
    #MHz, MHz, kHz
    defData = {
        "numRF_Out": 900.0, "numOutFreq": 900.0, "numNVal": 4500,
        "numR_Freq": 10.0, "numUsrPFD_Freq": 200.0, "numPFD_Freq": 200.0,
        "numWriteData": 0, "selAutoR": 1, "bAutoCntrRst": True, "selAutoN": 0,
    }
    infoLabels = simplePllInfoLabels
    defBoxStyle = dict(pos_hint={'x': 0.0, "top": 1.0}, size_hint_y=.08)
    defInnerTop = defBoxStyle["pos_hint"]["top"] - defBoxStyle["size_hint_y"]
    tabLayout = [
        ("Config Mode", "mainTab", (
            BorderedGridLayout, None, dict(cols=2, spacing=[10, 10], padding=[6]), [
                (Bordered, "layoutCfgMd", {}, [
                    (Label, None, dict(defBoxStyle, text="Configuration Mode")),
                    (GridLayout, "gridCfgMd", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 2, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})]),
                (Bordered, "layoutCfgInf", {}, [
                    (Label, None, dict(defBoxStyle, text="Configuration Info")),
                    (GridLayout, "gridCfgInf", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 1, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})])])),
        ("Main", "vcoTab", (
            BorderedGridLayout, "gridMain", dict(cols=2, spacing=[10,10], padding=[6]), [
                (Bordered, "layoutVCO", {"size_hint_x": 1.2}, [
                    (Label, None, dict(defBoxStyle, text="N Counter (VCO)")),
                    (GridLayout, "gridVCO", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 2, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})]),
                (Bordered, "layoutRef", {}, [
                    (Label, None, dict(defBoxStyle, text="Reference (PFD)")),
                    (GridLayout, "gridRef", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 2, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})])])),
        ("Config", "cfgTab", (
            BorderedGridLayout, "gridCfg", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("Charge Pump", "chrgTab", (
            BorderedGridLayout, "gridChrg", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("History", "histTab", (
            BorderedGridLayout, "gridHist", dict(cols=1, spacing=[10, 10], padding=[5, 5, 5, 5]))),
        ("Other", "otherTab", (
            BorderedGridLayout, "gridOther", dict(cols=3, spacing=[10, 10], padding=[20, 10, 5, 5])))
    ]
    mkDefaultTabs = True
    lstDataEntries = [
        (
            InfoListUserDataView(
                "selAutoR", lstAutoROpts, lstAutoRInfs, theInfoListMaker, False,
                "Autoset R From", "Autoset R From"
            ),
            "gridCfgMd"
        ),
        (
            InfoListUserDataView(
                "selAutoN", lstAutoNOpts, lstAutoNInfs, theInfoListMaker, False,
                "Autoset N From", "Autoset N From"
            ),
            "gridCfgMd"
        ),
        (
            NumberUserDataView(
                "numRF_Out", theDecUFloatMakerBig, (float, fromFloat), (350.0, 1800.0),
                "RF Output Frequency", "Desired VCO Output Freq", "MHz"
            ),
            "gridVCO"
        ),
        (
            ListUserDataView(
                "selPreScl", lstPreSclOpts, theListMaker, False,
                "Internal Prescaler (P)", "Internal Prescaler (P)"
            ).setUpstream(RegisterDownstream([(0, 22, 24)])),
            "gridVCO"
        ),
        (
            NumberUserDataView(
                "numNVal", theDecUIntMaker, (int, str), (24, 262143),
                "VCO Counter (N = BP+A)", "VCO Counter (N = BP+A)", None
            ).setMarkup(True),
            "gridVCO"
        ),
        (
            NumberUserDataView(
                "numACntr", theDecUIntMaker, (int, str), (0, 31),
                "A Counter", "A Counter", None
            ).setUpstream(RegisterDownstream([(2, 2, 7)])),
            "gridVCO"
        ),
        (
            NumberUserDataView(
                "numBCntr", theDecUIntMaker, (int, str), (3, 8191),
                "B Counter", "B Counter", None
            ).setUpstream(RegisterDownstream([(2, 8, 21)])),
            "gridVCO"
        ),
        (
            NumberUserDataView(
                "numOutFreq", None, (None, getFormatFn("%.15g")), None,
                "VCO Output Freq (N * F[sub][size=10]PFD[/size][/sub])",
                "VCO Output Freq (N * F[sub][size=10]PFD[/size][/sub])", "MHz", True
            ).setMarkup(True),
            "gridVCO"
        ),
        (
            NumberUserDataView(
                "numR_Freq", theDecUFloatMakerBig, (float, fromFloat), (10.0, 250.0),
                "Reference Frequency", "Reference Freq (F[sub][size=10]REF[/size][/sub])", "MHz"
            ).setMarkup(True),
            "gridRef"
        ),
        (
            NumberUserDataView(
                "numR_Count", theDecUIntMaker, (int, str), (1, 16383),
                "Reference Counter (R)",
                "REF Counter\n(R = F[sub][size=10]REF[/size][/sub]/F[sub][size=10]PFD[/size][/sub])", None
            ).setUpstream(RegisterDownstream([(1, 2, 16)])).setMarkup(True),
            "gridRef"
        ),
        (
            NumberUserDataView(
                "numUsrPFD_Freq", theDecUFloatMakerBig, (float, fromFloat), (10.0, 260000.0),
                "Desired PFD Freq", "Desired PFD Freq", "kHz"
            ),
            "gridRef"
        ),
        (
            NumberUserDataView(
                "numPFD_Freq", None, (None, getFormatFn("%.15g")), None,
                "Actual PFD Freq (F[sub][size=10]PFD[/size][/sub])",
                "Actual PFD Freq (F[sub][size=10]PFD[/size][/sub])", "kHz", True
            ).setMarkup(True),
            "gridRef"
        ),
        (
            ListUserDataView(
                "selMuxout", lstMuxoutOpts, theListMaker, False,
                "Muxout", "Muxout"
            ).setUpstream(RegisterDownstream([(0, 5, 8)])),
            "gridCfg"
        ),
        (
            ListUserDataView(
                "selLkDtcP", lstLkDtcPOpts, theListMaker, False,
                "Lock Detect Precision", "Lock Detect Precision"
            ).setUpstream(RegisterDownstream([(1, 18)])),
            "gridCfg"
        ),
        (
            ListUserDataView(
                "selPhDtcPol", lstPhDtcPolOpts, theListMaker, False,
                "Phase Detector Polarity", "Phase Detector Polarity"
            ).setUpstream(RegisterDownstream([(0, 8)])),
            "gridCfg"
        ),
        (
            ListUserDataView(
                "selCorePwr", lstCorePwrOpts, theListMaker, False,
                "Core Power Level", "Core Power Level"
            ).setUpstream(RegisterDownstream([(0, 2, 4)])),
            "gridCfg"
        ),
        (
            ChkBoxUserDataView(
                "bCntRst",
                "Counter Reset"
            ).setUpstream(RegisterDownstream([(0, 4)])).setDisplay(False),
            "gridCfg"
        ),
        (
            ChkBoxUserDataView(
                "bAutoCntrRst",
                "Reset Counters on Write"
            ),
            "gridCfg"
        ),
        (
            ListUserDataView(
                "selPwrDwn", lstPwrDwnOpts, theListMaker, False,
                "Power Down", "Power Down"
            ).setUpstream(RegisterDownstream([(0, 20, 22)])),
            "gridCfg"
        ),
        (
            ChkBoxUserDataView(
                "mtLkDtc",
                "Mute Until Lock Detect"
            ).setUpstream(RegisterDownstream([(0, 11)])),
            "gridCfg"
        ),
        (
            ChkBoxUserDataView(
                "div2",
                "Divide By 2"
            ).setUpstream(RegisterDownstream([(2, 22)])),
            "gridCfg"
        ),
        (
            ChkBoxUserDataView(
                "div2sel",
                "Divide By 2 select"
            ).setUpstream(RegisterDownstream([(2, 23)])),
            "gridCfg"
        ),
        (
            ListUserDataView(
                "selABPW", lstABPWOpts, theListMaker, False,
                "Anti-Backlash Pulse Width", "Anti-Backlash Pulse Width"
            ).setUpstream(RegisterDownstream([(1, 16, 18)])),
            "gridChrg"
        ),
        (
            ListUserDataView(
                "selBandClk", lstBandClkOpts, theListMaker, False,
                "Band Select Clock", "Band Select Clock"
            ).setUpstream(RegisterDownstream([(1, 20, 22)])),
            "gridCfg"
        ),
        (
            ListUserDataView(
                "selOutPwr", lstOutPwrOpts, theListMaker, False,
                "Output Power Level", "Output Power Level"
            ).setUpstream(RegisterDownstream([(0, 12, 14)])),
            "gridCfg"
        ),
        (
            ChkBoxUserDataView(
                "cpGain",
                "Charge Pump Gain 1"
            ).setUpstream(UnionUpstream([RegisterDownstream([(2, 21)]), RegisterDownstream([(0, 10)])])),
            "gridChrg"
        ),
        (
            ChkBoxUserDataView(
                "chrgPmp3St",
                "Charge Pump Tri-State"
            ).setUpstream(RegisterDownstream([(0, 9)])),
            "gridChrg"
        ),
        (
            ListUserDataView(
                "selICP1", lstICP1Opts, theListMaker, False,
                "Current Setting 1", "Current Setting 1"
            ).setUpstream(RegisterDownstream([(0, 14, 17)])),
            "gridChrg"
        ),
        (
            ListUserDataView(
                "selICP2", lstICP2Opts, theListMaker, False,
                "Current Setting 2", "Current Setting 2"
            ).setUpstream(RegisterDownstream([(0, 17, 20)])),
            "gridChrg"
        ),
        (
            NumberUserDataView(
                "numDelayWrite", theDecUFloatMaker, (float, getFormatFn("%.12g")), (0.0, 1e3),
                "Number of milliseconds to delay", "Write Delay", "ms"
            ),
            "gridCfg"
        ),
        (
            NumberUserDataView(
                "numWriteData", theHexUIntMaker, (fromHexStr, getFormatFn("%06X")), (0, 0xFFFFFF),
                "Enter Register Hex", "Enter Register Hex", None
            ),
            "gridOther"
        )
    ]
    cntrRstName = "bCntRst"
    autoCntrRstName = "bAutoCntrRst"
    addUserLog = "gridHist"
    binLayoutStyle = dict(pos_hint={'y': .02, "right": .99}, size_hint=(.79, .15), cols=3)
    regMaker = Reg24
    regWordSz = regMaker.wordSz
    regCtlSz = regMaker.ctlSz
    regDuals = [-1] * 3
    # self.progBtn.pos_hint={'y':0.045*app.defStyle["size_hint"][1],'x':.01}
    dctCalcSystemUpd = {
        "numR_Freq": ("AutoR", False),
        "numNVal": ("AutoN", False),
        "numACntr": ("AutoN", False),
        "numBCntr": ("AutoN", False),
        "numRF_Out": ("AutoN", False),
        "selPreScl": ("AutoN", False),
        "numUsrPFD_Freq": ("AutoR", False),
        "numR_Count": ("AutoR", False),
        "selAutoN": ("AutoN", True),
        "selAutoR": ("AutoR", True),
    }
    calcSystems = {
        "AutoN": theIntAB_Calc_8_32,
        "AutoR": (theMilliPFD_Calc, "AutoN"),
    }
    # self.gridOther.add_widget(sizedButton(text=" Write Register ",on_release=self.writeReg))

    tabNameNoWrite = {"otherTab"}

    def postBuildFirst(self):
        self.dctNamed["gridOther"].add_widget(Button(text="Write Register", on_release=self.writeSingleReg))
    def addAllRegs(self, binLayout):
        for regId in [0, 1, 2]:
            binLayout.add_widget(self.getReg(regId).build())
    def outputReg(self, regId):
        if regId == 2 or (isinstance(regId, tuple) and regId[0] == 2):
            num = self.getData("numDelayWrite")
            sleep(num / 1e3)
            self.userLog.log("delaying %s for N Counter Latch" % unitSeconds(num, -3))
        return super(ProgADF4360_X, self).outputReg(regId)
    def getWriteRegSequence(self):
        return [1, 0, 2]
#get_enterSetHex(ProgADF4360_X, "WriteData", (0, 0xFFFFFF), "Enter Register Hex")
@partClass("XM-A5P5-0609")
class ProgXM_A5P5(ProgADF4360_X):
    XM_partName = "XM-A5P5-0609"
    mfgPartName = "ADF4360-0"
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13dBm)", "5.0mA(-11dBm)", "7.5mA(-8dBm)", "11.0mA(-5dBm)"])
    lstDataEntries, defData = changeLstEntries(ProgADF4360_X.lstDataEntries, ProgADF4360_X.defData, {
        "selOutPwr": {
            "lstOpts": lstOutPwrOpts
        },
        "numRF_Out": {
            "varRange": (2400.0, 2725.0)
        }
    })
    defData.update({
        "numRF_Out": 2500.0, "numOutFreq": 2500.0, "numNVal": 12500})
    defRegs = [0x61A12, 0xFF104, 0x3000C9]
@partClass("XM-A5P6-0609")
class ProgXM_A5P6(ProgADF4360_X):
    XM_partName = "XM-A5P6-0609"
    mfgPartName = "ADF4360-1"
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13dBm)", "5.0mA(-10.5dBm)", "7.5mA(-8dBm)", "11.0mA(-5dBm)"])
    lstDataEntries, defData = changeLstEntries(ProgADF4360_X.lstDataEntries, ProgADF4360_X.defData, {
        "selOutPwr": {
            "lstOpts": lstOutPwrOpts
        },
        "numRF_Out": {
            "varRange": (2050.0, 2450.0)
        }
    })
    defData.update({
        "numRF_Out": 2400.0, "numOutFreq": 2400.0, "numNVal": 12000})
    defRegs = [0x5DC02, 0xFF108, 0x3000C9]
@partClass("XM-A5P7-0609")
class ProgXM_A5P7(ProgADF4360_X):
    #REFERENCE COUNTER intervals are weird in data sheet
    XM_partName = "XM-A5P7-0609"
    mfgPartName = "ADF4360-2"
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13dBm)", "5.0mA(-11dBm)", "7.5mA(-8dBm)", "11.0mA(-5dBm)"])
    lstDataEntries, defData = changeLstEntries(ProgADF4360_X.lstDataEntries, ProgADF4360_X.defData, {
        "selOutPwr": {
            "lstOpts": lstOutPwrOpts
        },
        "numRF_Out": {
            "varRange": (1850.0, 2170.0)
        }
    })
    defData.update({
        "numRF_Out": 2200.0, "numOutFreq": 2200.0, "numNVal": 11000})
    defRegs = [0x55F02, 0xFF108, 0x3000C9]
@partClass("XM-A5P8-0609")
class ProgXM_A5P8(ProgADF4360_X):
    #REFERENCE COUNTER intervals are weird in data sheet
    XM_partName = "XM-A5P8-0609"
    mfgPartName = "ADF4360-3"
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-12dBm)", "5.0mA(-9dBm)", "7.5mA(-6dBm)", "11.0mA(-3dBm)"])
    lstDataEntries, defData = changeLstEntries(ProgADF4360_X.lstDataEntries, ProgADF4360_X.defData, {
        "selOutPwr": {
            "lstOpts": lstOutPwrOpts
        },
        "numRF_Out": {
            "varRange": (1600.0, 1950.0)
        }
    })
    defData.update({
        "numRF_Out": 1800.0, "numOutFreq": 1800.0, "numNVal": 9000})
    defRegs = [0x46502, 0xFF108, 0x3000C9]
@partClass("XM-A5P9-0609")
class ProgXM_A5P9(ProgADF4360_X):
    XM_partName = "XM-A5P9-0609"
    mfgPartName = "ADF4360-4"
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13dBm)", "5.0mA(-10dBm)", "7.5mA(-7dBm)", "11.0mA(-4dBm)"])
    lstDataEntries, defData = changeLstEntries(ProgADF4360_X.lstDataEntries, ProgADF4360_X.defData, {
        "selOutPwr": {
            "lstOpts": lstOutPwrOpts
        },
        "numRF_Out": {
            "varRange": (1450.0, 1750.0)
        }
    })
    defData.update({
        "numRF_Out": 1600.0, "numOutFreq": 1600.0, "numNVal": 8000})
    defRegs = [0x3E802, 0xFF108, 0x3000C9]
@partClass("XM-A5R1-0609")
class ProgXM_A5R1(ProgADF4360_X):
    #REFERENCE COUNTER intervals are weird in data sheet
    XM_partName = "XM-A5R1-0609"
    mfgPartName = "ADF4360-5"
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13.5dBm)", "5.0mA(-10.5dBm)", "7.5mA(-7.5dBm)", "11.0mA(-4.5dBm)"])
    lstDataEntries, defData = changeLstEntries(ProgADF4360_X.lstDataEntries, ProgADF4360_X.defData, {
        "selOutPwr": {
            "lstOpts": lstOutPwrOpts
        },
        "numRF_Out": {
            "varRange": (1200.0, 1400.0)
        }
    })
    defData.update({
        "numRF_Out": 1400.0, "numOutFreq": 1400.0, "numNVal": 7000})
    defRegs = [0x56B02, 0xFF104, 0x3000C9]
@partClass("XM-A5R2-0609")
class ProgXM_A5R2(ProgADF4360_X):
    #REFERENCE COUNTER intervals are weird in data sheet
    XM_partName = "XM-A5R2-0609"
    mfgPartName = "ADF4360-6"
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13.5dBm)", "5.0mA(-10.5dBm)", "7.5mA(-7.5dBm)", "11.0mA(-4.5dBm)"])
    lstDataEntries, defData = changeLstEntries(ProgADF4360_X.lstDataEntries, ProgADF4360_X.defData, {
        "selOutPwr": {
            "lstOpts": lstOutPwrOpts
        },
        "numRF_Out": {
            "varRange": (1050.0, 1250.0)
        }
    })
    defData.update({
        "numRF_Out": 1200.0, "numOutFreq": 1200.0, "numNVal": 6000})
    defRegs = [0x2EE02, 0xFF104, 0x3000C9]
@partClass("XM-A5R3-0609")
class ProgXM_A5R3(ProgADF4360_X):
    XM_partName = "XM-A5R3-0609"
    mfgPartName = "ADF4360-7"
    defRegs = [0x023212, 0x0FF100, 0x1200C9]
class ProgADF4360_89(ProgADF4360_X):
    lstOutPwrOpts = map(isReserv,[
        "3.5mA(-13.5dBm)", "5.0mA(-10.5dBm)", "7.5mA(-7.5dBm)", "11.0mA(-4.5dBm)"])
    calcSystems, dctCalcSystemUpd = remWidCalcSys(ProgADF4360_X.calcSystems, ProgADF4360_X.dctCalcSystemUpd, {
        "selPreScl", "div2", "div2sel", "numACntr"}, [])
    setCalc(calcSystems, "AutoN", theIntB_Calc)
    setCalc(calcSystems, "AutoR", theMilliPFD_Calc)
    lstDataEntries, defData = changeLstEntries(ProgADF4360_X.lstDataEntries, ProgADF4360_X.defData, {
        "selOutPwr": {
            "lstOpts": lstOutPwrOpts
        },
        "numRF_Out": {
            "varRange": (65.0, 400.0)
        },
        "numNVal": {
            "shortTitle": "VCO Counter (N = B)",
            "longTitle": "VCO Counter (N = B)"
        },
        "selPreScl": "delete",
        "div2": "delete",
        "div2sel": "delete"
    })
@partClass("XM-A5R4-0609")
class ProgXM_A5R4(ProgADF4360_89):
    XM_partName = "XM-A5R4-0609"
    mfgPartName = "ADF4360-8"
    lstDataEntries, defData = changeLstEntries(ProgADF4360_89.lstDataEntries, ProgADF4360_89.defData, {
        "numACntr": "delete"
    })
    defData.update({
        "numRF_Out": 120.0, "numOutFreq": 120.0, "numNVal": 120,
        "numUsrPFD_Freq": 1000.0, "numPFD_Freq": 1000.0})
    defRegs = [0x7802, 0xFF104, 0x3000C9]
class DivoutCalc(BaseCalc):
    def updateMode(self, context):
        """
        :param ProgXM_ADF4360_9 context:
        """
    def update(self, context):
        """
        :param ProgXM_ADF4360_9 context:
        """
        selMuxout = context.getData("selMuxout")
        if selMuxout not in [5, 6]: return
        numOutFreq = context.getData("numOutFreq")
        numACntr = context.getData("numACntr")
        numDivOutFreq = numOutFreq / numACntr
        if selMuxout == 5: numDivOutFreq /= 2
        context.setData("numDivOutFreq", numDivOutFreq, DC_SYNCH)
theDivoutCalc = DivoutCalc()

@partClass("XM-A5R6-0609")
class ProgXM_ADF4360_9(ProgADF4360_89):
    XM_partName = "XM-A5R6-0609"
    mfgPartName = "ADF4360-9"
    lstMuxoutOpts = map(isReserv,[
        "DVdd", "Digital Lock Detect", "N Divider Output", "DVdd",
        "R Divider Output", "A Cntr/2 Out", "A Cntr Out", "DGND"])
    # muxoutTitle = "Divout"
    calcSystems = dict(ProgADF4360_89.calcSystems)
    dctCalcSystemUpd = dict(ProgADF4360_89.dctCalcSystemUpd)
    calcSystems.update({
        "Divout": theDivoutCalc
    })
    addAfterCalcSys(calcSystems, "AutoN", "Divout")
    dctCalcSystemUpd.update({
        "numACntr": ("Divout", False),
        "selMuxout": ("Divout", False)
    })
    lstDataEntries, defData = changeLstEntries(ProgADF4360_89.lstDataEntries, ProgADF4360_89.defData, {
        "numACntr": {
            "varRange": (2, 31),
            "shortTitle": "Divout Divider",
            "longTitle": "Divout Divider"
        },
        "selMuxout": {
            "shortTitle": "Divout Control",
            "longTitle": "Divout Control",
            "lstOpts": lstMuxoutOpts,
        }
    })
    lstDataEntries.extend([
        (
            NumberUserDataView(
                "numDivOutFreq", None, (None, fromFloat), None,
                "Freq at Divout", "Freq at Divout", "MHz", True
            ),
            "gridVCO"
        )
    ])
    defData.update({
        "numRF_Out": 360.0, "numOutFreq": 360.0, "numR_Freq": 19.2,
        "numUsrPFD_Freq": 1600.0, "numDivOutFreq": 90.0,
        "numNVal": 225, "numPFD_Freq": 1600.0})
    defRegs = [0xE10A, 0xFF1A4, 0x300031]
"""@partClass("XM-ADF4360(7*)-0000")
class ProgXM_ADF4360_7(ProgADF4360_X):
    XM_partName = "XM-ADF4360(7*)-0000"
    mfgPartName = "ADF4360-7"
    defData = dict(ProgADF4360_X.defData, **{
        "numRF_Out":900.0,"numR_Freq":10.0,"numUsrPFD_Freq":200.0})
    defRegs = [0x023212,0x0FF100, 0x1200C9]"""
