from BaseADF_PLL import *
from math import ceil, floor
import time

"""
Implements:
    ADF4355
    ADF5355
    ADF4355-2
    ADF4355-3
    ADF4155
    ADF4356
    ADF5356
"""

log = None

@initFunc(__name__.replace("ProgramIt.", ""))
def init(progLog):
    global log
    log = progLog


dBmCnv = "%.15gdBm"
dBmify = lambda x: isReservStr(dBmCnv, x)
class ProgADFX35X(PLL_MulDivPFD, PLL_AuxMod, PLL_IntFrac2, BaseProgADF_PLL):
    mfgName = "Analog Devices"
    interface = "3-wire"
    catId = CAT_PLL_VCO
    XM_partName = "Unknown"
    mfgPartName = "ADF*35*"
    #0382
    extraBindFuncs = {
        "addPhLkLine": ("addSmallLine", "gridPhLk"),
        "addPhLkChk": ("addSmallChkLine", "gridPhLk"),
        "mkSmPhLkChk": ("addSmChkLine", "gridPhLk"),
        "mkSmPhLkLst": ("addSmallLstLine", "gridPhLk"),
        "mkModChk": ("addChkLine", "gridMod"),
        "addModLine": ("addLine", "gridMod"),
        "addModLineM": ("addLineM", "gridMod"),
        "addModLst": ("addLstLine", "gridMod"),
        "addSmOtherLn": ("addSmallLine", "gridOther"),
        "mkSmCfgMdChk": ("addSmChkLine", "gridCfgMd"),
        "addCfgMdLst": ("addSmallLstLine", "gridCfgMd")}
    lstBindFuncs = dict(BaseProgADF_PLL.lstBindFuncs, **extraBindFuncs)
    #chk funs 7, 8, 9, 10
    lstChkFuncs = BaseProgADF_PLL.lstChkFuncs + [
        "addPhLkChk", "mkSmPhLkChk", "mkModChk", "mkSmCfgMdChk"]
    lstChkBigger = BaseProgADF_PLL.lstChkBigger + [7, 9]
    #line func 8, 9, 10, 11
    lstLineFuncs = BaseProgADF_PLL.lstLineFuncs + [
        "addPhLkLine", "addModLine", "addModLineM", "addSmOtherLn"]
    lstLineBig = BaseProgADF_PLL.lstLineBig + [9, 10]
    #lst func 4, 5, 6
    lstLstFuncs = BaseProgADF_PLL.lstLstFuncs + [
        "mkSmPhLkLst", "addModLst", "addCfgMdLst"]
    lstLstBig = BaseProgADF_PLL.lstLstBig + [5]
    lstPreSclOpts = map(isReserv,["4/5 (N > 23)", "8/9 (N > 75)"])
    lstChrgPmpVals = [.3125*(x+1) for x in xrange(16)]
    lstChrgPmpOpts = [isReservStr("%.15g mA",x) for x in lstChrgPmpVals]
    lstLDFOpts = map(isReserv,["Factional N", "Integer N"])
    lstLDPOpts = map(isReserv,["5ns", "6ns", "8ns", "12ns"])
    lstOutPwrOpts = map(dBmify, [-4, -1, 2, 5])
    lstAuxOutPwrOpts = map(dBmify, [-4, -1, 2, 5])
    lstDivRFOpts = [isReservStr(u"\u00F7%u", 2 ** x) for x in xrange(7)]
    lstMuxoutOpts = map(isReserv,[
        "Three-State Output", "DVdd", "SDgnd", "R Divider Output",
        "N Divider Output", "Analog Lock Detect", "Digital Lock Detect",
        "Reserved"])
    lstMuxVoltOpts = map(isReserv,["1.8V", "3.3V"])
    lstPhDtcPolOpts = map(isReserv,["Negative", "Positive"])
    lstAuxOutOpts = map(isReserv, ["Divided Output", "Fundamental"])
    lstFdBkSelOpts = map(isReserv, ["Divided", "Fundamental"])
    lstRefMdOpts = map(isReserv,["Single", "Differential"])
    lstLESyncOpts = map(isReserv,["Disabled", "REFin"])
    lstLDCycleCntOpts = [isReservStr("%u", 1024<<x) for x in xrange(4)]
    useFuncs = {
        "R_Freq": (2,"MHz",1),
        "PFD_Freq": (2,"MHz",0),
        "RF_Out": (0,"MHz",1),
        "SpaceCh": (9,"",0),
        "UsrSpaceCh": (9,"kHz",1),
        "FreqADC": (4,0),
        "BldCur": (5,0),
        "Mod2Val": (9,"",1),
        "IntVal": (0,"",1),
        "Frac1Val": (0,"",1),
        "Frac2Val": (0,"",1),
        "NVal": (0,"",1),
        "OutFreq": (0,"MHz",0),
        "Phase":(8,1),
        "phaseAdjust": (8,"updReg3Dat"),
        "PhaseTimeout": (8,0),
        "Muxout": (3,),
        "MuxVolt": (3,),
        "PreScl": (5,),
        "RefMd": (3,),
        "phaseResync": (8,"updReg3Dat"),
        "bAutoCal": (2,"updReg0Dat"),
        "bSDReset": (2,"updReg3Dat"),
        "R_Count": (2,"",1),
        "bADC_Conv": (2,"updReg10Dat"),
        "dblBfr": (6,"updReg4Dat"),
        "bADC_On": (2,"updReg10Dat"),
        "bLOL": (8,"updReg7Dat"),
        "PhaseRsynClk": (8,1),
        "ADC_ClkDiv": (4,1),
        "bRefDbl": (0,"","updPFD_Freq"),
        "bRefDiv": (0,"","updPFD_Freq"),
        "ChrgPmp": (2,),
        "LDF": (4,),
        "LDP": (4,),
        "PhDtcPol": (4,),
        "pwrDwn": (6,"updReg4Dat"),
        "BldCP": (5,1),
        "bGateBld": (4,"updReg6Dat"),
        "bNegBld": (4,"updReg6Dat"),
        "chrgPmp3St": (4,"updReg4Dat"),
        #"cntrRst": (2,"updReg4Dat"),
        "bAutoCntrRst": (2,None),
        "FdBkSel": (5,), "DivRF": (5,1),
        "bMTLD": (6,"updReg6Dat"),
        "bAuxOut": (6,"updReg6Dat"), "bRFOut": (6,"updReg6Dat"),
        "AuxOutPwr": (3,), "OutPwr": (3,),
        "LESync": (3,), "SynLkTimeout": (4,1),
        "WaitALC": (4,1), "DivBndVCO": (4,1), "Timeout": (4,1),
        "LDCycleCnt": (4,), "bAutoBld": (4,"updNegBld"),
        "bAutoADC": (2,"updADC"), "bAutoFast": (2,"updCalib"),
        "TotalCalTm": (4,0),
        "UsrPFD_Freq": (2,"MHz",1),
        "WriteData": (11,1),
        "bAutoSelDiv":(10,"updAutoSelDiv")}
    titles = {
        "R_Freq": ("Reference Freq",),
        "PFD_Freq": ("Actual PFD Freq",),
        "RF_Out": ("Desired Output Freq",),
        "R_Count": ("Ref Counter (R)",),
        "bRefDbl": ("Reference Doubler",),
        "bRefDiv": ("Reference /2",),
        "PreScl": ("Internal Prescaler",),
        "Mod2Val": ("Auxiliary Modulus",),
        "IntVal": ("Integer Value",),
        "Frac1Val": ("Fractional Value",),
        "Frac2Val": ("Auxiliary Fraction",),
        "RefMd": ("REFin Mode",),
        "phaseAdjust": ("Phase Adjust",),
        "Muxout": ("Muxout",),
        "bSDReset": ("SD Load Reset",),
        "bAutoCal": ("Autocal",),
        "phaseResync": ("Phase Resync Enable",),
        "Phase": ("Phase",),
        "dblBfr": ("Double Buffer",),
        "ChrgPmp": ("CP Current","Charge Pump Current Setting"),
        "BldCP": ("Bleed Value","Charge Pump Bleed Value"),
        "BldCur": ("Bleed Current",),
        "LDF": ("LDF", "Lock Detect Function"),
        "bGateBld": ("Gated Bleed",),
        "LDP": ("LDP", "Lock Detect Precision"),
        "bNegBld": ("Negative Bleed",),
        "PhDtcPol": ("PD Polarity", "Phase Detector Polarity"),
        "pwrDwn": ("Power Down",),
        "chrgPmp3St": ("Charge Pump 3-State",),
        #"cntrRst": ("Counter Reset",),
        "bAutoCntrRst": ("Reset Counters on Write",),
        "bLOL": ("LOL Enable",),
        "FdBkSel": ("Feedback Select",),
        "DivRF": ("RF Div Select",),
        "bADC_On": ("ADC Enable",),
        "bADC_Conv": ("ADC Conversion",),
        "PhaseRsynClk": ("Phase Resync",),
        "bMTLD": ("Mute Till LD", "Mute Till Lock Detect"),
        "bAuxOut": ("RF B OUT Enable",),
        "bRFOut": ("RF A OUT Enable",),
        "AuxOutPwr": ("RF B OUT Power", "RF B Output Power"),
        "OutPwr":("RF A OUT Power", "RF A Output Power"),
        "SpaceCh": ("Actual Channel Spacing",),
        "OutFreq": ("Actual Output Freq",),
        "MuxVolt": ("Mux Level", "Muxout Logic Voltage"),
        "SynLkTimeout": ("Synth. Lock Timout",),
        "DivBndVCO": ("VCO Band Div",),
        "Timeout": ("Timeout",),
        "WaitALC": ("ALC Timeout",),
        "ADC_ClkDiv": ("ADC Clock",),
        "FreqADC": ("ADC Frequency",),
        "LESync": ("LE Sync",),
        "LDCycleCnt": ("LD Cycle Count", "Lock Detect Cycle Count"),
        "NVal": ("N = ", "N Value"),
        "bAutoBld": ("Autoset Bleed",),
        "bAutoADC": ("Autoset ADC",),
        "bAutoFast": ("Autoset Fastest",),
        "TotalCalTm": ("Total Cal. Time",),
        "PhaseTimeout": ("Phase Timeout",),
        "UsrPFD_Freq": ("Desired PFD Freq",),
        "UsrSpaceCh": ("Desired Freq Step Size",),
        "WriteData": ("Enter Register Hex",),
        "bAutoSelDiv": ("Auto select RF Divider",)}
    lstTxtDatas = [
        ("R_Freq", "num", "%.15g"), ("RF_Out", "num", "%.15g"),
        ("UsrSpaceCh", "num", "%.15g"),
        ("DivBndVCO", "num"), ("Timeout", "num"), ("WaitALC", "num"),
        ("SynLkTimeout", "num"), ("ADC_ClkDiv", "num"),
        ("bAutoCal", "chkAutoCal"), ("bADC_On", "chkADC_On"),
        ("bADC_Conv", "chkADC_Conv"),
        #("cntrRst", "chkCntrRst"),
        ("bAutoCntrRst", "chkAutoCntrRst"),
        ("bSDReset", "chkSDReset"), ("UsrPFD_Freq", "num", "%.15g"),
        ("bGateBld", "chkGateBld"), ("bNegBld", "chkNegBld"),
        ("BldCP", "num"), ("ChrgPmp", "lst"), ("chrgPmp3St", "chkCP3St"),
        ("Muxout", "lst"), ("MuxVolt", "lst"),
        ("bAuxOut", "chkAuxOut"), ("AuxOutPwr", "lst"),
        ("bRFOut", "chkRFOut"), ("OutPwr", "lst"),
        ("RefMd", "lst"), ("LESync", "lst"),
        ("bMTLD", "chkMTLD"), ("dblBfr", "chkDblBfr"), ("pwrDwn", "chkPwrDwn"),
        ("PhaseRsynClk", "num"), ("PhaseTimeout", "num", Bind(unitSeconds,e10=-6)),
        ("Phase", "num"), ("LDF", "lst"), ("bLOL", "chkLOL"), ("LDP", "lst"),
        ("phaseResync", "chkPhaseResync"), ("LDCycleCnt", "lst"),
        ("phaseAdjust", "chkPhaseAdjust"), ("PhDtcPol", "lst"),
        ("bRefDbl", "chkRefDbl"), ("R_Count", "num"), ("bRefDiv", "chkRefDiv2"),
        ("PFD_Freq", "num", "%.15g"), ("NVal", "num", "%.17g"),
        ("IntVal", "num"), ("Frac1Val", "num"), ("Frac2Val", "num"),
        ("Mod2Val", "num"), ("OutFreq", "num", "%.15g"),
        ("bAutoFast", "chkAutoFast"), ("bAutoADC", "chkAutoADC"),
        ("bAutoBld", "chkAutoBld"),
        ("TotalCalTm", "num", Bind(unitSeconds,e10=-6)),
        ("SpaceCh", "num", Bind(unitHertz, e10=3,fmtBase="%.6g")),
        ("PreScl", "lst"), ("DivRF", "lst"),
        ("FdBkSel", "lst"), ("FreqADC", "num", Bind(unitHertz, e10=3)),
        ("BldCur", "num", u"%.15g \u03BCA"), ("WriteData", "num", "%08X"),
        ("bAutoSelDiv", "chkAutoSelDiv")]
    regDatas = {
        "bAutoCal":[(0,21)], "selPreScl":[(0,20)], "numIntVal":[(0,4,20)],
        "numFrac1Val":[(1,4,28)], "numFrac2Val":[(2,18,32)],
        "numMod2Val":[(2,4,18)], "numPhase":[(3,4,28)], "phaseAdjust":[(3,28)],
        "phaseResync":[(3,29)], "bSDReset":[(3,30)], "selMuxout":[(4,27,30)],
        "bRefDbl":[(4,26)], "bRefDiv":[(4,25)], "numR_Count":[(4,15,25)],
        "dblBfr":[(4,14)], "selChrgPmp":[(4,10,14)], "selRefMd":[(4,9)],
        "selMuxVolt":[(4,8)], "selPhDtcPol":[(4,7)], "pwrDwn":[(4,6)],
        "chrgPmp3St":[(4,5)], "cntrRst":[(4,4)], "bGateBld":[(6,30)],
        "bNegBld":[(6,29)], "selFdBkSel":[(6,24)], "selDivRF":[(6,21,24)],
        "numBldCP":[(6,13,21)], "bMTLD":[(6,11)], "bAuxOut":[(6,9)],
        "selAuxOutPwr":[(6,7,9)], "bRFOut":[(6,6)], "selOutPwr":[(6,4,6)],
        "selLESync":[(7,25)], "selLDCycleCnt":[(7,8,10)], "bLOL":[(7,7)],
        "selLDP":[(7,5,7)], "selLDF":[(7,4)], "numDivBndVCO":[(9,24,32)],
        "numTimeout":[(9,14,24)], "numWaitALC":[(9,9,14)],
        "numSynLkTimeout":[(9,4,9)], "numADC_ClkDiv":[(10,6,14)],
        "bADC_Conv":[(10,5)], "bADC_On":[(10,4)],
        "numPhaseRsynClk":[(12,16,32)]}
    lstEnterSets = [
        ("float", "R_Freq", (10.,600.), "updPFD_Freq"),
        ("float", "UsrPFD_Freq", (10.,260000.), "updPFD_Freq"),
        ("float", "RF_Out", (54.,13600.), "updAutoNOrder"),
        ("float", "UsrSpaceCh", (1.0,18000.), "updModSpace"),
        ("float", "NVal", (23.0,65536.), "updAutoNOrder"),
        ("int", "Mod2Val", (2, 16383), "updModSpace"),
        ("int", "IntVal", (23, 65535), "updAutoNOrder"),
        ("int", "Frac1Val", (0, 16777215), "updAutoNOrder"),
        ("int", "Frac2Val", (0, 16383), "updAutoNOrder"),
        ("int", "R_Count", (1, 1023), "updPFD_Freq"),
        ("float", "FreqADC", BaseRange(False)),
        ("float", "BldCur", BaseRange(False)),
        ("float", "PhaseTimeout", BaseRange(False)),
        ("lst", "PreScl", "updReg0Dat"),
        ("lst", "Muxout", "updReg4Dat"),
        ("lst", "MuxVolt", "updReg4Dat"),
        ("lst", "OutPwr", "updReg6Dat"),
        ("lst", "AuxOutPwr", "updReg6Dat"),
        ("lst", "ChrgPmp", "updNegBld"),
        ("lst", "LDF", "updReg7Dat"),
        ("lst", "LDP", "updReg7Dat"),
        ("lst", "PhDtcPol", "updReg4Dat"),
        ("lst", "FdBkSel", "updAutoNOrder"),
        ("lst", "DivRF", "updAutoNOrder"),
        ("lst", "RefMd", "updReg4Dat"),
        ("lst", "LESync", "updReg7Dat"),
        ("lst", "LDCycleCnt", "updReg7Dat"),
        ("int", "DivBndVCO", (1, 255), "updCalib"),
        ("int", "Timeout", (2, 1023), "updCalib"),
        ("int", "WaitALC", (1, 31), "updCalib"),
        ("int", "SynLkTimeout", (2, 31), "updCalib"),
        ("int", "ADC_ClkDiv", (1, 255), "updADC"),
        ("int", "PhaseRsynClk", (0, 65535), "updPhaseRsyn"),
        ("int", "Phase", (0, 16777215), "updReg3Dat"),
        ("int", "BldCP", (0, 255), "updNegBld"),
        ("hex", "WriteData", (0,2**32-1), None)]
    defRegs = [
        0x00200510, 0x06155551, 0x00040032, 0x3, 0x30008B84, 0x800025, 0x34018076,
        0x12000007, 0x102D0428, 0x2A29FCC9, 0xC03EBA, 0x61300B, 0x1041C]
    #remove last 4 elements in list to restore defaults
    defData = {
        #"numR_Freq":122.88,"numUsrPFD_Freq":61.44,
        "numR_Freq":100.,"numUsrPFD_Freq":100.,"numPFD_Freq":100.0,
        "numRF_Out":5000.0,"numOutFreq":5000.0,"numNVal":50.0,
        "numWriteData":0, "bAutoFast": True, "bAutoADC": True,
        "numSpaceCh": 1.1920928955078125e-05,
        "numFreqADC":99.8003992015968,
        "numBldCur":75.0,
        "numTotalCalTm": 2829.42,
        "numPhaseTimeout": 0.01,
        "bAutoBld":True,"bAutoCntrRst":True,
        "numUsrSpaceCh":200.,"numMod1Val":2**24,
        "bAutoSelDiv":True}
    #defData["numSpaceCh"] = 1.1920928955078125e-05
    rangeVCO_Freq = (3400.,6800.)
    nRegs = 13
    defBoxStyle = dict(pos_hint={'x': 0.0, "top": 1.0}, size_hint_y=.08)
    defInnerTop = defBoxStyle["pos_hint"]["top"] - defBoxStyle["size_hint_y"]
    tabLayout = [
        ("Config Mode", "mainTab", (
            BorderedGridLayout, None, dict(cols=2, spacing=[10, 10], padding=[6]), [
                (Bordered, "layoutCfgMd", {}, [
                    (Label, None, dict(defBoxStyle, text="Configuration Mode")),
                    (GridLayout, "gridCfgMd", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 2, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})]),
                (Bordered, "layoutCfgInf", {}, [
                    (Label, None, dict(defBoxStyle, text="Configuration Info")),
                    (GridLayout, "gridCfgInf", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 1, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})])])),
        ("Main", "vcoTab", (
            GridSelFrameLayout, "gridMain", dict(
                rows=1, spacing=[10, 10], padding=[6], side_names=["VCO", "PFD", "Misc\n   /\nMOD"]), [
                (Bordered, "layoutMod", {}, [
                    (Label, None, dict(defBoxStyle, text="Channel Spacing (Misc/Mod)")),
                    (GridLayout, "gridMod", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 3, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})], -4),
                (Bordered, "layoutRef", {}, [
                    (Label, None, dict(defBoxStyle, text="Reference (PFD)")),
                    (GridLayout, "gridRef", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 3, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})], -4),
                (Bordered, "layoutVCO", {"size_hint_x": 1.2}, [
                    (Label, "lblNCntr", dict(defBoxStyle, text="N Counter (VCO)")),
                    (GridLayout, "gridVCO", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 3, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})], -4)])),
        ("Config", "cfgTab", (
            BorderedGridLayout, "gridCfg", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("Phase/Lock", "phLkTab", (
            BorderedGridLayout, "gridPhLk", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("Charge Pump", "chrgTab", (
            BorderedGridLayout, "gridChrg", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("IN/OUT/PWR", "outTab", (
            BorderedGridLayout, "gridOut", dict(cols=4, spacing=[10, 10], padding=[20, 10, 5, 5]))),
        ("History", "histTab", (
            BorderedGridLayout, "gridHist", dict(cols=1, spacing=[10, 10], padding=[5, 5, 5, 5]))),
        ("Other", "otherTab", (
            BorderedGridLayout, "gridOther", dict(cols=3, spacing=[10, 10], padding=[20, 10, 5, 5])))
    ]
    mkDefaultTabs = True
    def __init__(self, fl=None, ver=0):
        if not hasattr(self,"regs"):
            self.regs = [Reg32_4(x) for x in xrange(self.nRegs)]
        #print self.data.__dict__.get("numR_Count", "Not Available")
        super(ProgADFX35X, self).__init__(fl, ver)
        self.buildFirst()
        self.tabNoWrite.add(self.otherTab)
        self.tabPanel.bind(current_tab=self.curTabEvt)
    def buildFirst(self):
        self.tabPanel = self.getTabPanel()
        defStyle = {"size_hint":(1.0,None), "height":32}
        dctLbls = {
            "AutoN": "Select the method of entering the N Counter value",
            "AutoMod": "Select the method of entering modulus",
            "AutoR": "Select the method of entering the R Counter value",
            "bAutoSelDiv": "check to auto select the RF Out divider",
            "CalOut": "select which output you want to calc. freq. for"
        }
        self.addCfgMdLbls(self.gridCfgInf, defStyle, dctLbls)
        self.fillGridHist()
        self.buildTxtData(True)
        self.gridOther.add_widget(sizedButton(
            text=" Write Register ",on_release=self.writeReg))
        self.gridOther.add_widget(sizedLabel(text=""))
        #self.gridOther.add_widget(sizedButton(text=" Load Into Program ", on_release=self.loadReg))
        self.gridOther.add_widget(Button(
            text="Reset To Defaults",
            on_release=self.resetToDefs,size_hint_y=None,height=36))
        #if hasattr(self, "inDivRF"): self.inDivRF.disabled = True
        self.loadTxtData()
        self.add12Regs()
        for c in xrange(len(self.regs)): self.updRegNDat(c)
        self.updAutoN()
        self.updAutoR()
        self.updAutoMod()
        self.updPFD_Freq()
        self.updAutoSelDiv()
    def updAutoSelDiv(self):
        self.inDivRF.disabled = self.data.bAutoSelDiv
        self.updAutoNOrder()
    def moveNCnt(self, obj):
        swapWidgets(self.gridSelFrame, self.layoutVCO)
    def moveMod(self, obj):
        swapWidgets(self.gridSelFrame, self.layoutMod)
    def moveRef(self, obj):
        swapWidgets(self.gridSelFrame, self.layoutRef)
    def add12Regs(self):
        self.binLayout = BorderedGridLayout(
            pos_hint={'y':.02, "right":.99},size_hint=(.79,.15),cols=4)
        for reg in self.regs:
            reg.curReg = 0
            self.binLayout.add_widget(reg.build())
    def build(self, app):
        self.curLayout = super(ProgADFX35X, self).build(app)
        if not hasattr(self, "binLayout"):self.buildFirst()
        self.curLayout.add_widget(self.tabPanel)
        self.curLayout.add_widget(self.binLayout)
        self.progBtn.pos_hint={'y':0.045*app.defStyle["size_hint"][1],'x':.01}
    def onhide(self, app):
        self.curLayout.remove_widget(self.binLayout)
        self.curLayout.remove_widget(self.tabPanel)
        super(ProgADFX35X, self).onhide(app)
    def writeReg(self, obj=None):
        tmp = self.data.numWriteData
        self.userLog.log("Directly writing 0x%08X to the part" % tmp)
        Len = (self.regs[0].wordSz + 7) / 8
        rtn = [0] * Len
        lsbfirst = self.spiBus.lsbfirst
        for c in xrange(Len):
            c = Len - c - 1
            rtn[c] = tmp & 0xFF
            if lsbfirst: rtn[c] = bitMapInt(rtn[c],8)[::-1]
            tmp >>= 8
        rtn = map(int, rtn)
        spiPulseData(self.spiBus, rtn)
    def loadReg(self, obj=None):
        if self.data.numWriteData & 15 >= len(self.regs): return
        self.setReg(self.data.numWriteData)
        self.onProgChg()
        self.loadDataFromRegs()
        self.loadTxtData()
    def setReg(self, regNum):
        regId = regNum & 15
        if regId >= len(self.regs): return
        self.regs[regId][4:] = regNum >> 4
    def updPhaseRsyn(self):
        self.data.numPhaseTimeout = self.data.numPhaseRsynClk / self.data.numPFD_Freq
        self.loadTxtAttr(*self.kTxtData.PhaseTimeout)
        self.updReg12Dat()
    def updOutFreq(self):
        super(ProgADFX35X, self).updOutFreq()
        self.updNegBld()
        self.updCalib()
        self.updPhaseRsyn()
    def updAutoNOrder(self):
        if "bAutoSelDiv" in self.data and self.data.bAutoSelDiv and self.data.selAutoN == 0:
            freq = {
                0: lambda: self.data.numRF_Out,
                1: lambda: self.getExtFreq(
                    self.data.numNVal *
                    (self.data.numPFD_Freq * 10 ** self.pfdFreqShift)),
                2: lambda: self.getExtFreq(
                    (self.data.numFrac1Val + self.data.numFrac2Val/self.data.numMod2Val)
                    / self.data.numMod1Val *
                    (self.data.numPFD_Freq * 10 ** self.pfdFreqShift))}[self.data.selAutoN]()
            fMin, fMax = self.rangeVCO_Freq
            count = 0
            while freq < fMin and count + 1 < len(self.lstDivRFOpts):
                count += 1
                freq *= 2
            while freq > fMax and count > 0:
                count -= 1
                freq /= 2
            if count != self.data.selDivRF:
                self.data.selDivRF = count
                self.loadTxtAttr(*self.kTxtData.DivRF)
        super(ProgADFX35X, self).updAutoNOrder()
    def getIntFreq(self, freq):
        if self.data.selFdBkSel == 1: # Fundamental
            nType = self.getAttrEnterSet1("NVal", "num")[0]
            toNum = {"int": int, "float": float, "Decimal": Decimal}[nType]
            freq *= toNum(2. ** self.data.selDivRF)
        return freq
    def getExtFreq(self, freq):
        if self.data.selFdBkSel == 1: # Fundamental
            nType = self.getAttrEnterSet1("NVal", "num")[0]
            toNum = {"int": int, "float": float, "Decimal": Decimal}[nType]
            freq /= toNum(2. ** self.data.selDivRF)
        return freq
    def updADC(self):
        self.txtADC_ClkDiv.disabled = self.data.bAutoADC
        if self.data.bAutoADC:
            val = (self.data.numPFD_Freq * 10 - 2) / 4
            if val % 1 > 0: val += 1
            val = int(val)
            if val < 0: val = 0
            elif val > 255: val = 255
            self.data.numADC_ClkDiv = val
            self.loadTxtAttr(*self.kTxtData.ADC_ClkDiv)
        denom = self.data.numADC_ClkDiv * 4 + 2
        numer = self.data.numPFD_Freq * 1000
        self.data.numFreqADC = numer / denom
        self.loadTxtAttr(*self.kTxtData.FreqADC)
        self.updReg10Dat()
    def updCalib(self):
        self.txtDivBndVCO.disabled = self.data.bAutoFast
        self.txtTimeout.disabled = self.data.bAutoFast
        self.txtWaitALC.disabled = self.data.bAutoFast
        self.txtSynLkTimeout.disabled = self.data.bAutoFast
        if self.data.bAutoFast:
            self.data.numSynLkTimeout = 12
            self.data.numWaitALC = 30
            num3 = int(restrict(ceil(self.data.numPFD_Freq / 2.4), 2, 255)+.5)
            self.data.numDivBndVCO = num3
            num4 = int(ceil(self.data.numPFD_Freq * 50 / self.data.numWaitALC))
            self.data.numTimeout = restrict(num4, 2, 1023)
            self.loadTxtAttr(*self.kTxtData.DivBndVCO)
            self.loadTxtAttr(*self.kTxtData.Timeout)
            self.loadTxtAttr(*self.kTxtData.WaitALC)
            self.loadTxtAttr(*self.kTxtData.SynLkTimeout)
        tmTotal = 5 * 11 * self.data.numWaitALC
        #tmTotal += self.data.numSynLkTimeout
        tmTotal *= self.data.numTimeout / self.data.numPFD_Freq
        tmTotal += 11 * 16 * self.data.numDivBndVCO / self.data.numPFD_Freq
        self.data.numTotalCalTm = tmTotal
        self.updReg9Dat()
        self.loadTxtAttr(*self.kTxtData.TotalCalTm)
    def updNegBld(self):
        self.txtBldCP.disabled = self.data.bAutoBld
        if self.data.bAutoBld:
            Icp = self.lstChrgPmpVals[self.data.selChrgPmp] * 1000
            val = restrict(int(Icp*4 / (float(self.data.numNVal)*3.75) + .5),0,255)
            self.data.numBldCP = val
            self.loadTxtAttr(*self.kTxtData.BldCP)
        self.data.numBldCur = self.data.numBldCP * 3.75
        self.updRegAttrDat(["selChrgPmp", "numBldCP"])
        self.loadTxtAttr(*self.kTxtData.BldCur)
    def updPFD_Freq(self):
        super(ProgADFX35X, self).updPFD_Freq()
        if "FreqADC" in self.titles: self.updADC()
    def outputReg(self, x):
        supObj = super(ProgADFX35X, self)
        doIt = ("cntrRst" in self.data) and self.regHasAttr(x, "cntrRst")
        if doIt and self.data.bAutoCntrRst:
            self.data.cntrRst = True
            self.userLog.log("Counter Reset: 1")
            self.updRegAttrDat(["cntrRst"])
            supObj.outputReg(x)
            self.data.cntrRst = False
            self.userLog.log("Counter Reset: 0")
            self.updRegAttrDat(["cntrRst"])
            supObj.outputReg(x)
        else: supObj.outputReg(x)
    def programIt(self, obj=None):
        seq = [(x,0) for x in xrange(len(self.regs))][::-1]
        numWait = 0
        strWait = ""
        lstPrn = map(self.fmtReg,seq)
        if "FreqADC" in self.titles:
            numWait = self.data.numPhaseTimeout * 16
            strWait = u"%.5g \u03BCs" % numWait
            lstPrn[-1:-1] = ["wait " + strWait]
        self.userLog.log("Sequence order: " + ", ".join(lstPrn))
        for x in seq:
            if x[0] == 0:
                if "FreqADC" in self.titles:
                    self.userLog.log("wait 16 ADC Cycles (%s)" % strWait)
                    time.sleep(numWait / 1e7)
            self.outputReg(x)
        super(ProgADFX35X, self).programIt(obj)
#get_enterSetHex(ProgADFX35X, "WriteData", (0, 0xFFFFFFFF), "Enter Register Hex")
@partClass("XM-A5M4-0409")
class ProgXM_A5M4(ProgADFX35X):
    boardImageName = images.baseImg+"0382A.png"
    XM_partName = "XM-A5M4-0409"
    mfgPartName = "ADF4355"
    DevOnly = False
    defRegs = fltrDefRegs(ProgADFX35X, 4, [
        0x00200510, 0x06155551, 0x08006002, 0x350184F6,
        0x00200320, 0x00000001, 0x00001F42, 0x30008984, 0x350284F6])
@partClass("XM-A741-0409")
class ProgXM_A741(ProgADFX35X):
    boardImageName = images.baseImg+"0382A.png"
    XM_partName = "XM-A741-0409"
    mfgPartName = "ADF5355"
    lstCalOutOpts = map(
        isReserv, ["Using RF Out A", "Using RF Out B"])
    lstTxtDatas, lstEnterSets, titles, useFuncs, regDatas = fltrFuncName(
        ["AuxOutPwr", "bAuxOut"], ProgADFX35X)
    lstEnterSets.append(("lst", "CalOut", "updCalOut"))
    lstTxtDatas.append(("bRF_OutB", "chkRF_OutB"))
    lstTxtDatas.append(("CalOut", "lst"))
    titles["bRF_OutB"]=("RF Out B Disable",)
    titles["OutPwr"] = ("RF Power", "RF Output Power")
    titles["CalOut"] = ("Represent Freq",)
    useFuncs["bRF_OutB"]=(6,"updReg6Dat")
    useFuncs["CalOut"] = (6,)
    regDatas["bRF_OutB"]=[(6,10)]
    defRegs = fltrDefRegs(ProgADFX35X, 4, [
        0x200440, 0x1, 0x1F42, 0x3401E476, 0x30008984, 0x2A29FCC9, 0x00C03EBA])
    defData = dict(ProgADFX35X.defData, **{
        "numRF_Out":13600.0, "numOutFreq":13600.0, "numNVal":68.0,
        "numFreqADC": 99.8003992015968, "numTotalCalTm": 2829.42,
        "numBldCur": 56.25, "selCalOut": 1})
    lstLblNCntr = [
        "Frequency at RF Out A (VCO)",
        "Frequency at RF Out B (VCO)"]
    def buildFirst(self):
        super(ProgXM_A741, self).buildFirst()
        self.updCalOut()
    def updCalOut(self):
        self.lblNCntr.text = self.lstLblNCntr[self.data.selCalOut]
        self.updAutoNOrder()
    def getIntFreq(self, freq):
        rtn = super(ProgXM_A741, self).getIntFreq(freq)
        if self.data.selCalOut: rtn /= 2
        return rtn
    def getExtFreq(self, freq):
        rtn = super(ProgXM_A741, self).getExtFreq(freq)
        if self.data.selCalOut: rtn *= 2
        return rtn
@partClass("XM-A3R8-0409")
class ProgXM_A3R8(ProgADFX35X):
    boardImageName = images.baseImg+"0382A.png"
    XM_partName = "XM-A3R8-0409"
    mfgPartName = "ADF4355-2"
    lstEnterSets = list(ProgADFX35X.lstEnterSets)
    chgEnterSet(lstEnterSets, ("float", "RF_Out", (54.,4400.), "updAutoNOrder"))
    """defRegs = fltrDefRegs(ProgADFX35X, 4, [
        0x200410, 0x1AAAAA1, 0x10006002, 0x3501E4F6, 0x2A29FCC9,
        0x00200280, 0x00000001, 0x00001F42, 0x30008984, 0x350324F6])
    defData = dict(ProgADFX35X.defData, **{
        "numRF_Out":4000.0, "numOutFreq":4000.0, "numNVal": 40.0,
        "numBldCur": 93.75, "numTotalCalTm": 2829.42,
        "numFreqADC": 99.8003992015968})"""
    defData = {
        'numNVal': 567.125, 'numFreqADC': 97.5609756097561, 'numMod1Val': 16777216,
        'numWriteData': 0, 'numR_Freq': 80.0, 'selAutoR': 0, 'numPFD_Freq': 8.0,
        'bAutoCntrRst': True, 'numSpaceCh': 1.1920928955078125e-05, 'selAutoN': 0,
        'selAutoMod': 1, 'numUsrPFD_Freq': 80.0, 'bAutoSelDiv': True, 'numOutFreq': 2268.5,
        'numRF_Out': 2268.5, 'numTotalCalTm': 2769.25, 'bAutoBld': True, 'numBldCur': 7.5,
        'numPhaseTimeout': 0.125, 'numUsrSpaceCh': 200.0, 'bAutoADC': True, 'bAutoFast': True}
    defRegs = [
        0x00302370, 0x02000001, 0x00000282, 0x00000003,
        0x30050984, 0x00800025, 0x352046B6, 0x12000007,
        0x102D0428, 0x04037CC9, 0x00C0053A, 0x0061300B, 0x0001041C]
@partClass("XM-A5M6-0409")
class ProgXM_A5M6(ProgADFX35X):
    boardImageName = images.baseImg + "0382A.png"
    XM_partName = "XM-A5M6-0409"
    mfgPartName = "ADF4355-3"
    lstMuxoutOpts = map(isReserv,[
        "Three-State Output", "DVdd", "SDgnd", "R Divider Output",
        "N Divider Output", "Reserved", "Digital Lock Detect", "Reserved"])
    lstTxtDatas, lstEnterSets, titles, useFuncs, regDatas = fltrFuncName(
        ["WaitALC"], ProgADFX35X)
    chgEnterSet(lstEnterSets, ("float", "RF_Out", (51.5625,6600.), "updAutoNOrder"))
    rangeVCO_Freq = (3300., 6600.)
    defRegs = fltrDefRegs(ProgADFX35X, 4, [
        0x200240, 0x1, 0x1F42, 0x350544F6, 0x100002E7,
        0x1A69A6B8, 0x60003EBA, 0x81200B, 0x1050C])
    defData = dict(ProgADFX35X.defData, **{
        "numRF_Out":3600.0, "numOutFreq": 3600.0, "numNVal":36.0,
        "numBldCur":157.5, "numTotalCalTm": 2829.42,
        "numFreqADC": 99.8003992015968})
    def updCalib(self):
        self.txtDivBndVCO.disabled = self.data.bAutoFast
        self.txtTimeout.disabled = self.data.bAutoFast
        self.txtSynLkTimeout.disabled = self.data.bAutoFast
        if self.data.bAutoFast:
            self.data.numSynLkTimeout = 12
            num3 = restrict(ceil(self.data.numPFD_Freq / 2.4), 2, 255)
            self.data.numDivBndVCO = int(num3+.5)
            num4 = self.data.numPFD_Freq * 5. / 3
            self.data.numTimeout = int(restrict(ceil(num4), 2, 1023)+.5)
            self.loadTxtAttr(*self.kTxtData.DivBndVCO)
            self.loadTxtAttr(*self.kTxtData.Timeout)
            self.loadTxtAttr(*self.kTxtData.SynLkTimeout)
        tmTotal = 5 * 11 * 30
        #tmTotal += self.data.numSynLkTimeout
        tmTotal *= self.data.numTimeout / self.data.numPFD_Freq
        tmTotal += 11 * 16 * self.data.numDivBndVCO / self.data.numPFD_Freq
        self.data.numTotalCalTm = tmTotal
        self.updReg9Dat()
        self.loadTxtAttr(*self.kTxtData.TotalCalTm)
    def updNegBld(self):
        self.txtBldCP.disabled = self.data.bAutoBld
        if self.data.bAutoBld:
            num3 = (self.data.selChrgPmp + 1) / 3.
            mul = 42
            if self.data.numPFD_Freq <= 80:
                mul = 39 * (self.data.numPFD_Freq / 61.44)
            self.data.numBldCP = restrict(int(floor(num3 * mul)), 0, 255)
            self.loadTxtAttr(*self.kTxtData.BldCP)
        self.data.numBldCur = self.data.numBldCP * 3.75
        self.updRegAttrDat(["selChrgPmp", "numBldCP"])
        self.loadTxtAttr(*self.kTxtData.BldCur)
@partClass("XM-ADF4155-0409")
class ProgXM_ADF4155(ProgADFX35X):
    XM_partName = "XM-ADF4155-0409"
    mfgPartName = "ADF4155"
    catId = CAT_PLL
    lstTxtDatas, lstEnterSets, titles, useFuncs, regDatas = fltrFuncName1(
        ["R_Freq", "PFD_Freq", "RF_Out", "R_Count", "bRefDbl", "bRefDiv",
         "PreScl", "Mod2Val", "IntVal", "Frac1Val", "Frac2Val", "RefMd",
         "Muxout", "Phase", "dblBfr", "BldCP", "BldCur", "LDF", "bNegBld",
         "PhDtcPol", "pwrDwn", "chrgPmp3St", "cntrRst", "bLOL", "DivRF",
         "bMTLD", "bRFOut", "OutPwr", "SpaceCh", "OutFreq", "MuxVolt", "AutoR",
         "LDCycleCnt", "NVal", "ChrgPmp", "bAutoBld", "UsrPFD_Freq", "AutoN",
         "UsrSpaceCh", "AutoMod", "bAutoCntrRst", "WriteData"],ProgADFX35X, True)
    lstSmABPWOpts = map(isReserv,["Use Pulse Bleed Delay", "1.6 ns"])
    lstABPWOpts = map(isReserv,["2.6ns", "3.6ns", "4.8ns", "6.0ns"])
    lstLDCycleCntOpts = map(isReserv,["2048", "4096", "8192", "16384"])
    defData = dict(ProgADFX35X.defData)
    defData.update({
        "selFdBkSel":1,#Fundamental
        "numUsrSpaceCh":60,
        "numNVal":50.0, "numOutFreq":5000.0,
        "numSpaceCh":1.1920928955078125e-06,
        "numBldCur":75.0, "numPFD_Freq":100.0})
    regDatas.update({
        "bCSR": [(5,19)], "selABPW": [(5,23)],
        "bPB": [(5,25)], "selSmABPW": [(5,26,28)],
        "selRefMd": [(6,30)], "bNegBld": [(6,12)],
        "numFrac2Val": [(3,4,18)], "numPhase": [(8,4,28)],
        "bDither1": [(8,28)], "bDither2": [(4,30)]})
    lstTxtDatas.extend([
        ("ABPW", "lst"), ("SmABPW", "lst"),
        ("bPB", "chkPB"), ("bCSR", "chkCSR"),
        ("bDither1", "chkDither1"), ("bDither2", "chkDither2")])
    titles.update({
        "bCSR": ("CSR", "Cycle Slip Reduction"),
        "ABPW": ("Pulse Bleed Delay",),
        "bPB": ("Use Pulse Bleed",),
        "SmABPW": ("Pulse Bleed Select",),
        "bDither1": ("Dither 1 (1st Stage)",),
        "bDither2": ("Dither 2 (2nd Stage)",)})
    useFuncs.update({
        "bCSR": (2,"updReg5Dat"), "ABPW": (1,),
        "bPB": (2,"updReg5Dat"), "SmABPW": (1,),
        "bDither1": (2,"updReg8Dat"), "bDither2": (2,"updReg4Dat")})
    lstEnterSets.extend([
        ("lst", "ABPW", "updReg5Dat"),
        ("lst", "SmABPW", "updReg5Dat")])
    chgEnterSet(lstEnterSets, ("lst", "PhDtcPol", "updReg4Dat"))
    chgEnterSet(lstEnterSets, ("lst", "RefMd", "updReg6Dat"))
    chgEnterSet(lstEnterSets, ("int", "Phase", (0,16777215), "updReg8Dat"))
    defRegs = [
        0x320, 0x1, 0x13882, 0x3, 0x70008984,
        0x5, 0x45029036, 0x167, 0x1EA5FE18]
    def __init__(self, fl=None, ver=0):
        self.regs = [Reg32_4(x) for x in xrange(9)]
        super(ProgXM_ADF4155, self).__init__(fl, ver)
    def updADC(self): pass
    def updCalib(self): pass
    def updPhaseRsyn(self): pass
@partClass("XM-A5Y8-0409")
class ProgXM_A5Y8(ProgADFX35X):#TODO: add Div Band Hold in a good place
    boardImageName = images.baseImg+"0382A.png"
    nRegs = 14
    defRegs = [
        0x200510, 0x6155551, 0x8006002, 0x3, 0x32008B84, 0x800025, 0x350300F6,
        0x60000E7, 0x15596568, 0x2719FCC9, 0xC026BA, 0x61200B, 0x15FC, 0xD]
    lstAuxOutOpts = map(isReserv, ["Mirror RF Out", "Fundamental"])
    XM_partName = "XM-A5Y8-0409"
    mfgPartName = "ADF4356"
    lstBldPolOpts = map(isReserv, [
        "Negative", "Positive"])
    lstLESyncEdgeOpts = map(isReserv, [
        "LE Sync To REF Falling Edge",
        "LE Sync To REF Rising Edge"])
    lstTxtDatas, lstEnterSets, titles, useFuncs, regDatas = fltrFuncName(
        [], ProgADFX35X)
    defData = dict(ProgADFX35X.defData)
    defData.update({
        "numNVal": Decimal('81.38020833333333333333333333'),
        "numRF_Out": Decimal("5000"),
        "numTotalCalTm": 2877.83203125,
        "numFreqADC": 99.41747572815534,
        "numOutFreq": Decimal('4999.999999999999814962829229'),
        "numSpaceCh": 2.384185791015625e-06,
        "numPhaseTimeout": 0.016276041666666668,
        "numBldCur": 90.0,
        "numPFD_Freq": 61.44,
        "numUsrPFD_Freq": 61.44,
        "numR_Freq": 122.88
    })
    regDatas.update({
        "selAuxOut": [(6, 25)],
        "selBldPol": [(6, 31)],
        "selLESyncEdge": [(7,27)],
        "numPhaseRsynClk":[(12,12,32)],
        "numFrac2Val": [(2, 18, 32), (13, 18, 32)],
        "numMod2Val": [(2, 4, 18), (13, 4, 18)],
        "bVCOBandHold": [(11, 24)]
    })
    useFuncs.update({
        "AuxOut": (3,),
        "LESyncEdge": (3,),
        "BldPol": (2,),
        "bVCOBandHold": (6, "updReg11Dat")
    })
    titles.update({
        "AuxOut": ("Aux Out Select", "Auxiliary Output Select"),
        "LESyncEdge": ("LE Sync Edge", "LE Sel Sync Edge"),
        "BldPol": ("Bleed Polarity",),
        "bVCOBandHold": ("VCO Band Hold",)
    })
    chgEnterSet(lstEnterSets, ("int", "PhaseRsynClk", (0, 1048575), "updPhaseRsyn"))
    chgEnterSet(lstEnterSets, ("int", "Frac2Val", (0, 268435455), "updAutoNOrder"))
    chgEnterSet(lstEnterSets, ("int", "Mod2Val", (0, 268435455), "updModSpace")),
    chgEnterSet(lstEnterSets, ("Decimal", "NVal", (Decimal(0), Decimal(16777216)), "updAutoNOrder"), "float")
    chgEnterSet(lstEnterSets, ("Decimal", "RF_Out", (Decimal(0), Decimal(16777216)), "updAutoNOrder"), "float")
    updTxtDatas(lstTxtDatas, [
        (-1, "RF_Out", "num", fromDecimal),
        (-1, "NVal", "num", fromDecimal),
        (-1, "OutFreq", "num", fromDecimal),
        (-1, "BldPol", "lst"),
        (-1, "LESyncEdge", "lst"),
        (-1, "AuxOut", "lst"),
        (-1, "bVCOBandHold", "chkVCOBandHold")
    ])
    lstEnterSets.extend([
        ("lst", "AuxOut", "updReg6Dat"),
        ("lst", "BldPol", "updReg6Dat"),
        ("lst", "LESyncEdge", "updReg7Dat")])
    def updNegBld(self):
        self.txtBldCP.disabled = self.data.bAutoBld
        if self.data.bAutoBld:
            Icp = self.lstChrgPmpVals[self.data.selChrgPmp]
            # int acts like floor but only returns integers
            val = restrict(int(24 * (self.data.numPFD_Freq / 61.44) * (Icp / .9375)), 0, 255)
            self.data.numBldCP = val
            self.loadTxtAttr(*self.kTxtData.BldCP)
        self.data.numBldCur = self.data.numBldCP * 3.75
        self.updRegAttrDat(["selChrgPmp", "numBldCP"])
        self.loadTxtAttr(*self.kTxtData.BldCur)
    def updCalib(self):
        self.txtDivBndVCO.disabled = self.data.bAutoFast
        self.txtTimeout.disabled = self.data.bAutoFast
        self.txtWaitALC.disabled = self.data.bAutoFast
        self.txtSynLkTimeout.disabled = self.data.bAutoFast
        if self.data.bAutoFast:
            self.data.numSynLkTimeout = 12
            self.data.numWaitALC = 30
            num3 = int(restrict(ceil(self.data.numPFD_Freq / 1.6), 2, 255)+.5)
            self.data.numDivBndVCO = num3
            num4 = int(ceil(self.data.numPFD_Freq * 50 / self.data.numWaitALC))
            self.data.numTimeout = restrict(num4, 2, 1023)
            self.loadTxtAttr(*self.kTxtData.DivBndVCO)
            self.loadTxtAttr(*self.kTxtData.Timeout)
            self.loadTxtAttr(*self.kTxtData.WaitALC)
            self.loadTxtAttr(*self.kTxtData.SynLkTimeout)
        tmTotal = 5 * 6 * self.data.numWaitALC
        #tmTotal += self.data.numSynLkTimeout
        tmTotal *= self.data.numTimeout / self.data.numPFD_Freq
        tmTotal += 11 * 16 * self.data.numDivBndVCO / self.data.numPFD_Freq
        self.data.numTotalCalTm = tmTotal
        self.updReg9Dat()
        self.loadTxtAttr(*self.kTxtData.TotalCalTm)
@partClass("XM-A5Y9-0409")
class ProgXM_A5Y9(ProgXM_A5Y8):
    XM_partName = "XM-A5Y9-0409"
    mfgPartName = "ADF5356"
    DevOnly = False
    lstTxtDatas, lstEnterSets, titles, useFuncs, regDatas = fltrFuncName(
        ["AuxOut", "AuxOutPwr"], ProgXM_A5Y8)
    regDatas.update({
        "bAuxOut": [(6,10)]})
    defRegs = [
        0x3006E0, 0xAD55551, 0x8006002, 0x3, 0x32008B84,
        0x800025, 0x35030076, 0x60000E7, 0x15596568,
        0x2719FCC9, 0xC026BA, 0x61200B, 0x15FC, 0xD]
    defData = dict(ProgXM_A5Y8.defData)
    defData.update({
        "numNVal": Decimal('110.6770833333333333333333333'),
        "numTotalCalTm": 2877.83203125,
        "numFreqADC": 99.41747572815534,
        "numPhaseTimeout": 0.016276041666666668,
        "numSpaceCh": 2.384185791015625e-06,
        "numBldCur": 90.0,
        "numOutFreq": Decimal('6799.999999999999748349447750'),
        "numRF_Out": Decimal("6800")})
