from BaseRegPart import *

constRegs = {
    0x02, 0x03, 0x04, 0x05, 0x06, 0x0D, 0x0F, 0x12,
    0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C,
    0x1D, 0x1E, 0x20, 0x21, 0x2F, 0x30, 0x31, 0x32,
    0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3D,
    0x3E, 0x3F, 0x40, 0x41, 0x42, 0x43, 0x44, 0x4C,
    0x4D, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D,
    0x5E, 0x5F, 0x6B, 0x6C, 0x6D
}

class TiPFD_Calc(BasePFD_Calc):
    pfdFreqShift = 0
    lstAutoROpts = map(isReserv, [
        "PLL_R", "Desired PFD Frequency"])
    lstAutoRInfs = [
        "PLL_R is manually entered.",
        "PLL_R is calculated from desired PFD frequency."]

    def __init__(self, dctMultVals):
        """
        :param dict[str|unicode,int|long|float|Decimal] dctMultVals:
        """
        super(TiPFD_Calc, self).__init__([
            # 0,          1,              2,                3
            "selAutoR", "numPLL_R", "numUsrPFD_Freq", "numPFD_Freq",
            # 4,           5,         6,         7
            "numR_Freq", "bOSC_2X", "selMULT", "numPLL_PRE_R"])
        self.dctMultVals = dctMultVals

    def getPFD_Freq(self, context):
        return context.getData(self.names[3]) * 10 ** self.pfdFreqShift

    def updateMode(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        selAutoR = context.getData(names[0])
        txtR_Count = context.getWidget(names[1])
        txtUsrPFD_Freq = context.getWidget(names[2])
        txtR_Count.disabled = selAutoR
        txtUsrPFD_Freq.disabled = not selAutoR

    def getR_Freq(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        rFreq = context.getData(names[4])
        if context.getData(names[5]):
            rFreq *= 2
        val = context.getData(names[6])
        if val == 0:
            log.Log(
                "Register DUMP (expecting to crash):\n  " +
                "\n  ".join(context.mapRegsWithKey(
                    lambda reg, regId: "regId=%r\n    value=%08X" % (regId, reg.word.num)
                )),
                2)
        rFreq = self.dctMultVals[val] * rFreq
        rFreq /= context.getData(names[7])
        return rFreq

    def update(self, context):
        """
        :param BaseProgWithRegs context:
        """
        names = self.names
        rFreq = self.getR_Freq(context)
        selAutoR = context.getData(names[0])
        val = None
        if selAutoR:
            minMax = getMinMaxCtx(context, names[1])
            pfdFreq = context.getData(names[2]) * 10 ** self.pfdFreqShift
            val = int(rFreq / pfdFreq + .5)
            val = restrict(val, minMax[0], minMax[1])
            context.setData(names[1], val, DC_SYNCH)
        if val is None:
            val = context.getData(names[1])
        context.setData(names[3], rFreq / (val * 10.0 ** self.pfdFreqShift), DC_SYNCH)
theTiPFD_Calc = TiPFD_Calc({1: 1, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7})
theTiMainModCalc = MainMod_Calc(getNamesModSpace(), [], ["numPLL_DEN"], "AutoR")
theTiIntFracCalc = IntFrac_Calc(getNamesNVal(), ["numPLL_DEN"], ["numPLL_N", "numPLL_NUM"], "AutoR")

@partClass("XM-A724-0609")
class ProgXM_A724(BaseProgWithRegs):
    XM_partName = "XM-A724-0609"
    mfgPartName = "LMX2594"
    catId = CAT_PLL_VCO
    mfgName = "Texas Instruments"
    interface = "3-wire"
    DevOnly = False
    ioType = IO_SPI
    infoLabels = simplePllInfoLabels
    tabPanelStyle = {"size_hint": (.98, .93)}
    progBtnPosHint = {'y': 0.01, 'x': 0.02}
    defBoxStyle = dict(pos_hint={'x': 0.0, "top": 1.0}, size_hint_y=.08)
    defInnerTop = defBoxStyle["pos_hint"]["top"] - defBoxStyle["size_hint_y"]
    tabLayout = [
        ("Config Mode", "mainTab", (
            BorderedGridLayout, None, dict(cols=2, spacing=[10, 10], padding=[6]), [
                (Bordered, "layoutCfgMd", {}, [
                    (Label, None, dict(defBoxStyle, text="Configuration Mode")),
                    (GridLayout, "gridCfgMd", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 2, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})]),
                (Bordered, "layoutCfgInf", {}, [
                    (Label, None, dict(defBoxStyle, text="Configuration Info")),
                    (GridLayout, "gridCfgInf", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 1, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})])])),
        ("Binary", "binLayoutTab", (
            BorderedGridLayout, "binLayout", dict(cols=7, spacing=[0], padding=[5]*4))),
        ("Main", "vcoTab", (
            GridSelFrameLayout, "gridMain", dict(
                rows=1, spacing=[10, 10], padding=[6], side_names=["VCO", "PFD", "Misc\n   /\nMOD"]), [
                (Bordered, "layoutMod", {}, [
                    (Label, None, dict(defBoxStyle, text="Channel Spacing (Misc/Mod)")),
                    (GridLayout, "gridMod", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 2, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})], -4),
                (Bordered, "layoutRef", {}, [
                    (Label, None, dict(defBoxStyle, text="Reference (PFD)")),
                    (GridLayout, "gridRef", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 2, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})], -4),
                (Bordered, "layoutVCO", {"size_hint_x": 1.2}, [
                    (Label, None, dict(defBoxStyle, text="N Counter (VCO)")),
                    (GridLayout, "gridVCO", {
                        "pos_hint": {'x': 0.0, "top": defInnerTop}, "padding": [6],
                        "cols": 2, "size_hint": (1.0, defInnerTop), "spacing": [10, 10]})], -4)])),
        ("Config", "cfgTab", (
            BorderedGridLayout, "gridCfg", dict(cols=4, spacing=[10, 10], padding=[5] * 4))),
        ("SysRef Config", "sysRefTab", (
            BorderedGridLayout, "gridSysRef", dict(cols=4, spacing=[10, 10], padding=[5] * 4))),
        ("MASH/LD", "mashLDTab", (
            BorderedGridLayout, "gridMashLD", dict(cols=4, spacing=[10, 10], padding=[5] * 4))),
        ("IN/OUT/PWR", "outTab", (
            BorderedGridLayout, "gridOut", dict(cols=4, spacing=[10, 10], padding=[5] * 4))),
        ("VCO Config", "cfgVCOTab", (
            BorderedGridLayout, "gridCfgVCO", dict(cols=4, spacing=[10, 10], padding=[5] * 4))),
        ("Ramp", "rampTab", (
            BorderedGridLayout, "gridRamp", dict(cols=4, spacing=[10, 10], padding=[5] * 4))),
        ("Ramp 0", "ramp0Tab", (
            BorderedGridLayout, "gridRamp0", dict(cols=4, spacing=[10, 10], padding=[5] * 4))),
        ("Ramp 1", "ramp1Tab", (
            BorderedGridLayout, "gridRamp1", dict(cols=4, spacing=[10, 10], padding=[5] * 4))),
        ("History", "histTab", (
            BorderedGridLayout, "gridHist", dict(cols=1, spacing=[10, 10], padding=[5] * 4))),
        ("Other", "otherTab", (
            BorderedGridLayout, "gridOther", dict(cols=3, spacing=[10, 10], padding=[5] * 4)))
    ]
    lstMuxoutOpts = map(isReserv, ["Readback", "Lock Detect"])
    lstClkDivOpts = map(isReserv, [
        "Clock Divider Off", "Fast Lock Divider", "Reserved", "Ramp Divider"])
    #lstChrgPmpVals = [.3125 * (x + 1) for x in xrange(16)]
    #lstChrgPmpOpts = map(isReserv, map(getFormatFn("%.15g mA"), lstChrgPmpVals))
    lstCPGOpts = map(isReserv, [
        "0 mA", "6 mA", "Reserved", "12 mA", "3 mA", "9 mA", "Reserved", "15 mA"])
    lstNegBldVals = [3.87, 11.6, 27, 58, 120, 243, 491, 986]
    lstNegBldOpts = map(isReserv, map(getFormatFn(u"%.15g \u03BCA"), lstNegBldVals))
    lstPreSclOpts = map(isReserv, ["4/5", "8/9"])
    lstNSelOpts = map(isReserv, [
        u"Load on \u03A3-\u0394CLK", "Load Delayed 4 Cycle"])
    # LocK DeTeCt Precision
    # Correction Lock Detect Precision
    lstLDPOpts = map(isReserv, ["24 PFD Cycles", "40 PFD Cycles"])
    # [Ph]ase [D]e[t]e[c]tor [Pol]arity
    lstPhDtcPolOpts = map(isReserv, ["Negative", "Positive"])
    lstLESelOpts = map(isReserv, ["LE From Pin", "LE Synch With RefB"])
    lstRmpStatOpts = {
        0: "Normal Operation", 2: "Readback To Muxout",
        3: "Ramp Complete To Muxout", 16: "Charge Pump Up",
        17: "Charge Pump Down"}
    lstSDModMdOpts = {
        0: "Normal Operation", 14: "Disabled When Frac = 0"}
    lstRmpMdOpts = map(isReserv, [
        "Continuous Sawtooth", "Continuous Triangular",
        "Single Sawtooth Burst", "Single Ramp Burst"])
    lstTXInterOpts = map(isReserv, [
        "Interrupt Off", "Load Channel - Continue Sweep",
        "Load Channel - Stop Sweep"])
    lstTXRmpClkOpts = map(isReserv, ["CLK Div", "TX Data"])
    # lstR0BitNumOpts = map(lambda x: isReserv(str(x)), range(32))
    lstDelClkOpts = map(isReserv, ["PFD CLK", "PFD CLK * CLK1"])
    lstExtDivOpts = [isReservStr(u"\u00F7%u", 2 ** x) for x in xrange(4)]
    lstAutoNOpts = theTiIntFracCalc.lstAutoNOpts
    lstAutoNInfs = theTiIntFracCalc.lstAutoNInfs
    lstAutoModOpts = theTiMainModCalc.lstAutoModOpts
    lstAutoModInfs = theTiMainModCalc.lstAutoModInfs
    lstAutoROpts = theTiPFD_Calc.lstAutoROpts
    lstAutoRInfs = theTiPFD_Calc.lstAutoRInfs
    lstMULTOpts = {
        0b00001: "Bypass",
        0b00011: "3X",
        0b00100: "4X",
        0b00101: "5X",
        0b00110: "6X",
        0b00111: "7X"
    }
    lstFCAL_HPFD_ADJOpts = map(isReserv, [
        u"fPD \u2264 100 MHz",
        u"100 MHz < fPD \u2264 150 MHz",
        u"150 MHz < fPD \u2264 200 MHz",
        u"200 MHz < fPD"
    ])
    lstFCAL_LPFD_ADJOpts = map(isReserv, [
        u"fPD \u2265 10 MHz",
        u"10 MHz > fPD \u2265 5 MHz",
        u"5 MHz > fPD \u2265 2.5 MHz",
        u"fPD < 2.5 MHz"
    ])
    lstCAL_CLK_DIVOpts = map(isReserv, [
        u"fOSC \u2264 200 MHz",
        u"200 MHz < fOSC \u2264 400 MHz",
        u"400 MHz < fOSC \u2264 800 MHz",
        u"fOSC > 800 MHz"
    ])
    lstVCO_SELOpts = map(isReserv, [
        "Reserved", "VCO1", "VCO2", "VCO3",
        "VCO4", "VCO5", "VCO6", "VCO7"
    ])
    lstMASH_ORDEROpts = {
        0: "Integer Mode",
        1: "First Order Modulator",
        2: "Second Order Modulator",
        3: "Third Order Modulator",
        4: "Fourth Order Modulator"
    }
    lstOUTA_MUXOpts = map(isReserv, [
        "Channel Divider",
        "VCO",
        "Reserved",
        "High Impedance"
    ])
    lstOUT_ISETOpts = map(isReserv, [
        "Maximum Output Power Boost",
        "More Output Power Boost",
        "Less Output Power Boost",
        "No Output Power Boost"
    ])
    lstOUTB_MUXOpts = map(isReserv, [
        "Channel Divider",
        "VCO",
        "SysRef (SYSREF_EN=1)",
        "High Impedance"
    ])
    lstINPIN_LVLOpts = map(isReserv, [
        "Vin/2",
        "Vin/4",
        "Vin",
        "Reserved"
    ])
    lstINPIN_FMTOpts = map(isReserv, [
        "SYNC = SysRefReq = CMOS",
        "SYNC = LVDS, SysRefReq = CMOS",
        "SYNC = CMOS, SysRefReq = LVDS",
        "SYNC = SysRefReq = LVDS",
        "SYNC = SysRefReq = CMOS",
        "SYNC = LVDS(filtered), SysRefReq = CMOS",
        "SYNC = CMOS, SysRefReq = LVDS(filtered)",
        "SYNC = SysRefReq = LVDS(filtered)"
    ])
    lstLD_TYPEOpts = map(isReserv, [
        "VCO Calibration Status",
        "VCO Calibration Status and Vtune",
    ])
    lstSTSREF_DIV_PREOpts = {
        2: "Divide By 2",
        4: "Divide By 4"
    }
    lstSYSREF_REPEATOpts = map(isReserv, [
        "Master Mode",
        "Repeater Mode"
    ])
    lstCHDIV_DIV2Opts = map(isReserv, [
        "Disabled (CHDIV=2)",
        "Enabled (CHDIV>2)"
    ])
    lstCHDIVOpts = {
        0: "2", 1: "4", 2: "6",
        3: "8", 4: "12", 5: "16",
        6: "24", 7: "32", 8: "48",
        9: "64", 10: "72", 11: "96",
        12: "128", 13: "192", 14: "256",
        15: "384", 16: "512", 17: "768"
    }
    lstRAMP_TRIGBOpts = {
        0: "Disabled",
        1: "RampCLK pin rising edge",
        2: "RampDIR pin rising edge",
        4: "Always triggered",
        9: "RampCLK pin falling edge",
        10: "RampDIR pin falling edge"
    }
    lstRAMP_TRIGAOpts = {
        0: "Disabled",
        1: "RampCLK pin rising edge",
        2: "RampDIR pin rising edge",
        4: "Always triggered",
        9: "RampCLK pin falling edge",
        10: "RampDIR pin falling edge"
    }
    lstRAMP_BURST_TRIGOpts = map(isReserv, [
        "Ramp Transition",
        "Trigger A",
        "Trigger B",
        "Reserved"
    ])
    lstRAMP0_DLYOpts = map(isReserv, [
        "Normal Ramp Length",
        "Double Ramp Length"
    ])
    lstRAMP1_DLYOpts = map(isReserv, [
        "Normal Ramp Length",
        "Double Ramp Length"
    ])
    lstRAMP0_NEXTOpts = map(isReserv, [
        "RAMP0",
        "RAMP1"
    ])
    lstRAMP0_NEXT_TRIGOpts = map(isReserv, [
        "RAMP0_LEN Timeout Counter",
        "Trigger A",
        "Trigger B",
        "Reserved"
    ])
    lstRAMP_MANUALOpts = map(isReserv, [
        "Automatic Ramping Mode",
        "Manual Ramping Mode"
    ])
    lstRAMP1_NEXTOpts = map(isReserv, [
        "RAMP0",
        "RAMP1"
    ])
    lstRAMP1_NEXT_TRIGOpts = map(isReserv, [
        "RAMP1_LEN Timeout Counter",
        "Trigger A",
        "Trigger B",
        "Reserved"
    ])
    defRegs = [
        0x700000,
        0x6F0000,
        0x6E0000,
        0x6D0000,
        0x6C0000,
        0x6B0000,
        0x6A0000,
        0x690021,
        0x680000,
        0x670000,
        0x660000,
        0x650011,
        0x640000,
        0x630000,
        0x620000,
        0x610888,
        0x600000,
        0x5F0000,
        0x5E0000,
        0x5D0000,
        0x5C0000,
        0x5B0000,
        0x5A0000,
        0x590000,
        0x580000,
        0x570000,
        0x560000,
        0x550000,
        0x540000,
        0x530000,
        0x520000,
        0x510000,
        0x500000,
        0x4F0000,
        0x4E0003,
        0x4D0000,
        0x4C000C,
        0x4B0800,
        0x4A0000,
        0x49003F,
        0x480001,
        0x470081,
        0x46C350,
        0x450000,
        0x4403E8,
        0x430000,
        0x4201F4,
        0x410000,
        0x401388,
        0x3F0000,
        0x3E0322,
        0x3D00A8,
        0x3C0000,
        0x3B0001,
        0x3A8001,
        0x390020,
        0x380000,
        0x370000,
        0x360000,
        0x350000,
        0x340820,
        0x330080,
        0x320000,
        0x314180,
        0x300300,
        0x2F0300,
        0x2E07FC,
        0x2DC0DF,
        0x2C1FA3,
        0x2B0000,
        0x2A0000,
        0x290000,
        0x280000,
        0x2703E8,
        0x260000,
        0x250404,
        0x240046,
        0x230004,
        0x220000,
        0x211E21,
        0x200393,
        0x1F43EC,
        0x1E318C,
        0x1D318C,
        0x1C0488,
        0x1B0002,
        0x1A0DB0,
        0x190624,
        0x18071A,
        0x17007C,
        0x160001,
        0x150401,
        0x14E048,
        0x1327B7,
        0x120064,
        0x11012C,
        0x100080,
        0x0F064F,
        0x0E1E70,
        0x0D4000,
        0x0C5001,
        0x0B0018,
        0x0A10D8,
        0x091604,
        0x082000,
        0x0740B2,
        0x06C802,
        0x0500C8,
        0x040A43,
        0x030642,
        0x020500,
        0x010808,
        0x00251E
    ]
    lstDataEntries = [
        (
            InfoListUserDataView(
                "selAutoN", lstAutoNOpts, lstAutoNInfs, theInfoListMaker, False,
                "Autoset Total N From", "Autoset Total N From"
            ),
            "gridCfgMd"
        ),
        (
            InfoListUserDataView(
                "selAutoMod", lstAutoModOpts, lstAutoModInfs, theInfoListMaker, False,
                "Autoset PLL_DEN", "Autoset PLL_DEN"
            ),
            "gridCfgMd"
        ),
        (
            InfoListUserDataView(
                "selAutoR", lstAutoROpts, lstAutoRInfs, theInfoListMaker, False,
                "Autoset PLL_R From", "Autoset PLL_R From"
            ),
            "gridCfgMd"
        ),
        (
            NumberUserDataView(
                "numRF_Out", theDecUFloatMakerBig, (float, fromFloat), (34.0, 15000.0),
                "Desired VCO Output Frequency", "Desired VCO Freq", "MHz"
            ),
            "gridVCO"
        ),
        (
            NumberUserDataView(
                "numNVal", theDecUFloatMakerBig, (float, fromFloat), (28, 524288),
                "N Value", "N Value", None
            ),
            "gridVCO"
        ),
        (
            NumberUserDataView(
                "numPLL_N", theDecUIntMaker, (int, str), (28, 524287),
                "Integer Part of N", "PLL_N", None
            ).setUpstream(RegisterDownstream([(36, 0, 16), (34, 0, 3)])),
            "gridVCO"
        ),
        (
            NumberUserDataView(
                "numPLL_NUM", theDecUIntMaker, (int, str), (0, (1 << 32) - 1),
                "Fractional Numerator of N", "PLL_NUM", None
            ).setUpstream(RegisterDownstream([(43, 0, 16), (42, 0, 16)])),
            "gridVCO"
        ),
        (
            NumberUserDataView(
                "numPLL_DEN", theDecUIntMaker, (int, str), (1, (1 << 32) - 1),
                "Fractional Denominator of N", "PLL_DEN", None
            ).setUpstream(RegisterDownstream([(39, 0, 16), (38, 0, 16)])),
            "gridMod"
        ),
        (
            NumberUserDataView(
                "numOutFreq", None, (None, getFormatFn("%.15g")), None,
                "Actual VCO Freq", "Actual VCO Freq", "MHz", True
            ),
            "gridVCO"
        ),
        (
            NumberUserDataView(
                "numR_Freq", theDecUFloatMakerBig, (float, fromFloat), (10.0, 260.0),
                "Reference Freq", "Reference Freq", "MHz"
            ),
            "gridRef"
        ),
        (
            NumberUserDataView(
                "numUsrSpaceCh", theDecUFloatMakerBig, (float, fromFloat), (1.0, 18000.0),
                "Desired Freq Step Size", "Desired Freq Step Size", "kHz"
            ),
            "gridMod"
        ),
        (
            NumberUserDataView(
                "numSpaceCh", None, (None, Bind(unitHertz, e10=3, fmtBase="%.6g")), None,
                "Actual Channel Spacing", "Actual Channel Spacing", None, True
            ),
            "gridMod"
        ),
        (
            ChkBoxUserDataView(
                "bOSC_2X",
                "Ref x2 Doubler"
            ).setUpstream(RegisterDownstream([(9, 12)])),
            "gridRef"
        ),
        (
            NumberUserDataView(
                "numPLL_PRE_R", theDecUIntMaker, (int, str), (1, 4095),
                "Pre R Divider", "PLL_PRE_R", None
            ).setUpstream(RegisterDownstream([(12, 0, 12)])),
            "gridRef"
        ),
        (
            ListUserDataView(
                "selMULT", lstMULTOpts, theListMaker, False,
                "Reference Multiplier", "Ref MULT"
            ).setUpstream(RegisterDownstream([(10, 7, 12)])),
            "gridRef"
        ),
        (
            NumberUserDataView(
                "numPLL_R", theDecUIntMaker, (int, str), (1, 255),
                "Post R Divider", "PLL_R", None
            ).setUpstream(RegisterDownstream([(11, 4, 12)])),
            "gridRef"
        ),
        (
            NumberUserDataView(
                "numUsrPFD_Freq", theDecUFloatMakerBig, (float, fromFloat), (0.1, 260.0),
                "Desired PFD Freq", "Desired PFD Freq", "MHz"
            ),
            "gridRef"
        ),
        (
            NumberUserDataView(
                "numPFD_Freq", None, (None, getFormatFn("%.15g")), None,
                "Actual PFD Freq", "Actual PFD Freq", "MHz", True
            ),
            "gridRef"
        ),
        (
            ChkBoxUserDataView(
                "bRAMP_EN",
                "RAMP_EN"
            ).setUpstream(RegisterDownstream([(0, 15)])),
            "gridCfg"
        ),
        (
            ChkBoxUserDataView(
                "bVCO_PHASE_SYNC",
                "VCO_PHASE_SYNC"
            ).setUpstream(RegisterDownstream([(0, 14)])),
            "gridCfgVCO"
        ),
        (
            ChkBoxUserDataView(
                "bOUT_MUTE",
                "OUT_MUTE"
            ).setUpstream(RegisterDownstream([(0, 9)])),
            "gridOut"
        ),
        (
            ListUserDataView(
                "selFCAL_HPFD_ADJ", lstFCAL_HPFD_ADJOpts, theListMaker, False,
                "FCAL_HPFD_ADJ", "FCAL_HPFD_ADJ"
            ).setUpstream(RegisterDownstream([(0, 7, 9)])),
            "gridCfg"
        ),
        (
            ListUserDataView(
                "selFCAL_LPFD_ADJ", lstFCAL_LPFD_ADJOpts, theListMaker, False,
                "FCAL_LPFD_ADJ", "FCAL_LPFD_ADJ"
            ).setUpstream(RegisterDownstream([(0, 5, 7)])),
            "gridCfg"
        ),
        (
            ChkBoxUserDataView(
                "bFCAL_EN",
                "FCAL_EN"
            ).setUpstream(RegisterDownstream([(0, 3)])),
            "gridCfg"
        ),
        (
            ListUserDataView(
                "selMuxout", lstMuxoutOpts, theListMaker, False,
                "Muxout", "Muxout"
            ).setUpstream(RegisterDownstream([(0, 2)])),
            "gridOut"
        ),
        (
            ChkBoxUserDataView(
                "bRESET",
                "RESET"
            ).setUpstream(RegisterDownstream([(0, 1)])),
            "gridCfg"
        ),
        (
            ChkBoxUserDataView(
                "bPOWERDOWN",
                "POWERDOWN"
            ).setUpstream(RegisterDownstream([(0, 0)])),
            "gridOut"
        ),
        (
            ListUserDataView(
                "selCAL_CLK_DIV", lstCAL_CLK_DIVOpts, theListMaker, False,
                "CAL_CLK_DIV", "CAL_CLK_DIV"
            ).setUpstream(RegisterDownstream([(1, 0, 3)])),
            "gridCfg"
        ),
        (
            ChkBoxUserDataView(
                "bOUT_FORCE",
                "OUT_FORCE"
            ).setUpstream(RegisterDownstream([(7, 14)])),
            "gridOut"
        ),
#        (
#            NumberUserDataView(
#                "numPhase", theDecSIntMaker, (int, str), (-2048, 2047),
#                "Phase", "Phase", None
#            ).setUpstream(RegisterDownstream([(1, 3, 15)], *signRegValFunc(12))),
#            "gridCfg"
#        ),
#        (
#            ChkBoxUserDataView(
#                "r3CntRst",
#                "Counter Reset"
#            ).setUpstream(RegisterDownstream([(3, 3)])).setDisplay(False),
#            "gridCfg"
#        ),
#        (
#            ChkBoxUserDataView(
#                "bAutoCntrRst",
#                "Reset Counters on Write"
#            ),
#            "gridCfg"
#        ),
        (
            ListUserDataView(
                "selCPG", lstCPGOpts, theListMaker, False,
                "Charge Pump Current Setting", "CPG"
            ).setUpstream(RegisterDownstream([(14, 4, 7)])),
            "gridCfg"
        ),
        (
            ChkBoxUserDataView(
                "bVCO_DACISET_FORCE",
                "VCO_DACISET_FORCE"
            ).setUpstream(RegisterDownstream([(8, 14)])),
            "gridCfgVCO"
        ),
        (
            ChkBoxUserDataView(
                "bVCO_CAPCTRL_FORCE",
                "VCO_CAPCTRL_FORCE"
            ).setUpstream(RegisterDownstream([(8, 11)])),
            "gridCfgVCO"
        ),
        (
            NumberUserDataView(
                "numVCO_DACISET", theDecUIntMaker, (int, str), (0, 511),
                "VCO_DACISET", "VCO_DACISET"
            ).setUpstream(RegisterDownstream([(16, 0, 9)])),
            "gridCfgVCO"
        ),
        (
            NumberUserDataView(
                "numVCO_DACISET_STRT", theDecUIntMaker, (int, str), (0, 511),
                "VCO_DACISET_STRT", "VCO_DACISET_STRT"
            ).setUpstream(RegisterDownstream([(17, 0, 9)])),
            "gridCfgVCO"
        ),
        (
            NumberUserDataView(
                "numVCO_CAPCTRL", theDecUIntMaker, (int, str), (0, 255),
                "VCO_CAPCTRL", "VCO_CAPCTRL"
            ).setUpstream(RegisterDownstream([(19, 0, 8)])),
            "gridCfgVCO"
        ),
        (
            ListUserDataView(
                "selVCO_SEL", lstVCO_SELOpts, theListMaker, False,
                "VCO Select", "VCO_SEL"
            ).setUpstream(RegisterDownstream([(20, 11, 14)])),
            "gridCfgVCO"
        ),
        (
            ChkBoxUserDataView(
                "bVCO_SEL_FORCE",
                "VCO_SEL_FORCE"
            ).setUpstream(RegisterDownstream([(20, 10)])),
            "gridCfgVCO"
        ),
        (
            ChkBoxUserDataView(
                "bMASH_SEED_EN",
                "MASH_SEED_EN"
            ).setUpstream(RegisterDownstream([(37, 15)])),
            "gridMashLD"
        ),
        (
            NumberUserDataView(
                "numPFD_DLY_SEL", theDecUIntMaker, (int, str), (0, 63),
                "PFD_DLY_SEL", "PFD_DLY_SEL"
            ).setUpstream(RegisterDownstream([(37, 8, 14)])),
            "gridCfg"
        ),
        (
            NumberUserDataView(
                "numMASH_SEED", theDecUIntMaker, (int, str), (0, (1 << 32) - 1),
                "MASH_SEED", "MASH_SEED"
            ).setUpstream(RegisterDownstream([(41, 0, 16), (40, 0, 16)])),
            "gridMashLD"
        ),
        (
            NumberUserDataView(
                "numOUTA_PWR", theDecUIntMaker, (int, str), (0, 63),
                "OUTA_PWR", "OUTA_PWR"
            ).setUpstream(RegisterDownstream([(44, 8, 14)])),
            "gridOut"
        ),
        (
            ChkBoxUserDataView(
                "bOUTB_PD",
                "OUTB_PD"
            ).setUpstream(RegisterDownstream([(44, 7)])),
            "gridOut"
        ),
        (
            ChkBoxUserDataView(
                "bOUTA_PD",
                "OUTA_PD"
            ).setUpstream(RegisterDownstream([(44, 6)])),
            "gridOut"
        ),
        (
            ChkBoxUserDataView(
                "bMASH_RESET_N",
                "MASH_RESET_N"
            ).setUpstream(RegisterDownstream([(44, 5)])),
            "gridMashLD"
        ),
        (
            ListUserDataView(
                "selMASH_ORDER", lstMASH_ORDEROpts, theListMaker, False,
                "MASH_ORDER", "MASH_ORDER"
            ).setUpstream(RegisterDownstream([(44, 0, 3)])),
            "gridMashLD"
        ),
        (
            ListUserDataView(
                "selOUTA_MUX", lstOUTA_MUXOpts, theListMaker, False,
                "OUTA_MUX", "OUTA_MUX"
            ).setUpstream(RegisterDownstream([(45, 11, 13)])),
            "gridOut"
        ),
        (
            ListUserDataView(
                "selOUT_ISET", lstOUT_ISETOpts, theListMaker, False,
                "OUT_ISET", "OUT_ISET"
            ).setUpstream(RegisterDownstream([(45, 9, 11)])),
            "gridOut"
        ),
        (
            NumberUserDataView(
                "numOUTB_PWR", theDecUIntMaker, (int, str), (0, 63),
                "OUTB_PWR", "OUTB_PWR"
            ).setUpstream(RegisterDownstream([(45, 0, 6)])),
            "gridOut"
        ),
        (
            ListUserDataView(
                "selOUTB_MUX", lstOUTB_MUXOpts, theListMaker, False,
                "OUTB_MUX", "OUTB_MUX"
            ).setUpstream(RegisterDownstream([(46, 0, 2)])),
            "gridOut"
        ),
        (
            ChkBoxUserDataView(
                "bINPIN_IGNORE",
                "INPIN_IGNORE"
            ).setUpstream(RegisterDownstream([(58, 15)])),
            "gridOut"
        ),
        (
            ChkBoxUserDataView(
                "bINPIN_HYST",
                "INPIN_HYST"
            ).setUpstream(RegisterDownstream([(58, 14)])),
            "gridOut"
        ),
        (
            ListUserDataView(
                "selINPIN_LVL", lstINPIN_LVLOpts, theListMaker, False,
                "INPIN_LVL", "INPIN_LVL"
            ).setUpstream(RegisterDownstream([(58, 12, 14)])),
            "gridOut"
        ),
        (
            ListUserDataView(
                "selINPIN_FMT", lstINPIN_FMTOpts, theListMaker, False,
                "INPIN_FMT", "INPIN_FMT"
            ).setUpstream(RegisterDownstream([(58, 9, 12)])),
            "gridOut"
        ),
        (
            ListUserDataView(
                "selLD_TYPE", lstLD_TYPEOpts, theListMaker, False,
                "LD_TYPE", "LD_TYPE"
            ).setUpstream(RegisterDownstream([(59, 0)])),
            "gridMashLD"
        ),
        (
            NumberUserDataView(
                "numLD_DLY", theDecUIntMaker, (int, str), (0, 65535),
                "LD_DLY", "LD_DLY"
            ).setUpstream(RegisterDownstream([(60, 0, 16)])),
            "gridMashLD"
        ),
        (
            NumberUserDataView(
                "numMASH_RST_COUNT", theDecUIntMaker, (int, str), (0, (1 << 32) - 1),
                "MASH_RST_COUNT", "MASH_RST_COUNT"
            ).setUpstream(RegisterDownstream([(70, 0, 16), (69, 0, 16)])),
            "gridMashLD"
        ),
        (
            ListUserDataView(
                "selSTSREF_DIV_PRE", lstSTSREF_DIV_PREOpts, theListMaker, False,
                "STSREF_DIV_PRE", "STSREF_DIV_PRE"
            ).setUpstream(RegisterDownstream([(71, 5, 8)])),
            "gridSysRef"
        ),
        (
            ChkBoxUserDataView(
                "bSYSREF_PULSE",
                "SYSREF_PULSE"
            ).setUpstream(RegisterDownstream([(71, 4)])),
            "gridSysRef"
        ),
        (
            ChkBoxUserDataView(
                "bSYSREF_EN",
                "SYSREF_EN"
            ).setUpstream(RegisterDownstream([(71, 3)])),
            "gridSysRef"
        ),
        (
            ListUserDataView(
                "selSYSREF_REPEAT", lstSYSREF_REPEATOpts, theListMaker, False,
                "SYSREF_REPEAT", "SYSREF_REPEAT"
            ).setUpstream(RegisterDownstream([(71, 2)])),
            "gridSysRef"
        ),
        (
            NumberUserDataView(
                "numSYSREF_DIV", theDecUIntMaker, (int, str), (0, 2047),
                "SYSREF_DIV", "SYSREF_DIV"
            ).setUpstream(RegisterDownstream([(72, 0, 11)])),
            "gridSysRef"
        ),
        (
            NumberUserDataView(
                "numJESD_DAC1_CTRL", theDecUIntMaker, (int, str), (0, 63),
                "JESD_DAC1_CTRL", "JESD_DAC1_CTRL"
            ).setUpstream(RegisterDownstream([(73, 0, 6)])),
            "gridSysRef"
        ),
        (
            NumberUserDataView(
                "numJESD_DAC2_CTRL", theDecUIntMaker, (int, str), (0, 63),
                "JESD_DAC2_CTRL", "JESD_DAC2_CTRL"
            ).setUpstream(RegisterDownstream([(73, 6, 12)])),
            "gridSysRef"
        ),
        (
            NumberUserDataView(
                "numJESD_DAC3_CTRL", theDecUIntMaker, (int, str), (0, 63),
                "JESD_DAC3_CTRL", "JESD_DAC3_CTRL"
            ).setUpstream(RegisterDownstream([(74, 0, 6)])),
            "gridSysRef"
        ),
        (
            NumberUserDataView(
                "numJESD_DAC4_CTRL", theDecUIntMaker, (int, str), (0, 63),
                "JESD_DAC4_CTRL", "JESD_DAC4_CTRL"
            ).setUpstream(RegisterDownstream([(74, 6, 12)])),
            "gridSysRef"
        ),
        (
            NumberUserDataView(
                "numSYSREF_PULSE_CNT", theDecUIntMaker, (int, str), (0, 15),
                "SYSREF_PULSE_CNT", "SYSREF_PULSE_CNT"
            ).setUpstream(RegisterDownstream([(74, 12, 16)])),
            "gridSysRef"
        ),
        (
            ListUserDataView(
                "selCHDIV_DIV2", lstCHDIV_DIV2Opts, theListMaker, False,
                "CHDIV_DIV2", "CHDIV_DIV2"
            ).setUpstream(RegisterDownstream([(31, 14)])),
            "gridVCO"
        ),
        (
            ListUserDataView(
                "selCHDIV", lstCHDIVOpts, theListMaker, False,
                "CHDIV", "CHDIV"
            ).setUpstream(RegisterDownstream([(75, 6, 11)])),
            "gridVCO"
        ),
        (
            NumberUserDataView(
                "numRAMP_THRESH", theDecUIntMaker, (int, str), (0, (1 << 33) - 1),
                "RAMP_THRESH", "RAMP_THRESH"
            ).setUpstream(RegisterDownstream([(80, 0, 16), (79, 0, 16), (78, 11)])),
            "gridCfg"
        ),
        (
            ChkBoxUserDataView(
                "bQUICK_RECAL_EN",
                "QUICK_RECAL_EN"
            ).setUpstream(RegisterDownstream([(78, 9)])),
            "gridCfgVCO"
        ),
        (
            NumberUserDataView(
                "numVCO_CAPCTRL_STRT", theDecUIntMaker, (int, str), (0, 183),
                "VCO_CAPCTRL_STRT", "VCO_CAPCTRL_STRT"
            ).setUpstream(RegisterDownstream([(78, 1, 9)])),
            "gridCfgVCO"
        ),
        (
            NumberUserDataView(
                "numRAMP_LIMIT_HIGH", theDecUIntMaker, (int, str), (0, (1 << 33) - 1),
                "RAMP_LIMIT_HIGH", "RAMP_LIMIT_HIGH"
            ).setUpstream(RegisterDownstream([(83, 0, 16), (82, 0, 16), (81, 0)])),
            "gridRamp"
        ),
        (
            NumberUserDataView(
                "numRAMP_LIMIT_LOW", theDecUIntMaker, (int, str), (0, (1 << 33) - 1),
                "RAMP_LIMIT_LOW", "RAMP_LIMIT_LOW"
            ).setUpstream(RegisterDownstream([(86, 0, 16), (85, 0, 16), (84, 0)])),
            "gridRamp"
        ),
        (
            ChkBoxUserDataView(
                "bRAMP_BURST_EN",
                "RAMP_BURST_EN"
            ).setUpstream(RegisterDownstream([(96, 15)])),
            "gridRamp"
        ),
        (
            NumberUserDataView(
                "numRAMP_BURST_COUNT", theDecUIntMaker, (int, str), (0, 8191),
                "RAMP_BURST_COUNT", "RAMP_BURST_COUNT"
            ).setUpstream(RegisterDownstream([(96, 2, 15)])),
            "gridRamp"
        ),
        (
            ChkBoxUserDataView(
                "bRAMP0_RST",
                "RAMP0_RST"
            ).setUpstream(RegisterDownstream([(97, 15)])),
            "gridRamp0"
        ),
        (
            ListUserDataView(
                "selRAMP_TRIGB", lstRAMP_TRIGBOpts, theListMaker, False,
                "RAMP_TRIGB", "RAMP_TRIGB"
            ).setUpstream(RegisterDownstream([(97, 7, 11)])),
            "gridRamp"
        ),
        (
            ListUserDataView(
                "selRAMP_TRIGA", lstRAMP_TRIGAOpts, theListMaker, False,
                "RAMP_TRIGA", "RAMP_TRIGA"
            ).setUpstream(RegisterDownstream([(97, 3, 7)])),
            "gridRamp"
        ),
        (
            ListUserDataView(
                "selRAMP_BURST_TRIG", lstRAMP_BURST_TRIGOpts, theListMaker, False,
                "RAMP_BURST_TRIG", "RAMP_BURST_TRIG"
            ).setUpstream(RegisterDownstream([(97, 0, 2)])),
            "gridRamp"
        ),
        (
            NumberUserDataView(
                "numRAMP0_INC", theDecUIntMaker, (int, str), (0, (1 << 30) - 1),
                "RAMP0_INC", "RAMP0_INC"
            ).setUpstream(RegisterDownstream([(99, 0, 16), (98, 2, 16)])),
            "gridRamp0"
        ),
        (
            ListUserDataView(
                "selRAMP0_DLY", lstRAMP0_DLYOpts, theListMaker, False,
                "RAMP0_DLY", "RAMP0_DLY"
            ).setUpstream(RegisterDownstream([(98, 0)])),
            "gridRamp0"
        ),
        (
            NumberUserDataView(
                "numRAMP0_LEN", theDecUIntMaker, (int, str), (0, 65535),
                "RAMP0_LEN", "RAMP0_LEN"
            ).setUpstream(RegisterDownstream([(100, 0, 16)])),
            "gridRamp0"
        ),
        (
            ListUserDataView(
                "selRAMP1_DLY", lstRAMP1_DLYOpts, theListMaker, False,
                "RAMP1_DLY", "RAMP1_DLY"
            ).setUpstream(RegisterDownstream([(101, 6)])),
            "gridRamp1"
        ),
        (
            ChkBoxUserDataView(
                "bRAMP1_RST",
                "RAMP1_RST"
            ).setUpstream(RegisterDownstream([(101, 5)])),
            "gridRamp1"
        ),
        (
            ListUserDataView(
                "selRAMP0_NEXT", lstRAMP0_NEXTOpts, theListMaker, False,
                "RAMP0_NEXT", "RAMP0_NEXT"
            ).setUpstream(RegisterDownstream([(101, 4)])),
            "gridRamp0"
        ),
        (
            ListUserDataView(
                "selRAMP0_NEXT_TRIG", lstRAMP0_NEXT_TRIGOpts, theListMaker, False,
                "RAMP0_NEXT_TRIG", "RAMP0_NEXT_TRIG"
            ).setUpstream(RegisterDownstream([(101, 0, 2)])),
            "gridRamp0"
        ),
        (
            NumberUserDataView(
                "numRAMP0_INC", theDecUIntMaker, (int, str), (0, (1 << 30) - 1),
                "RAMP0_INC", "RAMP0_INC"
            ).setUpstream(RegisterDownstream([(103, 0, 16), (102, 0, 14)])),
            "gridRamp0"
        ),
        (
            NumberUserDataView(
                "numRAMP1_LEN", theDecUIntMaker, (int, str), (0, 65535),
                "RAMP1_LEN", "RAMP1_LEN"
            ).setUpstream(RegisterDownstream([(104, 0, 16)])),
            "gridRamp1"
        ),
        (
            NumberUserDataView(
                "numRAMP_DLY_CNT", theDecUIntMaker, (int, str), (0, 1023),
                "RAMP_DLY_CNT", "RAMP_DLY_CNT"
            ).setUpstream(RegisterDownstream([(105, 6, 16)])),
            "gridRamp"
        ),
        (
            ListUserDataView(
                "selRAMP_MANUAL", lstRAMP_MANUALOpts, theListMaker, False,
                "RAMP_MANUAL", "RAMP_MANUAL"
            ).setUpstream(RegisterDownstream([(105, 5)])),
            "gridRamp"
        ),
        (
            ListUserDataView(
                "selRAMP1_NEXT", lstRAMP1_NEXTOpts, theListMaker, False,
                "RAMP1_NEXT", "RAMP1_NEXT"
            ).setUpstream(RegisterDownstream([(105, 4)])),
            "gridRamp1"
        ),
        (
            ListUserDataView(
                "selRAMP1_NEXT_TRIG", lstRAMP1_NEXT_TRIGOpts, theListMaker, False,
                "RAMP1_NEXT_TRIG", "RAMP1_NEXT_TRIG"
            ).setUpstream(RegisterDownstream([(105, 0, 2)])),
            "gridRamp1"
        ),
        (
            ChkBoxUserDataView(
                "bRAMP_TRIG_CAL",
                "RAMP_TRIG_CAL"
            ).setUpstream(RegisterDownstream([(106, 4)])),
            "gridRamp"
        ),
        (
            NumberUserDataView(
                "numRAMP_SCALE_COUNT", theDecUIntMaker, (int, str), (0, 7),
                "RAMP_SCALE_COUNT", "RAMP_SCALE_COUNT"
            ).setUpstream(RegisterDownstream([(106, 6, 16)])),
            "gridRamp"
        ),
#        (
#            ListUserDataView(
#                "selNegBld", lstNegBldOpts, theListMaker, False,
#                "Neg Bleed Current", "Neg Bleed Current"
#            ).setUpstream(RegisterDownstream([(3, 22, 25)])),
#            "gridChrg"
#        ),
#        (
#            ChkBoxUserDataView(
#                "calcNegBld",
#                "Autoset Neg Bleed"
#            ),
#            "gridChrg"
#        ),
#        (
#            ChkBoxUserDataView(
#                "r3ChrgPmp3St",
#                "CP 3-State"
#            ).setUpstream(RegisterDownstream([(3, 4)])),
#            "gridChrg"
#        ),
#        (
#            ChkBoxUserDataView(
#                "constNegBld",
#                "Const Neg Bleed"
#            ).setUpstream(RegisterDownstream([(3, 21)])),
#            "gridChrg"
#        ),
        (
            NumberUserDataView(
                "numWriteData", theHexUIntMaker, (fromHexStr, getFormatFn("%06X")), (0, 16777215),
                "Enter Register Hex", "Enter Register Hex", None
            ),
            "gridOther"
        )
    ]
    regDuals = [-1] * 113
    regMaker = RegWithoutAddr1
    regCtlSz = regMaker.ctlSz
    regWordSz = regMaker.wordSz
    defData = {
        "numOutFreq":14000.0, "numR_Freq":100.0, "numRF_Out":14000.0,
        "numPFD_Freq":200.0, "calcNegBld":True, "numWriteData":0,
        "bAutoCntrRst":True,
        "numNVal":120.0,"numUsrPFD_Freq":200,
        "selAutoN": 0, "selAutoR": 1}
    dctCalcSystemUpd = {
        "numR_Freq": ("AutoR", False),
        "numNVal": ("AutoN", False),
        "numPLL_N": ("AutoN", False),
        "numPLL_NUM": ("AutoN", False),
        "numUsrSpaceCh": ("AutoMod", False),
        "numPLL_DEN": ("AutoMod", False),
        "numRF_Out": ("AutoN", False),
        "numUsrPFD_Freq": ("AutoR", False),
        "bOSC_2X": ("AutoR", False),
        "numPLL_PRE_R": ("AutoR", False),
        "selMULT": ("AutoR", False),
        "numPLL_R": ("AutoR", False),
        "selChrgPmp": ("NegBld", False),
        "calcNegBld": ("NegBld", True),
        "selAutoN": ("AutoN", True),
        "selAutoR": ("AutoR", True),
        "selAutoMod": ("AutoMod", True)
    }
    calcSystems = {
        "AutoR": (theTiPFD_Calc, ["AutoMod"]),
        "AutoMod": (theTiMainModCalc, "AutoN"),
        "AutoN": theTiIntFracCalc,
        #"NegBld": theNegBld,
    }
    addUserLog = "gridHist"
    cntrRstName = None
    autoCntrRstName = None
    showBinaryOnMain = False
    binLayoutClsName = "named-binLayout"
    do_reset = 0
    do_fcal = 0
    tabNameNoWrite = {"otherTab"}

    def postBuildFirst(self):
        self.dctNamed["gridOther"].add_widget(Button(text="Write Register", on_release=self.writeSingleReg))

    def dumpRegs(self):
        return self.mapRegsWithKey(lambda reg, regId: ((regId if isinstance(regId, int) else regId[0]) << 16) | reg.word.num)

    def addAllRegs(self, binLayout):
        """
        :param Layout binLayout:
        """
        for regId in xrange(len(self.regDuals)):
            wid = self.getReg(regId).build(BorderedLabel, bdr=0, bdo=0)
            if regId in constRegs:
                wid.color = (.75, .75, .75, 1)
            binLayout.add_widget(wid)

    def outputReg(self, regId):
        if not isinstance(regId, (int, long)) and len(regId) >= 3 and regId[2] in ["RESET=1", "RESET=0", "FCAL_EN=1", "FCAL_EN=1 (again)"]:
            if self.do_reset == 0:
                self.do_reset = 2 if self.getData("bRESET") else 1
            if self.do_fcal == 0:
                self.do_fcal = 2 if self.getData("bFCAL_EN") else 1
            if regId[2] == "RESET=1":
                self.setData("bFCAL_EN", False, DC_SYNCH)
                if self.do_reset == 2:
                    self.setData("bRESET", True, DC_SYNCH)
                reg = super(ProgXM_A724, self).outputReg(0)
            elif regId[2] == "RESET=0" and self.do_reset:
                if self.do_reset == 2:
                    self.setData("bRESET", False, DC_SYNCH)
                reg = super(ProgXM_A724, self).outputReg(0)
            elif regId[2] == "FCAL_EN=1" and self.do_fcal:
                if self.do_fcal == 2:
                    self.setData("bFCAL_EN", True, DC_SYNCH)
                reg = super(ProgXM_A724, self).outputReg(0)
            elif regId[2] == "FCAL_EN=1 (again)":
                reg = super(ProgXM_A724, self).outputReg(0)
                self.setData("bRESET", self.do_reset == 2, DC_SYNCH)
                self.setData("bFCAL_EN", self.do_fcal == 2, DC_SYNCH)
                self.do_reset = 0
                self.do_fcal = 0
        else:
            reg = super(ProgXM_A724, self).outputReg(regId)
        if regId in constRegs:
            reg.lbl.color = (.75, .75, .75, 1)

    #takes a regId 2-tuple of (index, reg.curReg) and formats it according to
    #  reg.isDual
    # regId is now one of the following
    #   an int for the register number
    #   a 2-tuple (register number, dual number code)
    #   a 3-tuple (register number, dual number code, extra string)
    #   a 4-tuple (register number, dual number code, extra string, "is it really dual?" bool)
    #     if "is it really dual?" is false then word selects are not displayed
    @staticmethod
    def fmtReg(regId, short=True):
        """
        :param int|long|(int|long, int|long) regId:
        :param bool short:
        :rtype: str|unicode
        """
        if isinstance(regId, (int, long)):
            return "R%u (0x%02X)" % (regId, regId)
        elif short and (len(regId) < 4 or regId[3]):
            return "R%u ws %u" % (regId[0], regId[1])
        elif (len(regId) == 3 or len(regId) > 3 and regId[3]) and isinstance(regId[2], (str, unicode)):
            return "R%u (0x%02X) word select %u (%s)" % (regId[0], regId[0], regId[1], regId[2])
        elif len(regId) >= 4 and isinstance(regId[2], (str, unicode)):
            return "R%u (0x%02X) (%s)" % (regId[0], regId[0], regId[2])
        else:
            return "R%u (0x%02X) word select %u" % (regId[0], regId[0], regId[1])

    def getWriteRegSequence(self):
        return [
            (0x00, 0, "RESET=1", False),
            (0x00, 0, "RESET=0", False),
            0x70, 0x6F, 0x6E, 0x6D, 0x6C, 0x6B, 0x6A, 0x69, 0x68, 0x67, 0x66, 0x65, 0x64, 0x63, 0x62, 0x61, 0x60,
            0x5F, 0x5E, 0x5D, 0x5C, 0x5B, 0x5A, 0x59, 0x58, 0x57, 0x56, 0x55, 0x54, 0x53, 0x52, 0x51, 0x50, 0x4F,
            0x4E, 0x4D, 0x4C, 0x4B, 0x4A, 0x49, 0x48, 0x47, 0x46, 0x45, 0x44, 0x43, 0x42, 0x41, 0x40, 0x3F, 0x3E,
            0x3D, 0x3C, 0x3B, 0x3A, 0x39, 0x38, 0x37, 0x36, 0x35, 0x34, 0x33, 0x32, 0x31, 0x30, 0x2F, 0x2E, 0x2D,
            0x2C, 0x2B, 0x2A, 0x29, 0x28, 0x27, 0x26, 0x25, 0x24, 0x23, 0x22, 0x21, 0x20, 0x1F, 0x1E, 0x1D, 0x1C,
            0x1B, 0x1A, 0x19, 0x18, 0x17, 0x16, 0x15, 0x14, 0x13, 0x12, 0x11, 0x10, 0x0F, 0x0E, 0x0D, 0x0C, 0x0B,
            0x0A, 0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02, 0x01, # 0x00,
            (0x00, 0, "FCAL_EN=1", False),
            (0x00, 0, "FCAL_EN=1 (again)", False)
        ]

@partClass("XM-A7M3-0609")
class ProgXM_A7M3(ProgXM_A724):
    XM_partName = "XM-A7M3-0609"
    mfgPartName = "LMX2595"

    lstOUTA_MUXOpts = map(isReserv, [
        "Channel Divider",
        "VCO",
        "VCO2X (VCO2X_EN must be checked)",
        "High Impedance"
    ])
    lstDataEntries, defData = changeLstEntries(
        ProgXM_A724.lstDataEntries, ProgXM_A724.defData, {
            "selOUTA_MUX": {"lstOpts": lstOUTA_MUXOpts}
        }
    )
    lstDataEntries.insert(
        lstFind(lstDataEntries, "selOUTA_MUX", lambda x: x[0].name) + 2,
        (
            ChkBoxUserDataView(
                "bVCO2X_EN",
                "VCO2X_EN"
             ).setUpstream(RegisterDownstream([(27, 0)])),
            "gridOut"
        )
    )