from Base_S_DSA import *
from ..LibIO import enumerateAllSpiBusCs

log = None

@initFunc(__name__.replace("ProgramIt.", ""))
def init(progLog):
    # noinspection PyGlobalUndefined
    global log
    log = progLog


def lockAddrOnShared(sharing, nAddr, bus, cs, addr):
    """
    :param dict[(int|long, int|long),list[bool]] sharing:
    :param int|long bus:
    :param int|long cs:
    :param int|long addr:
    :rtype: list[bool]
    """
    allAddr = sharing[(bus, cs)]
    if allAddr is None:
        allAddr = sharing[(bus, cs)] = [False] * nAddr
    allAddr[addr] = True
    return allAddr

def freeAddrOnShared(sharing, nAddr, bus, cs, addr):
    allAddr = sharing[(bus, cs)]
    if allAddr is None:
        allAddr = sharing[(bus, cs)] = [False] * nAddr
        return allAddr
    allAddr[addr] = False
    return allAddr


class AddrPanelStorage(object):
    __slots__ = [
        "layout", "lbl", "gLayout", "widgets",
        "val", "data", "onChgAny", "chgArg",
        "sharedBus", "isSelected"
    ]
    def __init__(self, sizeX, padX, sizeY, posX, posY, onChgAny, chgArg, addr, sharedBus):
        """

        :param int|long|float sizeX:
        :param int|long|float padX:
        :param int|long|float sizeY:
        :param int|long|float posX:
        :param int|long|float posY:
        :param (T) -> any onChgAny:
        :param T chgArg:
        :param list[bool|int] addr:
        :param list[AddrPanelStorage|None] sharedBus:
        """
        self.layout = Bordered(
            size_hint=(sizeX - padX, sizeY), bdr=8,
            pos_hint={"top": posY, 'x': posX + sizeX + padX / 2})
        self.lbl = Label(
            text="Select Device Address", pos_hint={'x': 0.0, 'top': 1.0},
            size_hint=(1.0, .1))
        self.gLayout = GridLayout(
            size_hint=(1.0, .9), pos_hint={'x': 0.0, "top": .9}, cols=2, rows=4)
        self.layout.add_widget(self.lbl)
        self.layout.add_widget(self.gLayout)
        self.widgets = [
            ToggleButton(
                text="%u \"%s\"" % (c, valToBin(c, 3)[::-1]),
                group=self
            )
            for c in xrange(8)]
        for c in xrange(8):
            self.widgets[c].fbind("state", AddrPanelStorage.onSelAddr, c)
            self.widgets[c].background_down = self.widgets[c].background_normal
            self.gLayout.add_widget(self.widgets[c])
        self.val = lstBinToVal(addr)
        self.data = addr
        self.onChgAny = onChgAny
        self.chgArg = chgArg
        self.sharedBus = sharedBus
        self.isSelected = False
    def destroy(self):
        for c, btn in enumerate(self.widgets):
            btn.funbind("state", AddrPanelStorage.onSelAddr, c)
            btn.group = None
        if self.isSelected:
            self.sharedBus[self.val] = None
            self.isSelected = False
        self.sharedBus = None
        self.chgArg = None
    def updateShared(self, sharedBus):
        b = self.sharedBus is not sharedBus
        if b:
            if self.sharedBus[self.val] is self:
                self.sharedBus[self.val] = None
                self.isSelected = False
            self.sharedBus = sharedBus
        for c, btn in enumerate(self.widgets):
            locked = sharedBus[c] is not None
            isSelf = sharedBus[c] is self
            btn.disabled = locked and not isSelf
            btn.state = "down" if isSelf else "normal"
            btn.color = stRGB[isSelf]
        if b:
            self.setAddrIfNoLk(self.val)
    def setAddrIfNoLk(self, addr):
        """
        :param int|long addr:
        :rtype: bool
        """
        lock = self.sharedBus[addr]
        if lock is None:
            self.widgets[addr].state = "down"
        elif lock is not self:
            self.isSelected = False
            return False
        return True
    @staticmethod
    def onSelAddr(addr, inst, val):
        """
        :param int|long addr:
        :param ToggleButton inst:
        :param str|unicode val:
        """
        self = inst.group
        assert isinstance(self, AddrPanelStorage)
        val = fromState[val]
        if val:
            lock = self.sharedBus[addr]
            if lock is None:
                self.sharedBus[addr] = self
                self.isSelected = True
            elif lock is not self:
                log.Log("expected addr = %u to be unlocked" % addr, 2)
                return
            self.val = addr
            for c in xrange(len(self.data)):
                self.data[c] = addr & 1
                addr >>= 1
        elif self.isSelected:
            self.isSelected = False
            self.sharedBus[self.val] = None
        inst.color = stRGB[val]
        self.onChgAny(self.chgArg)


class ProgPE437XX(BaseProgPart):
    XM_partName = "Unknown"
    mfgPartName = "PE437**"
    catId = CAT_DSA
    mfgName = "Peregrine"
    interface = "3-wire"
    boardImageName = images.baseImg + "0299A.png"
    nAddrs = 8
    sharedAddrs = makeDctKeys(enumerateAllSpiBusCs(), makeListLen, [nAddrs])
    sharedAddrs.update(makeDctKeys([(0, 2), (0, 3), (1, 0), (1, 1), (1, 2), (1, 3)], makeListLen, [nAddrs]))
    top_custom_name = None

    def __init__(self, fl=None, ver=0):
        self.bus = 0
        self.cs = 0
        self.addrPanel = None
        super(ProgPE437XX, self).__init__(fl, ver)
        self.sharedBus = self.sharedAddrs[getSpiBusInfo(self.spiBus)]
        self.addrWord = 7
        self.attnWord = 0
        self.attn = [0] * 7
        self.addr = [1] * 3
        if fl is not None:
            self.attnWord = GetLongStrBytes(fl.read((len(self.attn) + 7) // 8))
            self.addrWord = GetLongStrBytes(fl.read((len(self.addr) + 7) // 8))
            for c in xrange(len(self.attn)):
                self.attn[c] = (self.attnWord >> (7 - c)) & 0x01
            for c in xrange(len(self.addr)):
                self.addr[c] = (self.addrWord >> (7 - c)) & 0x01
        self.oldBitsDisp = self.addr[::-1] + [0] + self.attn[::-1]
        self.buildFirst()
    def serialize(self, fl, ver):
        super(ProgPE437XX, self).serialize(fl, ver)
        self.attnWord = int("".join(map(bitCh, self.attn)) + "0", 2)
        self.addrWord = int("".join(map(bitCh, self.addr)) + "0" * 5, 2)
        fl.write(AlignStrBytesLong(self.attnWord, (len(self.attn) + 7) // 8))
        fl.write(AlignStrBytesLong(self.addrWord, (len(self.addr) + 7) // 8))
        fl.flush()
    def setSpiBus(self, spiBus):
        super(ProgPE437XX, self).setSpiBus(spiBus)
        key = getSpiBusInfo(spiBus)
        try:
            self.sharedBus = self.sharedAddrs[key]
        except KeyError:
            self.sharedBus = self.sharedAddrs[key] = makeListLen(self.nAddrs)
        if self.addrPanel is not None:
            self.addrPanel.updateShared(self.sharedBus)
        self.bus, self.cs = key
        self.spiBus.cshigh = True
    def colorizeBinary(self):
        curBitsDisp = self.addr[::-1] + [0] + self.attn[::-1]
        for c in xrange(len(self.binLbl)):
            if self.binLbl[c].text == "U": continue
            self.binLbl[c].color = cmpRGB[int(curBitsDisp[c] != self.oldBitsDisp[c])]
    def destroy(self, app):
        self.addrPanel.destroy()
        super(ProgPE437XX, self).destroy(app)
    def onhide(self, app):
        self.layout.remove_widget(self.attnPanel.layout)
        self.layout.remove_widget(self.addrPanel.layout)
        self.layout.remove_widget(self.binLayout)
        self.top_custom_name.remove(self.layout)
        super(ProgPE437XX, self).onhide(app)
    def build(self, app):
        layout = super(ProgPE437XX, self).build(app)
        if not hasattr(self, "binLayout"): self.buildFirst()
        sizeY = .61
        posY = .93
        posY1 = posY - sizeY - .03
        self.top_custom_name.add(layout)
        layout.add_widget(self.attnPanel.layout)
        layout.add_widget(self.addrPanel.layout)
        layout.add_widget(self.binLayout)
        self.addrPanel.updateShared(self.sharedBus)
        lstBitsDisp = self.addr[::-1] + [0] + self.attn[::-1]
        for c in xrange(11):
            self.binLbl[c].text = bitCh(lstBitsDisp[c])
            if c < 3 and not self.addrPanel.isSelected:
                self.binLbl[c].text = 'U'
                self.binLbl[c].color = StrRGB("F00") + (1.,)
        self.layout = layout
        self.progBtn.pos_hint = {"top": posY1, 'x': .1}

    def on_chg_custom_name(self, v):
        if self.top_custom_name is not None:
            self.top_custom_name.txt_in.text = v

    def buildFirst(self):
        sizeX = .4
        padX = .015
        #posX = .5 - sizeX/2
        posX = .1
        sizeY = .61
        posY = .93
        posY1 = posY - sizeY - .03
        #sizeX = 1.0 - 2*posX
        self.top_custom_name = TopCustomNameStorage(self)
        self.attnPanel = mkChkBoxList(
            {"size_hint": (sizeX - padX, sizeY), "bdr": 8,
             "pos_hint": {"top": posY, 'x': posX + padX / 2}},
            7, .25, self.attn, "Attenuation: ", "dB", "Attenuation Setting")
        addrPanel = self.addrPanel = AddrPanelStorage(
            sizeX, padX, sizeY, posX, posY, self.__class__.onProgChg, self, self.addr, self.sharedBus)
        onChgAny = lambda theObj: self.onProgChg()
        self.attnPanel.onChgAny = onChgAny
        self.binLayout = BorderedGridLayout(
            cols=12, rows=2, pos_hint={'x': .3, "top": posY1}, size_hint=(.595, .1), bdr=8)
        self.binLayout.add_widget(Label(text=""))
        self.binLbl = [None] * 11
        lstBitsDisp = self.addr[::-1] + [0] + self.attn[::-1]
        lstBitLbl = "A2 A1 A0 D7 D6 D5 D4 D3 D2 D1 D0".split(" ")[::-1]
        for c in xrange(11):
            self.binLayout.add_widget(Label(text=lstBitLbl[c]))
        self.binLayout.add_widget(Label(text="binary"))
        for c in xrange(11):
            self.binLbl[c] = Label()
            """self.binLbl[c].text=bitCh(lstBitsDisp[c])
            if c < 3 and not self.isAddrSel:
                self.binLbl[c].text = 'U'
                self.binLbl[c].color = StrRGB("F00") + (1.,)"""
            self.binLayout.add_widget(self.binLbl[c])
        self.binLbl[:] = self.binLbl[::-1]
        if addrPanel.sharedBus is not None:
            addrPanel.setAddrIfNoLk(addrPanel.val)
    def programIt(self, obj=None):
        if not self.addrPanel.isSelected:
            errMsg = "Select an address that matches the\n"
            errMsg += "physical configuration of this device.\n\n"
            errMsg += "The Physical address is determined by\n"
            errMsg += "solderable configuration bits on the PCB.\n"
            errMsg += "No solder = logic 1. The factory default\n"
            errMsg += "address is 111 (no bits soldered)."
            content = FloatLayout(size_hint=(1.0, 1.0), pos_hint={'x': 0.0, "top": 1.0})
            content.add_widget(Label(
                size_hint=(1.0, .8), pos_hint={'x': 0.0, 'y': .2}, text=errMsg))
            btn = Button(text="close", size_hint=(1.0, .2), pos_hint={'x': 0.0, 'y': 0.0})
            content.add_widget(btn)
            popup = Popup(
                title="Device Address Is Undefined", title_color=StrRGB("F00") + (1.,),
                title_size="18sp", separator_color=StrRGB("D0D0D0") + (1.,),
                size_hint=(.4, .5), title_align="center", content=content)
            popup.open()
            btn.bind(on_release=popup.dismiss)
            return
        self.oldBitsDisp = self.addr[::-1] + [0] + self.attn[::-1]
        self.colorizeBinary()
        self.spiBus.cshigh = True
        self.attnWord = int("".join(map(bitCh, self.attn)) + "0", 2)
        self.addrWord = int("".join(map(bitCh, self.addr)) + "0" * 5,2)
        spiPulseData(self.spiBus, [self.attnWord])
        spiPulseData(self.spiBus, [self.addrWord])
        super(ProgPE437XX, self).programIt(obj)
        #self.attnPanel.layout.bdc = StrRGB("CCFF26")
        #self.addrPanel.layout.bdc = StrRGB("CCFF26")
    def onProgChg(self):
        lstBitsDisp = self.addr[::-1] + [0] + self.attn[::-1]
        lblTxts = map(bitCh, lstBitsDisp)
        if not self.addrPanel.isSelected: lblTxts[0:3] = ['U', 'U', 'U']
        for lbl, txt in zip(self.binLbl, lblTxts):
            lbl.text = txt
            if txt == 'U': lbl.color = StrRGB("F00") + (1.,)
            else: lbl.color = StrRGB("FFF") + (1.,)
        self.colorizeBinary()
        super(ProgPE437XX, self).onProgChg()
        """self.attnPanel.layout.bdc = StrRGB("FFF")
        self.addrPanel.layout.bdc = StrRGB("FFF")"""
    def dispName(self):
        prn = "None"
        if self.addrPanel.isSelected:
            prn = str(lstBinToVal(self.addr))
        return self.XM_partName + " on address: " + prn
    def executeCMD(self, s):
        if s.lower() == "help":
            return "S" + "\n".join([
                "list of commands",
                "  write  -  writes the data to the device",
                "  setAttn <float>  -  sets the attenuation value in dB",
                "  setBits <unsigned integer>  -  sets the bit values with a number",
                "  minAttn  -  sets the attenuation to the minimum",
                "  maxAttn  -  sets the attenuation to the maximum",
                "  help  -  shows this help"
            ])
        elif s.lower() == "write":
            self.programIt(None)
            return "S" + "wrote data"
        elif s.lower().startswith("setbits "):
            numStr = s[8:].strip()
            base = 10
            if numStr.startswith("0b"):
                base = 2
                numStr = numStr[2:]
            elif numStr.startswith("0x"):
                base = 16
                numStr = numStr[2:]
            elif numStr.startswith("0o"):
                base = 8
                numStr = numStr[2:]
            tmp = int(numStr, base)
            for c, (box, lbl) in enumerate(self.attnPanel.widgets[:-1]):
                box.active = tmp & 1
                tmp >>= 1
            return "S" + "set attenuation bits to %s" % "".join(map(str, self.attn))
        elif s.lower() in ["minattn", "maxattn"]:
            x = s.lower() == "maxattn"
            for c, (box, lbl) in enumerate(self.attnPanel.widgets[:-1]):
                box.active = x
            return "S" + "set attenuation bits to %s" % "".join(map(str, self.attn))
        else:
            return "E" + "unrecognized command, type help to see list of commands"
@partClass("XM-A3K6-0604")
class ProgXM_A3K6(ProgPE437XX):
    XM_partName = "XM-A3K6-0604"
    mfgPartName = "PE43704"
@partClass("XM-A7D4-0604")
class ProgXM_A7D4(ProgPE437XX):
    XM_partName = "XM-A7D4-0604"
    mfgPartName = "F1956"
    mfgName = "IDT"
@partClass("XM-A3K7-0604")
class ProgXM_A3K7(ProgPE437XX):
    XM_partName = "XM-A3K7-0604"
    mfgPartName = "PE43705"
@partClass("XM-A3K8-0604")
class ProgXM_A3K8(ProgPE437XX):
    XM_partName = "XM-A3K8-0604"
    mfgPartName = "PE43712"
@partClass("XM-A3K9-0604")
class ProgXM_A3K9(ProgPE437XX):
    XM_partName = "XM-A3K9-0604"
    mfgPartName = "PE43713"
@partClass("XM-PE43711-0604")
class ProgPE_43711(BaseProg_S_DSA):
    allowChain = False
    invBits = False
    lsbFirst = True
    XM_partName = "XM-PE43711-0604"
    mfgName = "Peregrine"
    mfgPartName = "PE43711"
    nBits = 7
    lsbVal = .25