from ..LibBaseProgPart import *
# struct (imported by TinyLib) needed for MPACs


log = None

@initFunc(__name__.replace("ProgramIt.", ""))
def init(progLog):
    # noinspection PyGlobalUndefined
    global log
    log = progLog

def getLoadDat(ch, funcMap, numBits):
    return [ch.word[funcMap(x)] for x in xrange(numBits)]
stblMap = [0,1,11,12]
mapStbl = lambda pos: stblMap[pos]
mapPhase = lambda pos: pos + 2
mapAttn = lambda pos: pos + 7
class ProgPE461XX(BaseProgPart):
    XM_partName = "Unknown"
    mfgPartName = "PE461**"
    interface = "SPI"
    catId = CAT_MPAC
    mfgName = "Peregrine"
    top_custom_name = None
    def __init__(self, fl=None, ver=0):
        super(ProgPE461XX, self).__init__(fl, ver)
        self.old = ValuePack()
        self.old.ch0 = bitMapInt(0,14)
        self.old.ch1 = bitMapInt(0x2000,14)
        if fl is not None:
            self.old.ch0.num = GetLongStrBytes(fl.read(2))
            self.old.ch1.num = GetLongStrBytes(fl.read(2))
        self.stblVals = [.05, .1, .2]
        self.attnVals = [.5, 1.0, 2.0, 4.0]
        self.phaseVals = [2.8, 5.6, 11.25, 22.5, 45.0]
        self.ch0 = ValuePack(dict(lbl="RF_OUT1",word=bitMapInt(self.old.ch0.num,14),chnum=0))
        self.ch1 = ValuePack(dict(lbl="RF_OUT2",word=bitMapInt(self.old.ch1.num,14),chnum=1))
        self.binLbls = "S0 S1 P0 P1 P2 P3 P4 M0 M1 M2 M3 S2 S3 C0".split(" ")
        self.xMid = .265
        self.xLenR = .725
        self.xLeft = 0.01
        self.xLenL = .235
        self.yMid = .38
        self.yLenD = .3#y length down
        self.buildFirst()
    def serialize(self, fl, ver):
        super(ProgPE461XX, self).serialize(fl, ver)
        fl.write(AlignStrBytesLong(self.ch0.word.num, 2))
        fl.write(AlignStrBytesLong(self.ch1.word.num, 2))
        fl.flush()
    def onProgChg(self):
        super(ProgPE461XX, self).onProgChg()
    def colorizeBinary(self):
        for k in self.old:
            ch = getattr(self,k)
            curCmp = bitMapInt(self.old[k].num ^ ch.word.num ,14)
            for bit, c in zip(curCmp, xrange(len(curCmp))):
                ch.binLbls[c].color = cmpRGB[bit]
    def programIt(self, obj=None):
        self.spiBus.xfer2(map(ord, struct.pack(">H", self.ch0.word.flipNum())))
        self.spiBus.xfer2(map(ord, struct.pack(">H", self.ch1.word.flipNum())))
        self.old.ch0.num = self.ch0.word.num
        self.old.ch1.num = self.ch1.word.num
        self.colorizeBinary()
        super(ProgPE461XX, self).programIt(obj)
    def on_chg_custom_name(self, v):
        if self.top_custom_name is not None:
            self.top_custom_name.txt_in.text = v
    def buildFirst(self):
        self.top_custom_name = TopCustomNameStorage(self)
        self.bins = GridLayout(
            cols=15,rows=3,pos_hint={'x':0.0,"top":.7}, size_hint_y=.7)
        xMid = self.xMid
        xLenR = self.xLenR
        xLeft = self.xLeft
        xLenL = self.xLenL
        yMid = self.yMid
        yLenD = self.yLenD#y length down
        origSz = [.175, .1]
        origSz[0] /= xLenL
        origSz[1] /= yLenD
        self.binLayout = Bordered(
            pos_hint={'x':xMid,"top":yMid},size_hint=(xLenR, yLenD),bdr=8)
        self.binLayout.add_widget(Label(
            text="Binary Data",size_hint=(.175,.1),pos_hint={'x':0.0,"top":.95}))
        self.bins.add_widget(Label(text="",color=StrRGB("0FF") + (1.,)))
        for lbl in self.binLbls: self.bins.add_widget(Label(text=lbl))
        self.binLayout.add_widget(self.bins)
        self.progPanel = Bordered(
            pos_hint={'x':xLeft, "top":yMid},size_hint=(xLenL,yLenD),bdr=8)
        self.progPanel.add_widget(Label(
            text="Programming",size_hint_y=.05/yLenD,
            pos_hint={'x':0.0,"top":.99}))
        self.hexGrid = GridLayout(
            size_hint=(1.0,1.0-origSz[1]-.05/yLenD-.1),
            pos_hint={'x':0.0,"top":.99-.05/yLenD}, rows=2, cols=2)
        for ch, cols, posX, szX in [(self.ch0,1,xLeft,xLenL),(self.ch1,3,xMid,xLenR)]:
            self.bins.add_widget(Label(text=ch.lbl, size_hint_x=None, width=70))
            ch.binLbls = list()
            for txt in ch.word.toStr(14, True, 2):
                ch.binLbls.append(Label(text=txt))
                self.bins.add_widget(ch.binLbls[-1])
            ch.outerFrame = Bordered(
                size_hint=(szX,.575),pos_hint={'x':posX, "top":1.0},bdr=8)
            ch.frame = GridLayout(pos_hint={'x':0.0,"top":.94},
                size_hint=(1.0,.94),spacing=[10,10],padding=[6],cols=cols)
            ch.frame.add_widget(CheckBits(
                None, "Phase", getLoadDat(ch, mapPhase, 5), self.phaseVals,
                Bind(self.onChChgChk, ch, mapPhase),"Phase: ","\xc2\xb0",bdr=8))
            if ch.chnum == 1:
                ch.frame.add_widget(CheckBits(
                    None, "Attenuation", getLoadDat(ch, mapAttn, 4),
                    self.attnVals,Bind(self.onChChgChk, ch, mapAttn),
                    "Attenuation: ", "dB",bdr=8,size_hint_x=None,width = 170))
                ch.frame.add_widget(CheckBits(
                    None, "IL Stabilizer Setting", getLoadDat(ch, mapStbl, 3),
                    self.stblVals,Bind(self.onChChgChk, ch, mapStbl),
                    "IL Stabilizer Setting: ",size_hint_x=None,bdr=8,width=200))
            self.hexGrid.add_widget(Label(text=ch.lbl))
            ch.hexLbl = Label(text="0x" + ch.word.flipToStr(4,False,16))
            self.hexGrid.add_widget(ch.hexLbl)
            #startHexY -= hexLblH
            ch.outerFrame.add_widget(ch.frame)
            ch.outerFrame.add_widget(Label(
                text=ch.lbl,pos_hint={'x':0.0, "top":.99},size_hint=(1.0,.05)))
        self.progPanel.add_widget(self.hexGrid)
    def onhide(self, app):
        self.progPanel.remove_widget(self.progBtn)
        self.layout.remove_widget(self.binLayout)
        self.layout.remove_widget(self.progPanel)
        self.layout.remove_widget(self.ch0.outerFrame)
        self.layout.remove_widget(self.ch1.outerFrame)
        self.top_custom_name.remove(self.layout)
        super(ProgPE461XX, self).onhide(app)
    def build(self, app):
        layout = super(ProgPE461XX, self).build(app)
        if not hasattr(self, "binLayout"):self.buildFirst()
        xMid = self.xMid
        xLenR = self.xLenR
        xLeft = self.xLeft
        xLenL = self.xLenL
        yMid = self.yMid
        yLenD = self.yLenD#y length down
        layout.add_widget(self.binLayout)
        origSz = list(self.progBtn.size_hint)
        origSz[0] /= xLenL
        origSz[1] /= yLenD
        layout.remove_widget(self.progBtn)
        self.progBtn.pos_hint={'x':.5-origSz[0]/2,"top":(yMid - .27)/yLenD}
        self.progBtn.size_hint = tuple(origSz)
        self.progPanel.add_widget(self.progBtn)
        self.top_custom_name.add(layout)
        layout.add_widget(self.ch0.outerFrame)
        layout.add_widget(self.ch1.outerFrame)
        layout.add_widget(self.progPanel)
        self.layout = layout
        self.colorizeBinary()
    def onChChgChk(self, ch, funcMap, chkBits, index):
        bitNum = funcMap(index) 
        ch.word[bitNum] = chkBits.lstBits[index]
        ch.binLbls[bitNum].text = DIGITS[ch.word[bitNum]]
        self.onProgChg()
        ch.hexLbl.text = "0x" + ch.word.flipToStr(4,False,16)
        self.colorizeBinary()
    def destroy(self, app):
        super(ProgPE461XX, self).destroy(app)
@partClass("XM-A3L3-0404")
class ProgXM_A3L3(ProgPE461XX):
    XM_partName = "XM-A3L3-0404"
    mfgPartName = "PE46120"
@partClass("XM-A3L4-0404")
class ProgXM_A3L4(ProgPE461XX):
    XM_partName = "XM-A3L4-0404"
    mfgPartName = "PE46130"
@partClass("XM-A3L5-0404")
class ProgXM_A3L5(ProgPE461XX):
    XM_partName = "XM-A3L5-0404"
    mfgPartName = "PE46140"
