import tarfile
import sys
import os
import subprocess
import time
import traceback
import json

def GetTmFmt():
    Tm = time.localtime()
    return "%u-%02u-%02u %02u:%02u:%02u" % (Tm.tm_year, Tm.tm_mon, Tm.tm_mday, Tm.tm_hour, Tm.tm_min, Tm.tm_sec)

# bytecode is always compatible between versions x.y.z and x.y.(z+1)

def dctMerge(dct0, dct1, chooseFn):
    rtn = {}
    both = set(dct0)&set(dct1)
    only_dct0 = set(dct0)-set(dct1)
    only_dct1 = set(dct1)-set(dct0)
    for k in both: rtn[k] = chooseFn(k, dct0[k], dct1[k])
    for k in only_dct0: rtn[k] = dct0[k]
    for k in only_dct1: rtn[k] = dct1[k]
    return rtn

def metainf_merger(k, val0, val1):
    if k == "version": return val1
    elif k == "os_pkg_ver": return val0
    else: return val0


class UpdateInfo(object):
    def __init__(self, fName, fl, metainf_inf):
        self.fName = fName
        self.name = fName[fName.rfind("/")+1:]
        if metainf_inf is None:
            self.metainf = None
        else:
            flObj = fl.extractfile(metainf_inf)
            try:
                self.metainf = json.load(flObj)
            except ValueError:
                pass
            flObj.close()
        end = "update.tar.gz"
        if self.metainf is None and self.name.endswith(end):
            version = self.name[:len(self.name)-len(end)]
            self.metainf = {
                "version": version[version.rfind("v") + 1:],
                "os_pkg_ver": 0}
        if self.metainf is None:
            self.dispTxt = "<%s>" % self.name
        else:
            self.dispTxt = self.metainf["version"]


def main(argv):
    thisFile = os.path.abspath(__file__).replace("\\", "/")
    pos = thisFile.rfind("/")
    baseDir = thisFile[:pos + 1]
    logFl = open(baseDir+"Updater.log", "wb")
    def Log(Str):
        logFl.write(GetTmFmt() + " [Updater] " + Str + "\n")
    try:
        pos1 = thisFile.rfind("/", 0, pos)
        parDir = thisFile[pos1 + 1:pos]
        parParFullName = thisFile[:pos1]
        locs = locals()
        Log("Running Updater:\n  " + "\n  ".join(map(
            lambda x: "%s=%r" % (x, locs[x]), ["thisFile", "parDir", "parParFullName", "argv"])))
        if parDir not in ["bin", "src"]:
            print "Must be in either bin or src"
            raise SystemExit(1)
        assert len(argv) >= 3
        updateFile = argv[1]
        homeProgramIt = argv[2]
        assert homeProgramIt.startswith(parParFullName),\
            "homeProgramIt=%r, parParFullName=%r" % (homeProgramIt, parParFullName)
        End = 3
        dctWordOpts = {
            "umount": None,
            "wait": None
        }
        chToWord = {
            "u": "umount",
            "w": "wait"
        }
        while len(argv) > End:
            if argv[End].startswith("--"):
                word = argv[End][2:]
                if word not in dctWordOpts:
                    Log("Unrecognized option: --%s"%word)
                    return
                End += 1
                if len(argv) <= End:
                    Log("expected argument for option --%s"%word)
                    return
                dctWordOpts[word] = argv[End]
                End += 1
            elif argv[End].startswith("-"):
                ch = argv[End][1:]
                if len(ch) != 1:
                    Log("expected single character option after -, -%s"%ch)
                    return
                End += 1
                if len(argv) <= End:
                    Log("expected argument for option -%s"%ch)
                    return
                word = chToWord[ch]
                dctWordOpts[word] = argv[End]
                End += 1
            else: break
        if dctWordOpts["wait"] is not None:
            pid = int(dctWordOpts["wait"])
            waitTxt = "waiting on %u"%pid
            if os.name == "posix":
                waitDir = "/proc/%u/stat" % pid
                while os.path.exists(waitDir):
                    Log(waitTxt)
                    time.sleep(1)
                Log("Done waiting")
            elif os.name == "nt":
                import ctypes
                kernel32 = ctypes.windll.kernel32
                kernel32.OpenProcess.argtypes = [ctypes.c_ulong, ctypes.c_int, ctypes.c_ulong]
                kernel32.OpenProcess.restype = ctypes.c_void_p
                Log(waitTxt)
                try:
                    hProc = kernel32.OpenProcess(0x00100400, 0, pid)
                    if hProc is not None and hProc != 0:
                        os.waitpid(hProc, 0)
                        kernel32.CloseHandle(hProc)
                except OSError:
                    pass
                time.sleep(.25)
                Log("Done waiting")
        fl = tarfile.open(updateFile)
        lstFiles = fl.getmembers()
        lstFiles = filter(lambda x: x.path.startswith(parDir), lstFiles)
        metainf_inf = None
        metainf_name = parDir + "/metainf.json"
        metainf_home_name = "/".join([parParFullName,metainf_name])
        metainf_exist = os.path.isfile(metainf_home_name)
        for inf in lstFiles:
            if metainf_exist and metainf_inf is None and inf.name == metainf_name:
                metainf_inf = inf
                continue
            fl.extract(inf, parParFullName)
        data0, data1 = {}, {}
        updInfo = UpdateInfo(updateFile, fl, metainf_inf)
        if updInfo.metainf is not None:
            data1 = updInfo.metainf
        if metainf_exist:
            with open(metainf_home_name, "r") as flObj:
                try:
                    data0 = json.load(flObj)
                except ValueError:
                    pass
        with open(metainf_home_name, "w") as flObj:
            json.dump(dctMerge(data0, data1, metainf_merger), flObj)
        fl.close()
        if dctWordOpts["umount"] is not None:
            arg = dctWordOpts["umount"]
            if os.name == "posix":
                subprocess.call(["sudo", "umount", arg])
            else:
                Log("Other operating system types (not posix) may not be able to dismount, doing nothing")
        if len(argv) > End:
            subprocess.Popen(argv[End:], close_fds=True)
    except:
        Log("ERROR: " + traceback.format_exc()[:-1])
    finally:
        Log("Exitting Updater")
        logFl.close()


if __name__ == "__main__": main(sys.argv)
